(* Proof Accelerator extraction (http://proofos.sf.net/accel/)
 * Copyright (C) 2005  The Regents of the University of California
 * Author: Adam Chlipala
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *)

open Pp
open Names
open Libnames
open Nameops
open Miniml

let int n =
  if n >= 0 then
    int n
  else
    str "(" ++ int n ++ str ")"

type inductive_name = string

type inductive_info = {
    ciPrintAst : (ml_ast -> std_ppcmds) -> ml_ast -> std_ppcmds option;
      ciPrintMatch : (identifier list -> std_ppcmds
	-> std_ppcmds * std_ppcmds) list;
  }

let specialTypes = Hashtbl.create 10

let pp_let s1 expr s2 =
    (hv 0 
       (hv 0 
	  (hov 2 (str "let " ++ s1 ++ str " =" ++ spc () ++ expr) 
	     ++ spc () ++ str "in") ++ 
	  spc () ++ hov 0 s2))

let natName = "Coq.Init.Datatypes#<empty>#nat"

open Big_int

type nat_result =
  | NatPp of std_ppcmds
  | NatConst of big_int

let simplify_nat pp_exp e =
  let rec simp = function
    MLcons (_, ConstructRef((iname, _), 1), [])
    when string_of_kn iname = natName -> NatConst zero_big_int
  | MLcons (_, ConstructRef((iname, _), 2), [e])
    when string_of_kn iname = natName ->
      begin match simp e with
	NatPp pp -> NatPp (str "(CoqNat.succ (" ++ pp ++ str "))")
      | NatConst n -> NatConst (CoqNat.succ n)
      end
  | e ->
      NatPp (pp_exp e) in
  simp e

let big_max_int = big_int_of_int max_int
let big_min_int = big_int_of_int min_int

let pp_big_int n =
  if eq_big_int n zero_big_int then
    str "Big_int.zero_big_int"
  else if eq_big_int n unit_big_int then
    str "Big_int.unit_big_int"
  else if le_big_int big_min_int n && le_big_int n big_max_int then
    str "(CoqNat.of_int (" ++ str (string_of_big_int n) ++ str "))"
  else
    str "(CoqNat.of_string \"" ++ str (string_of_big_int n) ++ str "\")"

let _ = Hashtbl.add specialTypes natName
    {ciPrintAst = (fun pp_exp ast ->
      match simplify_nat pp_exp ast with
	NatPp pp -> Some pp
      | NatConst n -> Some (pp_big_int n));
      ciPrintMatch = [
      (fun args expr -> str "_zero when Big_int.eq_big_int _zero Big_int.zero_big_int", expr);
      (fun args expr ->
	let arg = match args with
	  [arg] -> arg
	| _ -> failwith "Wrong number of arguments for nat.S" in
	let parg = pr_id arg in
	let pargp1 = parg ++ str "_plus_one" in
	pargp1, pp_let parg (str "CoqNat.pred (" ++ pargp1 ++ str ")") expr)]}

let posName = "Coq.NArith.BinPos#<empty>#positive"

let simplify_pos pp_exp e =
  let rec simp = function
    MLcons (_, ConstructRef((iname, _), 3), [])
    when string_of_kn iname = posName -> NatConst unit_big_int
  | MLcons (_, ConstructRef((iname, _), 1), [e])
    when string_of_kn iname = posName ->
      begin match simp e with
	NatPp pp -> NatPp (str "CoqZ.xI (" ++ pp ++ str ")")
      | NatConst n -> NatConst (CoqZ.xI n)
      end
  | MLcons (_, ConstructRef((iname, _), 2), [e])
    when string_of_kn iname = posName ->
      begin match simp e with
	NatPp pp -> NatPp (str "CoqZ.xO (" ++ pp ++ str ")")
      | NatConst n -> NatConst (CoqZ.xO n)
      end
  | e ->
      NatPp (pp_exp e) in
  simp e

let _ = Hashtbl.add specialTypes posName
    {ciPrintAst = (fun pp_exp ast ->
      match simplify_pos pp_exp ast with
	NatPp pp -> Some pp
      | NatConst n -> Some (pp_big_int n));
      ciPrintMatch = [
      (fun args expr ->
	let arg = match args with
	  [arg] -> arg
	| _ -> failwith "Wrong number of arguments for positive.xI" in
	let parg = pr_id arg in
	let pargp1 = parg ++ str "_shifted" in
	pargp1 ++ str " when not (Big_int.eq_big_int " ++ pargp1 ++ str " Big_int.unit_big_int) && Big_int.eq_big_int (Big_int.mod_big_int (" ++ pargp1 ++ str ") CoqNat.two) Big_int.unit_big_int",
	pp_let parg (str "Big_int.div_big_int (" ++ pargp1 ++ str ") CoqNat.two") expr);
      (fun args expr ->
	let arg = match args with
	  [arg] -> arg
	| _ -> failwith "Wrong number of arguments for positive.xO" in
	let parg = pr_id arg in
	let pargp1 = parg ++ str "_shifted" in
	pargp1 ++ str " when not (Big_int.eq_big_int " ++ pargp1 ++ str " Big_int.unit_big_int)",
	pp_let parg (str "Big_int.div_big_int (" ++ pargp1 ++ str ") CoqNat.two") expr);
      (fun args expr -> str "_", expr)]}

let zName = "Coq.ZArith.BinInt#<empty>#Z"

let simplify_Z pp_exp = function
    MLcons (_, ConstructRef((iname, _), 1), [])
    when string_of_kn iname = zName -> NatConst zero_big_int
  | MLcons (_, ConstructRef((iname, _), 2), [e])
    when string_of_kn iname = zName ->
      simplify_pos pp_exp e
  | MLcons (_, ConstructRef((iname, _), 3), [e])
    when string_of_kn iname = zName ->
      begin match simplify_pos pp_exp e with
	NatPp pp -> NatPp (str "CoqZ.neg (" ++ pp ++ str ")")
      | NatConst n -> NatConst (CoqZ.neg n)
      end
  | _ -> failwith "Bad Z constructor"

let _ = Hashtbl.add specialTypes zName
    {ciPrintAst = (fun pp_exp ast ->
      match simplify_Z pp_exp ast with
	NatPp pp -> Some pp
      | NatConst n -> Some (pp_big_int n));
      ciPrintMatch = [
     (fun args expr -> str "_zero when Big_int.eq_big_int _zero Big_int.zero_big_int", expr);
     (fun args expr ->
       let arg = match args with
	 [arg] -> arg
       | _ -> failwith "Wrong number of arguments for Z.Zpos" in
       let parg = pr_id arg in
       parg ++ str " when Big_int.gt_big_int (" ++ parg ++ str ") Big_int.zero_big_int", expr);
     (fun args expr ->
       let arg = match args with
	 [arg] -> arg
       | _ -> failwith "Wrong number of arguments for Z.Zneg" in
       let parg = pr_id arg in
       parg, expr)
   ]}

let bvectorName = "Coq.Bool.Bvector#<empty>#vector"

type n64_result =
  | N64Pp of (unit -> std_ppcmds)
  | N64Const of int64

type bool_result =
  | BoolConst of bool
  | BoolPp of std_ppcmds

let boolName = "Coq.Init.Datatypes#<empty>#bool"

let simplify_bool pp_exp e =
  let simp = function
    MLcons (_, ConstructRef((iname, _), 1), [])
    when string_of_kn iname = boolName -> BoolConst true
  | MLcons (_, ConstructRef((iname, _), 2), [])
    when string_of_kn iname = boolName -> BoolConst false
  | e -> BoolPp (pp_exp e) in
  simp e

let simplify_bvector pp_exp e =
  let rec simp = function
    MLcons (_, ConstructRef((iname, _), 1), _)
    when string_of_kn iname = bvectorName -> N64Const Int64.zero
  | MLcons (_, ConstructRef((iname, _), 2), [b; _; e])
    when string_of_kn iname = bvectorName ->
      begin match simp e with
	N64Pp pp ->
	  begin match simplify_bool pp_exp b with
	    BoolConst true -> N64Pp (fun () -> str "CoqBvector.xI (" ++ pp () ++ str ")")
	  | BoolConst false -> N64Pp (fun () -> str "CoqBvector.xO (" ++ pp () ++ str ")")
	  | BoolPp bpp -> N64Pp (fun () -> str "if " ++ bpp ++ str " then CoqBvector.xI (" ++ pp () ++ str ") else CoqBvector.xO (" ++ pp () ++ str ")")
	  end
      | N64Const n ->
	  begin match simplify_bool pp_exp b with
	    BoolConst true -> N64Const (CoqBvector.xI n)
	  | BoolConst false -> N64Const (CoqBvector.xO n)
	  | BoolPp bpp -> N64Pp (fun () -> str "if " ++ bpp ++ str " then "
				   ++ str (Int64.to_string (CoqBvector.xI n))
				   ++ str "L else "
				   ++ str (Int64.to_string (CoqBvector.xO n))
				   ++ str "L")
	  end
      end
  | e ->
      N64Pp (fun () -> pp_exp e) in
  simp e

let _ = Hashtbl.add specialTypes bvectorName
    {ciPrintAst = (fun pp_exp ast ->
      match simplify_bvector pp_exp ast with
	N64Pp pp -> Some (pp ())
      | N64Const n ->
	  if n = Int64.zero then
	    Some (str "Int64.zero")
	  else if n = Int64.one then
	    Some (str "Int64.one")
	  else
	    Some (str (Int64.to_string n) ++ str "L"));
     ciPrintMatch = [
     (fun args expr -> str "0L", expr);
     (fun args expr ->
       let a, n, v = match args with
	 [v; n; a] -> a, n, v
       | _ -> failwith "Wrong number of arguments for vector.Vcons" in
       let pv = pr_id v in
       let pvs = pv ++ str "_shifted" in
       pvs ++ str " when Int64.compare " ++ pvs ++ str " Int64.zero > 0",
       pp_let (pr_id a) (str "Int64.logand " ++ pvs ++ str " Int64.one = Int64.one")
	 (pp_let (pr_id n) (str "0")
	    (pp_let pv (str "Int64.shift_right " ++ pvs ++ str " 1")
	       expr)))]}
    
let special_type k = Hashtbl.find specialTypes k

let is_special_type k = Hashtbl.mem specialTypes k

