(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Two's-complement arithmetic *)

Require Import Arith Bool Bvector ZArith.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitwise.
Require Import Asm.Bitvector.Convert.
Require Import Asm.Bitvector.Defs.
Require Import Asm.Bitvector.Tactics.

Set Implicit Arguments.

(** * Preliminary definitions *)

(** Used to check for the absense of overflow *)
Inductive belowMax (len : nat) (v : Bvector len) (n : nat) : Prop :=
  | belowMax_rule :
    bvec_to_nat v < pow2 len - n
    -> belowMax v n.

Theorem belowMax_dec : forall len (v : Bvector len) n,
  belowMax v n
  -> forall m, m < n
    -> belowMax v m.
  intros.
  inversion H.
  split.
  omega.
Qed.

Hint Resolve belowMax_dec : asm.


(** * Zero *)

Lemma bzero_btn : forall len,
  bvec_to_nat (bzero len) = 0.
  induction len; simpl; intuition.
Qed.

Lemma bzero_neg : forall len,
  bneg (bzero len) = bzero len.
  induction len; intuition.
  simpl.
  unfold bneg.
  simpl.
  replace (Vcons false (bzero len))
    with (Vcons false (bneg (bzero len))).
  trivial.
  congruence.
Qed.


(** * Increment *)

(** ** Conversion to [nat] *)

Lemma binc_btn : forall len (v : Bvector len),
  eqMod (pow2 len) (bvec_to_nat (binc v)) (S (bvec_to_nat v)).
  vec_ind; myrewrite; myeauto.
Qed.

Hint Resolve binc_btn : asm.

Lemma binc2_btn : forall len (v : Bvector len),
  eqMod (pow2 len) (bvec_to_nat (binc2 v)) (S (S (bvec_to_nat v))).
  intros.
  unfold binc2.
  apply eqMod_trans with (S (bvec_to_nat (binc v))).
  apply binc_btn.
  apply eqMod_S.
  apply binc_btn.
Qed.

Lemma binc3_btn : forall len (v : Bvector len),
  eqMod (pow2 len) (bvec_to_nat (binc3 v)) (S (S (S (bvec_to_nat v)))).
  intros.
  unfold binc3.
  apply eqMod_trans with (S (bvec_to_nat (binc2 v))).
  apply binc_btn.
  apply eqMod_S.
  apply binc2_btn.
Qed.

Lemma binc4_btn : forall len (v : Bvector len),
  eqMod (pow2 len) (bvec_to_nat (binc4 v)) (S (S (S (S (bvec_to_nat v))))).
  intros.
  unfold binc4.
  apply eqMod_trans with (S (bvec_to_nat (binc3 v))).
  apply binc_btn.
  apply eqMod_S.
  apply binc3_btn.
Qed.

(** ** Disequalities *)

Theorem binc_neq : forall len (v : Bvector (S len)),
  binc v <> v.
  intros; intro.
  apply eqMod_S_contra_pow2 with len (bvec_to_nat v).
  apply eqMod_trans with (bvec_to_nat (binc v)).
  rewrite H; apply eqMod_refl.
  apply binc_btn.
Qed.

Theorem binc2_neq : forall len (v : Bvector (S (S len))),
  binc2 v <> v.
  intros; intro.
  apply eqMod_contra with (pow2 (S (S len))) (bvec_to_nat v) 2.
  simpl; add Hpos (pow2_pos len); omega.
  apply eqMod_trans with (bvec_to_nat (binc2 v)).
  rewrite H; apply eqMod_refl.
  replace (bvec_to_nat v + 2) with (S (S (bvec_to_nat v))).
  apply binc2_btn.
  omega.
Qed.

Theorem binc3_neq : forall len (v : Bvector (S (S len))),
  binc3 v <> v.
  intros; intro.
  apply eqMod_contra with (pow2 (S (S len))) (bvec_to_nat v) 3.
  simpl; add Hpos (pow2_pos len); omega.
  apply eqMod_trans with (bvec_to_nat (binc3 v)).
  rewrite H; apply eqMod_refl.
  replace (bvec_to_nat v + 3) with (S (S (S (bvec_to_nat v)))).
  apply binc3_btn.
  omega.
Qed.

Theorem binc_neq' : forall len (v : Bvector (S len)),
  binc v = v -> False.
  apply binc_neq.
Qed.

Theorem binc2_neq' : forall len (v : Bvector (S (S len))),
  binc (binc v) = v -> False.
  apply binc2_neq.
Qed.

Theorem binc3_neq' : forall len (v : Bvector (S (S len))),
  binc (binc (binc v)) = v -> False.
  apply binc3_neq.
Qed.

(** ** Alternate [bvec_to_nat] versions, for use when we know that incrementing won't overflow *)

(** Pattern common to lemmas about different numbers of incrementations *)
Ltac binc_btn_below func lemma :=
  match goal with
    [ |- forall len (v : Bvector len), _ ] =>
    let Hbound := fresh "Hbound" in let Hbound2 := fresh "Hbound" in
      (intros len v Hbelow; destruct Hbelow; destruct (eqMod_nat (lemma len v)) as [x];
	add Hbound (btn_upper v); add Hbound2 (btn_upper (func len v));
	assert (x > 0 -> x * pow2 len >= pow2 len); [intros; induction x; simpl; intuition
	  | destruct x; omega])
  end.

Theorem binc_btn_belowMax : forall len (v : Bvector len),
  belowMax v 1
  -> bvec_to_nat (binc v) = S (bvec_to_nat v).
  binc_btn_below binc binc_btn.
Qed.

Theorem binc2_btn_belowMax : forall len (v : Bvector len),
  belowMax v 2
  -> bvec_to_nat (binc2 v) = S (S (bvec_to_nat v)).
  binc_btn_below binc2 binc2_btn.
Qed.

Theorem binc3_btn_belowMax : forall len (v : Bvector len),
  belowMax v 3
  -> bvec_to_nat (binc3 v) = S (S (S (bvec_to_nat v))).
  binc_btn_below binc3 binc3_btn.
Qed.

Theorem binc4_btn_belowMax : forall len (v : Bvector len),
  belowMax v 4
  -> bvec_to_nat (binc4 v) = S (S (S (S (bvec_to_nat v)))).
  binc_btn_below binc4 binc4_btn.
Qed.

(** ** ?? *)

Ltac binc_btn_below_inc func lemma :=
  match goal with
    [ |- forall len (v : Bvector len) n, _ ] =>
    let Hbound := fresh "Hbound" in let Hbound2 := fresh "Hbound" in
      (intros len v n Hbelow; destruct Hbelow; destruct (eqMod_nat (lemma len v)) as [x];
	add Hbound (btn_upper v); add Hbound2 (btn_upper (func len v));
	assert (x > 0 -> x * pow2 len >= pow2 len); [intros; induction x; simpl; intuition
	  | destruct x; split; omega])
  end.

Theorem belowMax_binc_dec : forall len (v : Bvector len) n,
  belowMax v (S n)
  -> belowMax (binc v) n.
  binc_btn_below_inc binc binc_btn.
Qed.

Theorem belowMax_binc2_dec : forall len (v : Bvector len) n,
  belowMax v (S (S n))
  -> belowMax (binc2 v) n.
  binc_btn_below_inc binc2 binc2_btn.
Qed.

Theorem belowMax_binc3_dec : forall len (v : Bvector len) n,
  belowMax v (S (S (S n)))
  -> belowMax (binc3 v) n.
  binc_btn_below_inc binc3 binc3_btn.
Qed.

Hint Resolve belowMax_binc_dec belowMax_binc2_dec belowMax_binc3_dec : asm.


(** * Decrement *)

Lemma bdec_btn' : forall len (v : Bvector len),
  bvec_to_nat v > 0
  -> bvec_to_nat (bdec v) = pred (bvec_to_nat v).
  vec_ind.
Qed.

Lemma bdec2_btn' : forall len (v : Bvector len),
  bvec_to_nat v > 1
  -> bvec_to_nat (bdec2 v) = pred (pred (bvec_to_nat v)).
  intros len v Hgt.
  unfold bdec2.
  repeat (rewrite bdec_btn'); trivial; omega.
Qed.

Lemma bdec3_btn' : forall len (v : Bvector len),
  bvec_to_nat v > 2
  -> bvec_to_nat (bdec3 v) = pred (pred (pred (bvec_to_nat v))).
  intros len v Hgt.
  unfold bdec3.
  repeat (rewrite bdec_btn'); trivial; omega.
Qed.

Lemma bdec4_btn' : forall len (v : Bvector len),
  bvec_to_nat v > 3
  -> bvec_to_nat (bdec4 v) = pred (pred (pred (pred (bvec_to_nat v)))).
  intros len v Hgt.
  unfold bdec4.
  repeat (rewrite bdec_btn'); trivial; omega.
Qed.

Lemma bdec_btn : forall len (v : Bvector len),
  bvec_to_nat v > 0
  -> bvec_to_nat (bdec v) = bvec_to_nat v - 1.
  intros.
  rewrite <- pred_of_minus.
  apply bdec_btn'; trivial.
Qed.

Lemma bdec2_btn : forall len (v : Bvector len),
  bvec_to_nat v > 1
  -> bvec_to_nat (bdec2 v) = bvec_to_nat v - 2.
  intros.
  rewrite bdec2_btn'; trivial.
  omega.
Qed.

Lemma bdec3_btn : forall len (v : Bvector len),
  bvec_to_nat v > 2
  -> bvec_to_nat (bdec3 v) = bvec_to_nat v - 3.
  intros.
  rewrite bdec3_btn'; trivial.
  omega.
Qed.

Lemma bdec4_btn : forall len (v : Bvector len),
  bvec_to_nat v > 3
  -> bvec_to_nat (bdec4 v) = bvec_to_nat v - 4.
  intros.
  rewrite bdec4_btn'; trivial.
  omega.
Qed.


(** * Negation *)

Theorem bneg_btn : forall len (v : Bvector len),
  eqMod (pow2 len) (bvec_to_nat (bneg v)) (pow2 len - bvec_to_nat v).
  unfold bneg.
  intros.
  replace (pow2 len - bvec_to_nat v) with (S (pow2 len - bvec_to_nat v) - 1);
    try omega.
  apply eqMod_trans with (S (bvec_to_nat (bnot v))); myauto.
  replace (S (pow2 len - bvec_to_nat v) - 1)
    with (S (pow2 len - 1 - bvec_to_nat v )); myrauto.
Qed.

Theorem bneg_btn_sub : forall len (v : Bvector len) n,
  bvec_to_nat v <= n
  -> eqMod (pow2 len) (n - bvec_to_nat v) (n + bvec_to_nat (bneg v)).
  intros len v n Hlt.
  apply eqMod_trans with (n + (pow2 len - bvec_to_nat v)).
  replace (n + (pow2 len - bvec_to_nat v))
    with (n + pow2 len - bvec_to_nat v); try (add Hbound (btn_upper v); omega).
  apply eqMod_plus_minus_mod; myauto.
  apply eqMod_plus; myauto.
  apply eqMod_sym; apply bneg_btn.
Qed.

Theorem bneg_btn_double : forall len (v : Bvector len),
  bneg (bneg v) = v.
  vec_ind;
  [unfold bneg; simpl; rewrite bnot_btn_double; trivial
    | unfold bneg; simpl; unfold bneg in IHlen; rewrite IHlen; trivial].
Qed.


(** * Addition *)

Lemma badd_zero' : forall len (v : Bvector len),
 badd (bzero len) v = v.
  intros.
  unfold badd, barith.
  rewrite bzero_btn.
  simpl.
  apply btn_invert.
Qed.


(** * Subtraction *)

Theorem bsub_btn : forall len (v1 v2 : Bvector len),
  eqMod (pow2 len) (bvec_to_nat (bsub v1 v2))
  (pow2 len + bvec_to_nat v1 - bvec_to_nat v2).
  intros len v1 v2.
  unfold bsub, barith.
  apply eqMod_trans with (bvec_to_nat v1 + bvec_to_nat (bneg v2)).
  apply ntb_invert_mod.
  apply eqMod_trans with (bvec_to_nat v1 + (pow2 len - bvec_to_nat v2)).
  apply eqMod_plus; myauto.
  apply bneg_btn.
  replace (bvec_to_nat v1 + (pow2 len - bvec_to_nat v2))
    with (pow2 len + bvec_to_nat v1 - bvec_to_nat v2).
  myauto.
  add Hupper (btn_upper v2).
  omega.
Qed.

Theorem bsub_btn_neg : forall len (v1 v2 : Bvector len),
  eqMod (pow2 len) (bvec_to_nat (bneg (bsub v1 v2)))
  (bvec_to_nat (bsub v2 v1)).
  intros len v1 v2.
  apply eqMod_trans with (pow2 len - bvec_to_nat (bsub v1 v2)).
  apply bneg_btn.
  apply eqMod_trans with (2 * pow2 len - (pow2 len + bvec_to_nat v1 - bvec_to_nat v2)).
  apply eqMod_minus; myauto.
  apply bsub_btn.
  apply btn_upper''.
  add Hupper (btn_upper v1); omega.
  replace (2 * pow2 len - (pow2 len + bvec_to_nat v1 - bvec_to_nat v2))
    with (2 * pow2 len - pow2 len + bvec_to_nat v2 - bvec_to_nat v1).
  replace (2 * pow2 len - pow2 len) with (pow2 len); try omega.
  apply eqMod_sym; apply bsub_btn.
  add Hbound (btn_upper v2); omega.
Qed.

Lemma bsub_zero : forall len (v : Bvector len),
  bvec_to_nat (bsub v (bzero len)) = bvec_to_nat v.
  intros.
  unfold bsub, barith.
  rewrite bzero_neg.
  rewrite bzero_btn.
  rewrite ntb_invert.
  omega.
  add Hup (btn_upper v); omega.
Qed.

Lemma mod2_bool_1 : mod2_bool 1 = true.
  unfold mod2_bool.
  unfold mod2.
  destruct (Euclid.eucl_dev 2 two_gt_0 1).
  destruct (eq_nat_dec r 0); trivial.
  wrong; omega.
Qed.

Lemma div2_1 : div2 1 = 0.
  unfold div2.
  destruct (Euclid.eucl_dev 2 two_gt_0 1).
  omega.
Qed.

Lemma mod2_bool_0 : mod2_bool 0 = false.
  unfold mod2_bool.
  unfold mod2.
  destruct (Euclid.eucl_dev 2 two_gt_0 0).
  destruct (eq_nat_dec r 0); trivial.
  wrong; omega.
Qed.

Lemma div2_0 : div2 0 = 0.
  unfold div2.
  destruct (Euclid.eucl_dev 2 two_gt_0 0).
  omega.
Qed.

Hint Rewrite mod2_bool_1 div2_1 mod2_bool_0 div2_0 : asm.

Lemma mod2_bool_add_S : forall n1 n2,
  mod2_bool (S (2 * n1 + S (2 * n2))) = false.
  intros.
  unfold mod2_bool, mod2.
  destruct (Euclid.eucl_dev 2 two_gt_0 (S (2 * n1 + S (2 * n2)))).
  destruct (eq_nat_dec r 0); trivial.
  wrong; omega.
Qed.

Lemma div2_add_S : forall n1 n2,
  div2 (S (2 * n1 + S (2 * n2))) = S (n1 + n2).
  intros.
  unfold div2.
  destruct (Euclid.eucl_dev 2 two_gt_0 (S (2 * n1 + S (2 * n2)))).
  omega.
Qed.

Lemma mod2_bool_add : forall n1 n2,
  mod2_bool (2 * n1 + S (2 * n2)) = true.
  intros.
  unfold mod2_bool, mod2.
  destruct (Euclid.eucl_dev 2 two_gt_0 (2 * n1 + S (2 * n2))).
  destruct (eq_nat_dec r 0); trivial.
  wrong; omega.
Qed.

Lemma div2_add : forall n1 n2,
  div2 (2 * n1 + S (2 * n2)) = n1 + n2.
  intros.
  unfold div2.
  destruct (Euclid.eucl_dev 2 two_gt_0 (2 * n1 + S (2 * n2))).
  omega.
Qed.

Hint Rewrite mod2_bool_add_S div2_add_S mod2_bool_add div2_add : asm.

Lemma zero_ntb : forall len,
  nat_to_bvec len 0 = bzero len.
  induction len; srewrite; intuition.
  rewrite IHlen.
  trivial.
Qed.

Hint Rewrite zero_ntb : asm.

Lemma bnot_bzero_mod : forall len,
  eqMod (pow2 len) (bvec_to_nat (bnot (bzero len))) (pow2 len - 1).
  induction len; srewrite; myauto.
  rewrite <- rewrite_S.
  myauto.
Qed.

Hint Resolve bnot_bzero_mod : asm.

Lemma bzero_S_sub : forall len n,
  eqMod (pow2 len) (S n + bvec_to_nat (bnot (bzero len))) n.
  intros.
  apply eqMod_trans with
    (S n + (pow2 len - 1)).
  apply eqMod_plus; myauto.
  replace (S n + (pow2 len - 1))
    with (pow2 len + (S n - 1)).
  apply eqMod_sym; apply eqMod_mod.
  replace (S n - 1) with n; try omega; myauto.
  add Hpos (pow2_pos len); omega.
Qed.

Lemma bdec_binc : forall len (v : Bvector len),
  bdec (binc v) = v.
  vec_ind.
  rewrite H0.
  trivial.
Qed.

Lemma binc_bdec : forall len (v : Bvector len),
  binc (bdec v) = v.
  vec_ind.
  rewrite H0.
  trivial.
Qed.

Lemma bzero_S_dec : forall len (v : Bvector len),
  bvec_to_nat v > 0
  -> eqMod (pow2 len) (bvec_to_nat v + bvec_to_nat (bnot (bzero len))) (bvec_to_nat (bdec v)).
  vec_ind; intro Hgt.
  destrEq Hz (bvec_to_nat (Vtail v)).
  clear IHlen H H0 Hgt Hz v.
  induction len; srewrite; intuition.
  exists 1%Z.
  trivial.
  myrewrite.
  replace (S (2 * 0)) with 1; trivial.
  replace (S (S (S (2 * bvec_to_nat (bnot (bzero len)) + S (2 * bvec_to_nat (bnot (bzero len)))))))
    with (2 * (S (S (2 * bvec_to_nat (bnot (bzero len)))))); try omega.
  apply eqMod_trans with (2 * (2 * 0)); myauto.
  
  intro Hv.
  replace (S (2 * S Hz + S (2 * bvec_to_nat (bnot (bzero len)))))
    with (2 * (S (S Hz + bvec_to_nat (bnot (bzero len))))); try omega.
  apply eqMod_mult_mod.
  apply eqMod_S.
  rewrite <- Hv.
  replace Hz with (bvec_to_nat (Vtail v) - 1); try omega.
  rewrite <- bdec_btn; try omega.
  apply IHlen; omega.

  replace (2 * bvec_to_nat (Vtail v) + S (2 * bvec_to_nat (bnot (bzero len))))
    with (S (2 * (bvec_to_nat (Vtail v) + bvec_to_nat (bnot (bzero len))))); try omega.
  apply eqMod_S.
  apply eqMod_mult_mod.
  apply IHlen; omega.
Qed.

Hint Resolve bzero_S_sub bzero_S_dec : asm.

Lemma bsub_bdec : forall len (v : Bvector len),
  bvec_to_nat v > 0
  -> bsub v (nat_to_bvec len 1) = bdec v.
  vec_ind; unfold bsub, barith; srewrite; intro Hv.

  rewrite (@ntb_mod len
    (S (bvec_to_nat (Vtail v) + bvec_to_nat (bnot (bzero len))))
    (bvec_to_nat (Vtail v))).
  rewrite btn_invert; trivial.
  rewrite <- plus_Sn_m.
  myauto.

  rewrite (@ntb_mod len
    (bvec_to_nat (Vtail v) + bvec_to_nat (bnot (bzero len)))
    (bvec_to_nat (bdec (Vtail v)))).
  rewrite btn_invert; trivial.
  apply bzero_S_dec; omega.
Qed.

Lemma eqMod_eq : forall m x1 x2,
  x1 = x2
  -> eqMod m x1 x2.
  intros; subst; myauto.
Qed.

Lemma badd_btn_neg : forall len (v1 v2 : Bvector len),
  eqMod (pow2 len)
  (bvec_to_nat (bneg (nat_to_bvec len
    (bvec_to_nat v1 + bvec_to_nat v2))))
  (bvec_to_nat (bneg v1) + bvec_to_nat (bneg v2)).
  intros.
  apply eqMod_trans with
    (pow2 len - bvec_to_nat (nat_to_bvec len (bvec_to_nat v1 + bvec_to_nat v2))).
  apply bneg_btn.
  apply eqMod_trans with
    (2 * pow2 len - bvec_to_nat (nat_to_bvec len (bvec_to_nat v1 + bvec_to_nat v2))).
  replace (2 * pow2 len - bvec_to_nat (nat_to_bvec len (bvec_to_nat v1 + bvec_to_nat v2)))
    with (pow2 len + (pow2 len - bvec_to_nat (nat_to_bvec len (bvec_to_nat v1 + bvec_to_nat v2)))).
  apply eqMod_mod; myauto.
  add Hup (btn_upper (nat_to_bvec len (bvec_to_nat v1 + bvec_to_nat v2))); omega.
  apply eqMod_trans with
    (2 * pow2 len -
      (bvec_to_nat v1 + bvec_to_nat v2)).
  apply eqMod_minus; myauto.
  apply ntb_invert_mod.
  add Hup (btn_upper (nat_to_bvec len (bvec_to_nat v1 + bvec_to_nat v2))); omega.
  add Hup1 (btn_upper v1); add Hup2 (btn_upper v2); omega.
  apply eqMod_trans with
    ((pow2 len - bvec_to_nat v1) + (pow2 len - bvec_to_nat v2)).
  apply eqMod_eq.
  add Hup1 (btn_upper v1); add Hup2 (btn_upper v2); omega.
  apply eqMod_plus; apply eqMod_sym; apply bneg_btn.
Qed.

Lemma bsub_btn_const : forall len (v1 : Bvector len) n1 n2,
  bsub v1 (nat_to_bvec len (n1 + n2))
  = bsub (bsub v1 (nat_to_bvec len n1)) (nat_to_bvec len n2).
  intros.
  unfold bsub, barith.
  apply ntb_mod.
  apply eqMod_trans with
    (bvec_to_nat v1 + bvec_to_nat (bneg (nat_to_bvec len n1)) +
      bvec_to_nat (bneg (nat_to_bvec len n2))).
  repeat (rewrite <- plus_assoc).
  apply eqMod_plus; myauto.
  apply eqMod_trans with
    (bvec_to_nat (bneg (nat_to_bvec len
      (bvec_to_nat (nat_to_bvec len n1) + bvec_to_nat (nat_to_bvec len n2))))).
  rewrite (@ntb_mod len
    (bvec_to_nat (nat_to_bvec len n1) + bvec_to_nat (nat_to_bvec len n2))
    (n1 + n2)).
  myauto.
  apply eqMod_sym; apply ntb_add.
  apply badd_btn_neg.
  apply eqMod_plus.
  apply eqMod_sym; apply ntb_invert_mod.
  myauto.
Qed.

Lemma bsub_bdec2 : forall len (v : Bvector len),
  bvec_to_nat v > 1
  -> bsub v (nat_to_bvec len 2) = bdec2 v.
  intros.
  replace 2 with (1+1); trivial.
  rewrite bsub_btn_const.
  unfold bdec2.
  repeat (rewrite bsub_bdec); trivial; try omega.
  rewrite bdec_btn; omega.
Qed.

Lemma bsub_bdec3 : forall len (v : Bvector len),
  bvec_to_nat v > 2
  -> bsub v (nat_to_bvec len 3) = bdec3 v.
  intros.
  replace 3 with (2+1); trivial.
  replace (bdec3 v) with (bdec (bdec2 v)); trivial.
  rewrite bsub_btn_const.
  rewrite bsub_bdec.
  rewrite bsub_bdec2; trivial; try omega.
  rewrite bsub_bdec2; try omega.
  rewrite bdec2_btn; omega.
Qed.

Lemma bsub_bdec4 : forall len (v : Bvector len),
  bvec_to_nat v > 3
  -> bsub v (nat_to_bvec len 4) = bdec4 v.
  intros.
  replace 4 with (3+1); trivial.
  replace (bdec4 v) with (bdec (bdec3 v)); trivial.
  rewrite bsub_btn_const.
  rewrite bsub_bdec.
  rewrite bsub_bdec3; trivial; try omega.
  rewrite bsub_bdec3; try omega.
  rewrite bdec3_btn; omega.
Qed.

Lemma badd_comm : forall len (v1 v2 : Bvector len),
  badd v1 v2 = badd v2 v1.
  intros len v1 v2.
  unfold badd, barith.
  rewrite plus_comm.
  trivial.
Qed.

Lemma badd32_comm : forall v1 v2,
  badd32 v1 v2 = badd32 v2 v1.
  unfold badd32; apply badd_comm.
Qed.

Lemma badd32_swap_neg : forall v1 v2,
  badd32 v2 v1 = bsub32 v1 (bneg v2).
  intros.
  unfold bsub32, badd32, bsub, badd, barith.
  rewrite bneg_btn_double.
  rewrite plus_comm.
  trivial.
Qed.
