(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Theorems about converting between bitvectors and naturals *)

Require Import Arith Bool Bvector Euclid.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Defs.
Require Import Asm.Bitvector.Tactics.

Set Implicit Arguments.

(** * Converting bitvectors to naturals *)

(** ** Upper bounds *)

(** A bitvector can't translate to a natural number that requires more bits than the original did. *)
Lemma btn_upper : forall len (v : Bvector len),
  bvec_to_nat v < pow2 len.
  vec_ind.
Qed.

Lemma btn_upper' : forall len (v : Bvector len),
  bvec_to_nat v <= pow2 len - 1.
  intros.
  add Hupper (btn_upper v).
  omega.
Qed.

Lemma btn_upper'' : forall len (v : Bvector len),
  bvec_to_nat v <= pow2 len.
  intros.
  add Hupper (btn_upper v).
  omega.
Qed.

Hint Immediate btn_upper btn_upper' : asm.

Lemma btn_diff_pos : forall len (v : Bvector len),
  pow2 len - bvec_to_nat v > 0.
  intros.
  add Hupper (btn_upper v).
  omega.
Qed.

Hint Resolve btn_diff_pos : asm.

Theorem btn_inj : forall n (t1 t2 : Bvector n), bvec_to_nat t1 = bvec_to_nat t2 -> t1 = t2.
  vec_ind; simpl; intuition;
    try (myVOtac; trivial);
      myVSntac Ht2 t2; rewrite Ht2 in H1; simpl in H1;
	destruct (Vhead t2); try (assert False; [omega | tauto]);
	  rewrite (H0 (Vtail t2)); trivial; omega.
Qed.

Theorem btn_invert : forall len (v : Bvector len),
  nat_to_bvec len (bvec_to_nat v) = v.
  vec_ind; myrewrite; rewrite IHlen; trivial.
Qed.

Lemma S_minus_1_pow2 : forall len (v : Bvector len),
  S (pow2 len - bvec_to_nat v) - 1 = S (pow2 len - bvec_to_nat v - 1).
  intros.
  assert (pow2 len - bvec_to_nat v > 0);
    [add Hupper (btn_upper v); omega | idtac].
  destruct (pow2 len - bvec_to_nat v); omega.
Qed.

Hint Rewrite S_minus_1_pow2 : asm.


(** * Converting naturals to bitvectors *)

Theorem ntb_mod : forall len n1 n2,
  eqMod (pow2 len) n1 n2
  -> nat_to_bvec len n1 = nat_to_bvec len n2.
  induction len; srewrite.
  intros.
  rewrite (IHlen (div2 n1) (div2 n2)).
  replace (mod2_bool n2) with (mod2_bool n1); trivial.
  eapply eqMod_mod2_bool; eauto.
  eapply eqMod_div2; eauto.
Qed.

Lemma mod2_div2 : forall n,
  n = 2 * div2 n + mod2 n.
  intros.
  unfold div2, mod2.
  destruct (eucl_dev 2 two_gt_0 n).
  omega.
Qed.

Lemma mod2_inj : forall n1 n2,
  mod2_bool n1 = mod2_bool n2
  -> mod2 n1 = mod2 n2.
  intros n1 n2 Hbool.
  unfold mod2_bool, mod2 in *; idtac.
  destruct (eucl_dev 2 two_gt_0 n1).
  destruct (eucl_dev 2 two_gt_0 n2).
  destruct (eq_nat_dec r 0);
    destruct (eq_nat_dec r0 0);
      try omega; discriminate.
Qed.

Theorem ntb_mod' : forall len n1 n2,
  nat_to_bvec len n1 = nat_to_bvec len n2
  -> eqMod (pow2 len) n1 n2.
  induction len; srewrite; intros n1 n2 Heq.
  myauto.
  rewrite (mod2_div2 n1).
  rewrite (mod2_div2 n2).
  assert (Htail : Vtail (Bcons (mod2_bool n1) len (nat_to_bvec len (div2 n1))) =
    Vtail (Bcons (mod2_bool n2) len (nat_to_bvec len (div2 n2)))).
  congruence.
  injection Heq; intros.
  destruct H.
  injection Htail; intros.
  rewrite (mod2_inj n1 n2); myauto.
Qed.

Theorem ntb_invert : forall len n,
  n < pow2 len
  -> bvec_to_nat (nat_to_bvec len n) = n.
  induction len; srewrite; intuition; srewrite.
  unfold mod2_bool, mod2, div2.
  destruct (eucl_dev 2 two_gt_0 n).
  destruct (eq_nat_dec r 0); rewrite IHlen; omega.
Qed.

Theorem ntb_invert_mod : forall len n,
  eqMod (pow2 len) (bvec_to_nat (nat_to_bvec len n)) n.
  induction len; srewrite; intuition; srewrite.
  unfold mod2_bool, mod2, div2.
  destruct (eucl_dev 2 two_gt_0 n).
  destruct (eq_nat_dec r 0); subst;
    [replace (q * 2 + 0) with (2 * q); try omega
      | replace (q * 2 + r) with (S (2 * q)); try omega; apply eqMod_S];
    apply eqMod_mult_mod; auto.
Qed.

Theorem ntb_add : forall len n1 n2,
  eqMod (pow2 len) (n1 + n2)
  (bvec_to_nat (nat_to_bvec len n1) + bvec_to_nat (nat_to_bvec len n2)).
  intros.
  apply eqMod_plus; apply eqMod_sym; apply ntb_invert_mod.
Qed.

Lemma ntb_invert_eqMod : forall len n1 n2 n k,
  bvec_to_nat (nat_to_bvec len n) <= n1
  -> n <= k * pow2 len
  -> eqMod (pow2 len) (n1 + k * pow2 len - n) n2
  -> eqMod (pow2 len) (n1 - bvec_to_nat (nat_to_bvec len n)) n2.
  intros len n1 n2 n k Hle Hpow2 Hmod.
  apply eqMod_trans with (n1 + k * pow2 len - n); trivial;
    apply eqMod_minus; myauto;
      [apply eqMod_trans with (n1 + pow2 len);
	[apply eqMod_add_mod
	  | apply eqMod_plus;
	    [apply eqMod_refl
	      | apply eqMod_mod_times]]
	| apply ntb_invert_mod
	| omega].
Qed.
