(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Splitting bitvectors into component pieces *)

Require Import Arith Bool Bvector Euclid.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.BArith.
Require Import Asm.Bitvector.Bitwise.
Require Import Asm.Bitvector.Convert.
Require Import Asm.Bitvector.Defs.

(** * The basic split operation *)

(** Split an even-length bitvector into its two halves *)
Definition split_word n (v : Bvector (2 * n)) :=
  eucl_dev (pow2 n) (pow2_pos n) (bvec_to_nat v).

Definition lower_half n (v : Bvector (2 * n)) :=
  match split_word n v with
    divex q r _ _ => nat_to_bvec n r
  end.

Definition upper_half n (v : Bvector (2 * n)) :=
  match split_word n v with
    divex q r _ _ => nat_to_bvec n q
  end.

Theorem split_sound : forall n (v : Bvector (2 * n)),
  bvec_to_nat v = bvec_to_nat (lower_half n v) + pow2 n * bvec_to_nat (upper_half n v).
  intros.
  unfold lower_half, upper_half.
  destruct (split_word n v).
  repeat (rewrite ntb_invert); trivial;
    [rewrite plus_comm; rewrite (mult_comm (pow2 n) q); trivial
      | idtac].
  add Hupper (btn_upper v).
  rewrite pow2_times2 in Hupper.
  rewrite e in Hupper.
  assert (q * pow2 n < pow2 n * pow2 n); try omega.
  add Hpos (pow2_pos n).
  apply demult with (pow2 n); omega.
Qed.


(** * Specialized to 32 bits *)

Definition byte1 (v : int32) : int8 := lower_half 8 (lower_half 16 v).
Definition byte2 (v : int32) : int8 := upper_half 8 (lower_half 16 v).
Definition byte3 (v : int32) : int8 := lower_half 8 (upper_half 16 v).
Definition byte4 (v : int32) : int8 := upper_half 8 (upper_half 16 v).

Theorem split32_sound : forall v,
  bvec_to_nat v
  = bvec_to_nat (byte1 v)
  + pow2 8 * bvec_to_nat (byte2 v)
  + pow2 16 * bvec_to_nat (byte3 v)
  + pow2 24 * bvec_to_nat (byte4 v).
  intro.
  assert (pow2 16 * bvec_to_nat (byte3 v) + pow2 24 * bvec_to_nat (byte4 v)
    = pow2 16 * (bvec_to_nat (byte3 v) + pow2 8 * bvec_to_nat (byte4 v))).
  replace (pow2 24) with (pow2 16 * pow2 8).
  generalize dependent (pow2 16).
  generalize dependent (pow2 8).
  intros.
  rewrite mult_plus_distr_l.
  rewrite mult_assoc.
  trivial.
  replace 24 with (16 + 8); try omega.
  apply pow2_plus.
  unfold byte1, byte2, byte3, byte4.
  rewrite <- split_sound.
  rewrite <- plus_assoc.
  unfold byte3, byte4 in H.
  rewrite H.
  rewrite <- split_sound.
  apply (split_sound 16).
Qed.
