(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Arithmetic *)

Require Import Coq.Arith.Arith ZArith.

Require Import Asm.Util.Tactics.

(** * Exponentiation *)

Fixpoint pow2 (n : nat) : nat :=
  match n with
    | O => 1
    | S n => 2 * pow2 n
  end.

(** * Rewrite rules *)

Lemma rewrite_plus_0 : forall e, e + 0 = e.
  intros; omega.
Qed.

Lemma rewrite_minus_0 : forall e, e - 0 = e.
  intros; omega.
Qed.

Lemma Z_rewrite_plus_0 : forall e, (e + 0 = e)%Z.
  intros; omega.
Qed.

Lemma Z_rewrite_times_0 : forall e, (0 * e = 0)%Z.
  intros; omega.
Qed.

Lemma rewrite_times_2 : forall e, e + e = 2 * e.
  intros; omega.
Qed.

Lemma rewrite_undistrib : forall x y z, x * y + x * z = x * (y + z).
  intros.
  rewrite (mult_comm x y).
  rewrite (mult_comm x z).
  rewrite (mult_comm x (y + z)).
  symmetry.
  apply mult_plus_distr_r.
Qed.

Lemma rewrite_undistrib_minus : forall x y z,
  x * y - x * z = x * (y - z).
  intros.
  rewrite (mult_comm x y).
  rewrite (mult_comm x z).
  rewrite (mult_comm x (y - z)).
  symmetry.
  apply mult_minus_distr_r.
Qed.

Lemma rewrite_minus_plus : forall x y z,
  x - (y + z) = x - y - z.
  intros; omega.
Qed.

Hint Rewrite rewrite_plus_0 rewrite_minus_0 rewrite_times_2
  rewrite_undistrib rewrite_undistrib_minus : asm.
Hint Rewrite Z_rewrite_plus_0 Z_rewrite_times_0 : asm.

Lemma rewrite_undistrib_minus_const : forall x y,
  x * y - x = x * (y - 1).
  intros.
  replace (x * y - x) with (x * y - x * 1); try omega.
  myrewrite.
Qed.

Hint Rewrite rewrite_undistrib_minus_const : asm.

Lemma sub_one_S : forall n m,
  n - 1 - S m = n - 2 - m.
  intros; omega.
Qed.

Hint Rewrite sub_one_S : asm.

Lemma double_minus : forall n,
  2 * S n - 1 = 2 * (S n - 1) + 1.
  intros; omega.
Qed.

Lemma double_minus' : forall n,
  n > 0
  -> 2 * n - 1 = 2 * (n - 1) + 1.
  intros.
  destruct n; omega.
Qed.

Hint Rewrite double_minus : asm.
Hint Rewrite double_minus' using (myauto; fail) : asm.

Lemma sub_one : forall k n,
  S k * S n - 1 = S k * (S n - 1) + k.
  induction k; simpl; intuition.
  myrewrite.
  replace (k * S n) with (k + k * n); try omega.
  replace (S n) with (n + 1); [ring | omega].
Qed.

Lemma pow2_pos : forall n, pow2 n > 0.
  induction n; simpl; intuition.
Qed.

Lemma sub_one_pow2 : forall k n,
  S k * pow2 n - 1 = S k * (pow2 n - 1) + k.
  intros.
  add Hpos (pow2_pos n).
  destruct (pow2 n); try omega.
  apply sub_one.
Qed.

Lemma minus_S_1 : forall n m,
  n - S m - 1 = n - m - 2.
  intros; omega.
Qed.

Hint Rewrite sub_one sub_one_pow2 minus_S_1 : asm.

Lemma S_minus_1 : forall n, S (S n) - 1 = S (S n - 1).
  intros; omega.
Qed.

Lemma percolate_minus_1 : forall n m,
  n - 1 - m = n - m - 1.
  intros; omega.
Qed.

Hint Rewrite S_minus_1 percolate_minus_1 : asm.


(** * Injectivity of [Z_of_nat] *)

Ltac Z_of_nat_tac :=
  match goal with
    | [ H : ?X = ?Y |- _] =>
      let H' := fresh in
	(assert (H' : Z_of_nat X = Z_of_nat Y);
	  [ congruence | clear H; rename H' into H])
    | [ H : ?X > ?Y |- _] =>
      let H' := fresh in
	(assert (H' : (Z_of_nat X > Z_of_nat Y)%Z);
	  [ apply inj_gt; trivial | clear H; rename H' into H])
  end.

Lemma Z_of_nat_bit : forall n,
  2 > n
  -> Z_of_nat n = 0%Z \/ Z_of_nat n = 1%Z.
  intros.
  destruct n; [tauto | destruct n; [tauto | omega]].
Qed.

Lemma Psucc_inj : forall n m,
  Psucc n = Psucc m
  -> n = m.
  induction n; induction m; simpl; intuition; try discriminate.
  rewrite (IHn m); congruence.
  destruct n; simpl in H; discriminate.
  congruence.
  destruct m; simpl in H; discriminate.
Qed.

Lemma P_of_succ_nat_inj : forall n m,
  P_of_succ_nat n = P_of_succ_nat m
  -> n = m.
  induction n; simpl; intuition.
  destruct m; trivial.
  assert False; [idtac | tauto].
  induction m; simpl in *; intuition.
  discriminate.
  destruct (Psucc (P_of_succ_nat m)); discriminate.

  destruct m.
  simpl in H.
  assert False; [idtac | tauto].
  induction n; simpl in *; intuition.
  discriminate.
  destruct (Psucc (P_of_succ_nat n)); discriminate.
  simpl in H.
  rewrite (IHn m); trivial.
  apply Psucc_inj; trivial.
Qed.

Lemma Z_of_nat_inj : forall n m,
  Z_of_nat n = Z_of_nat m
  -> n = m.
  intros n m HZ.
  unfold Z_of_nat in HZ.
  destruct n; destruct m; trivial; try discriminate.
  rewrite (P_of_succ_nat_inj n m); trivial.
  congruence.
Qed.


(** * Other lemmas *)

Lemma S_times_two : forall n,
  S (S (2 * n)) = 2 * (S n).
  intros; omega.
Qed.

Hint Rewrite S_times_two : asm.

Lemma pow2_times2 : forall n,
  pow2 (2 * n) = pow2 n * pow2 n.
  induction n; trivial.
  replace (2 * S n) with (S (S (2 * n))); try omega.
  unfold pow2; fold pow2.
  ring.
  omega.
Qed.

Theorem pow2_plus : forall n m,
  pow2 n * pow2 m = pow2 (n + m).
  induction n; intuition.
  replace (S n + m) with (S (n + m)); try omega.
  unfold pow2; fold pow2.
  add Hih (IHn m).
  replace (2 * pow2 n * pow2 m) with (2 * (pow2 n * pow2 m)).
  omega.
  ring.
Qed.

Theorem demult : forall k n m,
  k >= 1
  -> n * k < m * k
  -> n < m.
  induction n; intros m Hk Hnm.
  destruct m; omega.
  destruct m.
  simpl in Hnm.
  generalize dependent (k + n * k); intros; omega.
  assert (n < m); try omega.
  apply IHn; trivial.
  replace (S n * k) with (n * k + k) in Hnm.
  replace (S m * k) with (m * k + k) in Hnm.
  omega.
  induction m; simpl; intuition.
  induction n; simpl; intuition.
Qed.

Theorem rewrite_S : forall n,
  S n = n + 1.
  intros; omega.
Qed.

Lemma plus_minus : forall n m k,
  k <= m
  -> n + (m - k) = n + m - k.
  intros; omega.
Qed.

Lemma minus_plus' : forall n m k,
  n - (m + k) = n - k - m.
  intros; omega.
Qed.
