(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Modular arithmetic *)

Require Import Arith ZArith Euclid.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Tactics.

Set Implicit Arguments.

(** * Basic definitions *)

(** Equality of naturals modulo a natural *)
Definition eqMod m n1 n2 :=
  exists k,
    (Z_of_nat n1 = Z_of_nat n2 + k * Z_of_nat m)%Z.

Lemma two_gt_0 : 2 > 0.
  omega.
Qed.

Definition div2 n :=
  let (q, r, pfGt, pfEq) := eucl_dev 2 two_gt_0 n in
    q.

Definition mod2 n :=
  let (q, r, pfGt, pfEq) := eucl_dev 2 two_gt_0 n in
    r.

Definition mod2_bool n :=
  if eq_nat_dec (mod2 n) 0 then
    false
  else
    true.

(** * [eqMod] is an equivalence relation. *)

Theorem eqMod_refl : forall m n,
  eqMod m n n.
  intros.
  exists 0%Z.
  omega.
Qed.

Theorem eqMod_sym : forall m n1 n2,
  eqMod m n1 n2
  -> eqMod m n2 n1.
  intros.
  destruct H as [k].
  exists (-k)%Z.
  rewrite <- Zopp_mult_distr_l.
  omega.
Qed.

Theorem eqMod_trans : forall m n1 n2 n3,
  eqMod m n1 n2
  -> eqMod m n2 n3
  -> eqMod m n1 n3.
  intros.
  destruct H as [k1].
  destruct H0 as [k2].
  exists (k1+k2)%Z.
  rewrite H.
  rewrite H0.
  ring.
Qed.

(** * [eqMod] is a congruence for the basic arithmetic operations. *)

Theorem eqMod_plus : forall m n1 n2 k1 k2,
  eqMod m n1 n2
  -> eqMod m k1 k2
  -> eqMod m (n1 + k1) (n2 + k2).
  intros.
  destruct H as [d1].
  destruct H0 as [d2].
  exists (d1 + d2)%Z.
  repeat (rewrite inj_plus).
  rewrite H.
  rewrite H0.
  ring.
Qed.

Theorem eqMod_minus : forall m n1 n2 k1 k2,
  eqMod m n1 n2
  -> eqMod m k1 k2
  -> k1 <= n1
  -> k2 <= n2
  -> eqMod m (n1 - k1) (n2 - k2).
  intros m n1 n2 k1 k2 Hmn Hmk Hlt1 Hlt2.
  destruct Hmn as [d1].
  destruct Hmk as [d2].
  exists (d1 - d2)%Z.
  repeat (rewrite inj_minus1; trivial).
  rewrite H.
  rewrite H0.
  ring.
Qed.

Theorem eqMod_S : forall m n1 n2,
  eqMod m n1 n2
  -> eqMod m (S n1) (S n2).
  intros.
  rewrite rewrite_S.
  rewrite (rewrite_S n2).
  apply eqMod_plus; trivial.
  apply eqMod_refl.
Qed.

Theorem eqMod_mult : forall m n1 n2 k1 k2,
  eqMod m n1 n2
  -> eqMod m k1 k2
  -> eqMod m (n1 * k1) (n2 * k2).
  intros.
  destruct H as [d1].
  destruct H0 as [d2].
  exists (d1 * Z_of_nat k2 + d2 * Z_of_nat n2 + d1 * d2 * Z_of_nat m)%Z.
  repeat (rewrite inj_mult).
  rewrite H.
  rewrite H0.
  ring.
Qed.

(** * Specific properties of modular equality *)

Theorem eqMod_mult_mod : forall m n1 n2 k,
  eqMod m n1 n2
  -> eqMod (k * m) (k * n1) (k * n2).
  intros.
  destruct H as [d].
  exists d.
  repeat (rewrite inj_mult).
  rewrite H.
  ring.
Qed.

Theorem eqMod_one : forall n1 n2,
  eqMod 1 n1 n2.
  intros.
  exists (Z_of_nat n1 - Z_of_nat n2)%Z.
  simpl.
  omega.
Qed.

Theorem eqMod_mod : forall m n1 n2,
  eqMod m n1 n2
  -> eqMod m n1 (m + n2).
  intros m n1 n2 Hmod.
  inversion Hmod as [k].
  exists (k-1)%Z.
  rewrite inj_plus.
  rewrite H.
  ring.
Qed.

Theorem eqMod_plus_minus_mod : forall m n1 n21 n22,
  n22 <= n21
  -> eqMod m n1 (n21 - n22)
  -> eqMod m n1 (n21 + m - n22).
  intros n n1 n21 n22 Hlt Hmod.
  replace (n21 + n - n22) with (n + (n21 - n22)); try omega.
  apply eqMod_mod; trivial.
Qed.

Theorem eqMod_mod' : forall m n1 n2,
  eqMod m n1 (m + n2)
  -> eqMod m n1 n2.
  intros m n1 n2 Hmod.
  inversion Hmod as [k].
  exists (k+1)%Z.
  rewrite inj_plus in H.
  rewrite H.
  ring.
Qed.

Theorem eqMod_plus_minus_mod' : forall m n1 n21 n22,
  n22 <= n21
  -> n22 <= m
  -> eqMod m n1 (n21 + (m - n22))
  -> eqMod m n1 (n21 - n22).
  intros m n1 n21 n22 Hlt Hlt2 Hmod.
  destruct Hmod as [k].
  rewrite inj_plus in H.
  rewrite inj_minus1 in H; trivial.
  exists (k+1)%Z.
  rewrite inj_minus1; trivial.
  rewrite H.
  ring.
Qed.

Theorem eqMod_mod_times : forall m n,
  eqMod m m (n * m).
  intros.
  exists (1 - Z_of_nat n)%Z.
  rewrite inj_mult.
  ring.
Qed.

Theorem eqMod_add_mod : forall m n,
  eqMod m n (n + m).
  intros.
  exists (-1)%Z.
  rewrite inj_plus.
  omega.
Qed.

Lemma eqMod_mod_plus : forall m n1 n2 k,
  eqMod m n1 n2
  -> eqMod m (k * m + n1) n2.
  intros m n1 n2 k Hmod.
  destruct Hmod as [x].
  exists (x + Z_of_nat k)%Z.
  rewrite inj_plus; rewrite inj_mult.
  rewrite H.
  ring.
Qed.


(** * Alternative [nat]-based conclusion from [eqMod] *)

Lemma eqMod_nat : forall m n1 n2,
  eqMod m n1 n2
  -> exists k, (n1 = n2 + k * m \/ n2 = n1 + k * m).
  intros m n1 n2 Hmod.
  destruct Hmod.
  assert (x < 0 \/ x >= 0)%Z; [omega | idtac].
  destruct H0.

  assert (exists n, -x = Z_of_nat n)%Z.
  apply Z_of_nat_complete; omega.
  destruct H1 as [n].
  exists n.
  right.
  assert (Z_of_nat n2 = Z_of_nat (n1 + n * m)).
  rewrite inj_plus; rewrite inj_mult.
  rewrite H.
  rewrite <- H1.
  ring.
  apply Z_of_nat_inj; trivial.

  assert (exists n, x = Z_of_nat n)%Z.
  apply Z_of_nat_complete; omega.
  destruct H1 as [n].
  exists n.
  left.
  assert (Z_of_nat n1 = Z_of_nat (n2 + n * m)).
  rewrite inj_plus; rewrite inj_mult.
  rewrite H.
  rewrite <- H1.
  ring.
  apply Z_of_nat_inj; trivial.
Qed.

Lemma eqMod_inj : forall m n1 n2,
  eqMod m n1 n2
  -> n1 < n2 + m
  -> n2 < n1 + m
  -> n1 = n2.
  intros m n1 n2 Hmod Hn1 Hn2.
  destruct (eqMod_nat Hmod) as [x].
  destruct x.
  omega.
  assert (1 * m <= S x * m).
  apply mult_le_compat; omega.
  omega.
Qed.

Lemma eqMod_nat' : forall m n1 n2 k,
  (n1 = n2 + k * m \/ n2 = n1 + k * m)
  -> eqMod m n1 n2.
  intros m n1 n2 k Hmod.
  destruct Hmod.
  exists (Z_of_nat k).
  auto with zarith.
  exists (-Z_of_nat k)%Z.
  replace (Z_of_nat n2 + - Z_of_nat k * Z_of_nat m)%Z
    with (Z_of_nat n2 - Z_of_nat k * Z_of_nat m)%Z; auto with zarith.
  ring.
Qed.

Lemma eqMod_mod2_bool : forall m n1 n2,
  eqMod (2 * m) n1 n2
  -> mod2_bool n1 = mod2_bool n2.
  intros m n1 n2 Hmod.
  unfold mod2_bool, mod2.
  destruct (eucl_dev 2 two_gt_0 n1).
  destruct (eucl_dev 2 two_gt_0 n2).
  replace r0 with r; trivial.
  destruct (eqMod_nat Hmod) as [k].
  rewrite mult_assoc in H.
  rewrite (mult_comm k 2) in H.
  rewrite <- mult_assoc in H.
  omega.
Qed.

Lemma eqMod_div2 : forall m n1 n2,
  eqMod (2 * m) n1 n2
  -> eqMod m (div2 n1) (div2 n2).
  intros m n1 n2 Hmod.
  unfold div2.
  destruct (eucl_dev 2 two_gt_0 n1).
  destruct (eucl_dev 2 two_gt_0 n2).
  destruct (eqMod_nat Hmod) as [k].
  rewrite mult_assoc in H.
  rewrite (mult_comm k 2) in H.
  rewrite <- mult_assoc in H.
  apply eqMod_nat' with k.
  omega.
Qed.

Theorem eqMod_contra : forall m n k,
  0 < k < m
  -> eqMod m n (n + k)
  -> False.
  intros.
  destruct (eqMod_nat H0); clear H0.
  intuition.
  generalize dependent (x * m); intros; omega.
  destruct x.
  omega.
  assert (S x * m >= m).
  induction x; simpl; intuition.
  generalize dependent (S x * m); intros; omega.
Qed.

Theorem eqMod_S_contra : forall m n,
  m >= 2
  -> eqMod m n (S n)
  -> False.
  intros.
  apply eqMod_contra with m n 1.
  omega.
  rewrite plus_comm.
  rewrite <- S_to_plus_one.
  trivial.
Qed.

Theorem eqMod_S_contra_pow2 : forall m n,
  eqMod (pow2 (S m)) n (S n)
  -> False.
  intros.
  apply eqMod_S_contra with (pow2 (S m)) n; trivial.
  simpl; add Hpos (pow2_pos m); omega.
Qed.


(** * Properties of [mod2] and [div2] *)

Lemma mod2_S2 : forall n,
  mod2 (S (2 * n)) = 1.
  intros.
  unfold mod2.
  destruct (eucl_dev 2 two_gt_0 (S (2 * n))).
  omega.
Qed.

Lemma mod2_bool_S2 : forall n,
  mod2_bool (S (2 * n)) = true.
  intros.
  unfold mod2_bool.
  rewrite mod2_S2.
  destruct (eq_nat_dec 1 0); trivial; discriminate.
Qed.

Lemma div2_S2 : forall n,
  div2 (S (2 * n)) = n.
  intros.
  unfold div2.
  destruct (eucl_dev 2 two_gt_0 (S (2 * n))).
  omega.
Qed.

Lemma mod2_2 : forall n,
  mod2 (2 * n) = 0.
  intros.
  unfold mod2.
  destruct (eucl_dev 2 two_gt_0 (2 * n)).
  omega.
Qed.

Lemma mod2_bool_2 : forall n,
  mod2_bool (2 * n) = false.
  intros.
  unfold mod2_bool.
  rewrite mod2_2.
  destruct (eq_nat_dec 0 0); tauto.
Qed.

Lemma div2_2 : forall n,
  div2 (2 * n) = n.
  intros.
  unfold div2.
  destruct (eucl_dev 2 two_gt_0 (2 * n)).
  omega.
Qed.

Hint Rewrite mod2_S2 mod2_bool_S2 div2_S2 : asm.
Hint Rewrite mod2_2 mod2_bool_2 div2_2 : asm.

(** * Hints *)

Hint Resolve eqMod_refl eqMod_sym eqMod_trans eqMod_one
  eqMod_plus eqMod_S eqMod_mult eqMod_mult_mod
  eqMod_mod eqMod_plus_minus_mod eqMod_mod_times : asm.