(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(* Parsing opcodes *)

open AsmUtil
open X86Types
open X86Util
open Printf

exception Parse of string

let file_base_offset = ref 0

let read_instr bits =
  let byte, bits = read_uint bits 8 in

  let arith_to_rm aop =
    let dst, bits, spare = read_rm32_with_spare bits in
    Arith (aop, dst, Reg (int_to_reg32 spare)), bits in

  let arith_from_rm aop =
    let src, bits, spare = read_rm32_with_spare bits in
    Arith (aop, Reg (int_to_reg32 spare), src), bits in

  let shift doRest =
    let dst, bits, spare = read_rm32_with_spare bits in
    let sop = int_to_shift_op spare in
    let offset, bits = doRest bits in
    Shift (sop, dst, offset), bits in

  (*printf "Byte = %x\n" byte;*)
  if byte = 0x66 then
    (* 16-bit mode instruction *)
    raise (Parse "No 16-bit instructions yet!")
  else
    if byte >= 0x40 && byte < 0x40 + nRegs32 then
      Inc (Reg (int_to_reg32 (byte - 0x40))), bits
    else if byte >= 0x50 && byte < 0x50 + nRegs32 then
      Push (Reg (int_to_reg32 (byte - 0x50))), bits
    else if byte >= 0x58 && byte < 0x58 + nRegs32 then
      Pop (Reg (int_to_reg32 (byte - 0x58))), bits
    else if byte >= 0x70 && byte < 0x70 + nCcs then
      let loc = get_byte bits + 1 in
      let imm, bits = read_int32 bits 8 in
      Jcc (int_to_cc (byte - 0x70), Int64.add (Int64.of_int (!file_base_offset + loc)) imm), bits
    else if byte >= 0xB8 && byte < 0xB8 + nRegs32 then
      let imm, bits = read_uint32 bits 32 in
      Mov (Reg (int_to_reg32 (byte - 0xB8)), Imm imm), bits
    else match byte with
      0x01 ->
	let dst, bits, spare = read_rm32_with_spare bits in
	Arith (Add, dst, Reg (int_to_reg32 spare)), bits
    | 0x03 ->
	let src, bits, spare = read_rm32_with_spare bits in
	Arith (Add, Reg (int_to_reg32 spare), src), bits
    | 0x29 -> arith_to_rm Sub
    | 0x2B -> arith_from_rm Sub
    | 0x68 ->
	let imm, bits = read_uint32 bits 32 in
	Push (Imm imm), bits
    | 0x6A ->
	let imm, bits = read_uint32 bits 8 in
	Push (Imm imm), bits
    | 0x80 ->
	let dst, bits, spare = read_rm8_with_spare bits in
	let aop = int_to_arith_op spare in
	let disp, bits = read_uint32 bits 8 in
	Arithb (aop, dst, Imm disp), bits
    | 0x81 ->
	let dst, bits, spare = read_rm32_with_spare bits in
	let aop = int_to_arith_op spare in
	let disp, bits = read_uint32 bits 32 in
	Arith (aop, dst, Imm disp), bits
    | 0x83 ->
	let dst, bits, spare = read_rm32_with_spare bits in
	let disp, bits = read_uint32 bits 8 in
	if spare = 7 then
	  Cmp (dst, Imm disp), bits
	else
	  let aop = int_to_arith_op spare in
	  Arith (aop, dst, Imm disp), bits
    | 0x88 ->
	let dst, bits, spare = read_rm8_with_spare bits in
	Movb (dst, Reg (int_to_reg8 spare)), bits
    | 0x89 ->
	let dst, bits, spare = read_rm32_with_spare bits in
	Mov (dst, Reg (int_to_reg32 spare)), bits
    | 0x8A ->
	let src, bits, spare = read_rm8_with_spare bits in
	Movb (Reg (int_to_reg8 spare), src), bits
    | 0x8B ->
	let src, bits, spare = read_rm32_with_spare bits in
	Mov (Reg (int_to_reg32 spare), src), bits
    | 0x8D ->
	let gop, bits, spare = read_rm32_with_spare bits in
	begin match gop with
	  Address addr -> Lea (int_to_reg32 spare, addr), bits
	| _ -> raise (Parse "Weird LEA operand")
	end
    | 0xA0 ->
	let imm, bits = read_uint32 bits 32 in
	Movb (Reg AL, Address {addrDisp = imm; addrBase = None; addrIndex = None}), bits
    | 0xA1 ->
	let imm, bits = read_uint32 bits 32 in
	Mov (Reg EAX, Address {addrDisp = imm; addrBase = None; addrIndex = None}), bits
    | 0xA2 ->
	let imm, bits = read_uint32 bits 32 in
	Movb (Address {addrDisp = imm; addrBase = None; addrIndex = None}, Reg AL), bits
    | 0xA3 ->
	let imm, bits = read_uint32 bits 32 in
	Mov (Address {addrDisp = imm; addrBase = None; addrIndex = None}, Reg EAX), bits
    | 0xC1 -> shift (fun bits ->
	let imm, bits = read_uint32 bits 8 in
	Imm imm, bits)
    | 0xC3 -> Ret, bits
    | 0xC6 ->
	let dst, bits, spare = read_rm8_with_spare bits in
	begin match spare with
	  0 ->
	    let imm, bits = read_uint32 bits 8 in
	    Movb (dst, Imm imm), bits
	| _ -> raise (Parse "Unknown 0xC6 instruction")
	end
    | 0xC7 ->
	let dst, bits, spare = read_rm32_with_spare bits in
	begin match spare with
	  0 ->
	    let imm, bits = read_uint32 bits 32 in
	    Mov (dst, Imm imm), bits
	| _ -> raise (Parse "Unknown 0xC7 instruction")
	end
    | 0xC9 -> Leave, bits
    | 0xE8 ->
	let loc = get_byte bits + 4 in
	let imm, bits = read_int32 bits 32 in
	Call (Imm (Int64.add (Int64.of_int (!file_base_offset + loc)) imm)), bits
    | 0xE9 ->
	let loc = get_byte bits + 4 in
	let imm, bits = read_int32 bits 32 in
	Jmp (Int64.add (Int64.of_int (!file_base_offset + loc)) imm), bits
    | 0xEB ->
	let loc = get_byte bits + 1 in
	let imm, bits = read_int32 bits 8 in
	Jmp (Int64.add (Int64.of_int (!file_base_offset + loc)) imm), bits
    | 0xFF ->
	let gop, bits, spare = read_rm32_with_spare bits in
	begin match spare with
	  0 -> Inc gop, bits
	| 6 -> Push gop, bits
	| _ -> raise (Parse "Unknown 0xFF instruction")
	end
    | _ -> raise (Parse (Printf.sprintf "Unknown instruction 0x%x" byte))
	  
