(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Proving abstract machine safety by bounded abstract interpretation *)

Require Import List TheoryList.

Require Import Asm.Util.Specif.

Require Import Asm.Bitvector.Bitvector.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Machine.Machine.

Set Implicit Arguments.

Ltac caseEq e := generalize (refl_equal e); pattern e at -1; case e.
Ltac destrEq hyp e := caseEq e; intro hyp.
Ltac destrEq2 hyp e := caseEq e; intro; intro hyp.
Ltac clear_all :=
  match goal with
    | [ H : _ |- _ ] => clear H; clear_all
  end.


Definition menter (mstate mastate mcontext : Set)
  (mconc : mastate -> mstate -> mcontext -> Prop)
  ctx abs ctx' abs' :=
  forall st, mconc abs st ctx -> mconc abs' st ctx'.

(** An abstraction of a concrete machine *)
Module Type ABSTRACTION.
  Declare Module Mac : MACHINE.

  Parameter mastate : Set.
  Parameter mastate_eq_dec : forall (a1 a2 : mastate), {a1 = a2} + {a1 <> a2}.
  (** Abstract verification-time state *)

  Parameter mcontext : Set.
  (** Abstract state used in proofs only *)

  Parameter mconc : mastate -> Mac.mstate -> mcontext -> Prop.
  (** Relate abstract and concrete states *)

  Parameter maget_instr : mastate -> option (Mac.minstr * int32).
  Axiom maget_instr_sound : forall abs st ctx v,
    mconc abs st ctx
    -> maget_instr abs = Some v
    -> v = Mac.mget_instr st.
  (** Decode the instruction to  be executed from a given state *)

  Parameter mimported : list (list mastate * mastate).
  (** States we assume to be safe, probably because we verify them separately *)

  Parameter minit : list (list mastate * mastate).
  (** A proposed fixed point *)

  Parameter mstep : forall hyps abs insPc, {result : option (list mastate)
    | match result with
	| None => True
	| Some abss =>
	  forall st ctx, mconc abs st ctx
	    -> ((exists st' : Mac.mstate, Mac.mexec insPc st st')
	      /\ forall st' (Hexec : Mac.mexec insPc st st'),
		exists abs', ((In abs' abss \/ In abs' hyps) /\ mconc abs' st' ctx)
		  \/ exists ctx', exists hyps',
		    In (hyps', abs') (minit ++ mimported)
		    /\ AllS (fun need => exists have,
		      (In have abss \/ In have hyps) /\ menter mconc ctx' need ctx have) hyps'
		    /\ mconc abs' st' ctx')
      end}.
  (** A transition function for testing the fixed point *)
End ABSTRACTION.

Lemma In_AllS : forall (T : Set) (f : T -> Prop) (ls : list T),
  (forall x, In x ls -> f x)
  -> AllS f ls.
  induction ls; intuition.
Qed.

Section RelativelySafe.
  Variable mstate : Set.
  Variable minstr : Set.
  Variable mexec : minstr * int32 -> mstate -> mstate -> Prop.
  Variable mget_instr : mstate -> minstr * int32.

  Definition cstep := cstep mexec mget_instr.
  Definition rsafe := safe cstep.

  Variable mastate : Set.
  Variable mcontext : Set.
  Variable mconc : mastate -> mstate -> mcontext -> Prop.
  
  Inductive relatively_safe (minit : list (list mastate * mastate)) : mstate -> Prop :=
    | rsafe_known : forall st,
      rsafe st
      -> relatively_safe minit st
    | rsafe_step : forall st st',
      cstep st st'
      -> (forall st'', cstep st st'' -> step_safe minit st'')
      -> relatively_safe minit st

  with step_safe (minit :  list (list mastate * mastate)) : mstate -> Prop :=
    | step_rel : forall st,
      relatively_safe minit st
      -> step_safe minit st
    | step_fp : forall st hyps abs ctx,
      In (hyps, abs) minit
      -> mconc abs st ctx
      -> (forall hyp, In hyp hyps -> forall st', mconc hyp st' ctx -> relatively_safe minit st')
      -> step_safe minit st.

  Hint Constructors relatively_safe step_safe.

  Scheme relatively_safe_mind := Minimality for relatively_safe Sort Prop
    with step_safe_mind := Minimality for step_safe Sort Prop.

  Section FixedPoint.
    Variable minit : list (list mastate * mastate).

    Hypothesis init_safe : forall hyps abs,
      In (hyps, abs) minit
      -> forall ctx, AllS (fun hyp => forall st, mconc hyp st ctx -> relatively_safe minit st) hyps
	-> forall st, mconc abs st ctx
	  -> relatively_safe minit st.

    Theorem relativity : forall st, relatively_safe minit st -> rsafe st.
      refine
	(cofix rel (st : mstate) (rs : relatively_safe minit st) : rsafe st := _).
      inversion rs; subst; trivial.

      unfold rsafe; apply safe_rule with st'; trivial.
      intros st'' Hstep.
      generalize (H0 st'' Hstep).
      intro Hsafe.
      inversion Hsafe; subst.
      apply rel; auto.

      apply rel.
      eapply init_safe; eauto.
      apply In_AllS; trivial.
    Qed.
  End FixedPoint.

  Theorem link : forall fp1 fp2 st,
    incl fp1 fp2
    -> relatively_safe fp1 st
    -> relatively_safe fp2 st.
    intros fp1 fp2 st Hincl Hsafe.
    generalize dependent st.
    apply (relatively_safe_mind (minit := fp1)
      (relatively_safe fp2)
      (step_safe fp2)); eauto.
  Qed.
End RelativelySafe.

Module Type CHECKER.
  Declare Module Abs : ABSTRACTION.

  Parameter verify : nat ->
    pred_option (forall fp,
      incl (Abs.minit ++ Abs.mimported) fp
      -> forall hyps abs,
	In (hyps, abs) Abs.minit
	-> forall ctx, AllS (fun hyp => forall st, Abs.mconc hyp st ctx
	  -> relatively_safe Abs.Mac.mexec Abs.Mac.mget_instr Abs.mconc
	  fp st) hyps
	-> forall st, Abs.mconc abs st ctx
	  -> relatively_safe Abs.Mac.mexec Abs.Mac.mget_instr Abs.mconc
	  fp st).
End CHECKER.

Module Checker (Abs : ABSTRACTION) : CHECKER with Module Abs := Abs.
  Import Abs.
  Import Mac.
  Module S := Safety(Mac).
  Import S.

  Definition relatively_safe := relatively_safe Abs.Mac.mexec Abs.Mac.mget_instr Abs.mconc.
  Hint Unfold relatively_safe.

  Lemma In_dec : forall (abs : mastate) ls,
    {In abs ls} + {~In abs ls}.
    induction ls; simpl; auto.
    destruct (mastate_eq_dec a abs); subst; auto.
    destruct IHls; intuition.
  Qed.

  Lemma AllS_dec : forall T f (ls : list T),
    {AllS (fun x => f x = true) ls} + {~AllS (fun x => f x = true) ls}.
    Hint Constructors AllS.
    induction ls; auto.
    destruct IHls; intuition.
    destrEq Hf (f a); auto.
    right.
    intuition.
    inversion H; congruence.
    right.
    intuition.
    inversion H; auto.
  Qed.

  Lemma AllS_In : forall T f (ls : list T),
    AllS f ls
    -> forall x, In x ls
      -> f x.
    induction ls; intuition.
    inversion H0.
    inversion H; subst.
    inversion H0.
    subst; trivial.
    auto.
  Qed.

  Lemma AllS_imp : forall (T : Set) (p1 p2 : T -> Prop),
    (forall x, p1 x -> p2 x)
    -> forall ls,
      AllS p1 ls -> AllS p2 ls.
    Hint Constructors AllS.
    induction ls; auto.
    intros Hall; inversion Hall; subst; auto.
  Qed.

  Definition popt_to_bool P (po : pred_option P) :=
    match po with
      | PNone => false
      | _ => true
    end.

  Definition asafe fp ctx abs := forall st, mconc abs st ctx -> relatively_safe fp st.

  Ltac ver_fail := intros; apply PNone.

  Definition bounded_verify : nat -> forall hyps abs,
    pred_option (forall fp,
      incl (Abs.minit ++ Abs.mimported) fp
      -> forall ctx, AllS (asafe fp ctx) hyps -> asafe fp ctx abs).
    intro n.
    induction n.

    ver_fail.

    intros hyps abs.
    caseEq (maget_instr abs); [idtac | ver_fail].
    intros p Hmaget.
    destruct (mstep hyps abs p).
    destruct x; [idtac | ver_fail].

    destruct (AllS_dec (fun abs => popt_to_bool (IHn hyps abs)) l);
      [idtac | ver_fail].

    apply PSome.
    intros fp Hfp ctx Hall st Hconc.
    destruct (y st ctx Hconc) as [Hprog Hpres].
    clear y.
    
    destruct Hprog as [st' Hst'].
    unfold relatively_safe; apply rsafe_step with st'.
    red; red.
    rewrite <- (maget_instr_sound Hconc Hmaget); trivial.

    intros st'' Hstep.
    red in Hstep.
    unfold Safety.cstep in Hstep.
    replace (mget_instr st) with p in Hstep.
    destruct (Hpres st'' Hstep) as [abs' Habs'].

    intuition.

    (* Step to abstract successor *)
    apply step_rel.
    caseEq (IHn hyps abs').
    intros a0 IHabs.
    unfold asafe in a0.
    apply a0 with ctx; trivial.
    intro Hih.
    assert (Htrue : popt_to_bool (IHn hyps abs') = true).
    apply (AllS_In a); trivial.
    destruct (IHn hyps abs'); discriminate.

    (* Step to hypothesis *)
    apply step_rel.
    assert (asafe fp ctx abs').
    eapply AllS_In; eauto.
    apply H0; trivial.

    (* Step to fixed point *)
    destruct H as [ctx' [hyps' H]].
    destruct H as [Hin [Hall' Hconc']].
    apply step_fp with hyps' abs' ctx'; auto.
    clear n abs st Hconc Hstep Hmaget Hst' Hpres Hin Hconc'.
    induction hyps'; auto.
    intros hyp Hhyp; inversion Hhyp.
    inversion Hall'; subst.
    intros hyp Hhyp st'0 Hconc'.
    destruct Hhyp; subst.
    destruct H1 as [have [Hin Henter]].
    intuition.

    assert (asafe fp ctx have).
    assert (popt_to_bool (IHn hyps have) = true).
    apply (AllS_In a); trivial.
    destruct (IHn hyps have).
    auto.
    discriminate.
    apply H1.
    apply Henter; trivial.

    assert (asafe fp ctx have).
    apply (AllS_In Hall); trivial.
    apply H1; auto.

    eauto.

    rewrite <- (maget_instr_sound Hconc Hmaget); trivial.
  Qed.

  Definition verify_list : nat -> forall abss,
    pred_option (forall fp,
      incl (Abs.minit ++ Abs.mimported) fp
      -> forall hyps abs, In (hyps, abs) abss
	-> forall ctx, AllS (asafe fp ctx) hyps
	-> asafe fp ctx abs).
    intro n.
    induction abss.

    apply PSome.
    intros fp Hfp hyps abs Hin.
    simpl in Hin.
    tauto.

    simpl.
    destruct IHabss; [idtac | ver_fail].
    intuition.
    destruct (@bounded_verify n a1 b); [idtac | ver_fail].
    apply PSome.
    intuition.
    injection H1; intros; subst; intuition.
    eauto.
  Qed.

  Definition verify : nat ->
    pred_option (forall fp,
      incl (Abs.minit ++ Abs.mimported) fp
      -> forall hyps abs,
	In (hyps, abs) minit
	-> forall ctx, AllS (asafe fp ctx) hyps
	  -> asafe fp ctx abs).
    intro n.
    exact (verify_list n minit).
  Qed.

  Module Abs := Abs.
End Checker.
