(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Basic definitions *)

Require Import Arith List Bvector TheoryList.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Map.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Specif.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Env.

Require Import ProofOS.Machine.Stateless.
Require Import ProofOS.Machine.Arches.

Require Import ProofOS.AbsInt.Bounded.

Import Datatypes.

(** * Memory regions *)

Parameter stackStart : int32.
(** Lowest address of stack region *)

Parameter stackSize : nat.
(** Guaranteed among of stack space left above [stackStart] *)

Axiom stack_fits : bvec_to_nat stackStart + stackSize + 9 < pow2 32 - 4.

Definition stackEnd := nat_to_bvec 32 (bvec_to_nat stackStart + stackSize + 9).
(** The very last memory cell considered to be in the stack *)

Definition initialESP := nat_to_bvec 32 (bvec_to_nat stackStart + stackSize + 4).
(** Starting stack pointer *)

Parameter codeLen : nat.
(** Amount of space reserved for code *)

Axiom code_high : bvec_to_nat initialPc > 3.
(** The code segment starts high enough to avoid underflow. *)

Axiom code_space : bvec_to_nat initialPc + (codeLen + longestInstr) + 3 < bvec_to_nat stackStart.
(** The stack starts sufficiently after the code. *)

Axiom initialESP_location : bvec_to_nat stackStart + stackSize < bvec_to_nat initialESP.
(** The stack pointer starts sufficiently after the start of the stack. *)

Axiom initialESP_upper : bvec_to_nat initialESP < pow2 32 - 4.
(** The stack pointer starts sufficiently low to avoid overflow. *)

(** * Program counters *)

Inductive apc : Set :=
  | apcConst : int32 -> apc
  | apcRet : apc.

(** ** Characterizing valid program counters *)

Parameter actualCodeLen : nat.
(** The real length of the code segment *)

Axiom actual_code_bound : actualCodeLen < codeLen.

Theorem actual_code_space : bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 3
  < bvec_to_nat stackStart.
  generalize code_space.
  generalize actual_code_bound.
  omega.
Qed.

Definition pcInRange (pc : int32) := bvec_to_nat initialPc <= bvec_to_nat pc
  <= bvec_to_nat initialPc + actualCodeLen.

Definition int32_In_dec : forall (n : int32) ls,
  {In n ls} + {~In n ls}.
  induction ls.
  right; tauto.
  destruct (int32_eq_dec n a); subst; intuition.
  right.
  intuition.
  simpl in H.
  intuition.
Qed.

Definition pcInRange_dec : forall pc, {pcInRange pc} + {~pcInRange pc}.
  intro.
  destruct (le_lt_dec (bvec_to_nat initialPc) (bvec_to_nat pc));
    destruct (le_lt_dec (bvec_to_nat pc) (bvec_to_nat initialPc + actualCodeLen));
      [left; red; intuition
	| right; unfold pcInRange; intuition
	| right; unfold pcInRange; intuition
	| right; unfold pcInRange; intuition].
Defined.


(** * Customization parameters *)

Set Implicit Arguments.

Inductive word_type (typ : Set) : Set :=
  | TyTop : word_type typ
  | TyConst : int32 -> word_type typ
  | TyCustom : typ -> word_type typ.

Inductive word_hasType (typ ctxt : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
  : ctxt -> int32 -> word_type typ -> Prop :=
  | WHT_Top : forall ctx e,
    word_hasType hasTyp ctx e (TyTop typ)
  | WHT_Const : forall ctx n,
    word_hasType hasTyp ctx n (TyConst typ n)
  | WHT_Custom : forall ctx e t,
    hasTyp ctx e t
    -> word_hasType hasTyp ctx e (TyCustom t).

Definition single_rule_instr ins :=
  match ins with
    | Jcc _ _ => False
    | _ => True
  end.


(** * Setting up some maps to use in the abstract state type *)

Module NatFinite.
  Definition t := nat.
  Definition to_nat (n : nat) := n.
  Theorem inj : forall t1 t2, t1 = t2 <-> to_nat t1 = to_nat t2.
    intuition.
  Qed.
End NatFinite.
Module NatOrd := FiniteOrderedSet(NatFinite).
Module NatMap := TotalMap(NatOrd).


(** * Back to customization parameters *)

Record status (typ : Set) : Set := {
  genopType : genop32 -> word_type typ; (** Type-check genops *)
  addrInStack : address -> bool;        (** Check if an address is in the stack *)
  programCounter : option int32;        (** Current program counter *)
  stackUsed : option nat;               (** Number of confirmed usable stack slots
					  above current ESP *)
  stackLeft : option nat;               (** Number of confirmed unused stack slots
					  below current ESP *)
  hasRetPtr : address -> bool           (** Does this address contain a valid
					  return pointer? *)
}.

Record statusOk (ctxt typ : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
  (stat : status typ) (ctx : ctxt) (st : state) : Prop := {
    genopTypeOk : forall g, word_hasType hasTyp ctx (eval_genop32 g st) (genopType stat g);
    addrInStackOk : forall a, addrInStack stat a = true
      -> bvec_to_nat stackStart
      <= bvec_to_nat (eval_addr a st)
      <= bvec_to_nat stackEnd;
    programCounterOk : match programCounter stat with
			 | None => True
			 | Some pc => pc = stPc st
		       end;
    stackUsedOk : match stackUsed stat with
		    | None => True
		    | Some used => bvec_to_nat (get_reg32 st ESP) + used <= bvec_to_nat initialESP
		  end;
    stackLeftOk : match stackLeft stat with
		    | None => True
		    | Some rem => bvec_to_nat stackStart + rem < bvec_to_nat (get_reg32 st ESP)
		  end;
    hasRetPtrOk : forall a, hasRetPtr stat a = true
      -> pcInRange (get_mem32 st (eval_addr a st));
    sokCodeMem : agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
  }.

Record presult (astat typ : Set) : Set := {
  prAbs : astat;
  prGenops : list (genop32 * word_type typ)
}.

Record presultOk (ctxt astat typ : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
  (pconc : ctxt -> astat -> state -> Prop)
  (res : presult astat typ) (ctx : ctxt) (st : state) : Prop := {
    pokAbs : pconc ctx (prAbs res) st;
    pokGenops : AllS (fun gp => word_hasType hasTyp ctx (eval_genop32 (fst gp) st) (snd gp)) (prGenops res)
  }.

Section EntryPoints.
  Variables typ ctxt astat : Set.
  Variable pconc : ctxt -> astat -> state -> Prop.
  Variable hasTyp : ctxt -> int32 -> typ -> Prop.

  (** Entry points to the code, treated like functions *)
  Record entryPoint : Set := {
    epPc : int32;                  (** Program counter *)
    epStackSize : nat;             (** Amount of stack space required *)
    epArgs : list (word_type typ); (** Arguments to the function *)
    epRet : word_type typ;         (** Return type *)
    epAbsIn : astat;               (** Custom abstract state to hold upon calling *)
    epAbsOut : astat               (** State to hold upon return *)
  }.
End EntryPoints.

Module Type FIRSTORDER_PARAM.
  Parameter typ : Set.
  (** Custom types *)
  Parameter typ_eq_dec : forall (t1 t2 : typ), {t1 = t2} + {t1 <> t2}.

  Parameter ctxt : Set.
  (** Typing contexts, A.K.A. extra information needed to check typing
    assertions.  These are only used in the soundness proof and are not present
    at runtime. *)

  Parameter astat : Set.
  (** Additional verification-time state *)

  Parameter pconc : ctxt -> astat -> state -> Prop.
  (** Does the context accurately summarize the concrete state, using the
    given abstract state? *)

  Parameter astat_eq_dec : forall (abs1 abs2 : astat), {abs1 = abs2} + {abs1 <> abs2}.

  Parameter subAstat : forall abs1 abs2, pred_option (forall st ctx,
    pconc ctx abs1 st -> pconc ctx abs2 st).
  (** Check compatibility of abstract states. *)

  Parameter isSafe : instr -> state -> Prop.
  (** A safety policy, suitable for use with Policy.Stateless *)

  Parameter hasTyp : ctxt -> int32 -> typ -> Prop.
  (** Does a value have a type? *)

  Parameter kernel_astat : option astat.
  (** Is this for checking the kernel, such that we need one special entry
    point that isn't a function?  If so, this should be the right abstract
    state to use in checking it. *)

  Parameter entryPoints : list (entryPoint typ astat).
  (** A set of abstract entry points that cover all concrete entry points. *)

  Parameter imported : list (entryPoint typ astat).
  (** Entry points that we assume safe, probably because we will verify them
    separately *)

  Parameter subTyp : word_type typ -> word_type typ -> bool.
  (** Decidable subtyping *)
  Parameter subTyp_refl : forall t, subTyp t t = true.
  Parameter subTyp_sound : forall t1 t2,
    subTyp t1 t2 = true
    -> forall ctx v, word_hasType hasTyp ctx v t1
      -> word_hasType hasTyp ctx v t2.

  Parameter checkSafe : forall abs stat ins,
    pred_option (forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> isSafe ins st).
  (** Check if known type information implies that the safety policy is
    satisfied for one instruction step. *)

  Parameter typeof_read : forall abs stat addr,
    sig_option (fun ty => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx (get_mem32 st (eval_addr addr st)) ty).
  (** Determine the type of something read from memory. *)

  Parameter check_write : forall abs stat addr vty,
    pred_option (forall st ctx v,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx v vty
      -> not_too_high (eval_addr addr st)
      /\ ~(bvec_to_nat initialPc - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)
      /\ ~(bvec_to_nat stackStart - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat stackEnd + 4)).
  (** Check that a memory write doesn't change code or stack memory. *)

  Parameter typeof_arith : forall aop t1 t2,
    sig_option (fun ty => forall ctx v1 v2,
      word_hasType hasTyp ctx v1 t1
      -> word_hasType hasTyp ctx v2 t2
      -> word_hasType hasTyp ctx (eval_arith aop v1 v2) ty).
  (** Determine the result type of an arithmetic operation. *)

  Parameter preservation : forall abs stat ins nextPc,
    single_rule_instr ins
    -> sig_option (fun res => forall st ctx st',
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> exec (ins, nextPc) st st'
      -> presultOk hasTyp pconc res ctx st').
  (** Update abstract state according to the effect of an instruction. *)

  Parameter preservation_jcc : forall abs stat co pc nextPc,
    sig_option (fun ress => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> (if check_cc st co then
	    presultOk hasTyp pconc (fst ress) ctx (set_pc st pc)
          else
	    presultOk hasTyp pconc (snd ress) ctx (set_pc st nextPc))).
  (** [preservation] version for Jcc, the only instruction governed by
    multiple rules. *)
End FIRSTORDER_PARAM.


(** * Lemmas about addresses *)

Lemma apart_dec : forall n m,
  {apart n m} + {~apart n m}.
  intros.
  destruct (le_lt_dec (bvec_to_nat m + 4) (bvec_to_nat n));
    [left; apply apart_lower; trivial
      | destruct (le_lt_dec (bvec_to_nat n + 4) (bvec_to_nat m));
	[left; apply apart_higher; omega
	  | right; intro; inversion H; omega]].
Qed.

Definition nat_apart n m :=
  m + 4 <= n \/ n + 4 <= m.

Lemma nat_apart_dec : forall n m,
  {nat_apart n m} + {~nat_apart n m}.
  intros.
  destruct (le_lt_dec (m + 4) n);
    [left; red; tauto
      | destruct (le_lt_dec (n + 4) m);
	[left; red; tauto
	  | right; unfold nat_apart; omega]].
Qed.

Lemma arith_op_eq_dec : forall (a1 a2 : arith_op), {a1 = a2} + {a1 <> a2}.
  intros.
  decide equality a1 a2.
Defined.

Lemma ntb_invert_eqMod : forall len n1 n2 n k,
  bvec_to_nat (nat_to_bvec len n) <= n1
  -> n <= k * pow2 len
  -> eqMod (pow2 len) (n1 + k * pow2 len - n) n2
  -> eqMod (pow2 len) (n1 - bvec_to_nat (nat_to_bvec len n)) n2.
  intros len n1 n2 n k Hle Hpow2 Hmod.
  apply eqMod_trans with (n1 + k * pow2 len - n); trivial;
    apply eqMod_minus; myauto;
      [apply eqMod_trans with (n1 + pow2 len);
	[apply eqMod_add_mod
	  | apply eqMod_plus;
	    [apply eqMod_refl
	      | apply eqMod_mod_times]]
	| apply ntb_invert_mod
	| omega].
Qed.


(** Metadata *)

Record cutpoint_info (ty astat : Set) : Set := {
  ciPc : apc;                 (** The program counter *)
  ciStackSize : nat;          (** Minimum stack size *)
  ciRegs : list (reg32 * ty); (** Types of registers *)
  ciSlots : list (nat * ty);  (** Types of stack slots *)
  ciCust : astat              (** Verifier-specific description *)
}.

Record function_info (ty : Set) : Set := {
  fiRet : ty;                 (** Return type *)
  fiArgs : list ty            (** How many arguments? *)
}.


(** * Generic types *)

Module GenericTypes.
  Section GenericTypes.
    Variables typ astat ctxt : Set.  
    Variable hasTyp : ctxt -> int32 -> typ -> Prop.
    Variable pconc : ctxt -> astat -> state -> Prop.

    (** ** Types *)

    Inductive ty : Set :=
      | Top : ty
      | Const : int32 -> ty
      | Stack : nat -> ty
      | RetPtr : ty
      | OldEbp : ty
      | Custom : typ -> ty.

    Definition aregs := Reg32Map.map ty.
    Definition astack := NatMap.map ty.

    (** ** Abstract states *)

    Record astate : Set :=
      {aPc : apc;             (** Current program counter *)
	aRegs : aregs;        (** Types of registers *)
	aStackSize : nat;     (** Lower bound on stack space left below initial
			          ESP *)
	aStackSlots : astack; (** Types of stack slots *)
	aCustom : astat       (** Custom verification-time state *)
      }.

    (** ** Typing judgment *)

    (** *** Typing contexts *)
    Record context : Set := {
      sptr : int32;       (** Function initial ESP *)
      oldebp : int32;     (** EBP value expected by the caller on return *)
      memIn : mem;        (** Memory at the start of this function call *)
      ret : option int32; (** Return pointer *)
      custom : ctxt       (** Custom context information, from the functor parameter *)
    }.

    Definition stack_slot ctx n :=
      nat_to_bvec 32 (bvec_to_nat (sptr ctx) - n).

    Inductive hasType : context -> int32 -> ty -> Prop :=
      | HasType_Top : forall ctx e,
	hasType ctx e Top
      | HasType_Const : forall ctx i,
	hasType ctx i (Const i)
      | HasType_Stack : forall ctx n,
	n <= bvec_to_nat (sptr ctx)
	-> hasType ctx (stack_slot ctx n) (Stack n)
      | HasType_RetPtr : forall ctx pc,
	ret ctx = Some pc
	-> pcInRange pc
	-> hasType ctx pc RetPtr
      | HasType_OldEbp : forall ctx,
	hasType ctx (oldebp ctx) OldEbp
      | HasType_Custom : forall ctx v t,
	hasTyp (custom ctx) v t
	-> hasType ctx v (Custom t).


    (** ** The soundness relation *)

    Definition conc_pc abs st ctx :=
      match aPc abs with
	| apcConst pc => stPc st = pc
	| apcRet => Some (stPc st) = ret ctx
      end.

    Definition aget_reg abs r := Reg32Map.sel (aRegs abs) r.

    Definition conc_regs abs st ctx :=
      forall r, hasType ctx (get_reg32 st r) (aget_reg abs r).

    Definition aget_stack abs n := NatMap.sel (aStackSlots abs) n.

    Definition conc_slots abs st ctx :=
      forall n, 4 <= n < aStackSize abs
	-> hasType ctx (get_mem32 st (stack_slot ctx n)) (aget_stack abs n).

    Definition agree_range' (mem1 mem2 : mem) (addr1 addr2 : int32) :=
      forall addr', bvec_to_nat addr1 <= bvec_to_nat addr' < bvec_to_nat addr2
	-> Int32Map.sel mem1 addr' = Int32Map.sel mem2 addr'.

    Record conc' (abs : astate) (st : state) (ctx : context) : Prop :=
      {concPc : conc_pc abs st ctx;
	concPcRange : pcInRange (stPc st);
	concCodeMem : agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr);
	concRegs : conc_regs abs st ctx;
	concStackSize : bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx);
	concStackSlots : conc_slots abs st ctx;
	concSptr : bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP;
	concMemIn : agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd}.
    
    Record conc (abs : astate) (st : state) (ctx : context) : Prop := {
      concMain : conc' abs st ctx;
      concCustom : pconc (custom ctx) (aCustom abs) st
    }.

    Hint Constructors hasType.

    Definition word_type_to_ty : forall t,
      {t' : ty | forall ctx v,
	word_hasType hasTyp (custom ctx) v t
	-> hasType ctx v t'}.
      destruct t.
      exists Top; myauto.
      exists (Const i); intros; inversion H; subst; myauto.
      exists (Custom t); intros; inversion H; subst; myauto.
    Defined.

    Fixpoint initialSlots (args : list (word_type typ)) (pos : nat) {struct args} : list (nat * ty) :=
      match args with
	| nil => nil
	| t :: args => (pos, proj1_sig (word_type_to_ty t)) :: initialSlots args (pos + 4)
      end.

    Definition kernel_cp abs :=
      (@nil (cutpoint_info ty astat),
	@Build_cutpoint_info ty astat (apcConst initialPc) stackSize nil nil abs).

    Definition initial_cp (ep : entryPoint typ astat) :=
      (Build_cutpoint_info
	apcRet 0
	((EAX, proj1_sig (word_type_to_ty (epRet ep))) :: (EBP, OldEbp) :: (ESP, Stack (4 * length (epArgs ep))) :: nil)
	nil
	(epAbsOut ep) :: nil,
	Build_cutpoint_info
	(apcConst (epPc ep))
	(epStackSize ep + (4 + 4 * length (epArgs ep)))
	((EBP, OldEbp) :: (ESP, Stack (4 + 4 * length (epArgs ep))) :: nil)
	((4 + 4 * length (epArgs ep), RetPtr) :: initialSlots (epArgs ep) 4)
	(epAbsIn ep)).

    Fixpoint make_init_regs (rs : list (reg32 * ty)) : aregs :=
      match rs with
	| nil => Reg32Map.empty Top
	| cons (r, t) rs => Reg32Map.upd (make_init_regs rs) r t
      end.

    Fixpoint make_init_slots (sls : list (nat * ty)) : astack :=
      match sls with
	| nil => NatMap.empty Top
	| cons (sl, t) sls => NatMap.upd (make_init_slots sls) sl t
      end.

    Definition make_init_state (cpt : cutpoint_info ty astat) :=
      Build_astate (ciPc cpt)
      (make_init_regs (ciRegs cpt))
      (ciStackSize cpt)
      (make_init_slots (ciSlots cpt))
      (ciCust cpt).

    Definition make_init hypsAbs :=
      (map make_init_state (fst hypsAbs), make_init_state (snd hypsAbs)).
  End GenericTypes.
End GenericTypes.

Definition cstep := cstep exec get_instr.

(** The grand functor *)

Module Make (Param : FIRSTORDER_PARAM).
  Import Param.
  Import GenericTypes.

  Hint Constructors hasType conc' conc : asm.
  
  (** * Specialize the generic types for this parameter *)

  Definition ty := ty typ.
  Hint Unfold ty : asm.

  Definition aregs := aregs typ.
  Hint Unfold aregs : asm.

  Definition astack := astack typ.
  Hint Unfold astack : asm.

  Definition astate := astate typ astat.
  Hint Unfold astate : asm.

  Definition aset_pc (abs : astate) pc :=
    Build_astate (apcConst pc) (aRegs abs) (aStackSize abs)
    (aStackSlots abs) (aCustom abs).

  Definition aget_reg := @aget_reg typ astat.
  Hint Unfold aget_reg : asm.

  Definition aset_reg (abs : astate) r ty :=
    Build_astate (aPc abs) (Reg32Map.upd (aRegs abs) r ty)
    (aStackSize abs) (aStackSlots abs) (aCustom abs).

  Definition aget_stack := @aget_stack typ astat.
  Hint Unfold aget_stack : asm.

  Definition aset_stack (abs : astate) n ty :=
    Build_astate (aPc abs) (aRegs abs) (aStackSize abs)
    (NatMap.upd
      (NatMap.upd
	(NatMap.upd
	  (NatMap.upd
	    (NatMap.upd
	      (NatMap.upd
		(NatMap.upd (aStackSlots abs) (n+3) (Top typ))
		(n+2) (Top typ))
	      (n+1) (Top typ))
	    (n-3) (Top typ))
	  (n-2) (Top typ))
	(n-1) (Top typ))
      n ty) (aCustom abs).
  (** All this extra cruft is here to make sure we invalidate any overlapping
    stack slots. *)

  Definition aset_custom (abs : astate) (cust : astat) :=
    Build_astate (aPc abs) (aRegs abs) (aStackSize abs)
    (aStackSlots abs) cust.

  Definition context := context ctxt.
  Hint Unfold context : asm.

  Definition cset_sptr (ctx : context) sp :=
    Build_context sp (oldebp ctx) (memIn ctx) (ret ctx) (custom ctx).
  Definition cset_oldebp (ctx : context) sp :=
    Build_context (sptr ctx) sp (memIn ctx) (ret ctx) (custom ctx).
  Definition cset_memIn (ctx : context) mi :=
    Build_context (sptr ctx) (oldebp ctx) mi (ret ctx) (custom ctx).
  Definition cset_ret (ctx : context) rp :=
    Build_context (sptr ctx) (oldebp ctx) (memIn ctx) rp (custom ctx).
  Definition cset_custom (ctx : context) (cust : ctxt) :=
    Build_context (sptr ctx) (oldebp ctx) (memIn ctx) (ret ctx) cust.

  Definition stack_slot := @stack_slot ctxt.
  Hint Unfold stack_slot : asm.

  Definition hasType := @hasType typ ctxt hasTyp.
  Hint Unfold hasType : asm.

  Definition conc_pc := @conc_pc typ astat ctxt.
  Hint Unfold conc_pc : asm.

  Definition conc_regs := @conc_regs typ astat ctxt hasTyp.
  Hint Unfold conc_regs : asm.

  Definition conc_slots := @conc_slots typ astat ctxt hasTyp.
  Hint Unfold conc_slots : asm.

  Definition conc' := @conc' typ astat ctxt hasTyp.
  Hint Unfold conc' : asm.
  
  Definition conc := @conc typ astat ctxt hasTyp pconc.
  Hint Unfold conc : asm.


  (** * Type operations *)

  Definition ty_eq_dec : forall (t1 t2 : ty), {t1 = t2} + {t1 <> t2}.
    intros t1 t2.
    decide equality t1 t2.
    unfold int32 in *; apply bvec_eq_dec.
    apply eq_nat_dec.
    apply typ_eq_dec.
  Qed.

  Definition subType t1 t2 :=
    match t2 with
      | Top => true
      | Custom t2 =>
	match t1 with
	  | Custom t1 => subTyp (TyCustom t1) (TyCustom t2)
	  | Top => subTyp (TyTop typ) (TyCustom t2)
	  | Const n => subTyp (TyConst typ n) (TyCustom t2)
	  | _ => false
	end
      | Const n =>
	match t1 with
	  | Custom t1 => subTyp (TyCustom t1) (TyConst typ n)
	  | _ => sum_to_bool (ty_eq_dec t1 t2)
	end
      | _ => sum_to_bool (ty_eq_dec t1 t2)
    end.

  Definition subReg abs1 abs2 r :=
    subType (aget_reg abs1 r) (aget_reg abs2 r).

  Theorem subType_refl : forall t, subType t t = true.
    intro.
    unfold subType.
    destruct (ty_eq_dec t t); simpl; destruct t; try tauto.
    apply subTyp_refl.
  Qed.
  
  Theorem subType_sound : forall t1 t2,
    subType t1 t2 = true
    -> forall ctx v, hasType ctx v t1
      -> hasType ctx v t2.
    intros t1 t2 Hsub ctx v Ht1.
    destruct t2; try myeauto;
      try match goal with
	    | [ |- hasType ctx v ?T ] =>
	      simpl in Hsub; destruct (ty_eq_dec t1 T) as [e | e];
		[rewrite e in Ht1; trivial | discriminate]
	  end;
      simpl in Hsub; destruct t1; try discriminate;
	try match goal with
	      [Hsub : sum_to_bool (ty_eq_dec ?T1 ?T2) = true |- _] =>
	      destruct (ty_eq_dec T1 T2); try discriminate
	    end.
    inversion Ht1; subst.
    injection e; intros; subst; trivial.

    assert (Hword : word_hasType hasTyp (custom ctx) v (TyConst typ i)).
    eapply subTyp_sound; eauto.
    apply WHT_Custom.
    inversion Ht1; subst; trivial.
    inversion Hword; subst; trivial.
    myauto.

    assert (Hword : word_hasType hasTyp (custom ctx) v (TyCustom t)).
    eapply subTyp_sound; eauto.
    apply WHT_Top.
    red; apply HasType_Custom.
    inversion Hword; trivial.

    assert (Hword : word_hasType hasTyp (custom ctx) v (TyCustom t)).
    eapply subTyp_sound; eauto.
    inversion Ht1; subst.
    apply WHT_Const.
    red; apply HasType_Custom.
    inversion Hword; trivial.

    red; apply HasType_Custom.
    assert (Hword : word_hasType hasTyp (custom ctx) v (TyCustom t)).
    eapply subTyp_sound; eauto.
    apply WHT_Custom.
    inversion Ht1; trivial.
    inversion Hword; trivial.
  Qed.


  (** * Abstract state operations *)

  Ltac bool_elim' :=
    match goal with
      | [ H : ?X && ?Y = true |- _ ] =>
	let and1 := fresh "and1" in
	  let and2 := fresh "and2" in
	    destruct (andb_prop X Y H) as [and1 and2];
	      clear H
    end.

  Ltac bool_elim := repeat bool_elim'.

  Definition apc_eq_dec : forall (pc1 pc2 : apc), {pc1 = pc2} + {pc1 <> pc2}.
    intros.
    decide equality pc1 pc2.
    apply (@bvec_eq_dec 32).
  Qed.

  Lemma subAbs_pc : forall st abs1 abs2 ctx,
    conc_pc abs1 st ctx
    -> sum_to_bool (apc_eq_dec (aPc abs1) (aPc abs2)) = true
    -> conc_pc abs2 st ctx.
    intros st abs1 abs2 ctx Hpc Heq.
    destruct (apc_eq_dec (aPc abs1) (aPc abs2)).
    red in Hpc; red in Hpc.
    destruct (aPc abs1); subst.
    red; red.
    rewrite <- e; trivial.
    red; red.
    rewrite <- e.
    trivial.
    discriminate.
  Qed.
  
  Lemma hasType_sel : forall ctx v (abs : astate) r,
    GenericTypes.hasType hasTyp ctx v (GenericTypes.aget_reg abs r)
    -> hasType ctx v (Reg32Map.sel (aRegs abs) r).
    intuition.
  Qed.

  Hint Resolve hasType_sel : asm.

  Lemma subAbs_regs : forall ctx st abs1 abs2,
    conc_regs abs1 st ctx
    -> proj1_sig (Reg32Map.subMap subType
      subType_refl (aRegs abs1) (aRegs abs2)) = true
    -> conc_regs abs2 st ctx.
    intros ctx st abs1 abs2 Hconc Hsub; intro.
    destruct (Reg32Map.subMap subType subType_refl (aRegs abs1) (aRegs abs2)).
    destruct x; try discriminate.
    eapply subType_sound; myeauto.
    unfold GenericTypes.aget_reg; myeauto.
  Qed.

  Lemma subAbs_sptr : forall (ctx : context) (abs1 abs2 : astate),
    bvec_to_nat stackStart + aStackSize abs1 < bvec_to_nat (sptr ctx)
    -> sum_to_bool (le_lt_dec (aStackSize abs2) (aStackSize abs1)) = true
    -> bvec_to_nat stackStart + aStackSize abs2 < bvec_to_nat (sptr ctx).
    intros ctx abs1 abs2 Hstart Hle.
    destruct (le_lt_dec (aStackSize abs2) (aStackSize abs1));
      simpl in Hle; try discriminate; omega.
  Qed.

  Lemma conc_slots_generic : forall abs st ctx,
    conc_slots abs st ctx
    -> GenericTypes.conc_slots hasTyp abs st ctx.
    intuition.
  Qed.

  Hint Resolve conc_slots_generic : asm.

  Hint Unfold GenericTypes.aget_stack : asm.

  Lemma subAbs_slots : forall st ctx abs1 abs2,
    conc_slots abs1 st ctx
    -> proj1_sig
    (NatMap.subMap subType subType_refl (aStackSlots abs1)
      (aStackSlots abs2)) = true
    -> sum_to_bool (le_lt_dec (aStackSize abs2) (aStackSize abs1)) = true
    -> conc_slots abs2 st ctx.
    intros st ctx abs1 abs2 Hconc Hsub Hle n HstackSize.
    destruct (NatMap.subMap subType subType_refl (aStackSlots abs1) (aStackSlots abs2)).
    destruct x; try discriminate.
    apply subType_sound with (aget_stack abs1 n).
    unfold aget_stack; unfold GenericTypes.aget_stack; auto.
    apply Hconc; destruct (le_lt_dec (aStackSize abs2) (aStackSize abs1));
      [omega | discriminate].
  Qed.

  Definition pred_option_true (P : Prop) (po : pred_option P) :=
    match po with
      | PNone => false
      | PSome _ => true
    end.

  Lemma subAbs_pconc : forall ctx (abs1 abs2 : astate) st,
    pconc (custom ctx) (aCustom abs1) st
    -> pred_option_true (subAstat (aCustom abs1) (aCustom abs2)) = true
    -> pconc (custom ctx) (aCustom abs2) st.
    intros ctx abs1 abs2 st Hconc Htrue.
    destruct (subAstat (aCustom abs1) (aCustom abs2)).
    auto.
    discriminate.
  Qed.

  Hint Resolve subAbs_pc subAbs_regs subAbs_sptr subAbs_slots subAbs_pconc : asm.

  Lemma conc_pc_generic : forall abs st ctx,
    conc_pc abs st ctx
    -> GenericTypes.conc_pc abs st ctx.
    intuition.
  Qed.
  
  Hint Resolve conc_pc_generic : asm.

  Lemma conc_regs_generic : forall abs st ctx,
    conc_regs abs st ctx
    -> GenericTypes.conc_regs hasTyp abs st ctx.
    intuition.
  Qed.

  Hint Resolve conc_regs_generic : asm.

  Definition subAbs abs1 abs2 : {b : bool | b = true -> forall st ctx,
    conc abs1 st ctx -> conc abs2 st ctx}.
    intros abs1 abs2.
    exists (sum_to_bool (apc_eq_dec (aPc abs1) (aPc abs2))
      && sum_to_bool (le_lt_dec (aStackSize abs2) (aStackSize abs1))
      && proj1_sig (Reg32Map.subMap subType subType_refl (aRegs abs1) (aRegs abs2))
      && proj1_sig (NatMap.subMap subType subType_refl (aStackSlots abs1) (aStackSlots abs2))
      && pred_option_true (subAstat (aCustom abs1) (aCustom abs2))).
    intros Hchecks st ctx Hconc.
    inversion Hconc.
    inversion concMain0.
    bool_elim.
    split; try split; trivial; myeauto.
  Qed.

  Definition astate_eq_dec : forall (a1 a2 : astate), {a1 = a2} + {a1 <> a2}.
    intros.
    decide equality a1 a2.
    apply astat_eq_dec.
    unfold GenericTypes.astack in * ; idtac.
    apply NatMap.eq_dec.
    apply ty_eq_dec.
    apply eq_nat_dec.
    unfold GenericTypes.aregs in * ; idtac.
    apply Reg32Map.eq_dec.
    apply ty_eq_dec.
    apply apc_eq_dec.
  Qed.

  Definition astates_eq_dec : forall (a1 a2 : list astate), {a1 = a2} + {a1 <> a2}.
    induction a1; destruct a2; intuition;
      try (right; intros; discriminate).
    destruct (astate_eq_dec a a0); subst.
    destruct (IHa1 a2); subst.
    auto.
    right; congruence.
    right; congruence.
  Qed.

  Fixpoint goodHyps (hyps hyps' : list astate) {struct hyps'} : Prop :=
    match hyps' with
      | nil =>
	match hyps with
	  | nil => True
	  | _ => False
	end
      | cons hyp' hyps' =>
	match hyps with
	  | nil => False
	  | cons hyp hyps =>
	    proj1_sig (subAbs hyp' hyp) = true /\ goodHyps hyps hyps'
	end
    end.

  Definition goodHyps_dec : forall hyps' hyps,
    pred_option (goodHyps hyps hyps').
    induction hyps'; simpl; intuition.
    destruct hyps.
    apply PSome; trivial.
    apply PNone.
    destruct hyps.
    apply PNone.
    destruct (IHhyps' hyps).
    destruct (subAbs a a0).
    destruct x; [intuition | apply PNone].
    simpl.
    apply PSome.
    tauto.
    apply PNone.
  Qed.

  Definition abs_covered : forall abss hyps abs,
    {b : bool | b = true ->
      exists hyps', exists abs',
	In (hyps', abs') abss
	/\ proj1_sig (subAbs abs abs') = true
	/\ goodHyps hyps hyps'}.
    induction abss; intros hyps abs.

    exists false; intuition; discriminate.

    destruct (goodHyps_dec (fst a) hyps).
    caseEq (subAbs abs (snd a)); intros x c Ha; destruct x.

    exists true; intuition; exists (fst a); exists (snd a); intuition.
    destruct a; simpl; intuition.
    rewrite Ha; trivial.

    destruct (IHabss hyps abs) as [x]; destruct x.
    exists true; intuition.
    destruct H0 as [hyps' H0].
    destruct H0 as [abs'].
    exists hyps'.
    exists abs'.
    intuition.
    exists false; intuition; discriminate.

    destruct (IHabss hyps abs) as [x]; destruct x.
    exists true; intuition.
    destruct H0 as [hyps' H0].
    destruct H0 as [abs'].
    exists hyps'.
    exists abs'.
    intuition.
    exists false; intuition; discriminate.
  Qed.


  (** * Generic lemmas about the soundness relation *)

  (** ** Tactics *)

  Ltac unfolder := unfold get_reg32, aget_reg, set_reg32, aset_reg, stack_slot, aset_stack, aget_stack,
    GenericTypes.aget_reg, GenericTypes.stack_slot, GenericTypes.aget_stack.
  Ltac sel_eq_replace e1 e2 lemma :=
    replace e1 with e2;
      [idtac | symmetry; unfolder; simpl; apply lemma; trivial].
  Ltac sel_neq_replace e1 e2 lemma :=
    replace e1 with e2;
      [idtac | symmetry; unfolder; simpl; apply lemma; trivial].

  Ltac ntb_conv :=
    match goal with
      | [ |- nat_to_bvec _ ?N = nat_to_bvec _ ?M ] =>
	assert (N = M); [idtac | congruence]
    end.

  Hint Constructors exec write_genop32 : asm.

  (** * Lemmas to be used in the following sections *)

  Lemma get_set_stack_apart : forall abs s1 s2 t,
    nat_apart s1 s2
    -> aget_stack (aset_stack abs s1 t) s2 = aget_stack abs s2.
    intros abs s1 s2 t Hapart.
    unfolder; simpl;
      repeat (rewrite NatMap.sel_neq); trivial; inversion Hapart; omega.
  Qed.

  Lemma stack_slot_below : forall ctx slot n,
    belowMax (sptr ctx) n
    -> bvec_to_nat (stack_slot ctx slot) <= bvec_to_nat (sptr ctx).
    intros.
    unfold stack_slot; unfold GenericTypes.stack_slot.
    rewrite ntb_invert;
      [omega
	| inversion H; omega].
  Qed.

  Lemma stack_slot_belowMax : forall ctx slot n,
    belowMax (sptr ctx) n
    -> belowMax (stack_slot ctx slot) n.
    intros ctx slot n Hbelow.
    inversion Hbelow.
    add Hbelow2 (stack_slot_below ctx slot Hbelow).
    split; omega.
  Qed.

  Lemma stack_slot_belowMax' : forall (ctx : context),
    bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
    -> belowMax (sptr ctx) 4.
    intros.
    split.
    generalize initialESP_upper; omega.
  Qed.

  Hint Resolve stack_slot_belowMax stack_slot_belowMax' : asm.

  Lemma slot_not_too_high : forall ctx n abs st,
    conc' abs st ctx
    -> not_too_high (stack_slot ctx n).
    intros.
    red.
    destruct H.
    myeauto.
  Qed.

  Import Datatypes.

  Lemma plus4_apart : forall n m,
    nat_apart n m
    -> nat_apart (S (S (S (S n)))) (S (S (S (S m)))).
    intros n m Hapart.
    inversion Hapart; red; omega.
  Qed.

  Hint Resolve plus4_apart : asm.

  Lemma slots_apart : forall ctx n m,
    n < bvec_to_nat (sptr ctx)
    -> m < bvec_to_nat (sptr ctx)
    -> nat_apart n m
    -> apart (stack_slot ctx n) (stack_slot ctx m).
    intros ctx n m Hn Hm Hapart.
    unfolder; add Hbound (btn_upper (sptr ctx));
      inversion Hapart; [right | left];
	repeat (rewrite ntb_invert); omega.
  Qed.

  Hint Rewrite get_set_stack_apart : asm.

  Hint Rewrite invert_set_mem32 : asm.
  Hint Rewrite invert_set_mem32_neq using (myeauto; fail) : asm.

  Lemma apart_hasType : forall st ctx abs t n n0,
    conc' abs st ctx
    -> nat_apart n n0
    -> 4 <= n0 < aStackSize (aset_stack abs n t)
    -> hasType ctx (get_mem32 st (stack_slot ctx n0))
    (aget_stack (aset_stack abs n t) n0).
    intros st ctx abs t n n0 Hconc Hapart Hlt.
    unfolder; unfold aStackSlots at 1;
      repeat (rewrite NatMap.sel_neq; try (inversion Hapart; omega));
	inversion Hconc; red in concStackSlots0;
	  unfold stack_slot, GenericTypes.stack_slot, aget_stack, GenericTypes.aget_stack in concStackSlots0;
	    apply concStackSlots0; trivial.
  Qed.

  Lemma stackSize_lt : forall abs st ctx n,
    conc' abs st ctx
    -> n < aStackSize abs
    -> n < bvec_to_nat (sptr ctx).
    intros abs st ctx n Hconc Hlt.
    inversion Hconc; omega.
  Qed.

  Hint Resolve slot_not_too_high apart_hasType slots_apart stackSize_lt : asm.

  (** ** Generic lemmas about aspects of [conc] *)

  (** *** [conc'] *)

  Lemma conc_pc_set_pc : forall abs st ctx pc,
    conc_pc abs st ctx
    -> pcInRange pc
    -> conc_pc (aset_pc abs pc) (set_pc st pc) ctx.
    intros abs st ctx pc Hconc Hpc.
    red; red; simpl; trivial.
  Qed.

  Hint Resolve conc_pc_set_pc : asm.

  Lemma conc_pc_set_pc_cust : forall abs st ctx cust,
    conc_pc abs st ctx
    -> conc_pc (aset_custom abs cust) st ctx.
    intros.
    trivial.
  Qed.

  Hint Resolve conc_pc_set_pc_cust : asm.

  Lemma conc'_pc : forall abs st ctx pc abs',
    conc' abs st ctx
    -> pcInRange pc
    -> pconc (custom ctx) abs' (set_pc st pc)
    -> conc' (aset_custom (aset_pc abs pc) abs') (set_pc st pc) ctx.
    intros abs st ctx pc abs' Hconc Hpc Hpconc.
    inversion Hconc; split; intuition.
  Qed.

  Hint Resolve conc'_pc : asm.

  Lemma conc'_arith_flags : forall abs st ctx aop v1 v2,
    conc' abs st ctx
    -> pconc (custom ctx) (aCustom abs) (arith_flags st aop v1 v2)
    -> conc' abs (arith_flags st aop v1 v2) ctx.
    intros abs st ctx aop v1 v2 Hconc Hpconc.
    inversion Hconc.
    split; intuition.
  Qed.

  Hint Resolve conc'_arith_flags : asm.

  Lemma conc'_arith_flags_no_CF : forall abs st ctx aop v1 v2,
    conc' abs st ctx
    -> pconc (custom ctx) (aCustom abs) (arith_flags_no_CF st aop v1 v2)
    -> conc' abs (arith_flags_no_CF st aop v1 v2) ctx.
    intros abs st ctx aop v1 v2 Hconc Hpconc.
    inversion Hconc.
    split; intuition.
  Qed.

  Hint Resolve conc'_arith_flags_no_CF : asm.
  

  (** *** [conc_regs] *)

  Lemma conc_regs_upd : forall abs st ctx r v t,
    conc_regs abs st ctx
    -> hasType ctx v t
    -> conc_regs (aset_reg abs r t) (set_reg32 st r v) ctx.
    intros abs st ctx r v t Hconc Hty r0.
    destruct (reg32_eq_dec r r0); subst;
      [sel_eq_replace (get_reg32 (set_reg32 st r0 v) r0) v Reg32Map.sel_eq;
	sel_eq_replace (GenericTypes.aget_reg (aset_reg abs r0 t) r0) t Reg32Map.sel_eq
	| sel_neq_replace (get_reg32 (set_reg32 st r v) r0) (get_reg32 st r0) Reg32Map.sel_neq;
	  sel_neq_replace (GenericTypes.aget_reg (aset_reg abs r t) r0) (aget_reg abs r0) Reg32Map.sel_neq];
      auto.
  Qed.

  Lemma conc_regs_pc : forall abs st ctx pc pc',
    conc_regs abs st ctx
    -> conc_regs (aset_pc abs pc) (set_pc st pc') ctx.
    intuition.
  Qed.

  Lemma conc_regs_stack : forall abs st ctx r t v a v' n t',
    conc_regs (aset_reg abs r t) (set_reg32 st r v) ctx
    -> conc_regs (aset_reg (aset_stack abs n t') r t)
    (set_reg32 (set_mem32 st a v') r v) ctx.
    intuition.
  Qed.

  Lemma reg_hasType : forall abs st ctx r,
    conc_regs abs st ctx
    -> hasType ctx (eval_genop32 (Reg r) st) (aget_reg abs r).
    intros abs st ctx r Hconc.
    simpl.
    apply Hconc.
  Qed.

  Lemma reg_hasType' : forall abs st ctx r,
    conc' abs st ctx
    -> hasType ctx (get_reg32 st r) (aget_reg abs r).
    intros abs st ctx r Hconc.
    destruct Hconc.
    simpl.
    apply concRegs0.
  Qed.

  Hint Resolve conc_regs_upd conc_regs_pc conc_regs_stack reg_hasType reg_hasType' : asm.

  (** *** [conc_slots] *)

  Lemma conc_slots_upd : forall abs st ctx n v t,
    conc' abs st ctx
    -> hasType ctx v t
    -> 4 <= n < aStackSize abs
    -> conc_slots (aset_stack abs n t)
    (set_mem32 st (stack_slot ctx n) v) ctx.
    intros abs st ctx n v t Hconc Hty Hlt n0 Hbound.
    destruct (nat_apart_dec n n0);
      [rewrite invert_set_mem32_neq; myeauto
	| idtac].
    apply apart_hasType; myeauto.
    simpl in Hbound; inversion Hconc; apply slots_apart; myauto; omega.

    (* Reading the slot we just set *)
    assert (n0 = n-3 \/ n0 = n-2 \/ n0 = n-1 \/ n0 = n \/ n0 = n+1 \/ n0 = n+2 \/ n0 = n+3);
      [unfold nat_apart in n1; omega | idtac].
    unfold aset_stack, aget_stack, GenericTypes.aget_stack; simpl.

    Hint Rewrite NatMap.sel_eq using omega : conc_slots_upd.
    Hint Rewrite NatMap.sel_neq using omega : conc_slots_upd.
    Hint Rewrite invert_set_mem32 : conc_slots_upd.

    Ltac doit :=
      autorewrite with conc_slots_upd; try apply HasType_Top; trivial;
	match goal with
	  [ |- context[NatMap.sel (NatMap.upd _ ?N1 _) ?N2] ] =>
	  let He := fresh "compare" in
	    (destruct (eq_nat_dec N1 N2) as [He | He];
	      try rewrite <- He; autorewrite with conc_slots_upd;
		try apply HasType_Top; trivial)
	end.

    intuition; subst; repeat doit.
  Qed.

  Lemma conc_slots_pc : forall abs st ctx pc pc',
    conc_slots abs st ctx
    -> conc_slots (aset_pc abs pc) (set_pc st pc') ctx.
    intuition.
  Qed.

  Lemma conc_slots_reg : forall abs st ctx r r' t v,
    conc_slots abs st ctx
    -> conc_slots (aset_reg abs r t) (set_reg32 st r' v) ctx.
    intuition.
  Qed.

  Hint Resolve conc_slots_upd conc_slots_pc conc_slots_reg : asm.

  (** ** [agree_range] *)

  Lemma agree_range_write_above : forall st start len addr v,
    agree_range (stMem st) initialMem start len
    -> belowMax addr 3
    -> bvec_to_nat addr > bvec_to_nat start + len
    -> agree_range (stMem (set_mem32 st addr v)) initialMem start len.
    intros st start len addr v Hrange Hupper Hgt addr' Hrange'.
    simpl.
    repeat (rewrite Int32Map.sel_neq); auto.
    assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
    assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
      [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
      [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
      [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
  Qed.

  Lemma agree_range_write_below : forall st start len addr v,
    agree_range (stMem st) initialMem start len
    -> belowMax addr 3
    -> bvec_to_nat addr + 3 < bvec_to_nat start
    -> agree_range (stMem (set_mem32 st addr v)) initialMem start len.
    intros st start len addr v Hrange Hupper Hgt addr' Hrange'.
    simpl.
    repeat (rewrite Int32Map.sel_neq); auto.
    assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
    assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
      [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
      [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
      [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
  Qed.

  Lemma agree_range_write'_above : forall st memIn start finish addr v,
    agree_range' (stMem st) memIn start finish
    -> belowMax addr 3
    -> bvec_to_nat addr > bvec_to_nat finish
    -> agree_range' (stMem (set_mem32 st addr v)) memIn start finish.
    intros st memIn0 start len addr v Hrange Hupper Hgt addr' Hrange'.
    simpl.
    repeat (rewrite Int32Map.sel_neq); auto.
    assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
    assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
      [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
      [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
      [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
  Qed.

  Lemma agree_range_write'_below : forall st memIn start finish addr v,
    agree_range' (stMem st) memIn start finish
    -> belowMax addr 3
    -> bvec_to_nat addr + 3 < bvec_to_nat start
    -> agree_range' (stMem (set_mem32 st addr v)) memIn start finish.
    intros st memIn0 start len addr v Hrange Hupper Hgt addr' Hrange'.
    simpl.
    repeat (rewrite Int32Map.sel_neq); auto.
    assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
    assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
      [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
      [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
    assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
      [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
  Qed.

  Lemma agree_range_set_pc : forall pc st m1 m2 m3,
    agree_range (stMem st) m1 m2 m3
    -> agree_range (stMem (set_pc st pc)) m1 m2 m3.
    intuition.
  Qed.

  Lemma agree_range_set_reg : forall r v st m1 m2 m3,
    agree_range (stMem st) m1 m2 m3
    -> agree_range (stMem (set_reg32 st r v)) m1 m2 m3.
    intuition.
  Qed.

  Hint Resolve agree_range_set_pc agree_range_set_reg : asm.


  (** * Cutpoints *)

  (** An assertion of an invariant *)
  Definition cutpoint_info := cutpoint_info ty astat.

  Parameter cutpoints : list (list cutpoint_info * cutpoint_info).
  (** The cutpoints asserted for the current program.
    Each entry contains a list of assumed-safe states and the actual entry state.
    *)


  (** * Function signatures *)

  Definition function_info := function_info ty.

  Parameter functions : int32 -> option function_info.


  (** * Initialization and the root state set *)

  Definition initials' :=
    map (B := (list cutpoint_info * cutpoint_info)) (initial_cp hasTyp) entryPoints
    ++ cutpoints.

  Definition initials :=
    match kernel_astat with
      | None => initials'
      | Some abs => kernel_cp typ abs :: initials'
    end.
  (** All root states, represented as [cutpoint]s *)

  Definition init : list (list astate * astate) :=
    map (B := list astate * astate) (@make_init typ astat) initials.

  Definition importd' := map (B := (list cutpoint_info * cutpoint_info)) (initial_cp hasTyp) imported.

  Definition importd : list (list astate * astate) :=
    map (B := list astate * astate) (@make_init typ astat) importd'.


  Lemma undo_fourx : forall n,
    n + (n + (n + (n + 0))) = 4 * n.
    intros.
    omega.
  Qed.

  Lemma undo_fourx_S : forall n,
    S (n + S (n + S (n + S (n + 0)))) = 4 * S n.
    intros.
    omega.
  Qed.

  Lemma undo_plus4 : forall n,
    S (S (S (S (4 * n)))) = 4 + 4 * n.
    intros; omega.
  Qed.

  Hint Rewrite undo_fourx undo_fourx_S undo_plus4 : mysimpl.

  Ltac mysimpl := simpl; autorewrite with mysimpl.

  Definition init_ctx (ep : entryPoint typ astat) st (ctx : context) :=
    Build_context
    (nat_to_bvec 32 (bvec_to_nat (get_reg32 st ESP) + (4 + 4 * length (epArgs ep))))
    (get_reg32 st EBP)
    (stMem st)
    (Some (get_mem32 st (get_reg32 st ESP)))
    ctx.


  (** * Operands *)

  (** ** Reading *)

  Lemma double_sub : forall n v1 v2,
    bvec_to_nat v1 <= n
    -> bvec_to_nat (bsub32 v1 v2) < n
    -> eqMod (pow2 32)
    (n - bvec_to_nat (bsub32 v1 v2))
    (bvec_to_nat v2 + n - bvec_to_nat v1).
    intros n v1 v2 Hle Hlt.
    apply eqMod_trans
      with (n + bvec_to_nat (bneg (bsub32 v1 v2))).
    apply bneg_btn_sub; trivial.
    apply eqMod_trans
      with (n + bvec_to_nat (bsub32 v2 v1)).
    apply eqMod_plus.
    apply eqMod_refl.
    unfold bsub32; apply bsub_btn_neg.
    apply eqMod_trans
      with (n + (pow2 32 + bvec_to_nat v2 - bvec_to_nat v1)).
    apply eqMod_plus.
    apply eqMod_refl.
    unfold bsub32; apply bsub_btn.
    add Hup (btn_upper v1).
    replace (n + (pow2 32 + bvec_to_nat v2 - bvec_to_nat v1))
      with (pow2 32 + (bvec_to_nat v2 + n - bvec_to_nat v1)).
    apply eqMod_sym; apply eqMod_mod.
    apply eqMod_refl.
    omega.
  Qed.

  Lemma slot_addr : forall d r st ss ctx abs,
    conc_regs abs st ctx
    -> aget_reg abs r = Stack typ ss
    -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) d) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> eval_addr (Build_address d (Some r) None) st
    = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) d)).
    intros d r st ss ctx abs Hconc Hr Hlt Hsptr.
    add Hrconc (Hconc r).
    unfold aget_reg in Hr.
    rewrite Hr in Hrconc.
    inversion Hrconc as [| |Ha Hb Hle Hd Heq Hf| | |]; subst.
    unfold eval_addr, addrIndex, addrBase, addrDisp, stack_slot, GenericTypes.stack_slot, badd32, badd, barith.
    apply (ntb_mod 32).
    rewrite <- Heq.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; try (add Hup (btn_upper (sptr ctx)); omega).
    rewrite plus_minus; trivial.
    replace (bvec_to_nat d + bvec_to_nat (sptr ctx) - ss)
      with (bvec_to_nat d + bvec_to_nat (sptr ctx) - bvec_to_nat (nat_to_bvec 32 ss)).
    apply eqMod_sym; apply double_sub.
    rewrite ntb_invert; add Hup (btn_upper (sptr ctx)); omega.
    omega.
    rewrite ntb_invert; add Hup (btn_upper (sptr ctx)); omega.
  Qed.

  Ltac type_fail := intros; apply SNone.

  Hint Constructors word_hasType : asm.

  Definition ty_to_word_type : forall t,
    sig_option (fun t' => forall ctx v,
      hasType ctx v t
      -> word_hasType hasTyp (custom ctx) v t').
    intros.
    destruct t.
    apply SSome with (TyTop typ); myauto.
    apply SSome with (TyConst typ i); intros; inversion H; subst; myauto.
    apply SNone.
    apply SNone.
    apply SNone.
    apply SSome with (TyCustom t); intros; inversion H; subst; myauto.
  Qed.

  Definition lower_ty t :=
    match ty_to_word_type t with
      | SNone => TyTop typ
      | SSome t' _ => t'
    end.

  Definition simple_typeof_genop abs gop :=
    match gop with
      | Imm i => TyTop typ
      | Reg r => lower_ty (aget_reg abs r)
      | Address addr =>
	match addrBase addr with
	  | None => TyTop typ
	  | Some r =>
	    match aget_reg abs r with
	      | Stack sl =>
		match addrIndex addr with
		  | Some _ => TyTop typ
		  | None =>
		    let sl' := bvec_to_nat (bsub32 (nat_to_bvec 32 sl) (addrDisp addr)) in
		      if sum_to_bool (le_lt_dec 4 sl')
			&& negb (sum_to_bool (le_lt_dec (aStackSize abs) sl')) then
			  lower_ty (aget_stack abs sl')
			else
			  TyTop typ
		end
	      | _ => TyTop typ
	    end
	end
    end.

  Definition simple_addr_inStack abs addr :=
    match addrBase addr with
      | None => false
      | Some r =>
	match aget_reg abs r with
	  | Stack sl =>
	    match addrIndex addr with
	      | Some _ => false
	      | None =>
		let sl' := bvec_to_nat (bsub32 (nat_to_bvec 32 sl) (addrDisp addr)) in
		  sum_to_bool (le_lt_dec 4 sl')
		  && negb (sum_to_bool (le_lt_dec (aStackSize abs) sl'))
	    end
	  | _ => false
	end
    end.

  Lemma initialESP_stackEnd :
    bvec_to_nat initialESP + 5 = bvec_to_nat stackEnd.
    unfold initialESP, stackEnd.
    repeat (rewrite ntb_invert).
    omega.
    generalize stack_fits; omega.
    generalize stack_fits; omega.
  Qed.

  Definition simple_pc (abs : astate) :=
    match aPc abs with
      | apcConst pc => Some pc
      | _ => None
    end.

  Definition simple_used (abs : astate) :=
    match aget_reg abs ESP with
      | Stack sl => Some sl
      | _ => None
    end.

  Definition simple_left (abs : astate) :=
    match aget_reg abs ESP with
      | Stack sl =>
	if le_lt_dec sl (aStackSize abs)
	  then Some (aStackSize abs - sl)
	  else None
      | _ => None
    end.

  Lemma invert_RetPtr: forall ctx v,
    hasType ctx v (RetPtr typ)
    -> Some v = ret ctx.
    intros ctx v Hret.
    inversion Hret; subst.
    auto.
  Qed.

  Lemma invert_RetPtr': forall ctx v,
    hasType ctx v (RetPtr typ)
    -> pcInRange v.
    intros ctx v Hret.
    inversion Hret; subst.
    auto.
  Qed.

  Hint Resolve invert_RetPtr invert_RetPtr' : asm.

  Definition simple_hasRetPtr abs addr :=
    match addrBase addr with
      | None => false
      | Some r =>
	match aget_reg abs r with
	  | Stack sl =>
	    match addrIndex addr with
	      | Some _ => false
	      | None =>
		let sl' := bvec_to_nat (bsub32 (nat_to_bvec 32 sl) (addrDisp addr)) in
		  if sum_to_bool (le_lt_dec 4 sl')
		    && negb (sum_to_bool (le_lt_dec (aStackSize abs) sl'))
		    then match aget_stack abs sl' with
			   | RetPtr => true
			   | _ => false
			 end
		    else false
	    end
	  | _ => false
	end
    end.

  Definition simple_status : forall abs,
    {stat : status typ
      | forall st ctx,
	conc' abs st ctx
	-> statusOk hasTyp stat (custom ctx) st}.
    intros.
    exists (Build_status
      (simple_typeof_genop abs)
      (simple_addr_inStack abs)
      (simple_pc abs)
      (simple_used abs)
      (simple_left abs)
      (simple_hasRetPtr abs)).
    intros st ctx Hconc.
    split.

    (* genopType *)
    intro g.
    unfold simple_typeof_genop; unfold genopType.
    destruct g; myauto; unfold lower_ty.
    
    add Hr (concRegs Hconc r).
    destruct (ty_to_word_type (aget_reg abs r)).
    simpl.
    auto.
    myauto.

    destruct a; unfold Defs.addrBase, Defs.addrIndex, Defs.addrDisp.
    destruct addrBase; myauto.
    add Hr (concRegs Hconc r).
    caseEq (aget_reg abs r); myauto.
    intros sl Hsl.
    destruct addrIndex; myauto.
    destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp))).
    unfold sum_to_bool at 1.
    destruct (le_lt_dec (aStackSize abs)
      (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp))).
    simpl; myauto.
    simpl negb.
    simpl andb.
    destruct (ty_to_word_type (aget_stack abs
      (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)))).
    apply w.
    unfold eval_genop32.
    replace (eval_addr (Build_address addrDisp (Some r) None) st)
      with (stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp))).
    apply (concStackSlots Hconc).
    omega.
    symmetry.
    eapply slot_addr with abs; inversion Hconc; eauto.
    myauto.

    unfold sum_to_bool.
    simpl andb.
    simpl.
    myauto.

    (* addrInStack *)
    intros a Hin.
    unfold simple_addr_inStack, addrInStack in Hin.
    destruct a.
    unfold Defs.addrBase, Defs.addrIndex, Defs.addrDisp in Hin.
    destruct addrBase; try discriminate.
    caseEq (aget_reg abs r);
    try (intros Hr; rewrite Hr in Hin; discriminate);
      try (intros x Hr; rewrite Hr in Hin; discriminate).
    intros sl Hr.
    rewrite Hr in Hin.
    destruct addrIndex; try discriminate.
    destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp))); try discriminate.
    destruct (le_lt_dec (aStackSize abs) (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp))); try discriminate.
    clear Hin.
    replace (eval_addr (Build_address addrDisp (Some r) None) st)
      with (stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp))).
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; try (generalize (btn_upper (sptr ctx)); omega).
    inversion Hconc.
    generalize initialESP_stackEnd.
    omega.

    symmetry; eapply slot_addr with abs; inversion Hconc; eauto.

    (* programCounter *)
    simpl.
    unfold simple_pc.
    add Hpc (concPc Hconc).
    red in Hpc.
    destruct (aPc abs); auto.

    (* stackUsed *)
    unfold stackUsed.
    unfold simple_used.
    caseEq (aget_reg abs ESP); trivial.
    intros sl Hsl.
    add Hesp (concRegs Hconc ESP).
    unfold aget_reg in Hsl.
    rewrite Hsl in Hesp.
    inversion Hesp; subst.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    inversion Hconc; omega.
    generalize (btn_upper (sptr ctx)); omega.
    
    (* stackLeft *)
    simpl.
    unfold simple_left.
    caseEq (aget_reg abs ESP); trivial.
    intros sl Hsl.
    destruct (le_lt_dec sl (aStackSize abs)); trivial.
    add Hesp (concRegs Hconc ESP).
    unfold aget_reg in Hsl.
    rewrite Hsl in Hesp.
    inversion Hesp; subst.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    inversion Hconc; omega.
    generalize (btn_upper (sptr ctx)); omega.

    (* hasRetPtr *)
    simpl.
    unfold simple_hasRetPtr.
    intro a.
    destruct a.
    unfold Defs.addrBase, Defs.addrIndex, Defs.addrDisp.
    destruct addrBase; try (intros; discriminate).
    caseEq (aget_reg abs r); try (intros; discriminate).
    intros sl Hsl.
    destruct addrIndex; try (intros; discriminate).
    destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)));
      unfold sum_to_bool; [idtac | unfold andb, ifb; intros; discriminate].
    destruct (le_lt_dec (aStackSize abs)
      (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)));
    unfold andb, ifb, negb;
      [intros; discriminate | idtac].
    caseEq (aget_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)));
    try (intros; discriminate).
    intros Hstack Hdumb.
    clear Hdumb.

    assert (Hty : hasType ctx (get_mem32 st (eval_addr (Build_address addrDisp (Some r) None) st)) (RetPtr typ)).

    replace (eval_addr (Build_address addrDisp (Some r) None) st)
      with (stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp))).
    rewrite <- Hstack.
    apply (concStackSlots Hconc).
    tauto.

    symmetry; eapply slot_addr with abs; inversion Hconc; eauto.    

    myeauto.

    (* sokCodeMem *)
    inversion Hconc; trivial.
  Qed.

  Definition typeof_addr32_custom : forall abs addr,
    sig_option (fun ty =>
      forall st ctx,
	conc abs st ctx
	-> hasType ctx (eval_genop32 (Address reg32 addr) st) ty).
    intros.
    destruct (simple_status abs) as [stat Hstat].
    destruct (typeof_read (aCustom abs) stat addr) as [t Ht |]; [idtac | type_fail].
    destruct (word_type_to_ty hasTyp t) as [t' Ht'].
    apply SSome with t'.
    intros st ctx Hconc.
    simpl.
    apply Ht'.
    inversion Hconc.
    apply Ht; simpl; auto.
  Qed.

  Definition typeof_genop32 : forall abs gop,
    sig_option (fun ty =>
      forall st ctx,
	conc abs st ctx
	-> hasType ctx (eval_genop32 gop st) ty).
    intros abs gop.
    destruct gop.

    (* Imm *)
    apply SSome with (Const typ i); myeauto.

    (* Reg *)
    apply SSome with (aget_reg abs r); intros st ctx Hconc; simpl; inversion Hconc; inversion concMain0; myauto.
    
    (* Address *)
    Ltac try_custom := apply typeof_addr32_custom.
    Ltac try_custom' := intros; apply typeof_addr32_custom.

    destruct a.
    destruct addrBase; [idtac | try_custom].
    destruct addrIndex; [try_custom | idtac].
    caseEq (aget_reg abs r); [try_custom' | try_custom' | idtac | try_custom' | try_custom' | try_custom'].
    intros ss Hr.
    destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
      [idtac | apply SNone].
    destruct (le_lt_dec (aStackSize abs) (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
      [apply SNone | idtac].
    apply SSome with (aget_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
    intros st ctx Hconc.
    inversion Hconc; inversion concMain0.
    unfold eval_genop32.
    assert (eval_addr (Build_address addrDisp (Some r) None) st
      = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
    eapply slot_addr with abs; eauto.
    rewrite H.
    myauto.
  Qed.


  (** ** Writing *)

  (** *** Inversion lemmas, needed to be able to use inversion on [write_genop32]
    in sort [Set] *)

  Lemma invert_write_reg : forall r n st st',
    write_genop32 (Reg r) n st st'
    -> st' = set_reg32 st r n.
    inversion 1; trivial.
  Qed.

  Lemma invert_write_mem : forall a n st st',
    write_genop32 (Address reg32 a) n st st'
    -> st' = set_mem32 st (eval_addr a st) n.
    inversion 1; trivial.
  Qed.

  Lemma invert_write_imm : forall i n st st',
    write_genop32 (Imm reg32 i) n st st'
    -> False.
    inversion 1.
  Qed.

  (** *** Writing to the stack *)

  Lemma awrite_stack_high : forall abs st ctx r ss addrDisp,
    aget_reg abs r = Stack typ ss
    -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
    -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> belowMax (sptr ctx) 3
    -> bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st)
    >= bvec_to_nat stackStart.
    intros abs st ctx r ss addrDisp Hr Hss Hagree Hconc Hsptr Hbelow.
    assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
      = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
    eapply slot_addr; eauto.
    rewrite Hslot.
    unfold stack_slot, GenericTypes.stack_slot.
    add Hup (btn_upper (sptr ctx)).
    rewrite ntb_invert; omega.
  Qed.

  Lemma btn_conv : forall len (v1 v2 : Bvector len),
    bvec_to_nat v1 <> bvec_to_nat v2
    -> v1 <> v2.
    intros; congruence.
  Qed.

  Lemma stack_slot_belowMax'' : forall abs st ctx r ss addrDisp,
    aget_reg abs r = Stack typ ss
    -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> belowMax (sptr ctx) 3
    -> belowMax (eval_addr (Build_address addrDisp (Some r) None) st) 3.
    intros abs st ctx r ss addrDisp Hr Hsize Hconc Hsptr Hbelow.
    assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
      = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
    eapply slot_addr; eauto.
    rewrite Hslot.
    unfold stack_slot, GenericTypes.stack_slot.
    split.
    inversion Hbelow.
    rewrite ntb_invert; omega.
  Qed.

  Hint Resolve stack_slot_belowMax'' : asm.

  Lemma awrite_stack_agree_range : forall abs st ctx r ss addrDisp v,
    aget_reg abs r = Stack typ ss
    -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
    -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> belowMax (sptr ctx) 3
    -> agree_range
    (stMem
      (set_mem32 st (eval_addr (Build_address addrDisp (Some r) None) st) v))
    initialMem initialPc (actualCodeLen + longestInstr).
    intros abs st ctx r ss addrDisp v Hr Hss Hagree Hconc Hsptr Hbelow addr Hrange.
    unfold set_mem32, set_mem, stMem.
    assert (Hhigh : bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st) >=
      bvec_to_nat stackStart).
    eapply awrite_stack_high; eauto.
    add Hspace actual_code_space.
    repeat (rewrite Int32Map.sel_neq).
    auto.
    apply (@btn_conv 32); omega.
    apply (@btn_conv 32); rewrite binc_btn_belowMax; [omega | myeauto].
    apply (@btn_conv 32); rewrite binc2_btn_belowMax; [omega | myeauto].
    apply (@btn_conv 32); rewrite binc3_btn_belowMax; [omega | myeauto].
  Qed.

  Hint Resolve awrite_stack_agree_range : asm.

  Lemma ignore_four : forall k n m,
    k <= n < m
    -> n < m.
    intros; omega.
  Qed.

  Hint Resolve ignore_four : asm.

  Lemma stackSlot_below_sptr : forall r ss addrDisp abs ctx st,
    aget_reg abs r = Stack typ ss
    -> 4 <= bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> conc_regs abs st ctx
    -> bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st) + 3
    < bvec_to_nat (sptr ctx).
    intros r ss addrDisp abs ctx st Hr Hss Hsptr Hconc.
    assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
      = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
    eapply slot_addr; eauto.
    omega.
    rewrite Hslot.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    omega.
    generalize (btn_upper (sptr ctx)); omega.
  Qed.

  Lemma awrite_stack_memIn : forall abs st ctx r ss addrDisp v,
    aget_reg abs r = Stack typ ss
    -> 4 <= bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
    -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> belowMax (sptr ctx) 4
    -> agree_range'
    (stMem
      (set_mem32 st (eval_addr (Build_address addrDisp (Some r) None) st) v))
    (memIn ctx) (sptr ctx) stackEnd.
    intros abs st ctx r ss addrDisp v Hr Hss Hagree Hconc Hsptr Hbelow.
    assert (Hlow : bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st) + 3
      < bvec_to_nat (sptr ctx)).
    eapply stackSlot_below_sptr; eauto.
    apply agree_range_write'_below; myeauto.
  Qed.

  Hint Resolve awrite_stack_memIn : asm.

  Lemma awrite_stack_conc_slots : forall abs st ctx v t r ss addrDisp,
    conc' abs st ctx
    -> hasType ctx v t
    -> aget_reg abs r = Stack typ ss
    -> 4 <= bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> conc_slots abs st ctx
    -> conc_slots
    (aset_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)) t)
    (set_mem32 st (eval_addr (Build_address addrDisp (Some r) None) st) v)
    ctx.
    intros abs st ctx v t r ss addrDisp Hconc Ht Hr Hsize Hregs Hsptr Hslots.
    assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
      = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
    eapply slot_addr; eauto.
    omega.
    rewrite Hslot.
    apply conc_slots_upd; trivial.
  Qed.

  Hint Resolve awrite_stack_conc_slots : asm.


  (** *** Concrete write function *)

  Definition writable_genop32 (gop : genop32) :=
    match gop with
      | Imm _ => false
      | _ => true
    end.

  Definition write_concrete : forall (gop : genop32) (st : state) (v : int32)
    (writable : writable_genop32 gop = true), state.
    intros.
    destruct gop.

    discriminate.
    exact (set_reg32 st r v).
    exact (set_mem32 st (eval_addr a st) v).
  Defined.

  Lemma apart_or : forall a a',
    (bvec_to_nat a' + 4 <= bvec_to_nat a
      \/ bvec_to_nat a' >= bvec_to_nat a + 4)
    -> apart a a'.
    intros.
    destruct H.
    apply apart_lower; trivial.
    apply apart_higher; trivial.
  Qed.

  Lemma belowMax_stack_slot : forall ctx n,
    belowMax (sptr ctx) 3
    -> belowMax (stack_slot ctx n) 3.
    intros.
    inversion H.
    split.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; omega.
  Qed.

  Hint Resolve belowMax_stack_slot : asm.

  (** *** The main write function *)

  Lemma rewrite_get_mem : forall st addr,
    Int32Map.sel (stMem st) addr
    = get_mem st addr.
    trivial.
  Qed.

  Lemma cwrite_agree_range : forall st v addr,
    (not_too_high (eval_addr addr st) /\
      ~
      bvec_to_nat initialPc - 4 <= bvec_to_nat (eval_addr addr st) <
      bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4 /\
      ~
      bvec_to_nat stackStart - 4 <= bvec_to_nat (eval_addr addr st) <
      bvec_to_nat stackEnd + 4)
    -> agree_range (stMem st) initialMem initialPc
    (actualCodeLen + longestInstr)
    -> agree_range (stMem (set_mem32 st (eval_addr addr st) v)) initialMem
    initialPc (actualCodeLen + longestInstr).
    intros st v addr Hok Hagree addr' Haddr'.
    
    rewrite rewrite_get_mem.
    rewrite invert_set_mem32_neq'; auto; try tauto.
    rewrite <- rewrite_get_mem; auto; try tauto.
    apply apart_or.
    omega.
  Qed.

  Hint Resolve cwrite_agree_range : asm.

  Lemma cwrite_stack_slots : forall abs st v addr ctx,
    not_too_high (eval_addr addr st) /\
    ~
    bvec_to_nat initialPc - 4 <= bvec_to_nat (eval_addr addr st) <
    bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4 /\
    ~
    bvec_to_nat stackStart - 4 <= bvec_to_nat (eval_addr addr st) <
    bvec_to_nat stackEnd + 4
    -> conc' abs st ctx
    -> conc_slots abs (set_mem32 st (eval_addr addr st) v) ctx.
    intros abs st v addr ctx Hok Hconc slot Hslot.

    inversion Hconc.
    rewrite invert_set_mem32_neq; myeauto; try tauto.
    apply apart_or.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; [idtac | generalize (btn_upper (sptr ctx)); omega].
    generalize initialESP_stackEnd.
    simpl in Hslot.
    omega.
  Qed.

  Hint Resolve cwrite_stack_slots : asm.

  Lemma cwrite_memIn : forall abs st v addr ctx,
    (not_too_high (eval_addr addr st) /\
      ~
      bvec_to_nat initialPc - 4 <= bvec_to_nat (eval_addr addr st) <
      bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4 /\
      ~
      bvec_to_nat stackStart - 4 <= bvec_to_nat (eval_addr addr st) <
      bvec_to_nat stackEnd + 4)
    -> conc' abs st ctx
    -> agree_range' (stMem (set_mem32 st (eval_addr addr st) v)) 
    (memIn ctx) (sptr ctx) stackEnd.
    intros abs st v addr ctx Hok Hconc addr' Haddr'.
    inversion Hconc.
    
    rewrite rewrite_get_mem.
    rewrite invert_set_mem32_neq'; auto; try tauto.
    rewrite <- rewrite_get_mem; auto; try tauto.
    apply apart_or.
    omega.
  Qed.

  Hint Resolve cwrite_memIn : asm.

  Definition awrite_addr32_custom : forall abs addr ty,
    sig_option (fun abs' =>
      forall st st' ctx v,
	conc abs st ctx
	-> hasType ctx v ty
	-> write_genop32 (Address reg32 addr) v st st'
	-> conc' abs' st' ctx).
    intros abs addr vty.
    destruct (ty_to_word_type vty) as [vty' Hvty' |]; [idtac | type_fail].
    destruct (simple_status abs) as [stat Hstat].
    destruct (check_write (aCustom abs) stat addr vty') as [Hok |]; [idtac | type_fail].
    apply SSome with abs.
    intros st st' ctx v Hconc Hv Hwrite.
    inversion Hwrite; subst.

    inversion Hconc.
    assert (Hand : not_too_high (eval_addr addr st) /\
      ~
      bvec_to_nat initialPc - 4 <= bvec_to_nat (eval_addr addr st) <
      bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4 /\
      ~
      bvec_to_nat stackStart - 4 <= bvec_to_nat (eval_addr addr st) <
      bvec_to_nat stackEnd + 4).
    eauto.
    inversion concMain0.
    split; try apply cwrite_stack_slots; myeauto.
  Qed.

  Definition awrite_genop32 : forall abs gop ty,
    sig_option (fun abs' =>
      forall st st' ctx v,
	conc abs st ctx
	-> hasType ctx v ty
	-> write_genop32 gop v st st'
	-> conc' abs' st' ctx).
    intros abs gop t.
    destruct gop.

    (* Imm *)
    apply SNone.

    (* Reg *)
    apply SSome with (aset_reg abs r t).
    intros st st' ctx v Hconc Ht Hwrite.
    add Hst' (invert_write_reg Hwrite); subst; clear Hwrite.
    inversion Hconc; inversion concMain0; split; intuition.

    (* Address *)
    Ltac try_customw := apply awrite_addr32_custom.
    Ltac try_customw' := intros; try_customw.

    destruct a.
    destruct addrBase.

    (* Base register *)
    destruct addrIndex; [try_customw | idtac].
    caseEq (aget_reg abs r);
    [try_customw' | try_customw' | idtac | try_customw' | try_customw' | try_customw'].
    intros ss Hr.
    destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
      [idtac | apply SNone].
    destruct (le_lt_dec (aStackSize abs) (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
      [apply SNone | idtac].
    apply SSome with (aset_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)) t).
    intros st st' ctx v Hconc Ht Hwrite.
    add Hst' (invert_write_mem Hwrite); subst; clear Hwrite.
    inversion Hconc; inversion concMain0; split; myeauto.

    (* No base register *)
    try_customw.
  Qed.


  (** * Operations *)

  Lemma stack_slot_add : forall i n ctx,
    bvec_to_nat i <= n
    -> n <= bvec_to_nat (sptr ctx)
    -> badd32 i (stack_slot ctx n) = stack_slot ctx (n - bvec_to_nat i).
    intros i n ctx Hi Hsptr.
    simpl.
    apply btn_inj.
    unfold stack_slot, GenericTypes.stack_slot, badd32, badd, barith.
    add Hup (btn_upper (sptr ctx)).
    repeat (rewrite ntb_invert); omega.
  Qed.

  Lemma stack_slot_sub : forall i n ctx,
    n + bvec_to_nat i <= bvec_to_nat (sptr ctx)
    -> bsub32 (stack_slot ctx n) i = stack_slot ctx (n + bvec_to_nat i).
    intros i n ctx Hsptr.
    unfold bsub32, bsub, barith, stack_slot, GenericTypes.stack_slot.
    apply ntb_mod.
    add Hup (btn_upper (sptr ctx)).
    rewrite ntb_invert; try omega.
    apply eqMod_trans
      with (bvec_to_nat (sptr ctx) - n + (pow2 32 - bvec_to_nat i)).
    apply eqMod_plus; [apply eqMod_refl | apply bneg_btn].
    replace (bvec_to_nat (sptr ctx) - n + (pow2 32 - bvec_to_nat i))
      with (pow2 32 + (bvec_to_nat (sptr ctx) - n - bvec_to_nat i)).
    apply eqMod_mod'.
    rewrite rewrite_minus_plus.
    apply eqMod_refl.
    omega.
  Qed.

  Lemma badd_comm : forall len (v1 v2 : Bvector len),
    badd v1 v2 = badd v2 v1.
    intros len v1 v2.
    unfold badd, barith.
    rewrite plus_comm.
    trivial.
  Qed.

  Ltac give_up := exists (Top typ); intuition.

  Lemma badd32_swap_neg : forall v1 v2,
    badd32 v2 v1 = bsub32 v1 (bneg v2).
    intros.
    unfold bsub32, badd32, bsub, badd, barith.
    rewrite bneg_btn_double.
    rewrite plus_comm.
    trivial.
  Qed.

  (** ** The main operation functions *)

  Definition aeval_arith_custom : forall abs aop t1 t2,
    {t : ty
      | forall st ctx v1 v2,
	conc abs st ctx 
	-> hasType ctx v1 t1
	-> hasType ctx v2 t2
	-> hasType ctx (eval_arith aop v1 v2) t}.
    intros.
    destruct (ty_to_word_type t1) as [t1' Ht1' |]; [idtac | give_up].
    destruct (ty_to_word_type t2) as [t2' Ht2' |]; [idtac | give_up].
    destruct (typeof_arith aop t1' t2') as [t Ht |]; [idtac | give_up].
    exists (proj1_sig (word_type_to_ty hasTyp t)).
    intros st ctx v1 v2 Hconc Hv1 Hv2.
    destruct (word_type_to_ty hasTyp t) as [t' Ht'].
    simpl.
    apply Ht'.
    apply Ht; auto.
  Qed.

  Definition aeval_arith : forall abs aop t1 t2,
    {t : ty |
      forall st ctx v1 v2,
	conc abs st ctx 
	-> hasType ctx v1 t1
	-> hasType ctx v2 t2
	-> hasType ctx (eval_arith aop v1 v2) t}.
    intros abs aop t1 t2.

    Ltac custom_arith := apply aeval_arith_custom.
    Ltac custom_arith' := intros; custom_arith.

    destruct t1.

    (* Top vs. anything *)
    custom_arith.

    destruct t2; [custom_arith' | idtac | idtac | custom_arith' | custom_arith' | custom_arith'].

    (* Const vs. Const *)
    exists (Const typ (eval_arith aop i i0)).
    intros st ctx v1 v2 Hconc Hv1 Hv2.
    inversion Hv1; subst.
    inversion Hv2; subst.
    intuition.

    (* Const vs. Stack *)
    destruct aop; [idtac | custom_arith' | custom_arith' | custom_arith'].

    (* Addition *)
    destruct (le_lt_dec (bvec_to_nat i) n).

    (* Add positive *)
    exists (Stack typ (n - bvec_to_nat i)).
    intros st ctx v1 v2 Hconc Hv1 Hv2.
    inversion Hv1; subst.
    inversion Hv2; subst.
    simpl.
    rewrite stack_slot_add; unfold hasType, stack_slot; intuition; myauto.

    clear l.
    destruct (le_lt_dec (n + bvec_to_nat (bneg i)) (aStackSize abs)); [idtac | give_up].

    (* Add negative *)
    exists (Stack typ (n + bvec_to_nat (bneg i))).
    intros st ctx v1 v2 Hconc Hv1 Hv2.
    inversion Hv1; subst.
    inversion Hv2; subst.
    simpl.
    assert (Hsptr : aStackSize abs <= bvec_to_nat (sptr ctx)).
    inversion Hconc; inversion concMain0; omega.
    rewrite badd32_swap_neg.
    rewrite stack_slot_sub; unfold hasType, stack_slot; intuition.
    
    destruct t2; [custom_arith' | idtac | idtac | custom_arith' | custom_arith' | custom_arith'].

    (* Stack vs. Const *)
    destruct aop; [idtac | custom_arith' | custom_arith' | idtac].

    (* Addition *)
    destruct (le_lt_dec (bvec_to_nat i) n); [idtac | give_up].
    exists (Stack typ (n - bvec_to_nat i)).
    intros st ctx v1 v2 Hconc Hv1 Hv2.
    inversion Hv1; subst.
    inversion Hv2; subst.
    simpl.
    unfold badd32; rewrite badd_comm.
    rewrite stack_slot_add; unfold hasType, stack_slot; intuition.

    (* Subtraction *)
    destruct (le_lt_dec (n + bvec_to_nat i) (aStackSize abs)); [idtac | give_up].
    exists (Stack typ (n + bvec_to_nat i)).
    intros st ctx v1 v2 Hconc Hv1 Hv2.
    inversion Hv1; subst.
    inversion Hv2; subst.
    simpl.
    assert (Hsptr : aStackSize abs <= bvec_to_nat (sptr ctx));
      [inversion Hconc; inversion concMain0; omega
	| rewrite stack_slot_sub; unfold hasType, stack_slot; intuition].

    (* Stack vs. Stack *)
    custom_arith.

    (* RetPtr vs. Anything *)
    custom_arith.

    (* OldEbp vs. Anything *)
    custom_arith.

    (* Custom vs. Anything *)
    custom_arith.
  Qed.

  Definition aeval_addr : forall abs a,
    {t : ty |
      forall st ctx,
	conc abs st ctx 
	-> hasType ctx (eval_addr a st) t}.
    intros abs a.

    destruct a.
    destruct addrBase.
    destruct addrIndex.

    (* disp + base + scale * idx *)
    give_up.

    (* disp + base *)
    destruct (aeval_arith abs Add (Const typ addrDisp) (aget_reg abs r)) as [t Ht].
    exists t.
    intuition.
    replace (eval_addr (Build_address addrDisp (Some r) None) st)
      with (eval_arith Add addrDisp (get_reg32 st r)).
    inversion H; myeauto.
    simpl.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
    trivial.

    destruct addrIndex.

    (* disp + scale * idx *)
    give_up.

    (* disp *)
    exists (Const typ addrDisp).
    myeauto.
  Qed.


  (** * Call and return *)

  (** ** Switch point of view from caller to callee.
    Use this on the state right before a CALL instruction.
    *)

  Lemma enter_agree_range : forall abs ss st ctx ret,
    aget_reg abs ESP = Stack typ ss
    -> 4 <= S (S (S (S ss))) < aStackSize abs
    -> agree_range (stMem st) initialMem initialPc
    (actualCodeLen + longestInstr)
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> belowMax (sptr ctx) 4
    -> agree_range
    (stMem (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret))
    initialMem initialPc (actualCodeLen + longestInstr).
    intros abs ss st ctx ret' Hesp Hbound Hagree Hregs Hsptr Hbelow.
    replace (badd neg_four (get_reg32 st ESP))
      with (eval_addr (Build_address neg_four (Some ESP) None) st).
    eapply awrite_stack_agree_range; myeauto.
    replace (bsub32 (nat_to_bvec 32 ss) neg_four)
      with (badd (nat_to_bvec 32 ss) four).
    unfold badd, barith.
    unfold four.
    repeat (rewrite ntb_invert); try (inversion Hbelow; omega).
    unfold neg_four.
    unfold badd, bsub32, bsub.
    rewrite bneg_btn_double; trivial.

    unfold eval_addr, addrIndex, addrBase, addrDisp.
    trivial.
  Qed.

  Hint Resolve enter_agree_range : asm.

  Lemma enter_conc_regs : forall ss abs pc st ret ctx sls (args : list ty) ct cust,
    aget_reg abs ESP = Stack typ ss
    -> S (S (S (S ss))) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> conc_regs abs st ctx
    -> 4 * length args <= ss
    -> conc_regs
    (Build_astate (apcConst pc)
      (Reg32Map.upd (Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ (S (S (S (S (4 * length args))))))) EBP
	(OldEbp typ)) (aStackSize abs - (ss - 4 * length args))
      sls cust)
    (set_pc
      (set_reg32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret) ESP
	(bsub (get_reg32 st ESP) four)) pc)
    (Build_context (stack_slot ctx (ss - 4 * length args)) (get_reg32 st EBP)
      (stMem st) (Some ret) ct).
    intros ss abs pc st rt ctx sls args ct cust Hesp Hss Hsptr Hregs Hargs r.
    unfold aget_reg, GenericTypes.aget_reg, get_reg32; simpl.
    destruct (reg32_eq_dec EBP r); subst.

    rewrite Reg32Map.sel_eq.
    rewrite Reg32Map.sel_neq; try discriminate.
    replace (Reg32Map.sel (stRegs32 st) EBP)
      with (oldebp (Build_context (stack_slot ctx (ss - 4 * length args))
	(Reg32Map.sel (stRegs32 st) EBP) (stMem st) (Some rt) ct));
      trivial.
    myauto.

    destruct (reg32_eq_dec ESP r); subst.
    rewrite (Reg32Map.sel_neq (Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ (S (S (S (S (4 * length args))))))));
      try discriminate.
    repeat (rewrite Reg32Map.sel_eq).
    replace (bsub (Reg32Map.sel (stRegs32 st) ESP) four)
      with (stack_slot (Build_context (stack_slot ctx (ss - 4 * length args))
	(Reg32Map.sel (stRegs32 st) EBP)
        (stMem st) (Some rt) ct) (S (S (S (S (4 * length args)))))).
    unfold stack_slot; apply HasType_Stack.
    simpl sptr.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    add Hhigh code_high; add Hspace code_space; omega.
    add Hup (btn_upper (sptr ctx)); omega.
    unfold stack_slot, GenericTypes.stack_slot.
    unfold sptr at 1.
    rewrite ntb_invert.
    add Hr (Hregs ESP).
    unfold aget_reg in Hesp.
    rewrite Hesp in Hr.
    inversion Hr; subst.
    unfold get_reg32 in H.
    rewrite <- H.
    unfold stack_slot, GenericTypes.stack_slot.
    unfold four.
    rewrite bsub_bdec4.
    apply btn_inj.
    rewrite bdec4_btn.
    repeat (rewrite ntb_invert); trivial; try (add Hup (btn_upper (sptr ctx)); omega).
    rewrite ntb_invert; trivial; try (add Hup (btn_upper (sptr ctx)); omega).
    rewrite ntb_invert.
    omega.
    generalize (btn_upper (sptr ctx)); omega.
    generalize (btn_upper (sptr ctx)); omega.

    repeat (rewrite Reg32Map.sel_neq; auto with arith).
    rewrite Reg32Map.sel_empty.
    myauto.
  Qed.

  Hint Resolve enter_conc_regs : asm.

  Lemma enter_stackSize' : forall ss (abs : astate) ctx,
    S (S (S (S ss))) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> bvec_to_nat stackStart + (aStackSize abs - ss) <
    bvec_to_nat (stack_slot ctx ss).
    intros ss abs ctx Hss Hsptr.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    omega.
    generalize (btn_upper (sptr ctx)); omega.
  Qed.

  Lemma enter_stackSize : forall ss (abs : astate) ctx pc rs sls a b c (d : ctxt) args (cust : astat),
    S (S (S (S ss))) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> bvec_to_nat stackStart +
    aStackSize
    (Build_astate (typ := typ) pc rs (aStackSize abs - (ss - args)) sls cust) <
    bvec_to_nat
    (sptr
      (Build_context (stack_slot ctx (ss - args)) a b c d)).
    intros.
    simpl aStackSize; simpl sptr.
    apply enter_stackSize'; trivial.
    omega.
  Qed.

  Hint Resolve enter_stackSize : asm.

  Lemma stack_slot_neg_four : forall (abs : astate) ctx ss,
    S (S (S (S ss))) <= aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> badd neg_four (stack_slot ctx ss)
    = stack_slot ctx (S (S (S (S ss)))).
    intros abs ctx ss Hsize Hsptr.
    add Hbtn (btn_upper (sptr ctx)).
    replace (badd neg_four (stack_slot ctx ss))
      with (bsub (stack_slot ctx ss) four).
    assert (Habove : bvec_to_nat (stack_slot ctx ss) >= 4).
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; omega.
    unfold four.
    rewrite bsub_bdec4; try omega.
    apply btn_inj.
    rewrite bdec4_btn.
    unfold stack_slot, GenericTypes.stack_slot.
    repeat (rewrite ntb_invert); omega.
    omega.

    unfold badd, barith.
    rewrite plus_comm.
    unfold neg_four.
    trivial.
  Qed.

  Fixpoint argsStack (args : list ty) {struct args} : astack :=
    match args with
      | nil => NatMap.empty (Top typ)
      | cons ty rest => NatMap.upd (argsStack rest) (4 * length args) ty
    end.

  Definition robust_type (ty : ty) :=
    match ty with
      | Top => true
      | Const _ => true
      | Custom _ => true
      | _ => false
    end.

  Fixpoint checkArgs (args : list ty) (ss : nat) (sls : astack) {struct args} : bool :=
    match args with
      | nil => true
      | cons actual rest =>
	robust_type actual
	&& subType (NatMap.sel sls (ss + 4 * length args)) actual
	&& checkArgs rest ss sls
    end.

  Lemma btn_inj' : forall len (v1 v2 : Bvector len),
    bvec_to_nat v1 <> bvec_to_nat v2
    -> v1 <> v2.
    congruence.
  Qed.

  Lemma ntb_eqMod_inj : forall len n1 n2,
    ~eqMod (pow2 len) n1 n2
    -> nat_to_bvec len n1 <> nat_to_bvec len n2.
    intros len n1 n2 Hmod Heq.
    apply Hmod.
    apply ntb_mod'; trivial.
  Qed.

  Lemma ntb_O : forall len,
    bvec_to_nat (nat_to_bvec len 0) = 0.
    intros.
    rewrite ntb_invert; trivial.
    generalize (pow2_pos len); omega.
  Qed.

  Lemma btn_ntb_four : forall len,
    bvec_to_nat (nat_to_bvec (S (S (S len))) 4) = 4.
    intro.
    simpl.
    unfold mod2_bool, mod2, div2.
    destruct (Euclid.eucl_dev 2 two_gt_0 4).
    assert (q = 2); [omega | subst].
    assert (r = 0); [omega | subst].
    destruct (eq_nat_dec 0 0); try tauto.
    destruct (Euclid.eucl_dev 2 two_gt_0 2).
    assert (q = 1); [omega | subst].
    assert (r = 0); [omega | subst].
    destruct (eq_nat_dec 0 0); try tauto.
    destruct (Euclid.eucl_dev 2 two_gt_0 1).
    assert (q = 0); [omega | subst].
    assert (r = 1); [omega | subst].
    destruct (eq_nat_dec 1 0); try (wrong; omega).
    rewrite ntb_O.
    omega.
  Qed.

  Lemma enter_ret_neq : forall ss slot ctx len (abs : astate),
    len <= ss
    -> S (S (S (S ss))) < aStackSize abs
    -> 4 <= slot <= len
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> bvec_to_nat (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - (ss - len) - slot))
    >= bvec_to_nat (nat_to_bvec 32 (bvec_to_nat neg_four + bvec_to_nat (stack_slot ctx ss))) + 4.
    intros ss slot ctx len abs Hlen Hsize Hslot Hsize'.
    add Hbtn (btn_upper (sptr ctx)).
    assert (Heq : nat_to_bvec 32
      (bvec_to_nat neg_four + bvec_to_nat (stack_slot ctx ss))
      = nat_to_bvec 32
      (bvec_to_nat (stack_slot ctx ss) - bvec_to_nat four)).
    apply ntb_mod.
    rewrite plus_comm.
    unfold neg_four; apply eqMod_sym; apply bneg_btn_sub.
    unfold four; rewrite btn_ntb_four.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; omega.
    rewrite Heq.
    unfold four; rewrite btn_ntb_four.
    unfold stack_slot, GenericTypes.stack_slot.
    repeat (rewrite ntb_invert); try omega.
  Qed.

  Lemma enter_hasType : forall rt ctx v ctx',
    hasType ctx' v rt
    -> robust_type rt = true
    -> custom ctx = custom ctx'
    -> hasType ctx v rt.
    intros rt ctx v ctx' Hty Hrobust Hcust.
    destruct rt; try discriminate.
    myauto.
    inversion Hty; myauto.
    inversion Hty; subst; red; apply HasType_Custom; rewrite Hcust; trivial.
  Qed.

  Lemma ntb_inj : forall len n1 n2,
    n1 = n2
    -> nat_to_bvec len n1 = nat_to_bvec len n2.
    congruence.
  Qed.

  Lemma argsStack_aligned : forall k,
    1 <= k <= 3
    -> forall n args, NatMap.sel (argsStack args) (4 * n - k) = Top typ.
    intros k Hk.
    induction args; simpl.

    apply NatMap.sel_empty.

    rewrite undo_fourx_S.
    rewrite undo_fourx.
    destruct (eq_nat_dec (4 * S (length args)) (4 * n - k)).
    wrong; omega.
    
    rewrite NatMap.sel_neq; trivial.
  Qed.

  Lemma argsStack_robust : forall ss sls args slot,
    checkArgs args ss sls = true
    -> robust_type (NatMap.sel (argsStack args) slot) = true.
    induction args; simpl; intros slot Hcheck.

    rewrite NatMap.sel_empty; trivial.

    rewrite undo_fourx_S.
    rewrite undo_fourx_S in Hcheck.
    destruct (eq_nat_dec (4 * S (length args)) slot); subst.

    rewrite NatMap.sel_eq.
    bool_elim; trivial.

    rewrite NatMap.sel_neq; trivial.
    bool_elim; eauto.
  Qed.

  Lemma enter_checkArgs : forall ss abs st args c d ret slot ctx,
    aget_reg abs ESP = Stack typ ss
    -> conc_regs abs st ctx
    -> 4 * length args <= ss
    -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
    -> 4 <= slot <= 4 * length args
    -> S (S (S (S ss))) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> conc_slots abs st ctx
    -> hasType
    (Build_context (stack_slot ctx (ss - 4 * length args)) c d (Some ret) (custom ctx))
    (get_mem32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret)
      (stack_slot
	(Build_context (stack_slot ctx (ss - 4 * length args)) c d
          (Some ret) (custom ctx)) slot)) (NatMap.sel (argsStack args) slot).
    intros ss abs st args c d rt slot ctx Hesp Hregs Hlen Hargs Hslot Hsize Hsptr Hslots.
    add Hbtn (btn_upper (sptr ctx)).

    add Hr (Hregs ESP).
    unfold aget_reg in Hesp.
    rewrite Hesp in Hr.
    inversion Hr; subst.
    
    rewrite invert_set_mem32_neq.

    replace (stack_slot
      (Build_context (stack_slot ctx (ss - 4 * length args)) c d
	(Some rt) (custom ctx)) slot)
    with (stack_slot ctx (ss - 4 * length args + slot)).
    apply enter_hasType with ctx.

    clear Hesp Hr H2 H.
    generalize dependent ss.
    induction args; intros ss Hss Hcheck Hsize; simpl argsStack.

    rewrite NatMap.sel_empty.
    myauto.

    simpl length.
    rewrite undo_fourx_S.

    simpl in Hcheck.
    rewrite undo_fourx_S in Hcheck.
    bool_elim.

    destruct (eq_nat_dec (4 * S (length args)) slot); subst.

    rewrite NatMap.sel_eq.
    apply subType_sound
      with (NatMap.sel (aStackSlots abs)
	(ss - 4 * S (length args) + 4 * S (length args))); trivial.
    red in Hslots; unfold aget_stack, GenericTypes.aget_stack in Hslots.
    unfold hasType, stack_slot.
    red in Hslots.
    unfold GenericTypes.aget_stack in Hslots.
    apply Hslots.
    omega.

    rewrite NatMap.sel_neq; trivial.
    destruct (eq_nat_dec (4 * S (length args) - 1) slot);
      [subst; rewrite argsStack_aligned; [myauto | omega]
	| idtac].
    destruct (eq_nat_dec (4 * S (length args) - 2) slot);
      [subst; rewrite argsStack_aligned; [myauto | omega]
	| idtac].
    destruct (eq_nat_dec (4 * S (length args) - 3) slot);
      [subst; rewrite argsStack_aligned; [myauto | omega]
	| idtac].
    assert (Heq : ss - 4 * S (length args)
      = (ss - 4) - 4 * length args).
    omega.
    rewrite Heq.
    simpl length in Hslot, Hss.
    apply IHargs; try (rewrite <- Heq; trivial; fail); omega.

    eapply argsStack_robust; eauto.

    trivial.
    unfold stack_slot, GenericTypes.stack_slot.
    unfold sptr at 2.
    rewrite ntb_invert.
    apply ntb_inj.
    omega.
    omega.

    split.
    unfold badd, barith.
    rewrite plus_comm.
    unfold neg_four.
    assert (Heq : nat_to_bvec 32 (bvec_to_nat (get_reg32 st ESP) + bvec_to_nat (bneg four))
      = nat_to_bvec 32 (bvec_to_nat (get_reg32 st ESP) - bvec_to_nat four)).
    apply ntb_mod.
    apply eqMod_sym; apply bneg_btn_sub.
    rewrite <- H.
    unfold stack_slot, GenericTypes.stack_slot.
    unfold four; rewrite btn_ntb_four.
    rewrite ntb_invert; omega.
    rewrite <- H in Heq.
    rewrite Heq.
    assert (Hlt : bvec_to_nat (GenericTypes.stack_slot ctx ss) - bvec_to_nat four < pow2 32 - 3).
    unfold GenericTypes.stack_slot.
    rewrite ntb_invert; omega.
    unfold GenericTypes.stack_slot.
    unfold GenericTypes.stack_slot in Hlt.
    rewrite ntb_invert; omega.

    split.
    unfold stack_slot, GenericTypes.stack_slot.
    unfold sptr at 1.
    repeat (rewrite ntb_invert); omega.

    apply apart_higher.
    unfold stack_slot, GenericTypes.stack_slot.
    unfold sptr at 1.
    rewrite ntb_invert.
    
    replace (bvec_to_nat
      (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - (ss - 4 * length args))) - slot)
    with (bvec_to_nat (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - (ss - 4 * length args) - slot))).
    replace (bvec_to_nat (badd neg_four (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - ss))))
      with (bvec_to_nat (nat_to_bvec 32 (bvec_to_nat neg_four + bvec_to_nat (stack_slot ctx ss)))); trivial.
    eapply enter_ret_neq; eauto.
    rewrite ntb_invert.
    rewrite ntb_invert.
    trivial.
    omega.
    omega.
    rewrite ntb_invert; omega.
  Qed.

  Lemma argsStack_cleared : forall args slot,
    4 * length args < slot
    -> NatMap.sel (argsStack args) slot = Top typ.
    induction args; simpl; intros slot Hslot.
    apply NatMap.sel_empty.
    rewrite undo_fourx_S in Hslot.
    rewrite undo_fourx_S.
    destruct (eq_nat_dec (4 * S (length args)) slot); subst.
    wrong; omega.
    rewrite NatMap.sel_neq; trivial.
    apply IHargs.
    omega.
  Qed.

  Lemma enter_conc_slots' : forall abs pc rs ss st ret c d ctx args cust,
    pcInRange ret
    -> aget_reg abs ESP = Stack typ ss
    -> conc_regs abs st ctx
    -> S (S (S (S ss))) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> 4 * length args <= ss
    -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
    -> conc_slots abs st ctx
    -> conc_slots
    (Build_astate pc rs (aStackSize abs - (ss - 4 * length args))
      (NatMap.upd (argsStack args) (S (S (S (S (4 * length args))))) (RetPtr typ)) cust)
    (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret)
    (Build_context (stack_slot ctx (ss - 4 * length args)) c d (Some ret) (custom ctx)).
    intros until cust; intros Hret Hesp Hregs Hss Hsize Hargs HcheckArgs Hslots slot Hslot.
    simpl in Hslot.
    rewrite undo_fourx in Hslot.
    unfold aget_stack, GenericTypes.aget_stack; simpl.
    destruct (eq_nat_dec slot (S (S (S (S (4 * length args)))))); subst.
    rewrite NatMap.sel_eq.

    rewrite undo_fourx.
    replace (GenericTypes.stack_slot (Build_context (stack_slot ctx (ss - 4 * length args))
      c d (Some ret0) (custom ctx)) (S (S (S (S (4 * length args))))))
    with (GenericTypes.stack_slot ctx (S (S (S (S ss))))).
    add Hr (Hregs ESP).
    unfold aget_reg in Hesp.
    rewrite Hesp in Hr.
    inversion Hr; subst.
    rewrite (stack_slot_neg_four abs ctx); trivial; try omega.
    unfold stack_slot.
    rewrite invert_set_mem32.
    myauto.

    unfold stack_slot, GenericTypes.stack_slot.
    unfold sptr at 2.
    rewrite ntb_invert.
    assert (bvec_to_nat (sptr ctx) - S (S (S (S ss)))
      = bvec_to_nat (sptr ctx) - (ss - 4 * length args) - S (S (S (S (4 * length args))))).
    omega.
    congruence.
    generalize (btn_upper (sptr ctx)); omega.
    rewrite NatMap.sel_neq; auto with arith.
    rewrite undo_fourx.

    destruct (eq_nat_dec (4 * S (length args) - 1) slot);
      [subst; rewrite argsStack_aligned; try omega; myauto
	| idtac].
    destruct (eq_nat_dec (4 * S (length args) - 2) slot);
      [subst; rewrite argsStack_aligned; try omega; myauto
	| idtac].
    destruct (eq_nat_dec (4 * S (length args) - 3) slot);
      [subst; rewrite argsStack_aligned; try omega; myauto
	| idtac].
    destruct (le_lt_dec slot (4 * length args)).
    eapply enter_checkArgs; eauto.
    tauto.
    rewrite argsStack_cleared; myauto.
  Qed.

  Lemma get_mem32_pc : forall st pc addr,
    get_mem32 (set_pc st pc) addr = get_mem32 st addr.
    intros.
    unfold get_mem32.
    unfold get_mem.
    trivial.
  Qed.

  Lemma get_mem32_reg32 : forall st r v addr,
    get_mem32 (set_reg32 st r v) addr = get_mem32 st addr.
    intros.
    unfold get_mem32.
    unfold get_mem.
    trivial.
  Qed.

  Hint Rewrite get_mem32_pc get_mem32_reg32 : asm.

  Lemma conc_slots_spc : forall abs st ctx pc,
    conc_slots abs st ctx
    -> conc_slots abs (set_pc st pc) ctx.
    intuition.
  Qed.

  Lemma conc_slots_sreg : forall abs st ctx r v,
    conc_slots abs st ctx
    -> conc_slots abs (set_reg32 st r v) ctx.
    intuition.
  Qed.

  Hint Resolve conc_slots_spc conc_slots_sreg : asm.

  Lemma enter_conc_slots : forall abs pc rs ss st ret c d ctx args cust,
    pcInRange ret
    -> aget_reg abs ESP = Stack typ ss
    -> conc_regs abs st ctx
    -> S (S (S (S ss))) < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> 4 * length args <= ss
    -> conc_slots abs st ctx
    -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
    -> conc_slots
    (Build_astate (apcConst pc) rs (aStackSize abs - (ss - 4 * length args))
      (NatMap.upd (argsStack args) (S (S (S (S (4 * length args))))) (RetPtr typ)) cust)
    (set_pc
      (set_reg32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret) ESP
	(bsub (get_reg32 st ESP) four)) pc)
    (Build_context (stack_slot ctx (ss - 4 * length args)) c d (Some ret) (custom ctx)).
    intros.
    apply conc_slots_spc; apply conc_slots_sreg;
      eapply enter_conc_slots'; eauto.
  Qed.

  Hint Resolve enter_conc_slots : asm.

  Lemma enter_belowMax : forall ctx ss b c d (e : ctxt) (args : list ty),
    belowMax (sptr ctx) 4
    -> belowMax (sptr (Build_context (stack_slot ctx (ss - 4 * length args)) b c d e)) 4.
    simpl; myeauto.
  Qed.

  Hint Resolve enter_belowMax : asm.

  Lemma enter_memIn' : forall ss abs st ctx ret args,
    aget_reg abs ESP = Stack typ ss
    -> S (S (S (S ss))) < aStackSize abs
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> belowMax (sptr ctx) 4
    -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
    -> args <= ss
    -> agree_range'
    (stMem (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret))
    (stMem st) (stack_slot ctx (ss - args)) stackEnd.
    intros ss abs st ctx rt args Hss Hsize Hregs Hsptr Hbelow Hagree Hargs addr Hbound.
    add Hesp (Hregs ESP).
    unfold aget_reg in Hss.
    rewrite Hss in Hesp.
    inversion Hesp; subst.
    assert (Heq : bvec_to_nat (stack_slot ctx (ss - args)) = bvec_to_nat (sptr ctx) - ss + args).
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; trivial.
    generalize (btn_upper (sptr ctx)); omega.
    generalize (btn_upper (sptr ctx)); omega.
    rewrite Heq in Hbound.
    unfold set_mem32, set_mem.
    do 4 (unfold stMem at 1).
    rewrite (stack_slot_neg_four abs ctx); trivial; try omega.
    assert (Heq2 : bvec_to_nat (stack_slot ctx (S (S (S (S ss))))) = bvec_to_nat (sptr ctx) - S (S (S (S ss)))).
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert; trivial.
    generalize (btn_upper (sptr ctx)); omega.
    repeat (rewrite Int32Map.sel_neq); trivial; apply (@btn_inj' 32).
    omega.
    rewrite binc_btn_belowMax; [omega | apply belowMax_dec with 4; myauto].
    rewrite binc2_btn_belowMax; [omega | apply belowMax_dec with 4; myauto].
    rewrite binc3_btn_belowMax; [omega | apply belowMax_dec with 4; myauto].
  Qed.

  Lemma enter_memIn : forall ss abs st ctx pc b c d f g h ret args,
    aget_reg abs ESP = Stack typ ss
    -> 4 <= S (S (S (S ss))) < aStackSize abs
    -> conc_regs abs st ctx
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> belowMax (sptr ctx) 4
    -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
    -> args <= ss
    -> agree_range'
    (stMem
      (set_pc
	(set_reg32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret) ESP
          (bsub (get_reg32 st ESP) four)) pc))
    (memIn
      (Build_context b c (stMem st) d (custom ctx)))
    (sptr (Build_context (stack_slot ctx (ss - args)) f g h (custom ctx))) stackEnd.
    intros.
    add H' (@enter_memIn' ss abs st ctx ret0).
    simpl memIn; simpl sptr.
    apply H'; auto; omega.
  Qed.

  Hint Resolve enter_memIn : asm.

  Lemma enter_esp : forall ctx ss b c d size args,
    S (S (S (S ss))) < size
    -> bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
    -> bvec_to_nat stackStart + size < bvec_to_nat (sptr ctx)
    -> bvec_to_nat
    (sptr
      (Build_context (stack_slot ctx (ss - args)) b c d (custom ctx))) <= bvec_to_nat initialESP.
    intros ctx ss b c d size args Hss Hesp Hsptr.
    simpl sptr.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    omega.
    generalize (btn_upper (sptr ctx)); omega.
  Qed.

  Hint Resolve enter_esp : asm.

  Definition leave_context st ctx ss (args : list ty) :=
    Build_context (stack_slot ctx (ss - 4 * length args))
    (get_reg32 st EBP) (stMem st) (Some (stPc st))
    (custom ctx).

  Definition enter : forall abs ret pc ss args cust,
    pcInRange pc
    -> pcInRange ret
    -> aget_reg abs ESP = Stack typ ss
    -> S (S (S (S ss))) < aStackSize abs
    -> 4 * length args <= ss
    -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
    -> sig_option (fun abs' =>
      forall st ctx, conc abs st ctx
	-> pconc (custom ctx) cust (set_pc
	  (set_reg32
	    (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret)
	    ESP (bsub (get_reg32 st ESP) four))
	  pc)
	-> conc abs' (set_pc
	  (set_reg32
	    (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret)
	    ESP (bsub (get_reg32 st ESP) four))
	  pc)
	(leave_context (set_pc st ret) ctx ss args)).
    intros abs rt pc ss args cust Hpc Hret Hesp Hsize Hargs Hcheck.
    apply SSome with (Build_astate (apcConst pc)
      (Reg32Map.upd
	(Reg32Map.upd
	  (Reg32Map.empty (Top typ))
	  ESP (Stack typ (S (S (S (S (4 * length args)))))))
	EBP (OldEbp typ)) (aStackSize abs - (ss - 4 * length args))
      (NatMap.upd
	(argsStack args)
	(S (S (S (S (4 * length args))))) (RetPtr typ))
      cust).
    intros st ctx Hconc Hpconc.
    unfold leave_context.
    simpl.
    replace (get_reg32 (set_pc st rt) EBP)
      with (get_reg32 st EBP); trivial.
    assert (Hsize' : 4 <= S (S (S (S ss))) < aStackSize abs).
    omega.
    inversion Hconc; inversion concMain0; unfold leave_context; split; [split | idtac];
      try rewrite undo_fourx; myeauto.
    red; simpl; trivial.
  Qed.


  (** * Switch point of view from callee to caller.
    Use this on the state right before a RET instruction.
    *)

  Definition leave_abs (abs : astate) retType args := Build_astate
    apcRet
    (Reg32Map.upd
      (Reg32Map.upd
	(Reg32Map.upd
	  (Reg32Map.empty (Top typ))
	  EAX retType)
	ESP (Stack typ args))
      EBP (OldEbp typ)) 0
    (NatMap.empty (Top typ))
    (aCustom abs).

  Fixpoint keepValidSlots (n : nat) (stk : astack) {struct n} : astack :=
    match n with
      | O => NatMap.empty (Top typ)
      | S n => NatMap.upd (keepValidSlots n stk) n (NatMap.sel stk n)
    end.

  (** Forget information across a function call *)
  Definition ret_abs retType abs ss args := Build_astate
    (aPc abs)
    (Reg32Map.upd
      (Reg32Map.upd
	(Reg32Map.upd
	  (Reg32Map.empty (Top typ))
	  EAX retType)
	ESP (aget_reg abs ESP))
      EBP (aget_reg abs EBP))
    (aStackSize abs)
    (keepValidSlots (S (ss - args)) (aStackSlots abs))
    (aCustom abs).

  Lemma ret_regs : forall abs st ctx rt ss args,
    conc_regs abs st ctx
    -> hasType ctx (get_reg32 st EAX) rt
    -> conc_regs (ret_abs rt abs ss args) st ctx.
    intros abs st ctx rt ss args Hregs Heax r.
    unfold aget_reg, GenericTypes.aget_reg; simpl.

    destruct (reg32_eq_dec EBP r); subst.
    rewrite Reg32Map.sel_eq.
    auto.
    rewrite Reg32Map.sel_neq; auto.

    destruct (reg32_eq_dec ESP r); subst.
    rewrite Reg32Map.sel_eq.
    auto.
    rewrite Reg32Map.sel_neq; auto.
    
    destruct (reg32_eq_dec EAX r); subst.
    rewrite Reg32Map.sel_eq.
    trivial.
    rewrite Reg32Map.sel_neq; auto.

    rewrite Reg32Map.sel_empty.
    myauto.
  Qed.

  Hint Resolve ret_regs : asm.

  Lemma ret_slots' : forall abs n st ctx pc a cust,
    conc_slots abs st ctx
    -> conc_slots (Build_astate pc a (aStackSize abs) (keepValidSlots n (aStackSlots abs)) cust) st ctx.
    induction n; intros st ctx pc a cust Hslots sl Hsl; unfold aget_stack, GenericTypes.aget_stack; simpl.
    rewrite NatMap.sel_empty.
    myauto.

    simpl in Hsl.
    destruct (eq_nat_dec n sl); subst.
    rewrite NatMap.sel_eq.
    red in Hslots.
    unfold aget_stack, GenericTypes.aget_stack in Hslots.
    red in Hslots.
    unfold GenericTypes.aget_stack in Hslots.
    auto.
    
    rewrite NatMap.sel_neq; trivial.
    apply (IHn st ctx pc a cust Hslots sl); trivial.
  Qed.

  Lemma ret_slots : forall abs st ctx rt ss args,
    conc_slots abs st ctx
    -> conc_slots (ret_abs rt abs ss args) st ctx.
    intros.
    unfold ret_abs.
    apply ret_slots'; trivial.
  Qed.

  Hint Resolve ret_slots : asm.

  Lemma leave_pc : forall abs st ctx rt ss st' pc args,
    aPc abs = apcConst pc
    -> conc' abs st ctx
    -> conc' (leave_abs abs rt (4 * length args)) st' (leave_context st ctx ss args)
    -> conc_pc (ret_abs rt abs ss (4 * length args)) st' ctx.
    intros abs st ctx rt ss st' pc args Hpc Hconc Hconc'.
    add Hcpc (concPc Hconc).
    add Hcpc' (concPc Hconc').
    red in Hcpc, Hcpc'.
    simpl in Hcpc'.
    rewrite Hpc in Hcpc.
    red; red.
    simpl.
    rewrite Hpc.
    congruence.
  Qed.

  Hint Resolve leave_pc : asm.

  Lemma leave_hasType : forall rt st ctx ss v args,
    hasType (leave_context st ctx ss args) v rt
    -> robust_type rt = true
    -> hasType ctx v rt.
    intros.
    eapply enter_hasType; eauto.
  Qed.

  Hint Resolve leave_hasType : asm.

  Lemma hasType_generic : forall ctx v t,
    hasType ctx v t
    -> GenericTypes.hasType hasTyp ctx v t.
    intuition.
  Qed.

  Hint Resolve hasType_generic : asm.

  Lemma leave_conc_regs : forall abs st ctx rt ss st' args,
    aget_reg abs ESP = Stack typ ss
    -> ss < aStackSize abs
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> conc' abs st ctx
    -> conc' (leave_abs abs rt (4 * length args)) st' (leave_context st ctx ss args)
    -> conc_regs (leave_abs abs rt (4 * length args)) st' (leave_context st ctx ss args)
    -> robust_type rt = true
    -> 4 * length args <= ss
    -> conc_regs (ret_abs rt abs ss (4 * length args)) st' ctx.
    intros abs st ctx rt ss st' args Hesp Hss Hsize Hconc Hconc' Hregs Hrobust Hargs r.
    unfold aget_reg, GenericTypes.aget_reg; simpl.

    destruct (reg32_eq_dec EBP r); subst.
    rewrite Reg32Map.sel_eq.
    add Hebp' (Hregs EBP).
    unfold aget_reg, GenericTypes.aget_reg in Hebp'; simpl in Hebp'.
    rewrite Reg32Map.sel_eq in Hebp'.
    inversion Hebp'; subst.
    add Hregs' (concRegs Hconc).
    auto.

    unfold aget_reg, GenericTypes.aget_reg.
    rewrite Reg32Map.sel_neq; trivial.
    destruct (reg32_eq_dec ESP r); subst.
    rewrite Reg32Map.sel_eq.
    add Hesp' (Hregs ESP).
    unfold aget_reg, GenericTypes.aget_reg in Hesp'; simpl in Hesp'.
    rewrite Reg32Map.sel_neq in Hesp'; trivial.
    rewrite Reg32Map.sel_eq in Hesp'.
    inversion Hesp'; subst.
    rewrite undo_fourx.
    replace (GenericTypes.stack_slot (leave_context st ctx ss args) (4 * length args))
      with (GenericTypes.stack_slot ctx ss).
    unfold aget_reg, GenericTypes.aget_reg in Hesp.
    rewrite Hesp; myauto.
    apply HasType_Stack; omega.
    unfold stack_slot, GenericTypes.stack_slot; simpl sptr.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    apply ntb_inj.
    omega.
    generalize (btn_upper (sptr ctx)); omega.

    rewrite Reg32Map.sel_neq; trivial.
    destruct (reg32_eq_dec EAX r); subst.
    rewrite Reg32Map.sel_eq.
    assert (rt = aget_reg (leave_abs abs rt (4 * length args)) EAX).
    unfold aget_reg, GenericTypes.aget_reg; simpl.
    do 2 (rewrite Reg32Map.sel_neq; try discriminate);
      rewrite Reg32Map.sel_eq; trivial.
    rewrite H.
    rewrite H in Hrobust.
    apply hasType_generic.
    myeauto.
    
    rewrite Reg32Map.sel_neq; trivial.
    rewrite Reg32Map.sel_empty; myauto.
  Qed.

  Hint Resolve leave_conc_regs : asm.

  Lemma leave_stackSize : forall abs st ctx rt ss args,
    conc' abs st ctx
    -> bvec_to_nat stackStart + aStackSize (ret_abs rt abs ss args) < bvec_to_nat (sptr ctx).
    intros abs st ctx rt ss args Hconc.
    inversion Hconc.
    trivial.
  Qed.

  Hint Resolve leave_stackSize : asm.

  Lemma agree_range'_read32 : forall st1 st2 start finish addr,
    agree_range' (stMem st1) (stMem st2) start finish
    -> bvec_to_nat start <= bvec_to_nat addr < bvec_to_nat finish - 4
    -> belowMax finish 0
    -> get_mem32 st1 addr = get_mem32 st2 addr.
    intros st1 st2 start finish addr Hrange Haddr Hbelow.
    assert (Hbelow' : belowMax addr 4).
    destruct Hbelow; split; omega.
    clear Hbelow; rename Hbelow' into Hbelow.
    red in Hrange.
    unfold get_mem32, get_mem.
    repeat (rewrite Hrange); trivial.
    replace (binc (binc (binc addr))) with (binc3 addr); trivial;
      rewrite binc3_btn_belowMax;
	[omega
	  | apply belowMax_dec with 4; trivial; omega].
    replace (binc (binc addr)) with (binc2 addr); trivial;
      rewrite binc2_btn_belowMax;
	[omega
	  | apply belowMax_dec with 4; trivial; omega].
    rewrite binc_btn_belowMax;
      [omega
	| apply belowMax_dec with 4; trivial; omega].
    omega.
  Qed.

  Lemma ret_abs_stack' : forall ss sl,
    sl < ss
    -> forall slots, NatMap.sel slots sl = NatMap.sel (keepValidSlots ss slots) sl.
    induction ss; intros sl Hle slots; simpl.
    rewrite NatMap.sel_empty; wrong; omega.

    destruct (eq_nat_dec ss sl); subst.
    rewrite NatMap.sel_eq; trivial.
    rewrite NatMap.sel_neq; trivial.
    simpl in IHss.
    apply IHss.
    omega.
  Qed.

  Lemma ret_abs_stack : forall rt ss sl args,
    sl <= ss - args
    -> forall abs, aget_stack abs sl = aget_stack (ret_abs rt abs ss args) sl.
    intros rt ss sl args Hss abs.
    unfold aget_stack, GenericTypes.aget_stack.
    replace (aStackSlots (ret_abs rt abs ss args))
      with (keepValidSlots (S (ss - args)) (aStackSlots abs)); trivial.
    apply ret_abs_stack'; omega.
  Qed.

  Lemma ret_abs_stack_not' : forall ss sl,
    sl >= ss
    -> forall slots, NatMap.sel (keepValidSlots ss slots) sl = Top typ.
    induction ss; intros sl Hle slots; simpl.
    apply NatMap.sel_empty.
    rewrite NatMap.sel_neq; trivial.
    apply IHss; omega.
    omega.
  Qed.

  Lemma ret_abs_stack_not : forall rt ss sl args,
    sl > ss - args
    -> forall abs, aget_stack (ret_abs rt abs ss args) sl = Top typ.
    intros rt ss sl args Hss abs.
    unfold aget_stack, GenericTypes.aget_stack.
    replace (aStackSlots (ret_abs rt abs ss args))
      with (keepValidSlots (S (ss - args)) (aStackSlots abs)); trivial.
    rewrite ret_abs_stack_not'; auto with arith.
  Qed.

  Lemma leave_slots : forall abs st ctx rt ss st' args,
    ss < aStackSize abs
    -> conc' abs st ctx
    -> conc' (leave_abs abs rt (4 * length args)) st' (leave_context st ctx ss args)
    -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
    -> bvec_to_nat stackStart + aStackSize (leave_abs abs rt (4 * length args)) <
    bvec_to_nat (sptr (leave_context st ctx ss args))
    -> conc_slots abs st ctx
    -> conc_slots (leave_abs abs rt (4 * length args)) st' (leave_context st ctx ss args)
    -> agree_range' (stMem st') (memIn (leave_context st ctx ss args))
    (sptr (leave_context st ctx ss args)) stackEnd
    -> bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
    -> conc_slots (ret_abs rt abs ss (4 * length args)) st' ctx.
    intros abs st ctx rt ss st' args Hss Hconc Hconc' Hsize Hsize' Hslots Hslots' Hagree Hbelow sl Hsl.
    simpl in Hsl.
    destruct (dec_le sl (ss - 4 * length args)).

    replace (GenericTypes.aget_stack (ret_abs rt abs ss (4 * length args)) sl)
      with (GenericTypes.aget_stack abs sl).
    simpl sptr in Hagree; simpl memIn in Hagree.
    rewrite (@agree_range'_read32 st' st (stack_slot ctx (ss - 4 * length args)) stackEnd);
      myauto.
    
    unfold stack_slot, GenericTypes.stack_slot.
    repeat (rewrite ntb_invert); try (generalize (btn_upper (sptr ctx)); omega).
    generalize initialESP_stackEnd; omega.
    split.
    unfold stackEnd.
    rewrite ntb_invert;
      generalize stack_fits; omega.
    apply ret_abs_stack; omega.
    
    rewrite ret_abs_stack_not; myauto.
    omega.
  Qed.

  Hint Resolve leave_slots : asm.

  Lemma leave_memIn : forall st st' ctx ss args,
    bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
    -> agree_range' (stMem st') (memIn (leave_context st ctx ss args))
    (sptr (leave_context st ctx ss args)) stackEnd
    -> agree_range' (stMem st) (memIn ctx) 
    (sptr ctx) stackEnd
    -> agree_range' (stMem st') (memIn ctx) (sptr ctx) stackEnd.
    intros st st' ctx ss args Hsptr Hagree Hagree'.
    simpl in Hagree.
    intros addr Haddr.
    transitivity (Int32Map.sel (stMem st) addr).

    apply Hagree.
    unfold stack_slot, GenericTypes.stack_slot.
    rewrite ntb_invert.
    omega.
    generalize (btn_upper (sptr ctx)); omega.
    
    myauto.
  Qed.

  Hint Resolve leave_memIn : asm.

  Definition leave' : forall (abs : astate) (rt : ty) ss args,
    ss < aStackSize abs
    -> aget_reg abs ESP = Stack typ ss
    -> 4 * length args <= ss
    -> sig_option (fun abss =>
      forall st ctx, conc' abs st ctx
	-> menter conc (leave_context st ctx ss args) (snd abss) ctx (fst abss)).
    intros abs rt ss args Hss Hesp Hargs.

    caseEq (aPc abs); [idtac | type_fail].
    intros pc Hpc.
    destrEq Hrobust (robust_type rt); [idtac | type_fail].
    apply SSome with (ret_abs rt abs ss (4 * length args), leave_abs abs rt (4 * length args)).
    intros st ctx Hconc st' Hconc'.
    simpl; rewrite undo_fourx.
    inversion Hconc; inversion Hconc'; inversion concMain0.
    split; [split; try (eapply leave_slots; eauto) | idtac]; myeauto.
  Qed.

  Lemma conc'_pc' : forall abs st ctx pc,
    conc' abs st ctx
    -> pcInRange pc
    -> conc' (aset_pc abs pc) (set_pc st pc) ctx.
    intros abs st ctx pc Hconc Hpc.
    inversion Hconc; split; intuition.
  Qed.

  Hint Resolve conc'_pc' : asm.

  Definition leave : forall abs pc (rt : ty) ss args,
    aget_reg abs ESP = Stack typ ss
    -> S (S (S (S ss))) < aStackSize abs
    -> 4 * length args <= ss
    -> sig_option (fun abss =>
      forall st ctx, conc' abs st ctx
	-> menter conc
	(leave_context
	  (set_pc st pc) ctx ss args) (snd abss) ctx (fst abss)).
    intros abs pc rt ss args Hesp Hsize Hargs.
    destruct (pcInRange_dec pc); [idtac | type_fail].
    destruct (@leave' (aset_pc abs pc) rt ss args); trivial.
    simpl; omega.

    apply SSome with x.
    intros st ctx Hconc.
    apply m; myauto.

    type_fail.
  Qed.


  (** * Creating the successor state of a RET instruction *)

  Lemma ret_conc_pc : forall abs res st' st ctx cust,
    conc' abs st ctx
    -> hasType ctx
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
      st) (RetPtr typ)
    -> conc_pc
    (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
      (aStackSize abs) (aStackSlots abs) cust)
    (set_pc st'
      (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
    ctx.
    intros abs res st' st ctx cust Hconc Hret.
    red; red; simpl.
    replace (ret ctx)
      with (Some (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st));
      myauto.
  Qed.

  Hint Resolve ret_conc_pc : asm.

  Lemma ret_pcInRange : forall abs st ctx st',
    conc' abs st ctx
    -> hasType ctx
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
      st) (RetPtr typ)
    -> pcInRange
    (stPc
      (set_pc st'
	(eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
          st))).
    intros abs st ctx st' Hconc Hret.
    myeauto.
  Qed.

  Hint Resolve ret_pcInRange : asm.

  Lemma ret_agree : forall e st st' pc,
    write_genop32 (Reg ESP) e st st'
    -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
    -> agree_range
    (stMem
      (set_pc st' pc)) initialMem initialPc (actualCodeLen + longestInstr).
    intros e st st' pc Hwrite Hagree.
    inversion Hwrite; subst.
    trivial.
  Qed.

  Hint Resolve ret_agree : asm.

  Lemma conc_regs_sel : forall abs st ctx r,
    conc' abs st ctx
    -> GenericTypes.hasType hasTyp ctx (Reg32Map.sel (stRegs32 st) r)
    (Reg32Map.sel (aRegs abs) r).
    intros abs st ctx r Hconc.
    inversion Hconc.
    red in concRegs0.
    unfold get_reg32, GenericTypes.aget_reg in concRegs0.
    auto.
  Qed.

  Hint Resolve conc_regs_sel : asm.

  Lemma ret_conc_regs : forall abs st ctx res st' esp cust,
    conc' abs st ctx
    -> hasType ctx (eval_genop32 (Reg ESP) st) esp
    -> (forall (v1 v2 : int32),
      hasType ctx v1 (Const typ four) ->
      hasType ctx v2 esp -> hasType ctx (eval_arith Add v1 v2) res)
    -> write_genop32 (Reg ESP)
    (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
    -> conc_regs abs st ctx
    -> conc_regs
    (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
      (aStackSize abs) (aStackSlots abs) cust)
    (set_pc st'
      (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
    ctx.
    intros abs st ctx res st' esp cust Hconc Hesp Hres Hwrite Hregs r.
    simpl.
    unfold get_reg32, aget_reg, GenericTypes.aget_reg; simpl.

    destruct (reg32_eq_dec ESP r); subst.
    rewrite Reg32Map.sel_eq.
    inversion Hwrite; subst.
    simpl.
    rewrite Reg32Map.sel_eq.
    unfold eval_arith in Hres.
    replace (badd32 (get_reg32 st ESP) four)
      with (badd32 four (get_reg32 st ESP)).
    apply Hres; trivial.
    myauto.
    unfold eval_genop32 in Hesp; auto.
    unfold badd32, badd, barith.
    assert (bvec_to_nat four + bvec_to_nat (get_reg32 st ESP) =
      bvec_to_nat (get_reg32 st ESP) + bvec_to_nat four).
    omega.
    congruence.
    
    rewrite Reg32Map.sel_neq; trivial.
    inversion Hwrite; subst.
    simpl.
    rewrite Reg32Map.sel_neq; trivial.
    red in Hregs; unfold get_reg32, aget_reg, GenericTypes.aget_reg in Hregs; myeauto.
  Qed.

  Hint Resolve ret_conc_regs : asm.

  Lemma conc_slots_sel : forall (abs : astate) st ctx sl,
    4 <= sl < aStackSize abs
    -> GenericTypes.conc_slots hasTyp abs st ctx
    -> GenericTypes.hasType hasTyp ctx (get_mem32 st (stack_slot ctx sl))
    (NatMap.sel (aStackSlots abs) sl).
    intros abs st ctx sl Hsl Hconc.
    red in Hconc.
    unfold GenericTypes.aget_stack in Hconc.
    auto.
  Qed.

  Hint Resolve conc_slots_sel : asm.

  Lemma ret_conc_slots : forall abs st ctx st' res cust,
    write_genop32 (Reg ESP)
    (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
    -> conc_slots abs st ctx
    -> conc_slots
    (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
      (aStackSize abs) (aStackSlots abs) cust)
    (set_pc st'
      (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
    ctx.
    intros abs st ctx st' res cust Hwrite Hslots sl Hsl.
    simpl.
    unfold aget_stack, GenericTypes.aget_stack.
    simpl.
    inversion Hwrite; subst.
    replace (get_mem32
      (set_pc
	(set_reg32 st ESP
          (eval_arith Add four (eval_genop32 (Reg ESP) st)))
	(get_mem32 st (eval_addr (Build_address bzero32 (Some ESP) None) st)))
      (GenericTypes.stack_slot ctx sl))
    with (get_mem32 st (stack_slot ctx sl)); trivial.
    red in Hslots.
    unfold aget_stack, GenericTypes.aget_stack in Hslots.
    myauto.
  Qed.

  Hint Resolve ret_conc_slots : asm.

  Lemma ret_memIn : forall st st' (ctx : context),
    write_genop32 (Reg ESP)
    (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
    -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
    -> agree_range'
    (stMem
      (set_pc st'
	(eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
          st))) (memIn ctx) (sptr ctx) stackEnd.
    intros st st' ctx Hwrite Hagree.
    inversion Hwrite; subst.
    trivial.
  Qed.

  Hint Resolve ret_memIn : asm.

  Theorem ret_conc' : forall abs st ctx esp st' res nextPc cust,
    hasType ctx (eval_genop32 (Reg ESP) st) esp
    -> hasType ctx
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
      st) (RetPtr typ)
    -> (forall (v1 v2 : int32),
      hasType ctx v1 (Const typ four) ->
      hasType ctx v2 esp -> hasType ctx (eval_arith Add v1 v2) res)
    -> conc' abs st ctx
    -> write_genop32 (Reg ESP)
    (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
    -> exec (Ret, nextPc) st
    (set_pc st'
      (eval_genop32
	(Address reg32 (Build_address bzero32 (Some ESP) None)) st))
    -> conc'
    (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
      (aStackSize abs) (aStackSlots abs) cust)
    (set_pc st'
      (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
    ctx.
    intros abs st ctx esp st' res nextPc cust Hesp Hret Hadd Hconc Hwrite32 Hexec.
    inversion Hexec; subst.
    inversion Hconc; split;
      try (eapply ret_conc_slots; eauto); myeauto.
  Qed.

  Lemma invert_conc_conc' : forall abs st ctx,
    conc abs st ctx
    -> conc' abs st ctx.
    intros abs st ctx Hconc.
    inversion Hconc.
    trivial.
  Qed.

  Lemma invert_conc_pconc : forall abs st ctx,
    conc abs st ctx
    -> pconc (custom ctx) (aCustom abs) st.
    intros abs st ctx Hconc.
    inversion Hconc.
    trivial.
  Qed.

  Hint Resolve invert_conc_conc' invert_conc_pconc : asm.

  Theorem ret_conc : forall abs st ctx esp st' res nextPc cust,
    hasType ctx (eval_genop32 (Reg ESP) st) esp
    -> hasType ctx
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
      st) (RetPtr typ)
    -> (forall (v1 v2 : int32),
      hasType ctx v1 (Const typ four) ->
      hasType ctx v2 esp -> hasType ctx (eval_arith Add v1 v2) res)
    -> conc abs st ctx
    -> write_genop32 (Reg ESP)
    (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
    -> exec (Ret, nextPc) st
    (set_pc st'
      (eval_genop32
	(Address reg32 (Build_address bzero32 (Some ESP) None)) st))
    -> pconc (custom ctx) cust
    (set_pc st'
      (get_mem32 st (eval_addr (Build_address bzero32 (Some ESP) None) st)))
    -> conc
    (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
      (aStackSize abs) (aStackSlots abs) cust)
    (set_pc st'
      (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
    ctx.
    intros abs st ctx esp st' res nextPc cust Hesp Hret Hadd Hconc Hwrite32 Hexec Hpconc.
    split; trivial.
    eapply ret_conc'; myeauto.
  Qed.

  Hint Resolve ret_conc : asm.

  Hint Constructors exec write_genop32 : asm.


  (** * Initializing the stack pointer *)

  Definition isInitializingStack : forall (g1 g2 : genop32),
    sig_option (fun esp0 => g1 = Reg ESP /\ g2 = Imm reg32 esp0).
    intros.
    destruct g1; [apply SNone | idtac | apply SNone].
    destruct (reg32_eq_dec r ESP); [subst | apply SNone].
    destruct g2; [idtac | apply SNone | apply SNone].
    apply SSome with i.
    eauto.
  Qed.

  Lemma setESP_conc_regs : forall st nextPc mi cust custA,
    conc_regs
    (Build_astate (apcConst nextPc) (Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ 0))
      stackSize (NatMap.empty (Top typ)) custA)
    (set_pc (set_reg32 st ESP (eval_genop32 (Imm reg32 initialESP) st))
      nextPc)
    (Build_context initialESP initialESP mi None cust).
    intros st nextPc mi cust custA r.
    unfold get_reg32, aget_reg, GenericTypes.aget_reg; simpl.
    destruct (reg32_eq_dec r ESP); subst.

    repeat (rewrite Reg32Map.sel_eq).
    assert (hasType (Build_context initialESP initialESP mi None cust)
      (GenericTypes.stack_slot (Build_context initialESP initialESP mi None cust) 0)
      (Stack typ 0)).
    red; apply HasType_Stack.
    omega.
    replace (GenericTypes.stack_slot (Build_context initialESP initialESP mi None cust) 0) with initialESP in H.
    trivial.
    unfold GenericTypes.stack_slot.
    rewrite <- minus_n_O.
    rewrite btn_invert.
    trivial.
    repeat (rewrite Reg32Map.sel_neq); auto with arith.
    repeat (rewrite Reg32Map.sel_empty).
    myauto.
  Qed.

  Hint Resolve setESP_conc_regs : asm.

  Lemma setESP_conc_slots : forall nextPc rs ss st ctx cust,
    conc_slots
    (Build_astate nextPc rs ss (NatMap.empty (Top typ)) cust)
    st ctx.
    intros; intros slot Hbound.
    unfold aget_stack, GenericTypes.aget_stack; simpl.
    rewrite NatMap.sel_empty.
    myauto.
  Qed.

  Hint Resolve setESP_conc_slots : asm.

  Lemma setESP_belowMax : belowMax initialESP 4.
    split; apply initialESP_upper.
  Qed.

  Hint Resolve setESP_belowMax : asm.

  Lemma setESP_pc : forall nextPc r ss sl st ctx cust,
    conc_pc
    (Build_astate (apcConst nextPc) r ss sl cust)
    (set_pc st nextPc)
    ctx.
    intros.
    red; red; simpl.
    trivial.
  Qed.

  Hint Resolve setESP_pc : asm.

  Lemma setESP_memIn : forall v1 v2 m1 m2,
    bvec_to_nat v1 <= bvec_to_nat v2
    -> agree_range' m1 m2 v1 stackEnd
    -> agree_range' m1 m2 v2 stackEnd.
    intros v1 v2 m1 m2 Hr Hm addr Haddr.
    apply Hm.
    omega.
  Qed.
  
  Hint Resolve setESP_memIn : asm.

  Lemma setESP_stack : forall (cust : astat) (ctx : context) nextPc,
    bvec_to_nat stackStart +
    aStackSize
    (Build_astate (apcConst nextPc)
      (Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ 0)) stackSize
      (NatMap.empty (Top typ)) cust) <
    bvec_to_nat
    (sptr
      (Build_context initialESP initialESP (memIn ctx) None (custom ctx))).
    intros.
    unfold aStackSize, sptr.
    apply initialESP_location.
  Qed.

  Hint Resolve setESP_stack : asm.

  Lemma setESP' : forall ctx st st' abs nextPc cust,
    pcInRange nextPc
    -> conc' abs st ctx
    -> write_genop32 (Reg ESP) (eval_genop32 (Imm reg32 initialESP) st) st st'
    -> exec (Mov (Reg ESP) (Imm reg32 initialESP), nextPc) st (set_pc st' nextPc)
    -> agree_range' (stMem st) (memIn ctx) initialESP stackEnd
    -> pconc (custom ctx) cust (set_pc st' nextPc)
    -> conc' (Build_astate (apcConst nextPc) (Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ 0))
      stackSize (NatMap.empty (Top typ)) cust)
    (set_pc st' nextPc) (Build_context initialESP initialESP (memIn ctx) None (custom ctx)).
    intros ctx st st' abs nextPc cust Hpc Hconc Hwrite Hexec Hagree Hpconc.
    inversion Hwrite; subst.
    inversion Hexec; subst.
    inversion Hconc.
    split; myeauto.
  Qed.

  Hint Resolve setESP' : asm.

  Lemma setESP : forall ctx st st' abs nextPc abs' cust,
    pcInRange nextPc
    -> conc' abs st ctx
    -> write_genop32 (Reg ESP) (eval_genop32 (Imm reg32 initialESP) st) st st'
    -> exec (Mov (Reg ESP) (Imm reg32 initialESP), nextPc) st (set_pc st' nextPc)
    -> proj1_sig (subAbs
      (Build_astate (apcConst nextPc)
	(Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ 0)) stackSize
	(NatMap.empty (Top typ)) cust) abs') = true
    -> pconc (custom ctx) cust (set_pc st' nextPc)
    -> conc abs'
    (set_pc st' nextPc) (Build_context initialESP initialESP (memIn ctx) None (custom ctx)).
    intros ctx st st' abs nextPc abs' cust Hpc Hconc Hwrite Hexec Hsub.
    destruct (subAbs
      (Build_astate (apcConst nextPc)
	(Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ 0)) stackSize
	(NatMap.empty (Top typ)) cust) abs').
    simpl in Hsub; subst; intuition.
    apply H0.
    split; intuition.
    inversion Hconc.
    eapply setESP'; myeauto.
  Qed.

  Hint Resolve setESP : asm.


  (** * Reasoning about assumed-safe states *)

  Lemma AllS_trivial : forall (T : Set) (f : T -> Prop) ls,
    (forall x, In x ls -> f x)
    -> AllS f ls.
    induction ls; intuition.
  Qed.

  Lemma AllS_imp : forall (T : Set) (f g : T -> Prop),
    (forall x, f x -> g x)
    -> forall ls, AllS f ls
      -> AllS g ls.
    induction ls; intuition.
    inversion H0; subst.
    apply allS_cons; auto.
  Qed.

  Lemma AllS_goodHyps : forall hyps' hyps ctx ls,
    goodHyps hyps hyps'
    -> AllS
    (fun need : astate =>
      exists have : astate,
	(In have ls \/ In have hyps) /\ menter conc ctx need ctx have)
    hyps'.
    induction hyps'; intros hyps ctx ls Hhyps;
      destruct hyps; simpl in Hhyps; intuition.
    apply allS_cons.
    exists a0.
    intuition.
    destruct (subAbs a a0).
    simpl in H.
    unfold menter.
    auto.
    apply AllS_imp with
      (fun need : astate =>
	exists have : astate,
          (In have nil \/ In have hyps) /\
          menter conc ctx need ctx have).
    simpl.
    intros x Hx.
    destruct Hx as [have Hx].
    exists have.
    intuition.
    auto.
  Qed.

  Hint Resolve AllS_goodHyps : asm.

  Lemma menter_refl : forall ctx abs,
    menter conc' ctx abs ctx abs.
    unfold menter; intuition.
  Qed.


  (** * The safety-policy-instrumented concrete machine semantics *)

  Module M86_Arg <: M86_PARAM.
    Definition minitState := initState.
  End M86_Arg.
  Module M86_with_init := M86(M86_Arg).

  Module MS86_Arg <: STATELESS.
    Module Mac := M86_with_init.
    Definition isSafe := isSafe.
  End MS86_Arg.
  
  Module MS86 := Stateless(MS86_Arg).
  Import MS86.

  (** * Translating register types to be meaningful to the custom type system *)

  Lemma use_safe : forall abs ins nextPc st st' ctx,
    (forall (st : state) (ctx : ctxt),
      pconc ctx (aCustom abs) st
      -> statusOk hasTyp (proj1_sig (simple_status abs)) ctx st
      -> isSafe ins st)
    -> conc abs st ctx
    -> exec (ins, nextPc) st st'
    -> mexec (ins, nextPc) st st'.
    intros abs ins nextPc st st' ctx Hsafe Hconc Hexec.
    split.
    unfold MS86_Arg.isSafe.
    apply (Hsafe st (custom ctx)).
    inversion Hconc; trivial.
    destruct (simple_status abs) as [stat Hstat].
    destruct (Hstat st ctx (concMain Hconc)).
    split; myeauto.
    myeauto.
  Qed.

  Hint Resolve use_safe : asm.

  Definition single_rule_instr_dec ins : {single_rule_instr ins} + {~single_rule_instr ins}.
    destruct ins; simpl; auto.
  Qed.


  (** * Soundness relation rules for particular instructions *)

  Lemma conc'_arith : forall abs' st'0 a g st g0 ctx nextPc,
    conc' abs' st'0 ctx
    -> pcInRange nextPc
    -> conc' (aset_pc abs' nextPc)
    (set_pc (arith_flags st'0 a (eval_genop32 g st) (eval_genop32 g0 st))
      nextPc) ctx.
    intros until nextPc; intros Hconc Hpc.
    inversion Hconc; split; intuition.
  Qed.

  Hint Resolve conc'_arith : asm.

  Lemma conc'_inc : forall abs' st'0 g st ctx nextPc,
    conc' abs' st'0 ctx
    -> pcInRange nextPc
    -> conc' (aset_pc abs' nextPc)
    (set_pc (arith_flags_no_CF st'0 Add (eval_genop32 g st) one)
      nextPc) ctx.
    intros until nextPc; intros Hconc Hpc.
    inversion Hconc.
    split; intuition.
  Qed.

  Lemma conc'_set_pc : forall abs st ctx pc,
    conc' abs st ctx
    -> pcInRange pc
    -> conc' (aset_pc abs pc) (set_pc st pc) ctx.
    intros abs st ctx pc Hconc Hpc.
    inversion Hconc; split; intuition.
  Qed.

  Hint Resolve conc'_set_pc : asm.

  Lemma conc_regs_custom : forall abs st ctx cust,
    conc_regs abs st ctx
    -> conc_regs (aset_custom abs cust) st ctx.
    intuition.
  Qed.

  Hint Resolve conc_regs_custom : asm.

  Lemma conc'_lea : forall abs a r st ctx res nextPc,
    conc abs st ctx
    -> pcInRange nextPc
    -> (forall st ctx,
      conc abs st ctx 
      -> hasType ctx (eval_addr a st) res)
    -> conc' (aset_pc (aset_reg abs r res) nextPc)
    (set_pc (set_reg32 st r (eval_addr a st)) nextPc) ctx.
    intros until nextPc; intros Hconc Hpc Hty.
    inversion Hconc; inversion concMain0.
    split; myeauto.
  Qed.


  (** * Move these earlier once everything works *)

  Lemma invert_conc_main : forall abs st ctx,
    conc abs st ctx
    -> conc' abs st ctx.
    intuition.
  Qed.

  Lemma invert_conc_status : forall abs st ctx,
    conc abs st ctx
    -> statusOk hasTyp (proj1_sig (simple_status abs)) (custom ctx) st.
    intros abs st ctx Hconc.
    destruct (simple_status abs); simpl.
    myeauto.
  Qed.

  Hint Resolve invert_conc_status : asm.

  Lemma conc'_regs_upd : forall abs r v t st ctx,
    conc' abs st ctx
    -> hasType ctx v t
    -> conc' (aset_reg abs r t) (set_reg32 st r v) ctx.
    intros abs r v t st ctx Hconc Hty.
    inversion Hconc.
    split; intuition.
  Qed.

  Hint Resolve conc'_regs_upd : asm.

  Lemma prAbs_hint : forall res ctx st,
    presultOk hasTyp pconc res ctx st
    -> pconc ctx (prAbs res) st.
    intros.
    inversion H; trivial.
  Qed.

  Hint Resolve prAbs_hint : asm.


  (** * Incorporating the extension's genop type coercion requests *)

  Lemma conc_regs_upd' : forall abs st ctx r t,
    conc_regs abs st ctx
    -> hasType ctx (get_reg32 st r) t
    -> conc_regs (aset_reg abs r t) st ctx.
    intros abs st ctx r t Hconc Hty r0.
    destruct (reg32_eq_dec r r0); subst.

    unfold aget_reg, GenericTypes.aget_reg, aset_reg; simpl.
    rewrite Reg32Map.sel_eq.
    trivial.

    unfold aget_reg, GenericTypes.aget_reg, aset_reg; simpl.
    rewrite Reg32Map.sel_neq; trivial.
    red in Hconc.
    unfold aget_reg in Hconc.
    myeauto.
  Qed.

  Hint Resolve conc_regs_upd' : asm.

  Lemma conc_stack_upd' : forall abs st ctx sl t,
    conc' abs st ctx
    -> hasType ctx (get_mem32 st (stack_slot ctx sl)) t
    -> conc' (aset_stack abs sl t) st ctx.
    intros abs st ctx sl t Hconc Hty.
    inversion Hconc; split; intuition.
    intros sl' Hsl'.
    unfold aget_stack, GenericTypes.aget_stack, aset_stack; simpl.

    Ltac duh :=
      match goal with
	| [ |- GenericTypes.hasType _ _ _ (NatMap.sel (NatMap.upd _ ?V1 _) ?V2) ] =>
	  destruct (eq_nat_dec V1 V2); subst;
	    [rewrite NatMap.sel_eq; myauto
	      | rewrite NatMap.sel_neq; trivial]
      end.

    repeat duh.
    red in concStackSlots0.
    unfold aget_stack, GenericTypes.aget_stack in concStackSlots0.
    auto.
  Qed.

  Hint Resolve conc_stack_upd' : asm.

  Ltac abs_fail abs := exists abs; intuition.

  Definition aset_genop : forall abs g t,
    {abs' : astate
      | forall st ctx,
	conc' abs st ctx
	-> word_hasType hasTyp (custom ctx) (eval_genop32 g st) t
	-> conc' abs' st ctx}.
    intros.
    destruct g.

    abs_fail abs.

    destruct (word_type_to_ty hasTyp t) as [t' Ht'].
    exists (aset_reg abs r t').
    intros st ctx Hconc Hty.
    inversion Hconc; split; intuition.
    
    destruct a.
    destruct addrBase; [idtac | abs_fail abs].
    destruct addrIndex; [abs_fail abs | idtac].
    caseEq (aget_reg abs r); [abs_fail abs | abs_fail abs | idtac | abs_fail abs | abs_fail abs | abs_fail abs].
    intros sl Hsl.
    destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)));
      [idtac | abs_fail abs].
    destruct (le_lt_dec (aStackSize abs)
      (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)));
    [abs_fail abs | idtac].
    simpl.
    destruct (word_type_to_ty hasTyp t) as [t' Ht'].
    exists (aset_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)) t').
    intros st ctx Hconc Hty.
    apply conc_stack_upd'; trivial.
    replace (stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 sl) addrDisp)))
      with (eval_addr (Build_address addrDisp (Some r) None) st).
    myauto.
    eapply slot_addr with abs; inversion Hconc; eauto.
  Qed.

  Definition aset_genops : forall gts abs,
    {abs' : astate
      | forall st ctx,
	conc' abs st ctx
	-> AllS (fun gp => word_hasType hasTyp (custom ctx) (eval_genop32 (fst gp) st) (snd gp)) gts
	-> conc' abs' st ctx}.
    induction gts; intros.

    exists abs; trivial.

    destruct a as [g t].
    destruct (aset_genop abs g t) as [abs' Habs'].
    destruct (IHgts abs') as [abs'' Habs''].
    clear IHgts.
    exists abs''.
    intros st ctx Hconc Hall.
    inversion Hall; subst.
    auto.
  Qed.

  Lemma conc'_custom : forall abs cust st ctx,
    conc' abs st ctx
    -> conc' (aset_custom abs cust) st ctx.
    intros abs cust st ctx Hconc.
    inversion Hconc; split; intuition.
  Qed.

  Hint Resolve conc'_custom : asm.

  Lemma conc'_generic : forall abs st ctx,
    conc' abs st ctx
    -> GenericTypes.conc' hasTyp abs st ctx.
    intuition.
  Qed.

  Hint Resolve conc'_generic : asm.

  Definition ause_cust : forall abs res,
    {abs' : astate
      | forall st ctx,
	conc' abs st ctx
	-> presultOk hasTyp pconc res (custom ctx) st
	-> conc abs' st ctx}.
    intros.
    destruct (aset_genops (prGenops res) abs) as [abs' Habs'].
    exists (aset_custom abs' (prAbs res)).
    intros st ctx Hconc Hres.
    inversion Hres.
    split; myeauto.
  Qed.

  Definition find_mexec : forall insPc,
    pred_option (forall st, exists st', exec insPc st st').
    intros.
    destruct insPc as [instr nextPc].

    Ltac efailure := intros; apply PNone.
    Ltac valid := apply PSome; myeauto.
    Ltac dst g := destruct g;
      [efailure
	| valid
	| valid].

    destruct instr.
    dst g.
    efailure.
    valid.    
    dst g.
    dst g.
    apply PSome; intro;
      destrEq Hflag (check_cc st c); myeauto.
    valid.
    valid.
    valid.
    dst g.
    efailure.
    dst g.
    dst g.
    valid.
    efailure.
  Qed.


  (** * The main step function *)

  Definition fp := init ++ importd.

  Definition step : forall hyps abs insPc, {result : option (list astate)
    | match result with
	| None => True
	| Some abss =>
	  forall st ctx, conc abs st ctx
	    -> (exists st' : state, mexec insPc st st')
	    /\ forall st' (Hexec : mexec insPc st st'),
	      exists abs', ((In abs' abss \/ In abs' hyps) /\ conc abs' st' ctx)
		\/ exists ctx', exists hyps',
		  In (hyps', abs') fp
		  /\ AllS (fun need => exists have,
		    (In have abss \/ In have hyps) /\ menter conc ctx' need ctx have) hyps'
		  /\ conc abs' st' ctx'
      end}.
    intros hyps abs insPc.
    destruct insPc as [ins nextPc].

    Ltac failure := exists (@None (list astate)); trivial.
    Ltac assertOpt hyp e := destrEq hyp e; [idtac | failure].
    Ltac assertOptEq v H e := caseEq e; [intros v H | failure].
    Ltac sig_opt v H e := case e; [intros v H | failure].

    Ltac succ' abs' :=
      match goal with
	| [ nextPc : int32 |- _ ] =>
	  case (pcInRange_dec nextPc);
	    [intro HpcRange; exists (Some (cons abs' nil))
	      | failure]
      end.
    Ltac succ f abs :=
      let Hexec := fresh "Hexec"
	with HexecSafe := fresh "HexecSafe"
	with Hexec' := fresh "Hexec'"
	with aa := fresh "aa"
	with ba := fresh "ba"
	with ca := fresh "ca"
	with da := fresh "da" in
	(succ' abs; intros st ctx Hconc; split;
	  [exists (f st); myauto
	    | intros st' Hexec; inversion Hexec as [aa ba ca da HexecSafe Hexec']; subst;
	      clear Hexec; rename Hexec' into Hexec;
		exists abs; inversion Hexec; subst; left; split; [left; intu | idtac]]).

    Ltac succCust f abs cust :=
      succ f (proj1_sig (ause_cust abs cust)).

    Ltac destrConc' abs' :=
      match goal with
	[ H : conc' _ _ |- _ ] => destruct H as [abs']
      end.

    Ltac concrete g := destruct g; simpl; myauto; try discriminate;
      eapply use_safe; myeauto.
    Ltac concrete_arith rule g := destruct g; myauto; try discriminate;
      eapply use_safe; eauto; apply rule; simpl; myeauto.

    Ltac jmp :=
      match goal with
	[ hyps : list astate, abs : astate, i : int32, nextPc : int32 |- _] =>
	let Hx' := fresh "Hx'" in let c := fresh "c" in
 	  (case (pcInRange_dec i); [intro Hpc | failure];
	    case (abs_covered (proj1_sig init) hyps (aset_pc abs i));
	      intros x Hx; destruct x; [idtac | failure];
		add Hx' (Hx (refl_equal true)); clear Hx; rename Hx' into Hx;
		  exists (Some (@nil astate)); intros st ctx Hconc; split;
		    [exists (set_pc st i); myauto
		      | intros st' Hexec; inversion Hexec;
			subst; destruct Hx as [abs' Hx];
			  exists abs'; destruct (subAbs (aset_pc abs i) abs') as [b c];
			    exists ctx; exists hyps; intuition;
			      apply AllS_trivial; intros x Hx; exists x;
				unfold menter; intuition])
      end.

    (* Ask the extension to check for local progress. *)
    destruct (checkSafe (aCustom abs) (proj1_sig (simple_status abs)) ins) as [Hsafe |]; [idtac | failure].

    (* To handle preservation, we first check whether the instruction has only a
      single execution rule. *)
    destruct (single_rule_instr_dec ins) as [Hsingle | Hsingle].

    (* Single-rule instructions *)
    destruct (preservation (aCustom abs) (proj1_sig (simple_status abs)) ins nextPc Hsingle) as [cust Hcust |]; [idtac | failure].
    
    (* Preservation by case analysis on the instruction *)
    destruct ins.

    (* Arith *)
    sig_opt t Ht (typeof_genop32 abs g).
    sig_opt t0 Ht0 (typeof_genop32 abs g0).
    destruct (aeval_arith abs a t t0) as [res Hres].
    assertOpt writable (writable_genop32 g).
    sig_opt abs' Habs' (awrite_genop32 abs g res).

    succCust
    (fun st => set_pc
      (arith_flags (write_concrete g st (eval_arith a (eval_genop32 g st) (eval_genop32 g0 st)) writable)
	a (eval_genop32 g st) (eval_genop32 g0 st)) nextPc)
    (aset_pc abs' nextPc) cust;
    [concrete_arith XArith g
      | destruct (ause_cust (aset_pc abs' nextPc) cust) as [abs'' Habs'']; simpl;
	apply Habs''; [apply conc'_arith; myeauto | myeauto]].

    (* Arithb *)
    failure.

    (* Call *)
    destruct g; [idtac | failure | failure].
    destruct (pcInRange_dec i); [idtac | failure].
    destruct (pcInRange_dec nextPc); [idtac | failure].
    caseEq (aget_reg abs ESP); [failure | failure | idtac | failure | failure | failure].
    intros ss Hesp.
    destruct (le_lt_dec (aStackSize abs) (S (S (S (S ss))))); [failure | idtac].
    destruct (functions i); [idtac | failure].
    destruct (le_lt_dec (4 * length (fiArgs f)) ss); [idtac | failure].
    destrEq Hcheck (checkArgs (fiArgs f) (ss - 4 * length (fiArgs f)) (aStackSlots abs)); [idtac | failure].
    destruct (@enter abs nextPc i ss (fiArgs f) (prAbs cust)) as [absEnter |]; trivial; [idtac | failure].
    destruct (@leave abs nextPc (fiRet f) ss (fiArgs f)) as [absLeave |]; trivial; [idtac | failure].
    destruct (abs_covered fp (cons (snd absLeave) nil) absEnter).
    destruct x; [intuition | failure].

    exists (Some (cons (fst absLeave) nil)).
    destruct H as [hyps' H].
    destruct H as [abs' Habs'].
    destruct Habs' as [Hin Hsub].
    destruct Hsub as [Hsub Hhyps].
    destruct hyps' as [| hyp' hyps']; try (simpl in Hhyps; tauto).
    destruct hyps'; simpl in Hhyps; try tauto.

    intros st ctx Hconc.
    split.
    myeauto.
    intros st' Hexec.
    inversion Hexec as [aa ba ca da HexecSafe Hexec']; subst.
    inversion Hexec'; subst.
    exists abs'.
    right.
    exists (leave_context (set_pc st nextPc) ctx ss (fiArgs f)).
    exists (cons hyp' nil).
    intuition.
    apply allS_cons; [idtac | apply allS_nil].
    exists (fst absLeave).
    intuition.
    inversion H4; subst.
    inversion H1; subst.
    assert (Henter : menter conc (leave_context (set_pc st nextPc) ctx ss (fiArgs f))
      (snd absLeave) ctx (fst absLeave)).
    apply m; myauto.
    intros st' Hconc'.
    destruct (subAbs hyp' (snd absLeave)).
    simpl in H.
    unfold menter in Henter; apply Henter.
    auto.

    inversion H4; subst.
    inversion H1; subst.
    destruct (subAbs absEnter abs').
    simpl in Hsub.
    replace (eval_addr (Build_address neg_four (Some ESP) None) st)
      with (badd neg_four (get_reg32 st ESP)); trivial.
    replace (eval_arith Sub (eval_genop32 (Reg ESP) st) four)
      with (bsub (get_reg32 st ESP) four); trivial.
    replace (eval_genop32 (Imm reg32 i)
      (set_reg32 (set_mem32 st (get_reg32 st ESP) nextPc) ESP
	(bsub (get_reg32 st ESP) four)))
    with i; trivial.
    auto.
    apply c0; trivial.
    apply c; eauto.
    assert (Hpres : presultOk hasTyp pconc cust (custom ctx) (set_pc
      (set_reg32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) nextPc)
        ESP (bsub (get_reg32 st ESP) four)) i)).
    eapply Hcust; myeauto.
    myeauto.

    (* Cmp *)
    sig_opt t Ht (typeof_genop32 abs g).
    sig_opt t0 Ht0 (typeof_genop32 abs g0).
    destruct (aeval_arith abs Sub t t0) as [res Hres].

    succCust
    (fun st => set_pc
      (arith_flags st Sub (eval_genop32 g st) (eval_genop32 g0 st)) nextPc)
    (aset_pc abs nextPc) cust;
    [concrete_arith XCmp g
      | destruct (ause_cust (aset_pc abs nextPc) cust) as [abs' Habs']; simpl;
	apply Habs'; [apply conc'_arith; myeauto | myeauto]].

    (* Inc *)
    sig_opt t Ht (typeof_genop32 abs g).
    destruct (aeval_arith abs Add t (Const typ one)) as [res Hres].
    assertOpt writable (writable_genop32 g).
    sig_opt abs' Habs' (awrite_genop32 abs g res).

    succCust
    (fun st => set_pc
      (arith_flags_no_CF (write_concrete g st (eval_arith Add (eval_genop32 g st) one) writable)
	Add (eval_genop32 g st) one) nextPc)
    (aset_pc abs' nextPc) cust;
    [concrete_arith XInc g
      | destruct (ause_cust (aset_pc abs' nextPc) cust) as [abs'' Habs'']; simpl;
	apply Habs''; [apply conc'_inc; myeauto | myeauto]].

    (* Jcc (not possible) *)
    simpl in Hsingle; tauto.

    (* Jmp *)
    case (pcInRange_dec i); [intro Hpc | failure].
    case (abs_covered fp hyps
      (proj1_sig (ause_cust (aset_pc abs i) cust))).
    intros x Hx; destruct x; [idtac | failure].
    add Hx' (Hx (refl_equal true)); clear Hx; rename Hx' into Hx.
    exists (Some (@nil astate)); intros st ctx Hconc; split.
    myeauto.
    intros st' Hexec.
    inversion Hexec as [aa ba ca da HexecSafe Hexec']; subst.    
    inversion Hexec'; subst.
    destruct Hx as [hyps' Hx].
    destruct Hx as [abs' Hx].
    exists abs'.
    destruct (subAbs (proj1_sig (ause_cust (aset_pc abs i) cust)) abs') as [b c].
    right; exists ctx; exists hyps'; intuition.
    apply c; auto.
    destruct (ause_cust (aset_pc abs i) cust) as [abs'' Habs'']; simpl.
    apply Habs''.
    myeauto.
    myeauto.

    (* Lea *)
    destruct (aeval_addr abs a) as [res Hres].

    succCust
    (fun st => set_pc (set_reg32 st r (eval_addr a st)) nextPc)
    (aset_pc (aset_reg abs r res) nextPc) cust; myeauto.
    destruct (ause_cust (aset_pc (aset_reg abs r res) nextPc) cust) as [abs' Habs']; simpl;
      apply Habs'; [apply conc'_lea; myeauto | myeauto].

    (* Leave *)
    sig_opt ebp Hebp (typeof_genop32 abs (Reg EBP)).
    sig_opt t Ht (typeof_genop32 abs (Address reg32 (Build_address bzero32 (Some EBP) None))).
    destruct (aeval_arith abs Add (Const typ four) ebp) as [res Hres].

    succCust
    (fun st => set_pc
      (set_reg32
	(set_reg32 st EBP (eval_genop32 (Address reg32 (Build_address bzero32 (Some EBP) None)) st))
	ESP (eval_arith Add four (eval_genop32 (Reg EBP) st)))
      nextPc)
    (aset_pc (aset_reg (aset_reg abs EBP t) ESP res) nextPc) cust;
    try inversion H1; try inversion H4; subst; eauto 7 with asm.
    destruct (ause_cust (aset_pc (aset_reg (aset_reg abs EBP t) ESP res) nextPc)
      cust) as [abs' Habs']; simpl.
    apply Habs'; myeauto.
    apply conc'_pc'; myeauto.
    apply conc'_regs_upd; myeauto.
    inversion H0; subst; myeauto.
    eapply Hres; myeauto.
    replace (get_reg32 st EBP)
      with (eval_genop32 (Reg EBP) st); trivial.
    myauto.

    (* Mov *)
    destruct (isInitializingStack g g0) as [esp0 Hinit | ].

    (* We are initializing ESP from a constant. *)
    destruct Hinit as [Hg Hg0]; subst.
    destruct (pcInRange_dec nextPc); [idtac | failure].
    destruct (bvec_eq_dec esp0 initialESP); [subst | failure].
    destruct (abs_covered fp nil 
      (Build_astate (apcConst nextPc) (Reg32Map.upd (Reg32Map.empty (Top typ)) ESP (Stack typ 0))
	stackSize (NatMap.empty (Top typ)) (prAbs cust))).
    destruct x; [subst; intuition | failure].

    exists (Some (@nil astate)).
    destruct H as [hyps' H].
    destruct H as [abs' Habs'].
    destruct Habs' as [Hin Hproj].
    intros st ctx Hconc.
    split.
    exists (set_pc (set_reg32 st ESP initialESP) nextPc); eapply use_safe; myeauto.
    intros st' Hexec.
    inversion Hexec as [aa ba ca da HexecSafe Hexec']; subst.
    inversion Hexec'; subst.
    exists abs'.
    right.
    exists (Build_context initialESP initialESP (memIn ctx) None (custom ctx)).
    exists (@nil astate).
    intuition.
    destruct hyps'; simpl in H0; tauto.
    apply setESP with st abs (prAbs cust); myauto.
    myeauto.

    (* Other moves. *)
    sig_opt t0 Ht0 (typeof_genop32 abs g0).
    assertOpt writable (writable_genop32 g).
    sig_opt abs' Habs' (awrite_genop32 abs g t0).

    succCust
    (fun st => set_pc (write_concrete g st (eval_genop32 g0 st) writable) nextPc)
    (aset_pc abs' nextPc) cust;
    [concrete g
      | idtac].
    destruct (ause_cust (aset_pc abs' nextPc) cust) as [abs'' Habs'']; simpl;
      apply Habs''; myeauto.

    (* Movb *)
    failure.

    (* Pop *)
    sig_opt esp Hesp (typeof_genop32 abs (Reg ESP)).
    sig_opt t Ht (typeof_genop32 abs (Address reg32 (Build_address bzero32 (Some ESP) None))).
    destruct (aeval_arith abs Add (Const typ four) esp) as [res Hres].
    assertOpt writable (writable_genop32 g).
    sig_opt abs' Habs' (awrite_genop32 abs g t).

    destruct g; simpl in writable; try discriminate.

    succCust
    (fun st => set_pc
      (set_reg32
	(set_reg32 st r (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
	ESP (eval_arith Add four (eval_genop32 (Reg ESP) st)))
      nextPc)
    (aset_pc (aset_reg abs' ESP res) nextPc) cust;
    try inversion H1; try inversion H4; subst; myeauto.
    destruct (ause_cust (aset_pc (aset_reg abs' ESP res) nextPc) cust) as [abs'' Habs'']; simpl;
      apply Habs''; myeauto.
    apply conc'_pc'; myeauto.
    apply conc'_regs_upd; myeauto.
    eapply Hres; myeauto.
    replace (get_reg32 st ESP)
      with (eval_genop32 (Reg ESP) st); trivial.
    myauto.

    succCust
    (fun st => set_pc
      (set_reg32
	(set_mem32 st (eval_addr a st) (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
	ESP (eval_arith Add four (eval_genop32 (Reg ESP) st)))
      nextPc)
    (aset_pc (aset_reg abs' ESP res) nextPc) cust;
    try inversion H1; try inversion H4; subst; myeauto.
    destruct (ause_cust (aset_pc (aset_reg abs' ESP res) nextPc) cust) as [abs'' Habs'']; simpl;
      apply Habs''; myeauto.
    apply conc'_pc'; myeauto.
    apply conc'_regs_upd; myeauto.
    eapply Hres; myeauto.
    replace (get_reg32 st ESP)
      with (eval_genop32 (Reg ESP) st); trivial.
    myauto.

    (* Push *)
    sig_opt esp Hesp (typeof_genop32 abs (Reg ESP)).
    sig_opt t Ht (typeof_genop32 abs g).
    destruct (aeval_arith abs Add (Const typ neg_four) esp) as [res Hres].
    assertOpt writable (writable_genop32 (Address reg32 (Build_address neg_four (Some ESP) None))).
    sig_opt abs' Habs' (awrite_genop32 abs (Address reg32 (Build_address neg_four (Some ESP) None)) t).

    succCust
    (fun st => set_pc
      (set_reg32
	(set_mem32 st (eval_addr (Build_address neg_four (Some ESP) None) st) (eval_genop32 g st))
	ESP (eval_arith Add neg_four (eval_genop32 (Reg ESP) st)))
      nextPc)
    (aset_pc (aset_reg abs' ESP res) nextPc) cust; try inversion H1; subst; myeauto.
    inversion H4; subst.
    destruct (ause_cust (aset_pc (aset_reg abs' ESP res) nextPc) cust) as [abs'' Habs'']; simpl;
      apply Habs''; myeauto.
    apply conc'_pc'; myeauto.
    apply conc'_regs_upd; myeauto.
    eapply Hres; myeauto.
    replace (get_reg32 st ESP)
      with (eval_genop32 (Reg ESP) st); trivial.
    myauto.

    (* Ret *)
    destruct hyps as [| hyp hyps]; [failure | destruct hyps; [idtac | failure]].
    sig_opt esp Hesp (typeof_genop32 abs (Reg ESP)).
    sig_opt t Ht (typeof_genop32 abs (Address reg32 (Build_address bzero32 (Some ESP) None))).
    destruct t; [failure | failure | failure | idtac | failure | failure].
    destruct (aeval_arith abs Add (Const typ four) esp) as [res Hres].
    sig_opt abs' Habs' (awrite_genop32 abs (Reg ESP) res).
    pose (Build_astate
      apcRet
      (Reg32Map.upd (aRegs abs) ESP res)
      (aStackSize abs)
      (aStackSlots abs)
      (prAbs cust)).
    rename a into abs''.
    destruct (subAbs abs'' hyp); destruct x; [intuition | failure].

    exists (Some (@nil astate)).
    intros st ctx Hconc.
    split.
    myeauto.
    intros st' Hexec.
    inversion Hexec as [aa ba ca da HexecSafe Hexec']; subst.
    inversion Hexec'; subst.
    exists hyp.
    left.
    intuition.
    apply H.
    unfold abs''.
    eapply ret_conc; try (apply Hesp; trivial); myeauto.

    (* Shift *)
    failure.


    (* Double-rule instructions (currently only Jcc) *)
    destruct ins; simpl in Hsingle; try tauto; clear Hsingle.

    destruct (preservation_jcc (aCustom abs) (proj1_sig (simple_status abs))
      c i nextPc) as [cust Hcust |]; [idtac | failure].
    destruct cust as [custTrue custFalse]; simpl in Hcust.

    (* Jcc *)
    destruct (pcInRange_dec i); [idtac | failure].
    destruct (pcInRange_dec nextPc); [idtac | failure].
    case (abs_covered fp hyps (proj1_sig (ause_cust (aset_pc abs i) custTrue))).
    intros cov Hcov.
    destruct cov; [idtac | failure].
    intuition; rename H into Hcov.
    exists (Some (cons (proj1_sig (ause_cust (aset_pc abs nextPc) custFalse)) nil)).
    intros st ctx Hconc.
    destruct Hcov as [hyps' Hcov].
    destruct Hcov as [abs' Hcov].
    split.

    (* Progress *)
    destrEq Hc (check_cc st c).
    exists (set_pc st i); eapply use_safe; myeauto.
    exists (set_pc st nextPc); eapply use_safe; myeauto.

    (* Preservation *)
    intros st' Hexec.
    inversion Hexec as [aa ba ca da HexecSafe Hexec']; subst.
    add Hcust' (Hcust st (custom ctx)).
    clear Hcust.
    rename Hcust' into Hcust.
    inversion Hexec'; subst; rewrite H4 in Hcust.

    (* Taking the jump *)
    
    exists abs'.
    right; intuition.
    exists ctx.
    exists hyps'.
    intuition.

    destruct (subAbs (proj1_sig (ause_cust (aset_pc abs i) custTrue)) abs').
    destruct x; simpl in H1; try discriminate.
    apply c0; myeauto.
    destruct (ause_cust (aset_pc abs i) custTrue) as [abs'' Habs'']; simpl.
    apply Habs''.
    apply conc'_set_pc; myeauto.
    myeauto.

    (* Falling through *)
    exists (proj1_sig (ause_cust (aset_pc abs nextPc) custFalse)).
    simpl; intuition; myauto.
    left; split; intuition.
    destruct (ause_cust (aset_pc abs nextPc) custFalse) as [abs'' Habs'']; simpl.
    apply Habs''.
    apply conc'_set_pc; myeauto.
    myeauto.
  Qed.


  (** * Building the executable checker *)

  Module Abs <: ABSTRACTION.
    Module Mac := MS86.

    Definition mastate := astate.
    Definition mastate_eq_dec := astate_eq_dec.

    Definition mcontext := context.

    Definition mconc := conc.
    Definition maget_instr (abs : astate) :=
      match aPc abs with
	| apcConst pc => Some (decode_instr initialMem pc)
	| _ => None
      end.
    Theorem maget_instr_sound : forall abs st ctx v,
      conc abs st ctx
      -> maget_instr abs = Some v
      -> v = Mac.mget_instr st.
      intros abs st ctx v Hconc Hv.
      inversion Hconc.
      inversion concMain0.
      unfold maget_instr in Hv.
      unfold Mac.mget_instr.
      unfold MS86_Arg.Mac.mget_instr.
      unfold get_instr.
      red in concPc0.
      destruct (aPc abs); subst.
      rewrite (@decode_same (stMem st) initialMem (stPc st)).
      congruence.
      red in concCodeMem0.
      red.
      intros.
      apply concCodeMem0.
      red in concPcRange0.
      omega.
      discriminate.
    Qed.
    
    Definition minit := init.
    Definition mimported := importd.
    Definition mstep := step.
  End Abs.

  Module C := Checker(Abs).
End Make.

