(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Algebraic, detail-hiding re-packaging of FirstOrder *)

Require Import ProofOS.AbsInt.FirstOrder.

Require Import Arith List Peano_dec TheoryList.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Env.
Require Import ProofOS.Trusted.Safety.

Require Import ProofOS.Machine.Stateless.

Require Import ProofOS.AbsInt.Bounded.

Set Implicit Arguments.

Record abstraction : Type := {
  absTyp : Set;
  absAstat : Set;
  absCtxt : Set;
  absHasTyp : absCtxt -> int32 -> absTyp -> Prop;
  absPconc : absCtxt -> absAstat -> state -> Prop;
  absTyp_eq_dec : forall (t1 t2 : absTyp), {t1 = t2} + {t1 <> t2};
  absAstat_eq_dec : forall (abs1 abs2 : absAstat), {abs1 = abs2} + {abs1 <> abs2}
}.

Section FilterCovered.
  Variable T : Set.
  Hypothesis eq_dec : forall (x y : T), {x = y} + {x <> y}.

  Theorem In_dec : forall (x : T) ls, {In x ls} + {~In x ls}.
    induction ls; eauto.
    destruct (eq_dec x a); subst.
    simpl; auto.
    simpl; destruct IHls; auto.
    right.
    intro.
    destruct H; auto.
  Qed.

  (** Remove all members of ls1 that are also in ls2. *)
  Fixpoint filterCovered (ls1 ls2 : list T) {struct ls1} : list T :=
    match ls1 with
      | nil => nil
      | x :: ls1 =>
	if In_dec x ls2
	  then filterCovered ls1 ls2
	  else x :: filterCovered ls1 ls2
    end.

  Theorem In_filterCovered : forall x ls1 ls2,
    In x ls1
    -> In x (filterCovered ls1 ls2) \/ In x ls2.
    induction ls1; intros ls2 Hin; intuition.
    destruct Hin as [Hin | Hin]; subst; simpl.
    destruct (In_dec x ls2); simpl; auto.
    destruct (In_dec a ls2); simpl; auto.
    add Hdone (IHls1 ls2 Hin).
    tauto.
  Qed.

  Theorem In_filterCovered2 : forall x ls1 ls2,
    In x (filterCovered ls1 ls2)
    -> In x ls1.
    induction ls1; intros ls2 Hin; intuition.
    simpl in Hin.
    destruct (In_dec a ls2); simpl; eauto.
    destruct Hin; eauto.
  Qed.
End FilterCovered.

Section EntryPoints.
  Variable abstr : abstraction.

  Definition entryPoint_eq_dec : forall (x y : entryPoint (absTyp abstr) (absAstat abstr)),
    {x = y} + {x <> y}.
    intros.
    decide equality x y.
    apply (absAstat_eq_dec abstr).
    apply (absAstat_eq_dec abstr).
    decide equality epRet epRet0.
    apply (@bvec_eq_dec 32).
    apply (absTyp_eq_dec abstr).
    apply list_eq_dec.
    intros.
    decide equality x0 y0.
    apply (@bvec_eq_dec 32).
    apply (absTyp_eq_dec abstr).
    apply eq_nat_dec.
    apply (@bvec_eq_dec 32).
  Qed.
End EntryPoints.

Module Type FIRSTORDER.
  (** A certification that the [entryPoints] are safe if we know that
    [imported] is safe *)
  Parameter cert : forall (abstr : abstraction)
    (isSafe : instr -> state -> Prop)
    (kernel_astat : option (absAstat abstr))
    (entryPoints imported : list (entryPoint (absTyp abstr) (absAstat abstr))),
    Prop.

  (** From a certification with no assumptions, prove safety of a kernel *)
  Parameter kernel : forall abstr isSafe abs entryPoints,
    cert abstr isSafe (Some abs) entryPoints nil
    -> forall st ctx, initState st
      -> absPconc abstr ctx abs st
      -> safe (Safety.cstep (stateless_mexec exec isSafe) get_instr) st.

  (** Link together two certificates for the same abstraction *)
  Parameter link : forall abstr isSafe ka1 eps1 imp1 eps2 imp2,
    cert abstr isSafe ka1 eps1 imp1
    -> cert abstr isSafe None eps2 imp2
    -> cert abstr isSafe ka1 (eps1 ++ eps2)
    (filterCovered (entryPoint_eq_dec abstr) imp1 eps2 ++ imp2).

  (** The result of the verifier-building functor *)
  Module Type FIRSTORDER_VER.
    Declare Module Param : FIRSTORDER_PARAM.

    Parameter verify : nat ->
      pred_option
      (cert (Build_abstraction Param.hasTyp Param.pconc Param.typ_eq_dec Param.astat_eq_dec)
	Param.isSafe Param.kernel_astat
	Param.entryPoints Param.imported).
  End FIRSTORDER_VER.

  (** Construct a verifier *)
  Declare Module Make (Param : FIRSTORDER_PARAM)
    : FIRSTORDER_VER with Module Param := Param.
End FIRSTORDER.

Module FirstOrder : FIRSTORDER.
  Import GenericTypes.

  Definition asafe (typ astat ctxt : Set)
    (hasTyp : ctxt -> int32 -> typ -> Prop)
    (pconc : ctxt -> astat -> state -> Prop)
    (isSafe : instr -> state -> Prop)
    (fp : list (list (GenericTypes.astate typ astat) * GenericTypes.astate typ astat))
    hyps abs :=
    forall fp', incl fp fp'
      -> forall ctx : (GenericTypes.context ctxt),
      AllS
      (fun hyp : GenericTypes.astate typ astat =>
	forall st : state,
          GenericTypes.conc hasTyp pconc hyp st ctx ->
          relatively_safe (stateless_mexec exec isSafe) get_instr 
	  (GenericTypes.conc hasTyp pconc) fp' st) hyps ->
      forall st : state,
	GenericTypes.conc hasTyp pconc abs st ctx ->
	relatively_safe (stateless_mexec exec isSafe) get_instr
	(GenericTypes.conc hasTyp pconc) fp' st.

  Definition initials' (typ astat ctxt : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
    (entryPoints : list (entryPoint typ astat))
    (cutpoints : list (list (cutpoint_info (ty typ) astat) * cutpoint_info (ty typ) astat)) :=
    map (B := (list (cutpoint_info (ty typ) astat) * cutpoint_info (ty typ) astat)) (initial_cp hasTyp) entryPoints
    ++ cutpoints.

  Definition initials (typ astat ctxt : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
    (kernel_astat : option astat)
    (entryPoints : list (entryPoint typ astat))
    (cutpoints : list (list (cutpoint_info (ty typ) astat) * cutpoint_info (ty typ) astat)) :=
    match kernel_astat with
      | None => initials' hasTyp entryPoints cutpoints
      | Some abs => kernel_cp typ abs :: initials' hasTyp entryPoints cutpoints
    end.
    
  Definition entries (typ astat ctxt : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
    (kernel_astat : option astat)
    (entryPoints : list (entryPoint typ astat))
    (cutpoints : list (list (cutpoint_info (ty typ) astat) * cutpoint_info (ty typ) astat)) :=
    map (B := list (astate typ astat) * astate typ astat) (@make_init typ astat)
    (initials hasTyp kernel_astat entryPoints cutpoints).

  Definition importd' (typ astat ctxt : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
    (imported : list (entryPoint typ astat)) :=
    map (B := (list (cutpoint_info (ty typ) astat) * cutpoint_info (ty typ) astat))
    (initial_cp hasTyp) imported.

  Definition importd (typ astat ctxt : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
    (imported : list (entryPoint typ astat)) :=
    map (B := list (astate typ astat) * astate typ astat) (@make_init typ astat)
    (importd' hasTyp imported).

  Definition fixedPoint (typ astat ctxt : Set) (hasTyp : ctxt -> int32 -> typ -> Prop)
    (kernel_astat : option astat)
    (entryPoints : list (entryPoint typ astat))
    (imported : list (entryPoint typ astat))
    (cutpoints : list (list (cutpoint_info (ty typ) astat) * cutpoint_info (ty typ) astat)) :=
    entries hasTyp kernel_astat entryPoints cutpoints
    ++ importd hasTyp imported.

  Inductive _cert (abstr : abstraction)
    (isSafe : instr -> state -> Prop)
    (kernel_astat : option (absAstat abstr))
    (entryPoints imported : list (entryPoint (absTyp abstr) (absAstat abstr)))
    : Prop :=
    | cert_rule : forall (cutpoints : list (list (cutpoint_info (ty (absTyp abstr)) (absAstat abstr))
      * cutpoint_info (ty (absTyp abstr)) (absAstat abstr))),
      (forall fp,
	incl (fixedPoint (absHasTyp abstr) kernel_astat entryPoints imported cutpoints) fp
	-> forall (hyps : list (astate (absTyp abstr) (absAstat abstr))) (abs : astate (absTyp abstr) (absAstat abstr)),
	  In (hyps, abs) (entries (absHasTyp abstr) kernel_astat entryPoints cutpoints) ->
	  asafe (absHasTyp abstr) (absPconc abstr) isSafe fp hyps abs)
      -> _cert abstr isSafe kernel_astat entryPoints imported.

  Definition cert := _cert.
  
  Lemma entries_incl_fixedPoint : forall typ astat ctxt hasTyp kastat
    entryPoints imported cutpoints,
    incl
    (entries hasTyp kastat entryPoints cutpoints)
    (@fixedPoint typ astat ctxt hasTyp kastat entryPoints imported cutpoints).
    intros.
    unfold fixedPoint.
    apply incl_appl.
    apply incl_refl.
  Qed.

  Lemma entries_is_fixedPoint : forall typ astat ctxt hasTyp kastat
    entryPoints cutpoints,
    entries hasTyp kastat entryPoints cutpoints
    = @fixedPoint typ astat ctxt hasTyp kastat entryPoints nil cutpoints.
    intuition.
    unfold fixedPoint, importd.
    simpl.
    apply app_nil_end.
  Qed.

  Definition kernel_ctx (ctxt : Set) (ctx : ctxt) :=
    Build_context initialESP initialESP initialMem None ctx.

  Lemma kernel_conc_pc : forall (typ astat ctxt : Set) (a : astat) st (ctx : ctxt),
    initState st
    -> conc_pc (typ := typ)
    (make_init_state
      (Build_cutpoint_info (apcConst initialPc) stackSize nil nil a)) st
    (kernel_ctx ctx).
    intros.
    red; simpl; trivial.
    inversion H.
    trivial.
  Qed.

  Hint Resolve kernel_conc_pc : asm.

  Lemma kernel_pcInRange : forall st,
    initState st
    -> pcInRange (stPc st).
    intros st Hinit.
    inversion Hinit.
    rewrite initPc.
    split; omega.
  Qed.

  Hint Resolve kernel_pcInRange : asm.

  Lemma kernel_agree_range : forall st,
    initState st
    -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr).
    intros st Hinit.
    inversion Hinit.
    red; intros; congruence.
  Qed.

  Hint Resolve kernel_agree_range : asm.

  Lemma kernel_conc_regs : forall (typ astat ctxt : Set) hasTyp (a : astat) st (ctx : context ctxt),
    conc_regs (typ := typ) hasTyp
    (make_init_state
      (Build_cutpoint_info (apcConst initialPc) stackSize nil nil a)) st
    ctx.
    intros until ctx; intro r.
    unfold aget_reg, GenericTypes.aget_reg; simpl.
    rewrite Reg32Map.sel_empty.
    apply HasType_Top.
  Qed.

  Hint Resolve kernel_conc_regs : asm.

  Lemma kernel_stackLow : forall (typ astat ctxt : Set) (a : astat) (ctx : ctxt),
    bvec_to_nat stackStart +
    aStackSize (typ := typ)
    (make_init_state
      (Build_cutpoint_info (apcConst initialPc) stackSize nil nil a)) <
    bvec_to_nat (sptr (kernel_ctx ctx)).
    intros until ctx.
    unfold make_init_state, kernel_ctx, aStackSize, ciStackSize, sptr.
    apply initialESP_location.
  Qed.

  Hint Resolve kernel_stackLow : asm.

  Lemma kernel_conc_slots : forall (typ astat ctxt : Set) hasTyp (a : astat) st (ctx : context ctxt),
    conc_slots (typ := typ) hasTyp
    (make_init_state
      (Build_cutpoint_info (apcConst initialPc) stackSize nil nil a)) st
    ctx.
    intros until ctx; intros sl Hsl.
    unfold aget_stack, GenericTypes.aget_stack; simpl.
    rewrite NatMap.sel_empty.
    apply HasType_Top.
  Qed.

  Hint Resolve kernel_conc_slots : asm.

  Lemma kernel_agree_range' : forall (ctxt : Set) st (ctx : ctxt) x y,
    initState st
    -> agree_range' (stMem st) (memIn (kernel_ctx ctx)) x y.
    intros until y; intros Hinit.
    inversion Hinit.
    rewrite initMem.
    simpl.
    intros addr Haddr.
    trivial.
  Qed.

  Hint Resolve kernel_agree_range' : asm.

  Theorem kernel_conc : forall (typ astat ctxt : Set)
    hasTyp (pconc : ctxt -> astat -> state -> Prop) (a : astat) st (ctx : ctxt),
    initState st
    -> pconc ctx a st
    -> conc (typ := typ) hasTyp pconc
    (make_init_state
      (Build_cutpoint_info (apcConst initialPc) stackSize nil nil a)) st
    (kernel_ctx ctx).
    intros a st ctx Hpconc.
    split; trivial.
    split; myeauto.
  Qed.

  Definition kernel : forall abstr isSafe abs entryPoints,
    cert abstr isSafe (Some abs) entryPoints nil
    -> forall st ctx, initState st
      -> absPconc abstr ctx abs st
      -> safe (Safety.cstep (stateless_mexec exec isSafe) get_instr) st.
    intros until entryPoints.
    intros Hcert st ctx Hinit Hpconc.
    inversion Hcert as [cutpoints Hchecked].
    assert (Hsafe : rsafe (stateless_mexec exec isSafe) get_instr st).
    apply relativity with
      (astate (absTyp abstr) (absAstat abstr))
      (context (absCtxt abstr))
      (conc (absHasTyp abstr) (absPconc abstr))
      (fixedPoint (absHasTyp abstr) (Some abs) entryPoints nil cutpoints).

    (* Prove that the fixed point is verified *)
    intros hyps abs' Hin ctx' Hall st' Hconc.
    unfold asafe in Hchecked.
    apply Hchecked with (fixedPoint (absHasTyp abstr) (Some abs) entryPoints nil cutpoints) hyps abs' ctx'; myeauto.
    apply incl_refl.
    rewrite entries_is_fixedPoint; trivial.
    apply incl_refl.

    (* Prove that the kernel entry point is verified *)
    unfold asafe in Hchecked.
    apply Hchecked with
      (fixedPoint (absHasTyp abstr) (Some abs) entryPoints nil cutpoints)
      (fst (make_init (kernel_cp (absTyp abstr) abs)))
      (snd (make_init (kernel_cp (absTyp abstr) abs)))
      (kernel_ctx ctx); simpl; myeauto.
    apply incl_refl.
    apply incl_refl.
    apply allS_nil.
    apply kernel_conc; trivial.

    trivial.
  Qed.

  Lemma map_app : forall (T S : Set) (f : T -> S) (ls1 ls2 : list T),
    map f (ls1 ++ ls2) = map f ls1 ++ map f ls2.
    induction ls1; intuition.
    simpl.
    rewrite IHls1.
    trivial.
  Qed.
  
  Hint Resolve in_app_or in_or_app.

  Lemma link_initials' : forall abstr eps1 cp1 eps2 cp2 x,
    In x
    (initials' (absHasTyp abstr) (eps1 ++ eps2) (cp1 ++ cp2))
    -> In x (initials' (astat := absAstat abstr) (absHasTyp abstr) eps1 cp1)
    \/ In x (initials' (absHasTyp abstr) eps2 cp2).
  unfold initials'.
  intros until x.
  intro Hin.
  destruct (in_app_or _ _ _ Hin) as [Hin' | Hin'];
    [rewrite map_app in Hin' | idtac];
    destruct (in_app_or _ _ _ Hin') as [Hin'' | Hin'']; auto.
  Qed.

  Hint Resolve link_initials'.

  Lemma link_initials : forall abstr ka1 eps1 cp1 ka2 eps2 cp2 x,
    In x
    (initials (absHasTyp abstr) ka1 (eps1 ++ eps2) (cp1 ++ cp2))
    -> In x (initials (astat := absAstat abstr) (absHasTyp abstr) ka1 eps1 cp1)
    \/ In x (initials (absHasTyp abstr) ka2 eps2 cp2).
    unfold initials.
    intros until x.
    intro Hin.
    destruct ka1; destruct ka2; auto.

    destruct Hin as [Hin' | Hin']; subst; simpl; auto.
    add Hinit (link_initials' _ _ _ _ _ _ Hin').
    destruct Hinit as [Hinit' | Hinit']; auto.
    
    destruct Hin as [Hin' | Hin']; subst; simpl; auto.
    add Hinit (link_initials' _ _ _ _ _ _ Hin').
    destruct Hinit as [Hinit' | Hinit']; auto.
    
    add Hinit (link_initials' _ _ _ _ _ _ Hin).
    destruct Hinit as [Hinit' | Hinit']; simpl; auto.
  Qed.

  Lemma In_map : forall (T S : Set) (f : T -> S) ls x,
    In x (map f ls)
    -> exists y, In y ls
      /\ x = f y.
    induction ls; simpl; intuition.
    eauto.
    destruct (IHls x H0).
    exists x0.
    tauto.
  Qed.

  Hint Resolve in_map.

  Lemma link_entries : forall abstr ka1 eps1 cp1 ka2 eps2 cp2 x,
    In x
    (entries (absHasTyp abstr) ka1 (eps1 ++ eps2) (cp1 ++ cp2))
    -> In x (entries (astat := absAstat abstr) (absHasTyp abstr) ka1 eps1 cp1)
    \/ In x (entries (absHasTyp abstr) ka2 eps2 cp2).
    unfold entries.
    intros until x.
    intro Hin.
    destruct (In_map _ _ _ Hin) as [y [Hyin Hyeq]]; subst.
    destruct (link_initials _ _ _ _ ka2 _ _ _ Hyin) as [Hin' | Hin']; auto.
  Qed.

  Lemma link_importd' : forall abstr imp1 imp2 x,
    In x
    (importd' (absHasTyp abstr) (imp1 ++ imp2))
    -> In x (importd' (astat := absAstat abstr) (absHasTyp abstr) imp1)
    \/ In x (importd' (absHasTyp abstr) imp2).
    unfold importd'.
    intros until x.
    intro Hin.
    destruct (In_map _ _ _ Hin) as [y [Hyin Hyeq]]; subst.
    rewrite map_app in Hin.
    auto.
  Qed.

  Lemma link_importd : forall abstr imp1 imp2 x,
    In x
    (importd (absHasTyp abstr) (imp1 ++ imp2))
    -> In x (importd (astat := absAstat abstr) (absHasTyp abstr) imp1)
    \/ In x (importd (absHasTyp abstr) imp2).
    unfold importd.
    intros until x.
    intro Hin.
    destruct (In_map _ _ _ Hin) as [y [Hyin Hyeq]]; subst.
    destruct (link_importd' _ _ _ _ Hyin) as [Hin' | Hin']; auto.
  Qed.

  Lemma link_fixedPoint : forall abstr ka1 eps1 imp1 cp1 ka2 eps2 imp2 cp2 x,
    In x
    (fixedPoint (absHasTyp abstr) ka1 (eps1 ++ eps2)
      (filterCovered (entryPoint_eq_dec abstr) imp1 eps2 ++ imp2)
      (cp1 ++ cp2))
    -> In x (fixedPoint (absHasTyp abstr) ka1 eps1 imp1 cp1)
    \/ In x (fixedPoint (absHasTyp abstr) ka2 eps2 imp2 cp2).
    unfold fixedPoint.
    intros until x.
    intro Hin.
    destruct (in_app_or _ _ _ Hin) as [Hin' | Hin'].
    destruct (link_entries _ _ _ _ ka2 _ _ _ Hin') as [Hin'' | Hin'']; auto.
    destruct (link_importd _ _ _ _ Hin') as [Hin'' | Hin'']; auto.
    unfold importd, importd' in Hin''.
    destruct (In_map _ _ _ Hin'') as [y [Hyin Hyeq]]; subst.
    destruct (In_map _ _ _ Hyin) as [z [Hzin Hzeq]]; subst.
    add Hcov (In_filterCovered2 (entryPoint_eq_dec abstr) _ _ _ Hzin).
    left.
    apply in_or_app.
    right.
    unfold importd, importd'.
    auto.
  Qed.

  Lemma In_initials' : forall abstr eps1 cp1 eps2 cp2 x,
    In x (initials' (astat := absAstat abstr) (absHasTyp abstr) eps1 cp1)
    -> In x
    (initials' (absHasTyp abstr) (eps1 ++ eps2) (cp1 ++ cp2)).
    unfold initials'.
    intros until x.
    intro Hin.
    destruct (in_app_or (map (initial_cp (astat:=absAstat abstr) (absHasTyp abstr)) eps1)
      cp1 x Hin) as [Hin' | Hin'].
    rewrite map_app; auto.
    auto.
  Qed.

  Lemma In_initials : forall abstr ka1 eps1 cp1 eps2 cp2 x,
    In x (initials (astat := absAstat abstr) (absHasTyp abstr) ka1 eps1 cp1)
    -> In x
    (initials (absHasTyp abstr) ka1 (eps1 ++ eps2) (cp1 ++ cp2)).
    unfold initials.
    intros until x.
    intro Hin.
    destruct ka1.
    destruct Hin as [Hin' | Hin']; subst; simpl; auto.
    generalize (In_initials' _ _ _ eps2 cp2 _ Hin'); auto.
    generalize (In_initials' _ _ _ eps2 cp2 _ Hin); auto.
  Qed.

  Lemma In_entries : forall abstr ka1 eps1 cp1 eps2 cp2 x,
    In x (entries (astat := absAstat abstr) (absHasTyp abstr) ka1 eps1 cp1)
    -> In x
    (entries (absHasTyp abstr) ka1 (eps1 ++ eps2) (cp1 ++ cp2)).
    unfold entries.
    intros until x.
    intro Hin.
    destruct (In_map _ _ _ Hin) as [y [Hyin Hyeq]]; subst.
    generalize (In_initials _ _ _ _ eps2 cp2 _ Hyin); auto.
  Qed.

  Lemma In_importd' : forall abstr imp1 imp2 x,
    In x (importd' (astat := absAstat abstr) (absHasTyp abstr) imp1)
    -> In x
    (importd' (absHasTyp abstr) (imp1 ++ imp2)).
    unfold importd'.
    intros until x.
    intro Hin.
    rewrite map_app; auto.
  Qed.

  Lemma In_importd : forall abstr imp1 imp2 x,
    In x (importd (astat := absAstat abstr) (absHasTyp abstr) imp1)
    -> In x
    (importd (absHasTyp abstr) (imp1 ++ imp2)).
    unfold importd.
    intros until x.
    intro Hin.
    destruct (In_map _ _ _ Hin) as [y [Hyin Hyeq]]; subst.
    generalize (In_importd' _ _ imp2 _ Hyin); auto.
  Qed.

  Lemma In_initials'2 : forall abstr eps1 cp1 eps2 cp2 x,
    In x (initials' (astat := absAstat abstr) (absHasTyp abstr) eps2 cp2)
    -> In x
    (initials' (absHasTyp abstr) (eps1 ++ eps2) (cp1 ++ cp2)).
    unfold initials'.
    intros until x.
    intro Hin.
    destruct (in_app_or (map (initial_cp (astat:=absAstat abstr) (absHasTyp abstr)) eps2)
      cp2 x Hin) as [Hin' | Hin'].
    rewrite map_app; auto.
    auto.
  Qed.

  Lemma In_initials2 : forall abstr ka1 eps1 cp1 eps2 cp2 x,
    In x (initials (astat := absAstat abstr) (absHasTyp abstr) None eps2 cp2)
    -> In x
    (initials (absHasTyp abstr) ka1 (eps1 ++ eps2) (cp1 ++ cp2)).
    unfold initials.
    intros until x.
    intro Hin.
    destruct ka1.
    generalize (In_initials'2 _ eps1 cp1 _ _ _ Hin); simpl; tauto.
    generalize (In_initials'2 _ eps1 cp1 _ _ _ Hin); simpl; tauto.
  Qed.

  Lemma In_entries2 : forall abstr ka1 eps1 cp1 eps2 cp2 x,
    In x (entries (astat := absAstat abstr) (absHasTyp abstr) None eps2 cp2)
    -> In x
    (entries (absHasTyp abstr) ka1 (eps1 ++ eps2) (cp1 ++ cp2)).
    unfold entries.
    intros until x.
    intro Hin.
    destruct (In_map _ _ _ Hin) as [y [Hyin Hyeq]]; subst.
    generalize (In_initials2 _ ka1 eps1 cp1 _ _ _ Hyin); auto.
  Qed.

  Lemma In_importd'2 : forall abstr imp1 imp2 x,
    In x (importd' (astat := absAstat abstr) (absHasTyp abstr) imp2)
    -> In x
    (importd' (absHasTyp abstr) (imp1 ++ imp2)).
    unfold importd'.
    intros until x.
    intro Hin.
    rewrite map_app; auto.
  Qed.

  Lemma In_importd2 : forall abstr imp1 imp2 x,
    In x (importd (astat := absAstat abstr) (absHasTyp abstr) imp2)
    -> In x
    (importd (absHasTyp abstr) (imp1 ++ imp2)).
    unfold importd.
    intros until x.
    intro Hin.
    destruct (In_map _ _ _ Hin) as [y [Hyin Hyeq]]; subst.
    generalize (In_importd'2 _ imp1 _ _ Hyin); auto.
  Qed.

  Lemma In_fixedPoint : forall abstr ka1 eps1 imp1 cp1 eps2 imp2 cp2 x,
    In x (fixedPoint (astat := absAstat abstr) (absHasTyp abstr) ka1 eps1 imp1 cp1)
    -> In x
    (fixedPoint (absHasTyp abstr) ka1 (eps1 ++ eps2)
      (filterCovered (entryPoint_eq_dec abstr) imp1 eps2 ++ imp2)
      (cp1 ++ cp2)).
    unfold fixedPoint.
    intros until x.
    intro Hin.
    destruct (in_app_or _ _ _ Hin) as [Hin' | Hin'].
    generalize (In_entries _ _ _ _ eps2 cp2 _ Hin'); auto.
    apply in_or_app.
    unfold importd in Hin'.
    destruct (In_map _ _ _ Hin') as [y [Hyin Hyeq]]; subst.
    unfold importd in Hyin.
    destruct (In_map _ _ _ Hyin) as [z [Hzin Hzeq]]; subst.
    add Hwhich (@In_filterCovered (entryPoint (absTyp abstr) (absAstat abstr)) (entryPoint_eq_dec abstr) z
      imp1 eps2 Hzin).
    destruct Hwhich as [Hin'' | Hin''].
    right.
    unfold importd, importd'; auto.
    left.
    apply In_entries2.
    unfold entries, initials, initials'.
    auto.
  Qed.

  Lemma incl_fixedPoint : forall abstr ka1 eps1 imp1 cp1 eps2 imp2 cp2 fp,
    incl
    (fixedPoint (absHasTyp abstr) ka1 (eps1 ++ eps2)
      (filterCovered (entryPoint_eq_dec abstr) imp1 eps2 ++ imp2)
      (cp1 ++ cp2)) fp
    -> incl (fixedPoint (absHasTyp abstr) ka1 eps1 imp1 cp1) fp.
    intros until fp.
    intros Hincl x Hin.
    add Hfix (In_fixedPoint _ _ _ _ _ eps2 imp2 cp2 _ Hin).
    auto.
  Qed.

  Hint Resolve incl_fixedPoint.
  
  Lemma In_fixedPoint2 : forall abstr ka1 eps1 imp1 cp1 eps2 imp2 cp2 x,
    In x (fixedPoint (astat := absAstat abstr) (absHasTyp abstr) None eps2 imp2 cp2)
    -> In x
    (fixedPoint (absHasTyp abstr) ka1 (eps1 ++ eps2)
      (filterCovered (entryPoint_eq_dec abstr) imp1 eps2 ++ imp2)
      (cp1 ++ cp2)).
    unfold fixedPoint.
    intros until x.
    intro Hin.
    destruct (in_app_or _ _ _ Hin) as [Hin' | Hin'].
    generalize (In_entries2 _ ka1 eps1 cp1 _ _  _ Hin'); auto.
    apply in_or_app.
    unfold importd in Hin'.
    destruct (In_map _ _ _ Hin') as [y [Hyin Hyeq]]; subst.
    unfold importd in Hyin.
    destruct (In_map _ _ _ Hyin) as [z [Hzin Hzeq]]; subst.
    unfold importd, importd'.
    auto 6.
  Qed.

  Lemma incl_fixedPoint2 : forall abstr ka1 eps1 imp1 cp1 eps2 imp2 cp2 fp,
    incl
    (fixedPoint (absHasTyp abstr) ka1 (eps1 ++ eps2)
      (filterCovered (entryPoint_eq_dec abstr) imp1 eps2 ++ imp2)
      (cp1 ++ cp2)) fp
    -> incl (fixedPoint (absHasTyp abstr) None eps2 imp2 cp2) fp.
    intros until fp.
    intros Hincl x Hin.
    add Hfix (In_fixedPoint2 _ ka1 eps1 imp1 cp1 _ _ _ _ Hin).
    auto.
  Qed.

  Hint Resolve incl_fixedPoint2.

  Definition link : forall abstr isSafe ka1 eps1 imp1 eps2 imp2,
    cert abstr isSafe ka1 eps1 imp1
    -> cert abstr isSafe None eps2 imp2
    -> cert abstr isSafe ka1 (eps1 ++ eps2)
    (filterCovered (entryPoint_eq_dec abstr) imp1 eps2 ++ imp2).
    intros until imp2; intros Hcert1 Hcert2.
    destruct Hcert1 as [cp1 Hcert1].
    destruct Hcert2 as [cp2 Hcert2].
    apply cert_rule with (cp1 ++ cp2).
    unfold asafe in *; idtac.
    intros fp' Hfp' hyps abs Hin fp Hfp ctx Hall st Hconc.

    destruct (link_entries _ _ _ _ None _ _ _ Hin) as [Hin' | Hin'].
    eapply Hcert1; eauto.
    eapply Hcert2; eauto.
  Qed.

  Module Type FIRSTORDER_VER.
    Declare Module Param : FIRSTORDER_PARAM.
    
    Parameter verify : nat ->
      pred_option
      (cert (Build_abstraction Param.hasTyp Param.pconc Param.typ_eq_dec Param.astat_eq_dec)
	Param.isSafe Param.kernel_astat
	Param.entryPoints Param.imported).
  End FIRSTORDER_VER.

  Module Make2 (Param : FIRSTORDER_PARAM)
    : FIRSTORDER_VER with Module Param := Param.
    Module M := Make(Param).

    Definition verify : nat ->
      pred_option
      (cert (Build_abstraction Param.hasTyp Param.pconc Param.typ_eq_dec Param.astat_eq_dec)
	Param.isSafe Param.kernel_astat
	Param.entryPoints Param.imported).
      intro n.
      destruct (M.C.verify n) as [Hver |]; [apply PSome | apply PNone].
      apply cert_rule with M.cutpoints.
      unfold asafe.
      intros.
      eapply Hver; eauto.
      eapply incl_tran; eauto.
    Qed.

    Module Param := Param.
  End Make2.

  Module Make := Make2.
End FirstOrder.
