(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** A simple instantiation of FirstOrder, tracking parity of values *)

Require Import Arith Bvector.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.FirstOrder.
Require Import ProofOS.AbsInt.FirstOrderAlg.
Require Import ProofOS.Examples.SimpleFo.

Require Import ProofOS.Trusted.Env.

Import Datatypes.

Set Implicit Arguments.

Hint Constructors word_hasType.

Definition is_odd len (v : Bvector (S len)) :=
  match v with
    | Vcons true _ _ => true
    | _ => false
  end.

Definition is_even len (v : Bvector (S len)) :=
  match v with
    | Vcons false _ _ => true
    | _ => false
  end.

Module Param <: SIMPLE_FIRSTORDER_PARAM.
  Inductive typi : Set :=
    | Even
    | Odd.

  Definition typ := typi.

  Theorem typ_eq_dec : forall (t1 t2 : typ), {t1 = t2} + {t1 <> t2}.
    intros.
    decide equality t1 t2.
  Qed.

  Definition hasTyp (v : int32) t :=
    match t with
      | Even => is_even v = true
      | Odd => is_odd v = true
    end.

  Definition subTyp (t1 t2 : word_type typ) :=
    match t2 with
      | TyTop => true
      | TyConst i =>
	match t1 with
	  | TyConst i' => sum_to_bool (bvec_eq_dec i i')
	  | _ => false
	end
      | TyCustom t2 =>
	match t1 with
	  | TyCustom t1 => sum_to_bool (typ_eq_dec t1 t2)
	  | TyConst i =>
	    match t2 with
	      | Even => is_even i
	      | Odd => is_odd i
	    end
	  | _ => false
	end
    end.

  Theorem subTyp_refl : forall t, subTyp t t = true.
    destruct t; trivial.
    simpl; destruct (bvec_eq_dec i i);
      [trivial
	| wrong; tauto].
    simpl; destruct (typ_eq_dec t t);
      [trivial
	| wrong; tauto].
  Qed.

  Hint Constructors word_hasType.

  Theorem subTyp_sound : forall t1 t2,
    subTyp t1 t2 = true
    -> forall v, word_hasType (extraParam hasTyp) tt v t1
      -> word_hasType (extraParam hasTyp) tt v t2.
    intros t1 t2 Htyp v Hword.
    functional induction subTyp t1 t2; subst; inversion Htyp; inversion Hword; subst; auto.
    destruct (bvec_eq_dec i i') as [e | e];
      [rewrite e; trivial
	| discriminate].
    simpl in Htyp.
    destruct (typ_eq_dec t3 t0) as [e | e]; simpl in H0; try discriminate.
    rewrite <- e.
    trivial.
  Qed.
    
  Definition isSafe (_ : instr) (_ : state) := True.

  Theorem checkSafe : forall stat ins,
    pred_option (forall st,
      statusOk (extraParam hasTyp) stat tt st
      -> isSafe ins st).
    intros.
    apply PSome.
    intuition.
    red.
    trivial.
  Qed.

  Definition typeof_read : forall stat addr,
    sig_option (fun ty => forall st,
      statusOk (extraParam hasTyp) stat tt st
      -> word_hasType (extraParam hasTyp) tt (get_mem32 st (eval_addr addr st)) ty).
    intros.
    apply SNone.
  Qed.

  Definition check_write : forall stat addr vty,
    pred_option (forall st v,
      statusOk (extraParam hasTyp) stat tt st
      -> word_hasType (extraParam hasTyp) tt v vty
      -> not_too_high (eval_addr addr st)
      /\ ~(bvec_to_nat initialPc - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)
      /\ ~(bvec_to_nat stackStart - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat stackEnd + 4)).
    intros.
    apply PNone.
  Qed.
  
  Lemma mod2_bool_add_even1 : forall n m,
    mod2_bool (2 * n + 2 * m) = false.
    intros.
    unfold mod2_bool.
    unfold mod2.
    destruct (Euclid.eucl_dev 2 two_gt_0 (2 * n + 2 * m)).
    destruct (eq_nat_dec r 0); trivial; wrong; omega.
  Qed.

  Lemma mod2_bool_add_odd2 : forall n m,
    mod2_bool (2 * n + (1 + 2 * m)) = true.
    intros.
    unfold mod2_bool.
    unfold mod2.
    destruct (Euclid.eucl_dev 2 two_gt_0 (2 * n + (1 + 2 * m))).
    destruct (eq_nat_dec r 0); trivial; wrong; omega.
  Qed.

  Lemma mod2_bool_add_odd1 : forall n m,
    mod2_bool (1 + 2 * n + 2 * m) = true.
    intros.
    unfold mod2_bool.
    unfold mod2.
    destruct (Euclid.eucl_dev 2 two_gt_0 (1 + 2 * n + 2 * m)).
    destruct (eq_nat_dec r 0); trivial; wrong; omega.
  Qed.

  Lemma mod2_bool_add_even2 : forall n m,
    mod2_bool (1 + 2 * n + (1 + 2 * m)) = false.
    intros.
    unfold mod2_bool.
    unfold mod2.
    destruct (Euclid.eucl_dev 2 two_gt_0 (1 + 2 * n + (1 + 2 * m))).
    destruct (eq_nat_dec r 0); trivial; wrong; omega.
  Qed.

  Definition add_typ : forall t1 t2,
    {t : typ
      | forall v1 v2,
	hasTyp v1 t1
	-> hasTyp v2 t2
	-> hasTyp (badd32 v1 v2) t}.
    unfold int32; intros.
    destruct t1; destruct t2;
      [exists Even | exists Odd | exists Odd | exists Even];
      simpl; intros v1 v2 Hv1 Hv2;
	inversion Hv1 as [Hev1];
	  inversion Hv2 as [Hev2];
	    myVSntac Hv1' v1; rewrite Hv1' in Hev1; simpl in Hev1;
	      destruct (Vhead v1); try discriminate;
		myVSntac Hv2' v2; rewrite Hv2' in Hev2;
		  simpl in Hev2; destruct (Vhead v2); try discriminate;
		    unfold bvec_to_nat, is_even, is_odd; fold bvec_to_nat.
    rewrite mod2_bool_add_even1; trivial.
    rewrite mod2_bool_add_odd2; trivial.
    rewrite mod2_bool_add_odd1; trivial.
    rewrite mod2_bool_add_even2; trivial.
  Qed.

  Definition add_type : forall t1 t2,
    {t : word_type typ
      | forall v1 v2,
	hasTyp v1 t1
	-> hasTyp v2 t2
	-> word_hasType (extraParam hasTyp) tt (badd32 v1 v2) t}.
    intros.
    exists (TyCustom (proj1_sig (add_typ t1 t2))).
    intros.
    destruct (add_typ t1 t2).
    simpl.
    apply WHT_Custom.
    red.
    auto.
  Qed.

  Definition trans_const : forall i,
    {t : typ
      | forall v,
	word_hasType (extraParam hasTyp) tt v (TyConst typ i)
	-> hasTyp v t}.
    unfold int32 in *; intro.
    myVSntac Hi i.
    destruct (Vhead i).

    exists Odd.
    intros v Hv.
    inversion Hv.
    simpl; trivial.

    exists Even.
    intros v Hv.
    inversion Hv.
    simpl; trivial.
  Qed.

  Definition typeof_arith : forall aop t1 t2,
    sig_option (fun ty => forall v1 v2,
      word_hasType (extraParam hasTyp) tt v1 t1
      -> word_hasType (extraParam hasTyp) tt v2 t2
      -> word_hasType (extraParam hasTyp) tt (eval_arith aop v1 v2) ty).
    intros.
    destruct aop.

    Ltac no_idea := apply SSome with (TyTop typ); intuition.

    (* Add *)
    destruct t1; destruct t2.
    no_idea.
    no_idea.
    no_idea.
    no_idea.
    no_idea.
  
    (* Const vs. Custom *)
    destruct (trans_const i) as [ti Hti].
    destruct (add_type ti t) as [t' Ht'].
    apply SSome with t'.
    intros v1 v2 Hv1 Hv2.
    inversion Hv2; subst.
    auto.
  
    no_idea.
  
    (* Custom vs. Const *)
    destruct (trans_const i) as [ti Hti].
    destruct (add_type t ti) as [t' Ht'].
    apply SSome with t'.
    intros v1 v2 Hv1 Hv2.
    inversion Hv1; subst.
    auto.

    (* Custom vs. Custom *)
    destruct (add_type t t0) as [t' Ht'].
    apply SSome with t'.
    intros v1 v2 Hv1 Hv2.
    inversion Hv1; subst.
    inversion Hv2; subst.
    auto.

    (* Ignore other operators for now.... *)
    no_idea.
    no_idea.
    no_idea.
  Qed.
End Param.

Module VerParam := Make(Param).
Module Ver := FirstOrderAlg.FirstOrder.Make(VerParam).
