(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Pointers to integers in the data segment *)

Require Import Arith Bvector List Peano_dec TheoryList ZArith.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.FirstOrder.
Require Import ProofOS.AbsInt.FirstOrderAlg.

Require Import ProofOS.Machine.Stateless.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Env.

Import Datatypes.

Set Implicit Arguments.

Hint Constructors word_hasType.

Module Param <: FIRSTORDER_PARAM.
  Inductive _typ : Set :=
    | Int : _typ
    | List : _typ -> _typ
    | NeList : _typ -> _typ.

  Definition typ := _typ.

  Definition typ_eq_dec : forall (t1 t2 : typ), {t1 = t2} + {t1 <> t2}.
    intros.
    decide equality t1 t2.
  Qed.

  Definition ctxt := int32 -> option typ.
  (** We need to know which memory locations contain lists, and what type lists
    they are. *)

  Inductive _hasTyp (ctx : ctxt) : int32 -> typ -> Prop :=
    | HT_Int : forall v,
      _hasTyp ctx v Int

    | HT_List_real : forall t v,
      ctx v = Some t
      -> _hasTyp ctx v (List t)
    | HT_List_null : forall t,
      _hasTyp ctx bzero32 (List t)

    | HT_NeList_real : forall t v,
      ctx v = Some t
      -> _hasTyp ctx v (NeList t).

  Definition hasTyp := _hasTyp.

  Definition astat := option genop32.
  (** We only need to know which operand was checked for nullness. *)

  Definition astatOk (ropt : astat) (st : state) :=
    match ropt with
      | None => True
      | Some g => get_flag st ZF = eval_ZF Sub (eval_genop32 g st) bzero32
    end.
  
  Definition tl a := badd four a.

  Definition inData (addr : int32) :=
    bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4 < bvec_to_nat addr
    /\ bvec_to_nat addr + 8 < bvec_to_nat stackStart.

  Definition inData_dec addr : {inData addr} + {~inData addr}.
    intros.
    unfold inData.
    destruct (le_lt_dec (bvec_to_nat addr)
      (bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)).
    right; omega.
    destruct (le_lt_dec (bvec_to_nat stackStart)
      (bvec_to_nat addr + 8)).
    right; omega.
    left; omega.
  Qed.

  Definition consAligned (n : int32) :=
    exists k, bvec_to_nat n = 8 * k.

  Record cellOk (ctx : ctxt) (st : state) (a : int32) (t : typ) : Prop := {
    cellInData : inData a;
    cellAligned : consAligned a;
    hdTyp : hasTyp ctx (get_mem32 st a) t;
    tlTyp : hasTyp ctx (get_mem32 st (tl a)) (List t)
  }.

  Definition cellsOk (ctx : ctxt) (st : state) :=
    forall a,
      match ctx a with
	| None => True
	| Some t => cellOk ctx st a t
      end.
    
  Record _pconc (ctx : ctxt) (ropt : astat) (st : state) : Prop := {
    pconcAstat : astatOk ropt st;
    pconcCells : cellsOk ctx st
  }.

  Definition pconc := _pconc.

  Hint Unfold pconc.

  Definition genop32_eq_dec : forall (g1 g2 : genop32), {g1 = g2} + {g1 <> g2}.
    intros.
    decide equality g1 g2.
    apply (@bvec_eq_dec 32).
    apply reg32_eq_dec.
    decide equality a a0.
    decide equality addrIndex addrIndex0.
    decide equality a1 p.
    apply reg32_eq_dec.
    decide equality a2 s.
    decide equality addrBase addrBase0.
    apply reg32_eq_dec.
    apply (@bvec_eq_dec 32).
  Qed.

  Definition astat_eq_dec : forall (abs1 abs2 : astat), {abs1 = abs2} + {abs1 <> abs2}.
    intros.
    decide equality abs1 abs2.
    apply genop32_eq_dec.
  Qed.

  Hint Unfold astatOk.

  Definition subAstat : forall abs1 abs2, pred_option (forall st ctx,
    pconc ctx abs1 st -> pconc ctx abs2 st).
    intros.
    destruct abs2.
    destruct abs1.
    destruct (genop32_eq_dec g g0); [subst | apply PNone].
    apply PSome; trivial.
    apply PNone.
    apply PSome; intuition.
    split; intuition.
    inversion H; trivial.
  Qed.

  Definition minitState := initState.

  Definition kernel_astat := Some (@None genop32).

  Definition isSafe (_ : instr) (_ : state) := True.

  Definition entryPoints := @nil (entryPoint typ astat).
  Definition imported := @nil (entryPoint typ astat).

  Definition subTyp (t1 t2 : word_type typ) :=
    match t2 with
      | TyTop => true
      | TyConst n =>
	match t1 with
	  | TyConst n' => sum_to_bool (bvec_eq_dec n n')
	  | _ => false
	end
      | TyCustom Int => true
      | TyCustom (List s2) =>
	match t1 with
	  | TyConst n => sum_to_bool (bvec_eq_dec n bzero32)
	  | TyCustom (List s1) => sum_to_bool (typ_eq_dec s1 s2)
	  | TyCustom (NeList s1) => sum_to_bool (typ_eq_dec s1 s2)
	  | _ => false
	end
      | TyCustom (NeList s2) =>
	match t1 with
	  | TyCustom (NeList s1) => sum_to_bool (typ_eq_dec s1 s2)
	  | _ => false
	end
    end.

  Theorem subTyp_refl : forall t, subTyp t t = true.
    destruct t; simpl; intuition.
    destruct (bvec_eq_dec i i); intuition.
    destruct t; trivial;
      destruct (typ_eq_dec t t); intuition.
  Qed.

  Hint Constructors _hasTyp.

  Theorem subTyp_sound : forall t1 t2,
    subTyp t1 t2 = true
    -> forall ctx v, word_hasType hasTyp ctx v t1
      -> word_hasType hasTyp ctx v t2.
    unfold hasTyp.
    intros t1 t2 Hsub.
    functional induction subTyp t1 t2; subst; simpl in Hsub; try discriminate; myauto.
    destruct (bvec_eq_dec n n') as [e | e]; intuition;
      [rewrite e; trivial
	| discriminate].
    destruct (bvec_eq_dec n bzero32) as [e | e]; intuition;
      [rewrite e in H; inversion H; intuition
	| discriminate].
    destruct (typ_eq_dec s1 s2) as [e | e];
      [rewrite e; auto
	| discriminate].
    destruct (typ_eq_dec s1 s2) as [e | e];
      [rewrite e; intros ctx v Ht1; inversion Ht1; inversion H2; auto
	| discriminate].
    destruct (typ_eq_dec s1 s2) as [e | e];
      [rewrite e; auto
	| discriminate].
  Qed.

  Definition checkSafe : forall abs stat ins,
    pred_option (forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> isSafe ins st).
    intros.
    apply PSome.
    intuition.
    red; trivial.
  Qed.

  Ltac type_fail := intros; apply SNone.

  Definition typeof_read : forall abs stat addr,
    sig_option (fun ty => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx (get_mem32 st (eval_addr addr st)) ty).
    intros.
    destruct addr.

    destruct addrIndex; [type_fail | idtac].
    destruct addrBase; [idtac | type_fail].
    caseEq (genopType stat (Reg r)); [type_fail | type_fail | idtac].
    intros t Ht.
    destruct t; [type_fail | type_fail | idtac].

    (* Reading the data cell *)
    destruct (bvec_eq_dec addrDisp bzero32); [subst | idtac].
    apply SSome with (TyCustom t).
    intros st ctx Hpconc Hstat.
    add Hr (genopTypeOk Hstat (Reg r)).
    rewrite Ht in Hr.
    inversion Hr; subst.
    inversion H2; subst.
    simpl in H1.
    assert (Heq : eval_addr (Build_address bzero32 (Some r) None) st = get_reg32 st r).
    unfold eval_addr, addrIndex, addrBase, addrDisp.
    unfold badd32, bzero32; rewrite badd_zero'; trivial.
    rewrite Heq.
    add Haddr (pconcCells Hpconc (eval_addr (Build_address bzero32 (Some r) None) st)).
    rewrite Heq in Haddr.
    rewrite H1 in Haddr.
    inversion Haddr; auto.
    
    (* Reading the next-pointer cell *)
    clear n.
    destruct (bvec_eq_dec addrDisp four); [subst | idtac].
    apply SSome with (TyCustom (List t)).
    intros st ctx Hpconc Hstat.
    add Hr (genopTypeOk Hstat (Reg r)).
    rewrite Ht in Hr.
    inversion Hr; subst.
    inversion H2; subst.
    simpl in H1.
    assert (Heq : eval_addr (Build_address four (Some r) None) st = badd32 four (get_reg32 st r));
      trivial.
    rewrite Heq.
    add Haddr (pconcCells Hpconc (get_reg32 st r)).
    rewrite H1 in Haddr.
    inversion Haddr; auto.

    type_fail.
  Qed.

  Ltac pred_fail := intros; apply PNone.

  Theorem badd_btn : forall len (v1 v2 : Bvector len),
    eqMod (pow2 len) (bvec_to_nat (badd v1 v2))
    (bvec_to_nat v1 + bvec_to_nat v2).
    intros len v1 v2.
    unfold badd, barith.
    apply ntb_invert_mod.
  Qed.

  Lemma ntb_O : forall len,
    bvec_to_nat (nat_to_bvec len 0) = 0.
    intros.
    rewrite ntb_invert; trivial.
    generalize (pow2_pos len); omega.
  Qed.

  Lemma btn_ntb_four : forall len,
    bvec_to_nat (nat_to_bvec (S (S (S len))) 4) = 4.
    intro.
    simpl.
    unfold mod2_bool, mod2, div2.
    destruct (Euclid.eucl_dev 2 two_gt_0 4).
    assert (q = 2); [omega | subst].
    assert (r = 0); [omega | subst].
    destruct (eq_nat_dec 0 0); try tauto.
    destruct (Euclid.eucl_dev 2 two_gt_0 2).
    assert (q = 1); [omega | subst].
    assert (r = 0); [omega | subst].
    destruct (eq_nat_dec 0 0); try tauto.
    destruct (Euclid.eucl_dev 2 two_gt_0 1).
    assert (q = 0); [omega | subst].
    assert (r = 1); [omega | subst].
    destruct (eq_nat_dec 1 0); try (wrong; omega).
    rewrite ntb_O.
    omega.
  Qed.

  Lemma ge_plus : forall x y,
    x + y >= x.
    induction x; intuition; omega.
  Qed.

  Lemma mult_S : forall x y,
    S x * y >= y.
    intros.
    simpl.
    apply ge_plus.
  Qed.

  Lemma eqMod_eq : forall m x y,
    eqMod m x y
    -> x < m
    -> y < m
    -> x = y.
    intros m x y Hmod Hx Hy.
    destruct (eqMod_nat Hmod) as [k Hk].
    destruct k.
    omega.
    add HS (mult_S k m).
    omega.
  Qed.

  Lemma tail_addr : forall a,
    bvec_to_nat a + 8 < bvec_to_nat stackStart
    -> bvec_to_nat (badd32 four a) =
    4 + bvec_to_nat a.
    intros a Hlt.
    add HstackFits stack_fits.
    add Hfour (@btn_ntb_four 29).
    replace (4 + bvec_to_nat a) with (bvec_to_nat four + bvec_to_nat a).
    assert (Hmod : eqMod (pow2 32)
      (bvec_to_nat (badd32 four a))
      (bvec_to_nat four + bvec_to_nat a)).
    unfold badd32.
    apply badd_btn.
    apply eqMod_eq with (pow2 32).
    trivial.
    apply btn_upper.
    unfold four.
    rewrite btn_ntb_four.
    omega.
    unfold four.
    omega.
  Qed.

  Definition check_write : forall abs stat addr vty,
    pred_option (forall st ctx v,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx v vty
      -> not_too_high (eval_addr addr st)
      /\ ~(bvec_to_nat initialPc - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)
      /\ ~(bvec_to_nat stackStart - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat stackEnd + 4)).
    intros.
    destruct addr.
    destruct addrIndex; [pred_fail | idtac].
    destruct addrBase; [idtac | pred_fail].
    caseEq (genopType stat (Reg r)); [pred_fail | pred_fail | idtac].
    intros t Ht.
    destruct t; [pred_fail | pred_fail | idtac].
    destruct (bvec_eq_dec addrDisp bzero32); [subst | idtac].

    (* Writing to the data cell *)
    apply PSome.
    intros st ctx v Hpconc Hstat Hvty.
    assert (Heq : eval_addr (Build_address bzero32 (Some r) None) st = get_reg32 st r).
    unfold eval_addr, addrIndex, addrBase, addrDisp.
    unfold badd32, bzero32; rewrite badd_zero'; trivial.
    rewrite Heq.
    add Hr (genopTypeOk Hstat (Reg r)).
    rewrite Ht in Hr.
    simpl in Hr.
    inversion Hr; subst.
    clear Hr; rename H2 into Hr.
    inversion Hr; subst.

    add Haddr (pconcCells Hpconc (get_reg32 st r)).
    rewrite H1 in Haddr.
    inversion Haddr.
    unfold inData in cellInData0.
    add HstackFits stack_fits.
    split;
      [red; split; omega
	| omega].

    destruct (bvec_eq_dec addrDisp four); [subst | idtac].

    (* Writing to the next-pointer cell *)
    apply PSome.
    intros st ctx v Hpconc Hstat Hvty.
    assert (Heq : eval_addr (Build_address four (Some r) None) st = badd32 four (get_reg32 st r)).
    unfold eval_addr, addrIndex, addrBase, addrDisp; trivial.
    rewrite Heq.
    add Hr (genopTypeOk Hstat (Reg r)).
    rewrite Ht in Hr.
    simpl in Hr.
    inversion Hr; subst.
    clear Hr; rename H2 into Hr.
    inversion Hr; subst.

    add Haddr (pconcCells Hpconc (get_reg32 st r)).
    rewrite H1 in Haddr.
    inversion Haddr.
    unfold inData in cellInData0.
    add HstackFits stack_fits.
    assert (Heq' : bvec_to_nat (badd32 four (get_reg32 st r)) = 4 + bvec_to_nat (get_reg32 st r)).
    apply tail_addr; try tauto.
    rewrite Heq'.
    split;
      [red; split; omega
	| omega].

    pred_fail.
  Qed.

  Definition typeof_arith : forall aop t1 t2,
    sig_option (fun ty => forall ctx v1 v2,
      word_hasType hasTyp ctx v1 t1
      -> word_hasType hasTyp ctx v2 t2
      -> word_hasType hasTyp ctx (eval_arith aop v1 v2) ty).
    intros.
    apply SNone.
  Qed.

  Lemma get_flag_reg : forall st r v f,
    get_flag (set_reg32 st r v) f = get_flag st f.
    trivial.
  Qed.

  Lemma get_mem32_reg : forall st r v a,
    get_mem32 (set_reg32 st r v) a = get_mem32 st a.
    trivial.
  Qed.

  Lemma cellsOk_reg : forall ctx st r v,
    cellsOk ctx st
    -> cellsOk ctx (set_reg32 st r v).
    intros ctx st r v Hcells.
    intro a.
    add Ha (Hcells a).
    destruct (ctx a).
    destruct Ha.
    split; trivial.
    trivial.
  Qed.

  Hint Resolve cellsOk_reg.

  Hint Constructors _pconc.
  
  Lemma pconc_reg : forall ctx abs st r v,
    pconc ctx abs st
    -> pconc ctx None (set_reg32 st r v).
    intros ctx abs st r v Hpconc.
    destruct Hpconc.
    intuition.
  Qed.

  Hint Resolve pconc_reg.

  Ltac awrite_fail := exists (@None genop32); intuition; eauto.

  Definition awrite_reg : forall abs r,
    {abs' : astat
      | forall ctx st v,
	pconc ctx abs st
	-> pconc ctx abs' (set_reg32 st r v)}.
    intros.
    destruct abs.
    destruct g.

    awrite_fail.

    destruct (reg32_eq_dec r0 r); subst.
    awrite_fail.
    exists (Some (Reg r0)).
    intros ctx st v Hpconc.
    inversion Hpconc.
    split; auto.
    red.
    red in pconcCells0.
    replace (get_flag (set_reg32 st r v) ZF)
      with (get_flag st ZF); trivial.
    unfold get_reg32, set_reg32; simpl.
    unfold get_reg32; simpl.
    rewrite Reg32Map.sel_neq; auto with arith.

    destruct a.
    destruct addrBase.
    destruct (reg32_eq_dec r r0); [awrite_fail | idtac].
    destruct addrIndex.
    destruct p.
    destruct (reg32_eq_dec r r1); [awrite_fail | idtac].

    exists (Some (Address reg32 (Build_address addrDisp (Some r0) (Some (s, r1))))).
    intros ctx st v Hpconc.
    inversion Hpconc.
    split.
    red; simpl.
    red in pconcAstat0; simpl in pconcAstat0.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp in Hpconc.
    replace (get_reg32 (set_reg32 st r v) r0)
      with (get_reg32 st r0).
    replace (get_reg32 (set_reg32 st r v) r1)
      with (get_reg32 st r1).
    trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.
    auto.

    exists (Some (Address reg32 (Build_address addrDisp (Some r0) None))).
    intros ctx st v Hpconc.
    inversion Hpconc.
    split; [idtac | auto].
    red; simpl.
    red in pconcCells0; simpl in pconcCells0.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp in Hpconc.
    replace (get_reg32 (set_reg32 st r v) r0)
      with (get_reg32 st r0).
    trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.

    destruct addrIndex.
    destruct p.
    destruct (reg32_eq_dec r r0); [awrite_fail | idtac].
    exists (Some (Address reg32 (Build_address addrDisp None (Some (s, r0))))).
    intros ctx st v Hpconc.
    inversion Hpconc.
    split; [idtac | auto].
    red; simpl.
    red in pconcCells0; simpl in pconcCells0.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp in Hpconc.
    replace (get_reg32 (set_reg32 st r v) r0)
      with (get_reg32 st r0).
    trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.

    exists (Some (Address reg32 (Build_address addrDisp None None))).
    intros ctx st v Hpconc.
    inversion Hpconc.
    split; [idtac | auto].
    red; simpl.
    red in pconcCells0; simpl in pconcCells0.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    trivial.

    awrite_fail.
  Qed.

  Lemma apart_or : forall a a',
    (bvec_to_nat a' + 4 <= bvec_to_nat a
      \/ bvec_to_nat a' >= bvec_to_nat a + 4)
    -> apart a a'.
    intros.
    destruct H.
    apply apart_lower; trivial.
    apply apart_higher; trivial.
  Qed.

  Lemma pow2_ge_four : forall len,
    pow2 (S (S len)) >= 4.
    intros.
    simpl.
    generalize (pow2_pos len).
    omega.
  Qed.

  Ltac addr_fail := intros; apply SNone.

  Lemma btn_inj_contra : forall len (v1 v2 : Bvector len),
    v1 <> v2
    -> bvec_to_nat v1 <> bvec_to_nat v2.
    intros len v1 v2 Hv Hbv.
    apply Hv.
    apply btn_inj; trivial.
  Qed.

  Lemma subTyp_custom : forall ctx v t t',
    subTyp t (TyCustom t') = true
    -> word_hasType hasTyp ctx v t
    -> hasTyp ctx v t'.
    intros ctx v t t' Hsub Hword.
    assert (Ht' : word_hasType hasTyp ctx v (TyCustom t')).
    eapply subTyp_sound; eauto.
    inversion Ht'; trivial.
  Qed.

  Hint Resolve subTyp_custom.

  Definition awrite_addr : forall abs stat addr t,
    sig_option (fun abs' =>
      forall ctx st v,
	pconc ctx abs st
	-> statusOk hasTyp stat ctx st
	-> word_hasType hasTyp ctx v t
	-> pconc ctx abs' (set_mem32 st (eval_addr addr st) v)).
    intros.
    destrEq Hstack (addrInStack stat addr).

    (* Stack write *)
    apply SSome with (@None genop32).
    intros ctx st v Hpconc Hstat Ht.
    inversion Hpconc.
    split; intuition.
    intro addr'.
    add Haddr' (pconcCells0 addr').
    destruct (ctx addr'); trivial.
    inversion Haddr'.
    add Hstack' (addrInStackOk Hstat addr Hstack).
    split.
    trivial.
    trivial.

    (* Head *)
    rewrite invert_set_mem32_neq; trivial.
    split.
    add Hfits stack_fits.
    unfold stackEnd in Hstack'.
    rewrite ntb_invert in Hstack'.
    omega.
    omega.
    red in cellInData0.
    split.
    generalize stack_fits.
    omega.
    apply apart_or.
    red in cellInData0.
    generalize (addrInStackOk Hstat addr Hstack); omega.

    (* Tail *)
    rewrite invert_set_mem32_neq; trivial.
    split.
    add Hfits stack_fits.
    unfold stackEnd in Hstack'.
    rewrite ntb_invert in Hstack'.
    omega.
    omega.
    red in cellInData0.
    split.
    unfold tl; rewrite tail_addr.
    generalize stack_fits.
    omega.
    tauto.
    unfold tl.
    replace (badd four addr') with (badd32 four addr'); trivial.
    apply apart_or; red in cellInData0.
    rewrite tail_addr.
    generalize (addrInStackOk Hstat addr Hstack); omega.
    omega.

    (* Non-stack write *)
    destruct addr.
    destruct addrBase; [idtac | addr_fail].
    destruct addrIndex; [addr_fail | idtac].
    caseEq (genopType stat (Reg r)); [addr_fail | addr_fail | idtac].
    intros t' Ht'.
    destruct t'; [addr_fail | addr_fail | idtac].

    destruct (bvec_eq_dec addrDisp bzero32); subst.

    (* Head *)
    destrEq Hsub (subTyp t (TyCustom t')); [idtac | addr_fail].
    apply SSome with (@None genop32).
    intros ctx st v Hpconc Hstat Ht.
    add Hr (genopTypeOk Hstat (Reg r)).
    rewrite Ht' in Hr.
    inversion Hr; subst.
    inversion H2; subst.
    simpl in H1.
    inversion Hpconc.
    split; auto.
    intro a.
    add Ha (pconcCells0 a).
    caseEq (ctx a); trivial.
    intros t'' Hctx.
    rewrite Hctx in Ha.
    replace (eval_addr (Build_address bzero32 (Some r) None) st)
      with (get_reg32 st r).
    destruct (bvec_eq_dec a (get_reg32 st r)); subst.
    
    (* Same cell *)
    rewrite H1 in Hctx.
    injection Hctx; intros; subst.
    inversion Ha.
    split.
    trivial.
    trivial.
    rewrite invert_set_mem32; eauto.
    rewrite invert_set_mem32_neq; trivial.
    split; red in cellInData0; generalize stack_fits; omega.
    red in cellInData0; generalize stack_fits; split; unfold tl; rewrite tail_addr; omega.
    apply apart_higher.
    unfold tl; rewrite tail_addr.
    omega.
    red in cellInData0; generalize stack_fits; omega.

    (* Different cells *)
    add Hcell (pconcCells0 (get_reg32 st r)).
    rewrite H1 in Hcell.
    inversion Ha.
    inversion Hcell.
    split.
    trivial.
    trivial.

    rewrite invert_set_mem32_neq; trivial.
    split; red in cellInData1; generalize stack_fits; omega.
    split; red in cellInData0; generalize stack_fits; omega.
    destruct cellAligned0 as [ka Hka].
    destruct cellAligned1 as [kr Hkr].
    apply apart_or.
    generalize (btn_inj_contra n); omega.
    
    rewrite invert_set_mem32_neq; trivial.
    split; red in cellInData1; generalize stack_fits; omega.
    split; unfold tl; rewrite tail_addr; red in cellInData0; generalize stack_fits; omega.
    apply apart_or.
    destruct cellAligned0 as [ka Hka].
    destruct cellAligned1 as [kr Hkr].
    unfold tl; rewrite tail_addr.
    generalize (btn_inj_contra n); omega.
    red in cellInData0; generalize stack_fits; omega.

    unfold eval_addr, addrIndex, addrBase, addrDisp.
    unfold badd32, bzero32; rewrite badd_zero'; trivial.

    destruct (bvec_eq_dec addrDisp four); subst.

    (* Tail *)
    destrEq Hsub (subTyp t (TyCustom (List t'))); [idtac | addr_fail].
    apply SSome with (@None genop32).
    intros ctx st v Hpconc Hstat Ht.
    add Hr (genopTypeOk Hstat (Reg r)).
    rewrite Ht' in Hr.
    inversion Hr; subst.
    inversion H2; subst.
    simpl in H1.
    inversion Hpconc.
    split; auto.
    intro a.
    add Ha (pconcCells0 a).
    caseEq (ctx a); trivial.
    intros t'' Hctx.
    rewrite Hctx in Ha.
    replace (eval_addr (Build_address four (Some r) None) st)
      with (badd32 four (get_reg32 st r)); trivial.
    destruct (bvec_eq_dec a (get_reg32 st r)); subst.
    
    (* Same cell *)
    rewrite H1 in Hctx.
    injection Hctx; intros; subst.
    inversion Ha.
    split.
    trivial.
    trivial.
    rewrite invert_set_mem32_neq; trivial.
    split; rewrite tail_addr; red in cellInData0; generalize stack_fits; omega.
    split; red in cellInData0; generalize stack_fits; omega.
    apply apart_lower.
    rewrite tail_addr; red in cellInData0; generalize stack_fits; omega.
    rewrite invert_set_mem32; trivial.
    assert (Hwty : word_hasType hasTyp ctx v (TyCustom (List t''))).
    eapply subTyp_sound; eauto.
    inversion Hwty; trivial.

    (* Different cells *)
    add Hcell (pconcCells0 (get_reg32 st r)).
    rewrite H1 in Hcell.
    inversion Ha.
    inversion Hcell.
    split.
    trivial.
    trivial.

    rewrite invert_set_mem32_neq; trivial.
    split; rewrite tail_addr; red in cellInData1; generalize stack_fits; omega.
    split; red in cellInData0; generalize stack_fits; omega.
    destruct cellAligned0 as [ka Hka].
    destruct cellAligned1 as [kr Hkr].
    apply apart_or.
    rewrite tail_addr.
    generalize (btn_inj_contra n); omega.
    red in cellInData1; generalize stack_fits; omega.
    
    rewrite invert_set_mem32_neq; trivial.
    split; unfold tl; rewrite tail_addr; red in cellInData1; generalize stack_fits; omega.
    split; unfold tl; rewrite tail_addr; red in cellInData0; generalize stack_fits; omega.
    apply apart_or.
    destruct cellAligned0 as [ka Hka].
    destruct cellAligned1 as [kr Hkr].
    unfold tl; repeat (rewrite tail_addr).
    generalize (btn_inj_contra n0); omega.
    red in cellInData1; generalize stack_fits; omega.
    red in cellInData0; generalize stack_fits; omega.

    addr_fail.
  Qed.

  Definition awrite_genop : forall abs stat g t,
    sig_option (fun abs' =>
      forall ctx st v st',
	pconc ctx abs st
	-> statusOk hasTyp stat ctx st
	-> write_genop32 g v st st'
	-> word_hasType hasTyp ctx v t
	-> pconc ctx abs' st').
    intros.
    destruct g.

    apply SSome with (@None genop32); intuition.
    inversion H1.

    destruct (awrite_reg abs r) as [abs' Habs'].
    apply SSome with abs'.
    intros ctx st v st' Hpconc Hstat Hwrite Ht.
    inversion Hwrite; subst.
    apply Habs'.
    apply Hpconc.

    destruct (awrite_addr abs stat a t) as [abs' Habs' |];
      [idtac | apply SNone].
    apply SSome with abs'.
    intros ctx st v st' Hpconc Hstat Hwrite Ht.
    inversion Hwrite; subst.
    apply Habs'; trivial.
  Qed.

  Lemma get_flag_eq : forall st f v,
    get_flag (set_flag st f v) f = v.
    intros.
    unfold get_flag, set_flag; simpl.
    apply FlagMap.sel_eq.
  Qed.

  Lemma get_flag_neq : forall st f v f',
    f <> f'
    -> get_flag (set_flag st f v) f' = get_flag st f'.
    intros.
    unfold get_flag, set_flag; simpl.
    apply FlagMap.sel_neq; trivial.
  Qed.

  Lemma presultOk_nil : forall ctx abs st,
    pconc ctx abs st
    -> presultOk hasTyp pconc (Build_presult abs nil) ctx st.
    intros ctx abs st Hpconc.
    split; trivial.
    simpl; apply allS_nil.
  Qed.

  Hint Resolve presultOk_nil.

  Lemma cellsOk_set_pc : forall ctx st pc,
    cellsOk ctx st
    -> cellsOk ctx (set_pc st pc).
    intros ctx st pc Hcells a.
    add Ha (Hcells a).
    destruct (ctx a); trivial.
    inversion Ha.
    split; intuition.
  Qed.

  Hint Resolve cellsOk_set_pc.

  Lemma cellsOk_arith_flags : forall ctx st a x y,
    cellsOk ctx st
    -> cellsOk ctx (arith_flags st a x y).
    intros ctx st a x y Hcells ad.
    add Ha (Hcells ad).
    destruct (ctx ad); trivial.
    inversion Ha.
    split; intuition.
  Qed.

  Hint Resolve cellsOk_arith_flags.

  Lemma cellsOk_arith_flags_no_CF : forall ctx st a x y,
    cellsOk ctx st
    -> cellsOk ctx (arith_flags_no_CF st a x y).
    intros ctx st a x y Hcells ad.
    add Ha (Hcells ad).
    destruct (ctx ad); trivial.
    inversion Ha.
    split; intuition.
  Qed.

  Hint Resolve cellsOk_arith_flags_no_CF.

  Lemma pconc_pc : forall ctx abs st pc,
    pconc ctx abs st
    -> pconc ctx abs (set_pc st pc).
    intros ctx abs st pc Hpconc.
    inversion Hpconc.
    split; intuition.
  Qed.

  Hint Resolve pconc_pc.

  Lemma pconc_cellsOk : forall ctx abs st,
    pconc ctx abs st
    -> cellsOk ctx st.
    intros ctx abs st Hpconc.
    inversion Hpconc; trivial.
  Qed.

  Lemma pconc_astatOk : forall ctx abs st,
    pconc ctx abs st
    -> astatOk abs st.
    intros ctx abs st Hpconc.
    inversion Hpconc; trivial.
  Qed.

  Hint Resolve pconc_cellsOk pconc_astatOk.

  Definition genop_typ : forall abs stat g,
    {t : typ
      | forall st ctx,
	pconc ctx abs st
	-> statusOk hasTyp stat ctx st
	-> hasTyp ctx (eval_genop32 g st) t}.
    unfold hasTyp; intros.
    caseEq (genopType stat g); [exists Int; intuition | exists Int; intuition | idtac].
    intros t Ht.
    exists t; intuition.
    add Ht' (genopTypeOk H0 g).
    rewrite Ht in Ht'.
    inversion Ht'; intuition.
  Qed.

  Hint Unfold hasTyp.

  Definition preservation : forall abs stat ins nextPc,
    single_rule_instr ins
    -> sig_option (fun res => forall st ctx st',
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> exec (ins, nextPc) st st'
      -> presultOk hasTyp pconc res ctx st').
    intros abs stat ins nextPc Hsingle.
    destruct ins; simpl in Hsingle; try tauto; clear Hsingle.

    Ltac answer_is e := apply SSome with (Build_presult (typ := typ) e nil);
      intros st ctx st' Hpconc Hregs Hexec;
	inversion Hexec; subst.

    Ltac give_up := answer_is (@None genop32); intuition; eauto 8.
    
    (* Arith *)
    destruct (awrite_genop abs stat g (TyCustom Int)) as [abs' Habs' |];
      [idtac | apply SNone].
    answer_is (@None genop32).
    eauto 11.

    (* Arithb *)
    give_up.

    (* Call *)
    destruct (awrite_addr abs stat (Build_address neg_four (Some ESP) None) (TyCustom Int)) as [abs' Habs' |];
      [idtac | apply SNone].
    destruct (awrite_reg abs' ESP) as [abs'' Habs''].
    answer_is (@None genop32).
    inversion H1; subst.
    inversion H4; subst.
    clear H1 H4 Hexec.
    eauto 7.

    (* Cmp *)
    destruct g0; [idtac | give_up | give_up].
    destruct (bvec_eq_dec i bzero32); [subst | give_up].
    answer_is (Some g).
    apply presultOk_nil.
    apply pconc_pc.
    split.
    simpl.
    replace (eval_genop32 g (arith_flags st Sub (eval_genop32 g st) bzero32))
      with (eval_genop32 g st); trivial.
    unfold arith_flags.
    do 2 (rewrite get_flag_neq; try discriminate).
    apply get_flag_eq.
    eauto.

    (* Inc *)
    destruct (awrite_genop abs stat g (TyCustom Int)) as [abs' Habs' |];
      [idtac | apply SNone].
    answer_is (@None genop32).
    eauto 11.

    (* Jmp *)
    answer_is abs; auto.

    (* Lea *)
    destruct (awrite_reg abs r) as [abs' Habs'].
    answer_is abs'; auto.

    (* Leave *)
    destruct (awrite_reg abs EBP) as [abs' Habs'].
    destruct (awrite_reg abs' ESP) as [abs'' Habs''].
    answer_is (@None genop32).
    inversion H0; subst.
    inversion H1; subst.
    eauto.

    (* Mov *)
    destruct (awrite_genop abs stat g (genopType stat g0)) as [abs' Habs' |];
      [idtac | apply SNone].
    answer_is abs'.
    generalize (genopTypeOk Hregs g0).
    eauto.

    (* Movb *)
    give_up.

    (* Pop *)
    destruct (awrite_genop abs stat g (TyCustom Int)) as [abs' Habs' |];
      [idtac | apply SNone].
    destruct (awrite_addr abs' stat (Build_address bzero32 (Some ESP) None) (TyCustom Int)) as [abs'' Habs'' |];
      [idtac | apply SNone].
    answer_is (@None genop32).
    inversion H4; subst.
    eauto 7.
    
    (* Push *)
    destruct (awrite_addr abs stat (Build_address neg_four (Some ESP) None) (TyCustom Int)) as [abs' Habs' |];
      [idtac | apply SNone].
    destruct (awrite_reg abs' ESP) as [abs'' Habs''].
    answer_is (@None genop32).
    inversion H1; subst.
    inversion H4; subst.
    eauto 7.

    (* Ret *)
    destruct (awrite_reg abs ESP) as [abs' Habs'].
    answer_is (@None genop32).
    inversion H0; subst.
    eauto.

    (* Shift *)
    give_up.
  Qed.

  Definition preservation_jcc : forall abs stat co pc nextPc,
    sig_option (fun ress => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> (if check_cc st co then
	    presultOk hasTyp pconc (fst ress) ctx (set_pc st pc)
          else
	    presultOk hasTyp pconc (snd ress) ctx (set_pc st nextPc))).
    intros.

    Ltac default :=
      intros; apply SSome with (Build_presult (typ := typ) (@None genop32) nil,
	Build_presult (typ := typ) (@None genop32) nil);
	intros st ctx Hpconc Hregs; simpl;
	  match goal with
	    | [ |- if ?B then _ else _ ] =>
	      destruct B; eauto 8
	  end.

    destruct co.
    destruct b; [idtac | default].
    destruct c; [default | default | idtac | default | default | default | default| default].
    destruct abs; [idtac | default].
    caseEq (genopType stat g); [default | default | idtac].
    intros t Ht.
    destruct t; [default | idtac | default].
    apply SSome with (Build_presult (typ := typ) (@None genop32) nil,
	Build_presult (typ := typ) (@None genop32) (cons (g, TyCustom (NeList t)) nil)).
    intros st ctx Hpconc Hlocs.
    unfold check_cc; simpl.
    inversion Hpconc.
    red in pconcAstat0.
    rewrite pconcAstat0.
    replace (eqb (eval_ZF Sub (eval_genop32 g st) bzero32) true)
      with (eval_ZF Sub (eval_genop32 g st) bzero32).
    destrEq Hzf (eval_ZF Sub (eval_genop32 g st) bzero32).
    auto.

    split.
    simpl; trivial.
    auto.
    simpl.
    apply allS_cons; simpl.
    add Hg (genopTypeOk Hlocs g).
    rewrite Ht in Hg.
    replace (eval_genop32 g (set_pc st nextPc))
      with (eval_genop32 g st); trivial.
    inversion Hg; subst.
    inversion H2; subst.
    auto.

    rewrite <- H0 in Hzf.
    unfold eval_ZF, eval_arith, bsub32, bzero32, bvec_eq in Hzf.
    destruct (bvec_eq_dec (bsub (bzero 32) (bzero 32)) (bzero 32)) as [e | e].
    discriminate.
    wrong; apply e.
    apply btn_inj.
    apply bsub_zero.
    intuition.

    destruct (eval_ZF Sub (eval_genop32 g st) bzero32); trivial.
  Qed.
End Param.

Module Ver := FirstOrderAlg.FirstOrder.Make(Param).
