(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Pointers to integers in the data segment *)

Require Import Arith Bvector List Peano_dec TheoryList.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.FirstOrder.
Require Import ProofOS.AbsInt.FirstOrderAlg.

Require Import ProofOS.Machine.Stateless.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Env.

Import Datatypes.

Set Implicit Arguments.

Hint Constructors word_hasType.

Module Param <: FIRSTORDER_PARAM.
  Inductive _typ : Set :=
    | Ptr
    | NnPtr.

  Definition typ := _typ.

  Definition typ_eq_dec : forall (t1 t2 : typ), {t1 = t2} + {t1 <> t2}.
    intros.
    decide equality t1 t2.
  Qed.

  Definition ctxt := unit.

  Definition astat := option genop32.
  (** We only need to know which operand was checked for nullness. *)

  Definition pconc (_ : ctxt) (ropt : astat) (st : state) :=
    match ropt with
      | None => True
      | Some g => get_flag st ZF = eval_ZF Sub (eval_genop32 g st) bzero32
    end.

  Definition genop32_eq_dec : forall (g1 g2 : genop32), {g1 = g2} + {g1 <> g2}.
    intros.
    decide equality g1 g2.
    apply (@bvec_eq_dec 32).
    apply reg32_eq_dec.
    decide equality a a0.
    decide equality addrIndex addrIndex0.
    decide equality a1 p.
    apply reg32_eq_dec.
    decide equality a2 s.
    decide equality addrBase addrBase0.
    apply reg32_eq_dec.
    apply (@bvec_eq_dec 32).
  Qed.

  Definition astat_eq_dec : forall (abs1 abs2 : astat), {abs1 = abs2} + {abs1 <> abs2}.
    intros.
    decide equality abs1 abs2.
    apply genop32_eq_dec.
  Qed.

  Definition subAstat : forall abs1 abs2, pred_option (forall st ctx,
    pconc ctx abs1 st -> pconc ctx abs2 st).
    intros.
    destruct abs2.
    destruct abs1.
    destruct (genop32_eq_dec g g0); [subst | apply PNone].
    apply PSome; trivial.
    apply PNone.
    apply PSome; intuition.
    red; trivial.
  Qed.

  Hint Unfold pconc.

  Definition inData (addr : int32) :=
    bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4 < bvec_to_nat addr
    /\ bvec_to_nat addr + 4 < bvec_to_nat stackStart.

  Definition inData_dec addr : {inData addr} + {~inData addr}.
    intros.
    unfold inData.
    destruct (le_lt_dec (bvec_to_nat addr)
      (bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)).
    right; omega.
    destruct (le_lt_dec (bvec_to_nat stackStart)
      (bvec_to_nat addr + 4)).
    right; omega.
    left; omega.
  Qed.

  Inductive _hasTyp (ctx : ctxt) : int32 -> typ -> Prop :=
    | HT_Ptr_real : forall v,
      inData v
      -> _hasTyp ctx v Ptr
    | HT_Ptr_null : _hasTyp ctx bzero32 Ptr
    | HT_NnPtr : forall v,
      inData v
      -> _hasTyp ctx v NnPtr.

  Definition hasTyp := _hasTyp.

  Definition minitState := initState.

  Definition kernel_astat := Some (@None genop32).

  Definition isSafe (_ : instr) (_ : state) := True.

  Definition entryPoints := @nil (entryPoint typ astat).
  Definition imported := @nil (entryPoint typ astat).

  Definition subTyp (t1 t2 : word_type typ) :=
    match t2 with
      | TyTop => true
      | TyConst n =>
	match t1 with
	  | TyConst n' => sum_to_bool (bvec_eq_dec n n')
	  | _ => false
	end
      | TyCustom Ptr =>
	match t1 with
	  | TyConst n =>
	    sum_to_bool (bvec_eq_dec n bzero32)
	    || sum_to_bool (inData_dec n)
	  | TyCustom Ptr => true
	  | TyCustom NnPtr => true
	  | _ => false
	end
      | TyCustom NnPtr =>
	match t1 with
	  | TyConst n => sum_to_bool (inData_dec n)
	  | TyCustom NnPtr => true
	  | _ => false
	end
    end.

  Theorem subTyp_refl : forall t, subTyp t t = true.
    destruct t; simpl; intuition.
    destruct (bvec_eq_dec i i); intuition.
    destruct t; trivial.
  Qed.

  Hint Constructors _hasTyp.

  Theorem subTyp_sound : forall t1 t2,
    subTyp t1 t2 = true
    -> forall ctx v, word_hasType hasTyp ctx v t1
      -> word_hasType hasTyp ctx v t2.
    unfold hasTyp.
    intros t1 t2 Hsub.
    functional induction subTyp t1 t2; subst; simpl in Hsub; try discriminate; myauto.
    destruct (bvec_eq_dec n n'); intuition;
      [rewrite e; trivial
	| discriminate].
    destruct (bvec_eq_dec n bzero32); intuition;
      [rewrite e in H; inversion H; intuition
	| idtac].
    destruct (inData_dec n); intuition;
      [inversion H; intuition
	| discriminate].
    intros ctx v Hty.
    inversion Hty; subst.
    inversion H2; myauto.
    intros ctx v Hvty.
    destruct (inData_dec n);
      [inversion Hvty; intuition
	| discriminate].
  Qed.

  Definition checkSafe : forall abs stat ins,
    pred_option (forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> isSafe ins st).
    intros.
    apply PSome.
    intuition.
    red; trivial.
  Qed.

  Ltac type_fail := intros; apply SNone.

  Definition typeof_read : forall abs stat addr,
    sig_option (fun ty => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx (get_mem32 st (eval_addr addr st)) ty).
    intros.
    destruct addr.
    destruct (bvec_eq_dec addrDisp bzero32); [subst | type_fail].
    destruct addrIndex; [type_fail | idtac].
    destruct addrBase; [idtac | type_fail].
    caseEq (genopType stat (Reg r)); [type_fail | type_fail | idtac].
    intros t Ht.
    destruct t; [type_fail | idtac].
    apply SSome with (TyTop typ).
    myauto.
  Qed.

  Ltac pred_fail := intros; apply PNone.
    
  Definition check_write : forall abs stat addr vty,
    pred_option (forall st ctx v,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx v vty
      -> not_too_high (eval_addr addr st)
      /\ ~(bvec_to_nat initialPc - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)
      /\ ~(bvec_to_nat stackStart - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat stackEnd + 4)).
    intros.
    destruct addr.
    destruct (bvec_eq_dec addrDisp bzero32); [subst | pred_fail].
    destruct addrIndex; [pred_fail | idtac].
    destruct addrBase; [idtac | pred_fail].
    caseEq (genopType stat (Reg r)); [pred_fail | pred_fail | idtac].
    intros t Ht.
    destruct t; [pred_fail | idtac].
    apply PSome.

    intros st ctx v Hpconc Hloc Hvty.
    add Hr (genopTypeOk Hloc (Reg r)).
    rewrite Ht in Hr.
    simpl in Hr.
    inversion Hr; subst.
    clear Hr; rename H2 into Hr.
    inversion Hr; subst.
    unfold eval_addr, addrIndex, addrBase, addrDisp.
    unfold badd32, bzero32.
    rewrite badd_zero'.
    unfold inData in H.
    split.

    red.
    add HstackFits stack_fits.
    split.
    omega.

    omega.
  Qed.

  Definition typeof_arith : forall aop t1 t2,
    sig_option (fun ty => forall ctx v1 v2,
      word_hasType hasTyp ctx v1 t1
      -> word_hasType hasTyp ctx v2 t2
      -> word_hasType hasTyp ctx (eval_arith aop v1 v2) ty).
    intros.
    apply SNone.
  Qed.

  Lemma get_flag_reg : forall st r v f,
    get_flag (set_reg32 st r v) f = get_flag st f.
    trivial.
  Qed.

  Lemma get_mem32_reg : forall st r v a,
    get_mem32 (set_reg32 st r v) a = get_mem32 st a.
    trivial.
  Qed.

  Ltac awrite_fail := exists (@None genop32); intuition; red; trivial.

  Definition awrite_reg : forall abs r,
    {abs' : astat
      | forall ctx st v,
	pconc ctx abs st
	-> pconc ctx abs' (set_reg32 st r v)}.
    intros.
    destruct abs.
    destruct g.

    awrite_fail.

    destruct (reg32_eq_dec r0 r); subst.
    awrite_fail.
    exists (Some (Reg r0)).
    intros ctx st v Hpconc.
    red.
    red in Hpconc.
    replace (get_flag (set_reg32 st r v) ZF)
      with (get_flag st ZF); trivial.
    unfold get_reg32, set_reg32; simpl.
    unfold get_reg32; simpl.
    rewrite Reg32Map.sel_neq; auto with arith.

    destruct a.
    destruct addrBase.
    destruct (reg32_eq_dec r r0); [awrite_fail | idtac].
    destruct addrIndex.
    destruct p.
    destruct (reg32_eq_dec r r1); [awrite_fail | idtac].

    exists (Some (Address reg32 (Build_address addrDisp (Some r0) (Some (s, r1))))).
    intros ctx st v Hpconc.
    red; simpl.
    red in Hpconc; simpl in Hpconc.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp in Hpconc.
    replace (get_reg32 (set_reg32 st r v) r0)
      with (get_reg32 st r0).
    replace (get_reg32 (set_reg32 st r v) r1)
      with (get_reg32 st r1).
    trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.

    exists (Some (Address reg32 (Build_address addrDisp (Some r0) None))).
    intros ctx st v Hpconc.
    red; simpl.
    red in Hpconc; simpl in Hpconc.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp in Hpconc.
    replace (get_reg32 (set_reg32 st r v) r0)
      with (get_reg32 st r0).
    trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.

    destruct addrIndex.
    destruct p.
    destruct (reg32_eq_dec r r0); [awrite_fail | idtac].
    exists (Some (Address reg32 (Build_address addrDisp None (Some (s, r0))))).
    intros ctx st v Hpconc.
    red; simpl.
    red in Hpconc; simpl in Hpconc.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
    unfold eval_addr, addrIndex, addrBase, Defs.addrDisp in Hpconc.
    replace (get_reg32 (set_reg32 st r v) r0)
      with (get_reg32 st r0).
    trivial.
    unfold get_reg32, set_reg32; simpl.
    symmetry; apply Reg32Map.sel_neq; trivial.

    exists (Some (Address reg32 (Build_address addrDisp None None))).
    intros ctx st v Hpconc.
    red; simpl.
    red in Hpconc; simpl in Hpconc.
    rewrite get_flag_reg.
    rewrite get_mem32_reg.
    trivial.

    awrite_fail.
  Qed.
    
  Definition awrite_genop : forall abs g,
    {abs' : astat
      | forall ctx st v st',
	pconc ctx abs st
	-> write_genop32 g v st st'
	-> pconc ctx abs' st'}.
    intros.
    destruct g.

    exists (@None genop32); intuition.

    destruct (awrite_reg abs r) as [abs' Habs'].
    exists abs'.
    intros ctx st v st' Hpconc Hwrite.
    inversion Hwrite; subst.
    auto.

    exists (@None genop32); intuition.    
  Qed.

  Lemma pconc_pc : forall ctx abs st pc,
    pconc ctx abs st
    -> pconc ctx abs (set_pc st pc).
    intuition.
  Qed.

  Hint Resolve pconc_pc.

  Lemma get_flag_eq : forall st f v,
    get_flag (set_flag st f v) f = v.
    intros.
    unfold get_flag, set_flag; simpl.
    apply FlagMap.sel_eq.
  Qed.

  Lemma get_flag_neq : forall st f v f',
    f <> f'
    -> get_flag (set_flag st f v) f' = get_flag st f'.
    intros.
    unfold get_flag, set_flag; simpl.
    apply FlagMap.sel_neq; trivial.
  Qed.

  Lemma presultOk_nil : forall ctx abs st,
    pconc ctx abs st
    -> presultOk hasTyp pconc (Build_presult abs nil) ctx st.
    intros ctx abs st Hpconc.
    split; trivial.
    simpl; apply allS_nil.
  Qed.

  Hint Resolve presultOk_nil.

  Definition preservation : forall abs stat ins nextPc,
    single_rule_instr ins
    -> sig_option (fun res => forall st ctx st',
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> exec (ins, nextPc) st st'
      -> presultOk hasTyp pconc res ctx st').
    intros abs locs ins nextPc Hsingle.
    destruct ins; simpl in Hsingle; try tauto; clear Hsingle.

    Ltac answer_is e := apply SSome with (Build_presult (typ := typ) e nil);
      intros st ctx st' Hpconc Hregs Hexec;
	inversion Hexec; subst.

    Ltac give_up := answer_is (@None genop32); intuition.
    
    (* Arith *)
    give_up.

    (* Arithb *)
    give_up.

    (* Call *)
    give_up.

    (* Cmp *)
    destruct g0; [idtac | give_up | give_up].
    destruct (bvec_eq_dec i bzero32); [subst | give_up].
    answer_is (Some g).
    apply presultOk_nil.
    apply pconc_pc.
    red.
    simpl.
    replace (eval_genop32 g (arith_flags st Sub (eval_genop32 g st) bzero32))
      with (eval_genop32 g st); trivial.
    unfold arith_flags.
    do 2 (rewrite get_flag_neq; try discriminate).
    apply get_flag_eq.

    (* Inc *)
    give_up.

    (* Jmp *)
    answer_is abs; auto.

    (* Lea *)
    destruct (awrite_reg abs r) as [abs' Habs'].
    answer_is abs'; auto.

    (* Leave *)
    give_up.
    
    (* Mov *)
    destruct (awrite_genop abs g) as [abs' Habs'].
    answer_is abs'.
    apply presultOk_nil.
    apply pconc_pc.
    apply Habs' with st (eval_genop32 g0 st); eauto.

    (* Movb *)
    give_up.

    (* Pop *)
    give_up.
    
    (* Push *)
    give_up.

    (* Ret *)
    give_up.

    (* Shift *)
    give_up.
  Qed.

  Definition preservation_jcc : forall abs stat co pc nextPc,
    sig_option (fun ress => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> (if check_cc st co then
	    presultOk hasTyp pconc (fst ress) ctx (set_pc st pc)
          else
	    presultOk hasTyp pconc (snd ress) ctx (set_pc st nextPc))).
    intros.

    Ltac default :=
      intros; apply SSome with (Build_presult (typ := typ) (@None genop32) nil,
	Build_presult (typ := typ) (@None genop32) nil);
	intros st ctx Hpconc Hregs; simpl;
	  match goal with
	    | [ |- if ?B then _ else _ ] =>
	      destruct B; auto
	  end.

    destruct co.
    destruct b; [idtac | default].
    destruct c; [default | default | idtac | default | default | default | default| default].
    destruct abs; [idtac | default].
    caseEq (genopType stat g); [default | default | idtac].
    intros t Ht.
    destruct t; [idtac | default].
    apply SSome with (Build_presult (typ := typ) (@None genop32) nil,
	Build_presult (typ := typ) (@None genop32) (cons (g, TyCustom NnPtr) nil)).
    intros st ctx Hpconc Hlocs.
    unfold check_cc; simpl.
    red in Hpconc.
    rewrite Hpconc.
    replace (eqb (eval_ZF Sub (eval_genop32 g st) bzero32) true)
      with (eval_ZF Sub (eval_genop32 g st) bzero32).
    destrEq Hzf (eval_ZF Sub (eval_genop32 g st) bzero32).
    auto.
  
    split.
    simpl; trivial.
    simpl.
    apply allS_cons; simpl.
    add Hg (genopTypeOk Hlocs g).
    rewrite Ht in Hg.
    replace (eval_genop32 g (set_pc st nextPc))
      with (eval_genop32 g st); trivial.
    inversion Hg; subst.
    inversion H2; subst.
    apply WHT_Custom.
    apply HT_NnPtr; trivial.

    rewrite <- H0 in Hzf.
    unfold eval_ZF, eval_arith, bsub32, bzero32, bvec_eq in Hzf.
    destruct (bvec_eq_dec (bsub (bzero 32) (bzero 32)) (bzero 32)) as [e | e].
    discriminate.
    wrong; apply e.
    apply btn_inj.
    apply bsub_zero.
    intuition.

    destruct (eval_ZF Sub (eval_genop32 g st) bzero32); trivial.
  Qed.
End Param.

Module Ver := FirstOrderAlg.FirstOrder.Make(Param).
