(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** A further simplification of FirstOrder *)

Require Import Arith Bvector List TheoryList.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.FirstOrder.
Require Import ProofOS.Machine.Stateless.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Env.

Import Datatypes.

Set Implicit Arguments.

Definition extraParam (typ : Set) (hasTyp : int32 -> typ -> Prop)
  (ctx : unit) (v : int32) (t : typ) :=
  hasTyp v t.

Module Type SIMPLE_FIRSTORDER_PARAM.
  Parameter typ : Set.
  (** Custom types *)
  Parameter typ_eq_dec : forall (t1 t2 : typ), {t1 = t2} + {t1 <> t2}.

  Parameter hasTyp : int32 -> typ -> Prop.
  (** Does a value have a type? *)

  Parameter subTyp : word_type typ -> word_type typ -> bool.
  (** Decidable subtyping *)
  Parameter subTyp_refl : forall t, subTyp t t = true.
  Parameter subTyp_sound : forall t1 t2,
    subTyp t1 t2 = true
    -> forall v, word_hasType (extraParam hasTyp) tt v t1
      -> word_hasType (extraParam hasTyp) tt v t2.

  Parameter isSafe : instr -> state -> Prop.
  (** A safety policy, suitable for use with Policy.Stateless *)

  Parameter checkSafe : forall stat ins,
    pred_option (forall st,
      statusOk (extraParam hasTyp) stat tt st
      -> isSafe ins st).
  (** Check if known type information implies that the safety policy is
    satisfied for one instruction step. *)

  Parameter typeof_read : forall stat addr,
    sig_option (fun ty => forall st,
      statusOk (extraParam hasTyp) stat tt st
      -> word_hasType (extraParam hasTyp) tt (get_mem32 st (eval_addr addr st)) ty).
  (** Determine the type of something read from memory. *)

  Parameter check_write : forall stat addr vty,
    pred_option (forall st v,
      statusOk (extraParam hasTyp) stat tt st
      -> word_hasType (extraParam hasTyp) tt v vty
      -> not_too_high (eval_addr addr st)
      /\ ~(bvec_to_nat initialPc - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)
      /\ ~(bvec_to_nat stackStart - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat stackEnd + 4)).
  (** Check that a memory write doesn't change code or stack memory. *)

  Parameter typeof_arith : forall aop t1 t2,
    sig_option (fun ty => forall v1 v2,
      word_hasType (extraParam hasTyp) tt v1 t1
      -> word_hasType (extraParam hasTyp) tt v2 t2
      -> word_hasType (extraParam hasTyp) tt (eval_arith aop v1 v2) ty).
  (** Determine the result type of an arithmetic operation. *)

End SIMPLE_FIRSTORDER_PARAM.

Module Make (Param : SIMPLE_FIRSTORDER_PARAM) <: FIRSTORDER_PARAM.
  Import Param.

  Definition typ := typ.
  Definition typ_eq_dec := typ_eq_dec.

  Definition ctxt := unit.
  Definition astat := unit.

  Parameter entries : list (entryPoint typ astat).

  Definition pconc (_ : ctxt) (_ : astat) (_ : state) := True.
  Definition astat_eq_dec : forall (abs1 abs2 : astat), {abs1 = abs2} + {abs1 <> abs2}.
    intros.
    decide equality abs1 abs2.
  Qed.

  Definition subAstat : forall abs1 abs2, pred_option (forall st ctx,
    pconc ctx abs1 st -> pconc ctx abs2 st).
    intros.
    apply PSome; trivial.
  Qed.

  Definition hasTyp := extraParam hasTyp.

  Definition kernel_astat := @None unit.

  Definition entryPoints := entries.

  Definition subTyp := subTyp.
  Lemma subTyp_refl : forall t, subTyp t t = true.
    apply subTyp_refl.
  Qed.

  Lemma word_hasType_tt : forall v t ctx,
    word_hasType (extraParam Param.hasTyp) tt v t
    -> word_hasType hasTyp ctx v t.
    intros v t ctx Hty.
    destruct ctx.
    inversion Hty; subst; myauto.
  Qed.
    
  Lemma word_hasType_tt' : forall v t ctx,
    word_hasType hasTyp ctx v t
    -> word_hasType (extraParam Param.hasTyp) tt v t.
    intros v t ctx Hty.
    destruct ctx.
    inversion Hty; subst; myauto.
  Qed.

  Hint Resolve word_hasType_tt word_hasType_tt'.

  Lemma subTyp_sound : forall t1 t2,
    subTyp t1 t2 = true
    -> forall ctx v, word_hasType hasTyp ctx v t1
      -> word_hasType hasTyp ctx v t2.
    intros.
    apply word_hasType_tt.
    eapply subTyp_sound; eauto.
  Qed.

  Definition isSafe := isSafe.

  Definition checkSafe : forall abs stat ins,
    pred_option (forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> isSafe ins st).
    intros.
    destruct (checkSafe stat ins); [idtac | apply PNone].
    apply PSome.
    intros st ctx Hpconc Hregs.
    destruct ctx; eauto.
    eapply i; eauto.
  Qed.

  Definition typeof_read : forall abs stat addr,
    sig_option (fun ty => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx (get_mem32 st (eval_addr addr st)) ty).
    intros.
    destruct (typeof_read stat addr) as [t Ht |]; [idtac | apply SNone].
    apply SSome with t.
    intros; destruct ctx.
    eauto.
  Qed.

  Definition check_write : forall abs stat addr vty,
    pred_option (forall st ctx v,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> word_hasType hasTyp ctx v vty
      -> not_too_high (eval_addr addr st)
      /\ ~(bvec_to_nat initialPc - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)
      /\ ~(bvec_to_nat stackStart - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat stackEnd + 4)).
    intros.
    destruct (check_write stat addr vty) as [t Ht |]; [idtac | apply PNone].
    apply PSome.
    destruct ctx; eauto.
  Qed.

  Definition typeof_arith : forall aop t1 t2,
    sig_option (fun ty => forall ctx v1 v2,
      word_hasType hasTyp ctx v1 t1
      -> word_hasType hasTyp ctx v2 t2
      -> word_hasType hasTyp ctx (eval_arith aop v1 v2) ty).
    intros.
    destruct (typeof_arith aop t1 t2) as [t Ht |]; [idtac | apply SNone].
    apply SSome with t.
    eauto.
  Qed.

  Definition pres := Build_presult (typ := typ) tt nil.

  Definition preservation : forall abs stat ins nextPc,
    single_rule_instr ins
    -> sig_option (fun res => forall st ctx st',
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> exec (ins, nextPc) st st'
      -> presultOk hasTyp pconc res ctx st').
    intros.
    apply SSome with pres.
    intros; split; trivial.
    simpl.
    apply allS_nil.
  Qed.

  Definition preservation_jcc : forall abs stat co pc nextPc,
    sig_option (fun ress => forall st ctx,
      pconc ctx abs st
      -> statusOk hasTyp stat ctx st
      -> (if check_cc st co then
	    presultOk hasTyp pconc (fst ress) ctx (set_pc st pc)
          else
	    presultOk hasTyp pconc (snd ress) ctx (set_pc st nextPc))).
    intros.
    apply SSome with (pres, pres).
    intros.
    destruct (check_cc st co); trivial;
      simpl; split; trivial; simpl; apply allS_nil.
  Qed.

  Definition imported := @nil (entryPoint typ astat).
End Make.
