(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** The simplest instantiation of FirstOrder, with no custom types *)

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.FirstOrder.
Require Import ProofOS.AbsInt.FirstOrderAlg.
Require Import ProofOS.Examples.SimpleFo.

Require Import ProofOS.Trusted.Env.

Module Param <: SIMPLE_FIRSTORDER_PARAM.
  Definition typ := unit.

  Theorem typ_eq_dec : forall (t1 t2 : typ), {t1 = t2} + {t1 <> t2}.
    intros.
    decide equality t1 t2.
  Qed.

  Definition hasTyp (_ : int32) (_ : typ) := False.

  Definition subTyp (t1 t2 : word_type typ) :=
    match t2 with
      | TyTop => true
      | TyConst i =>
	match t1 with
	  | TyConst i' => sum_to_bool (bvec_eq_dec i i')
	  | _ => false
	end
      | TyCustom _ =>
	match t1 with
	  | TyCustom _ => true
	  | _ => false
	end
    end.

  Theorem subTyp_refl : forall t, subTyp t t = true.
    destruct t; trivial.
    simpl.
    destruct (bvec_eq_dec i i);
      [trivial
	| wrong; tauto].
  Qed.

  Hint Constructors word_hasType.

  Theorem subTyp_sound : forall t1 t2,
    subTyp t1 t2 = true
    -> forall v, word_hasType (extraParam hasTyp) tt v t1
      -> word_hasType (extraParam hasTyp) tt v t2.
    intros t1 t2 Htyp v Hword.
    functional induction subTyp t1 t2; subst; inversion Htyp; inversion Hword; subst; auto.
    destruct (bvec_eq_dec i i') as [e | e];
      [rewrite e; trivial
	| discriminate].
  Qed.
    
  Definition isSafe (_ : instr) (_ : state) := True.

  Theorem checkSafe : forall stat ins,
    pred_option (forall st,
      statusOk (extraParam hasTyp) stat tt st
      -> isSafe ins st).
    intros.
    apply PSome.
    intuition.
    red; trivial.
  Qed.

  Definition typeof_read : forall stat addr,
    sig_option (fun ty => forall st,
      statusOk (extraParam hasTyp) stat tt st
      -> word_hasType (extraParam hasTyp) tt (get_mem32 st (eval_addr addr st)) ty).
    intros.
    apply SNone.
  Qed.

  Definition check_write : forall stat addr vty,
    pred_option (forall st v,
      statusOk (extraParam hasTyp) stat tt st
      -> word_hasType (extraParam hasTyp) tt v vty
      -> not_too_high (eval_addr addr st)
      /\ ~(bvec_to_nat initialPc - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 4)
      /\ ~(bvec_to_nat stackStart - 4
	<= bvec_to_nat (eval_addr addr st)
	< bvec_to_nat stackEnd + 4)).
    intros.
    apply PNone.
  Qed.

  Definition typeof_arith : forall aop t1 t2,
    sig_option (fun ty => forall v1 v2,
      word_hasType (extraParam hasTyp) tt v1 t1
      -> word_hasType (extraParam hasTyp) tt v2 t2
      -> word_hasType (extraParam hasTyp) tt (eval_arith aop v1 v2) ty).
    intros.
    apply SNone.
  Qed.
End Param.

Module VerParam := Make(Param).
Module Ver := FirstOrderAlg.FirstOrder.Make(VerParam).
