(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Operations on abstract states *)

Require Import Arith Bool List.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Env.

Require Import ProofOS.Kernel.Defs.
Require Import ProofOS.Kernel.Types.

Hint Constructors hasType : asm.

Ltac bool_elim' :=
  match goal with
    | [ H : ?X && ?Y = true |- _ ] =>
      let and1 := fresh "and1" in
	let and2 := fresh "and2" in
	  destruct (andb_prop X Y H) as [and1 and2];
	    clear H
  end.

Ltac bool_elim := repeat bool_elim'.

Definition apc_eq_dec : forall (pc1 pc2 : apc), {pc1 = pc2} + {pc1 <> pc2}.
  intros.
  decide equality pc1 pc2.
  apply (@bvec_eq_dec 32).
Qed.

Lemma subAbs_pc : forall st abs1 abs2 ctx,
  conc_pc abs1 st ctx
  -> sum_to_bool (apc_eq_dec (aPc abs1) (aPc abs2)) = true
  -> conc_pc abs2 st ctx.
  intros st abs1 abs2 ctx Hpc Heq.
  destruct (apc_eq_dec (aPc abs1) (aPc abs2)).
  red in Hpc.
  destruct (aPc abs1); subst.
  red.
  rewrite <- e; trivial.
  red.
  rewrite <- e.
  trivial.
  discriminate.
Qed.

Lemma subAbs_regs : forall ctx st abs1 abs2,
  conc_regs abs1 st ctx
  -> proj1_sig (Reg32Map.subMap subType
    subType_refl (aRegs abs1) (aRegs abs2)) = true
  -> conc_regs abs2 st ctx.
  intros ctx st abs1 abs2 Hconc Hsub; intro.
  destruct (Reg32Map.subMap subType subType_refl (aRegs abs1) (aRegs abs2)).
  destruct x; try discriminate.
  eapply subType_sound; eauto.
  unfold aget_reg.
  auto.
Qed.

Lemma subAbs_sptr : forall ctx abs1 abs2,
  bvec_to_nat stackStart + aStackSize abs1 < bvec_to_nat (sptr ctx)
  -> sum_to_bool (le_lt_dec (aStackSize abs2) (aStackSize abs1)) = true
  -> bvec_to_nat stackStart + aStackSize abs2 < bvec_to_nat (sptr ctx).
  intros ctx abs1 abs2 Hstart Hle.
  destruct (le_lt_dec (aStackSize abs2) (aStackSize abs1));
    simpl in Hle; try discriminate; omega.
Qed.

Lemma subAbs_slots : forall st ctx abs1 abs2,
  conc_slots abs1 st ctx
  -> proj1_sig
  (NatMap.subMap subType subType_refl (aStackSlots abs1)
    (aStackSlots abs2)) = true
  -> sum_to_bool (le_lt_dec (aStackSize abs2) (aStackSize abs1)) = true
  -> conc_slots abs2 st ctx.
  intros st ctx abs1 abs2 Hconc Hsub Hle n HstackSize.
  destruct (NatMap.subMap subType subType_refl (aStackSlots abs1) (aStackSlots abs2)).
  destruct x; try discriminate.
  apply subType_sound with (aget_stack abs1 n).
  unfold aget_stack; auto.
  apply Hconc; destruct (le_lt_dec (aStackSize abs2) (aStackSize abs1));
    [omega | discriminate].
Qed.

Hint Resolve subAbs_pc subAbs_regs subAbs_sptr subAbs_slots : asm.

Definition subAbs abs1 abs2 : {b : bool | b = true -> forall st ctx,
  conc' abs1 st ctx -> conc' abs2 st ctx}.
  intros abs1 abs2.
  exists (sum_to_bool (apc_eq_dec (aPc abs1) (aPc abs2))
    && sum_to_bool (le_lt_dec (aStackSize abs2) (aStackSize abs1))
    && proj1_sig (Reg32Map.subMap subType subType_refl (aRegs abs1) (aRegs abs2))
    && proj1_sig (NatMap.subMap subType subType_refl (aStackSlots abs1) (aStackSlots abs2))).
  intros Hchecks st ctx Hconc.
  inversion Hconc.
  bool_elim.
  split; trivial; myeauto.
Qed.

Definition astate_eq_dec : forall (a1 a2 : astate), {a1 = a2} + {a1 <> a2}.
  intros.
  decide equality a1 a2.
  unfold astack in * ; idtac.
  apply NatMap.eq_dec.
  apply ty_eq_dec.
  apply eq_nat_dec.
  unfold aregs in * ; idtac.
  apply Reg32Map.eq_dec.
  apply ty_eq_dec.
  apply apc_eq_dec.
Qed.

Definition astates_eq_dec : forall (a1 a2 : list astate), {a1 = a2} + {a1 <> a2}.
  induction a1; destruct a2; intuition;
    try (right; intros; discriminate).
  destruct (astate_eq_dec a a0); subst.
  destruct (IHa1 a2); subst.
  auto.
  right; congruence.
  right; congruence.
Qed.

Fixpoint goodHyps (hyps hyps' : list astate) {struct hyps'} : Prop :=
  match hyps' with
    | nil =>
      match hyps with
	| nil => True
	| _ => False
      end
    | cons hyp' hyps' =>
      match hyps with
	| nil => False
	| cons hyp hyps =>
	  proj1_sig (subAbs hyp' hyp) = true /\ goodHyps hyps hyps'
      end
  end.

Definition goodHyps_dec : forall hyps' hyps,
  pred_option (goodHyps hyps hyps').
  induction hyps'; simpl; intuition.
  destruct hyps.
  apply PSome; trivial.
  apply PNone.
  destruct hyps.
  apply PNone.
  destruct (IHhyps' hyps).
  destruct (subAbs a a0).
  destruct x; [intuition | apply PNone].
  simpl.
  apply PSome.
  tauto.
  apply PNone.
Qed.

Definition abs_covered : forall abss hyps abs,
  {b : bool | b = true ->
    exists hyps', exists abs',
      In (hyps', abs') abss
      /\ proj1_sig (subAbs abs abs') = true
      /\ goodHyps hyps hyps'}.
  induction abss; intros hyps abs.

  exists false; intuition; discriminate.

  destruct (goodHyps_dec (fst a) hyps).
  caseEq (subAbs abs (snd a)); intros x c Ha; destruct x.

  exists true; intuition; exists (fst a); exists (snd a); intuition.
  destruct a; simpl; intuition.
  rewrite Ha; trivial.

  destruct (IHabss hyps abs) as [x]; destruct x.
  exists true; intuition.
  destruct H0 as [hyps' H0].
  destruct H0 as [abs'].
  exists hyps'.
  exists abs'.
  intuition.
  exists false; intuition; discriminate.

  destruct (IHabss hyps abs) as [x]; destruct x.
  exists true; intuition.
  destruct H0 as [hyps' H0].
  destruct H0 as [abs'].
  exists hyps'.
  exists abs'.
  intuition.
  exists false; intuition; discriminate.
Qed.
