(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Function calls and returns *)

Require Import Arith Bool Bvector List.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Specif.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Env.

Require Import ProofOS.AbsInt.Bounded.

Require Import ProofOS.Kernel.Defs.
Require Import ProofOS.Kernel.Conc.
Require Import ProofOS.Kernel.Operands.
Require Import ProofOS.Kernel.Misc.
Require Import ProofOS.Kernel.Types.
Require Import ProofOS.Kernel.Abs.

Ltac type_fail := intros; apply SNone.

(** * Switch point of view from caller to callee.
  Use this on the state right before a CALL instruction.
  *)

Import Datatypes.

Lemma enter_agree_range : forall abs ss st ctx ret,
  aget_reg abs ESP = Stack ss
  -> 4 <= S (S (S (S ss))) < aStackSize abs
  -> agree_range (stMem st) initialMem initialPc
  (actualCodeLen + longestInstr)
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 4
  -> agree_range
  (stMem (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret))
  initialMem initialPc (actualCodeLen + longestInstr).
  intros abs ss st ctx ret Hesp Hbound Hagree Hregs Hsptr Hbelow.
  replace (badd neg_four (get_reg32 st ESP))
    with (eval_addr (Build_address neg_four (Some ESP) None) st).
  eapply awrite_stack_agree_range; myeauto.
  replace (bsub32 (nat_to_bvec 32 ss) neg_four)
    with (badd (nat_to_bvec 32 ss) four).
  unfold badd, barith.
  unfold four.
  repeat (rewrite ntb_invert); try (inversion Hbelow; omega).
  unfold neg_four.
  unfold badd, bsub32, bsub.
  rewrite bneg_btn_double; trivial.

  unfold eval_addr, addrIndex, addrBase, addrDisp.
  trivial.
Qed.

Hint Resolve enter_agree_range : asm.

Lemma enter_conc_regs : forall ss abs pc st ret ctx sls (args : list ty),
  aget_reg abs ESP = Stack ss
  -> S (S (S (S ss))) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> conc_regs abs st ctx
  -> 4 * length args <= ss
  -> conc_regs
  (Build_astate (apcConst pc)
    (Reg32Map.upd (Reg32Map.upd (Reg32Map.empty Top) ESP (Stack (S (S (S (S (4 * length args))))))) EBP
      OldEbp) (aStackSize abs - (ss - 4 * length args))
    sls)
  (set_pc
    (set_reg32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret) ESP
      (bsub (get_reg32 st ESP) four)) pc)
  (Build_context (stack_slot ctx (ss - 4 * length args)) (4 * length args) (get_reg32 st EBP)
    (stMem st) (Some ret)).
  intros ss abs pc st ret ctx sls args Hesp Hss Hsptr Hregs Hargs r.
  unfold aget_reg, get_reg32; simpl.
  destruct (reg32_eq_dec EBP r); subst.

  rewrite Reg32Map.sel_eq.
  rewrite Reg32Map.sel_neq; try discriminate.
  replace (Reg32Map.sel (stRegs32 st) EBP)
    with (oldebp (Build_context (stack_slot ctx (ss - 4 * length args))
      (4 * length args) (Reg32Map.sel (stRegs32 st) EBP) (stMem st) (Some ret)));
    trivial.
  myauto.

  destruct (reg32_eq_dec ESP r); subst.
  rewrite (Reg32Map.sel_neq (Reg32Map.upd (Reg32Map.empty Top) ESP (Stack (S (S (S (S (4 * length args))))))));
    try discriminate.
  repeat (rewrite Reg32Map.sel_eq).
  replace (bsub (Reg32Map.sel (stRegs32 st) ESP) four)
    with (stack_slot (Build_context (stack_slot ctx (ss - 4 * length args))
      (4 * length args) (Reg32Map.sel (stRegs32 st) EBP)
        (stMem st) (Some ret)) (S (S (S (S (4 * length args)))))).
  apply HasType_Stack.
  simpl sptr.
  unfold stack_slot.
  rewrite ntb_invert.
  add Hhigh code_high; add Hspace code_space; omega.
  add Hup (btn_upper (sptr ctx)); omega.
  unfold stack_slot.
  unfold sptr at 1.
  rewrite ntb_invert.
  add Hr (Hregs ESP).
  rewrite Hesp in Hr.
  inversion Hr; subst.
  unfold get_reg32 in H.
  rewrite <- H.
  unfold stack_slot.
  unfold four.
  rewrite bsub_bdec4.
  apply btn_inj.
  rewrite bdec4_btn.
  repeat (rewrite ntb_invert); trivial; try (add Hup (btn_upper (sptr ctx)); omega).
  rewrite ntb_invert; trivial; try (add Hup (btn_upper (sptr ctx)); omega).
  rewrite ntb_invert.
  omega.
  generalize (btn_upper (sptr ctx)); omega.
  generalize (btn_upper (sptr ctx)); omega.

  repeat (rewrite Reg32Map.sel_neq; auto with arith).
  rewrite Reg32Map.sel_empty.
  myauto.
Qed.

Hint Resolve enter_conc_regs : asm.

Lemma enter_stackSize' : forall ss abs ctx,
  S (S (S (S ss))) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> bvec_to_nat stackStart + (aStackSize abs - ss) <
  bvec_to_nat (stack_slot ctx ss).
  intros ss abs ctx Hss Hsptr.
  unfold stack_slot.
  rewrite ntb_invert.
  omega.
  generalize (btn_upper (sptr ctx)); omega.
Qed.

Lemma enter_stackSize : forall ss abs ctx pc rs sls a b c d args,
  S (S (S (S ss))) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> bvec_to_nat stackStart +
  aStackSize
  (Build_astate pc rs (aStackSize abs - (ss - args)) sls) <
  bvec_to_nat
  (sptr
    (Build_context (stack_slot ctx (ss - args)) a b c d)).
  intros.
  simpl aStackSize; simpl sptr.
  apply enter_stackSize'; trivial.
  omega.
Qed.

Hint Resolve enter_stackSize : asm.

Lemma stack_slot_neg_four : forall abs ctx ss,
  S (S (S (S ss))) <= aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> badd neg_four (stack_slot ctx ss)
  = stack_slot ctx (S (S (S (S ss)))).
  intros abs ctx ss Hsize Hsptr.
  add Hbtn (btn_upper (sptr ctx)).
  replace (badd neg_four (stack_slot ctx ss))
    with (bsub (stack_slot ctx ss) four).
  assert (Habove : bvec_to_nat (stack_slot ctx ss) >= 4).
  unfold stack_slot.
  rewrite ntb_invert; omega.
  unfold four.
  rewrite bsub_bdec4; try omega.
  apply btn_inj.
  rewrite bdec4_btn.
  unfold stack_slot.
  repeat (rewrite ntb_invert); omega.
  omega.

  unfold badd, barith.
  rewrite plus_comm.
  unfold neg_four.
  trivial.
Qed.

Fixpoint argsStack (args : list ty) {struct args} : astack :=
  match args with
    | nil => NatMap.empty Top
    | cons ty rest => NatMap.upd (argsStack rest) (4 * length args) ty
  end.

Definition robust_type ty :=
  match ty with
    | Top => true
    | Const _ => true
    | _ => false
  end.

Fixpoint checkArgs (args : list ty) (ss : nat) (sls : astack) {struct args} : bool :=
  match args with
    | nil => true
    | cons actual rest =>
      robust_type actual
      && subType (NatMap.sel sls (ss + 4 * length args)) actual
      && checkArgs rest ss sls
  end.

Lemma undo_fourx : forall n,
  n + (n + (n + (n + 0))) = 4 * n.
  intros.
  omega.
Qed.

Lemma btn_inj' : forall len (v1 v2 : Bvector len),
  bvec_to_nat v1 <> bvec_to_nat v2
  -> v1 <> v2.
  congruence.
Qed.

Lemma ntb_eqMod_inj : forall len n1 n2,
  ~eqMod (pow2 len) n1 n2
  -> nat_to_bvec len n1 <> nat_to_bvec len n2.
  intros len n1 n2 Hmod Heq.
  apply Hmod.
  apply ntb_mod'; trivial.
Qed.

Lemma ntb_O : forall len,
  bvec_to_nat (nat_to_bvec len 0) = 0.
  intros.
  rewrite ntb_invert; trivial.
  generalize (pow2_pos len); omega.
Qed.

Lemma btn_ntb_four : forall len,
  bvec_to_nat (nat_to_bvec (S (S (S len))) 4) = 4.
  intro.
  simpl.
  unfold mod2_bool, mod2, div2.
  destruct (Euclid.eucl_dev 2 two_gt_0 4).
  assert (q = 2); [omega | subst].
  assert (r = 0); [omega | subst].
  destruct (eq_nat_dec 0 0); try tauto.
  destruct (Euclid.eucl_dev 2 two_gt_0 2).
  assert (q = 1); [omega | subst].
  assert (r = 0); [omega | subst].
  destruct (eq_nat_dec 0 0); try tauto.
  destruct (Euclid.eucl_dev 2 two_gt_0 1).
  assert (q = 0); [omega | subst].
  assert (r = 1); [omega | subst].
  destruct (eq_nat_dec 1 0); try (wrong; omega).
  rewrite ntb_O.
  omega.
Qed.

Lemma enter_ret_neq : forall ss slot ctx len abs,
  len <= ss
  -> S (S (S (S ss))) < aStackSize abs
  -> 4 <= slot <= len
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> bvec_to_nat (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - (ss - len) - slot))
  >= bvec_to_nat (nat_to_bvec 32 (bvec_to_nat neg_four + bvec_to_nat (stack_slot ctx ss))) + 4.
  intros ss slot ctx len abs Hlen Hsize Hslot Hsize'.
  add Hbtn (btn_upper (sptr ctx)).
  assert (Heq : nat_to_bvec 32
    (bvec_to_nat neg_four + bvec_to_nat (stack_slot ctx ss))
    = nat_to_bvec 32
    (bvec_to_nat (stack_slot ctx ss) - bvec_to_nat four)).
  apply ntb_mod.
  rewrite plus_comm.
  unfold neg_four; apply eqMod_sym; apply bneg_btn_sub.
  unfold four; rewrite btn_ntb_four.
  unfold stack_slot.
  rewrite ntb_invert; omega.
  rewrite Heq.
  unfold four; rewrite btn_ntb_four.
  unfold stack_slot.
  repeat (rewrite ntb_invert); try omega.
Qed.

Lemma undo_fourx_S : forall n,
  S (n + S (n + S (n + S (n + 0)))) = 4 * S n.
  intros.
  omega.
Qed.

Lemma enter_hasType : forall rt ctx v ctx',
  hasType ctx' v rt
  -> robust_type rt = true
  -> hasType ctx v rt.
  intros rt ctx v ctx' Hty Hrobust.
  destruct rt; try discriminate.
  myauto.
  inversion Hty; myauto.
Qed.

Lemma ntb_inj : forall len n1 n2,
  n1 = n2
  -> nat_to_bvec len n1 = nat_to_bvec len n2.
  congruence.
Qed.

Lemma argsStack_aligned : forall k,
  1 <= k <= 3
  -> forall n args, NatMap.sel (argsStack args) (4 * n - k) = Top.
  intros k Hk.
  induction args; simpl.

  apply NatMap.sel_empty.

  rewrite undo_fourx_S.
  rewrite undo_fourx.
  destruct (eq_nat_dec (4 * S (length args)) (4 * n - k)).
  wrong; omega.
  
  rewrite NatMap.sel_neq; trivial.
Qed.

Lemma argsStack_robust : forall ss sls args slot,
  checkArgs args ss sls = true
  -> robust_type (NatMap.sel (argsStack args) slot) = true.
  induction args; simpl; intros slot Hcheck.

  rewrite NatMap.sel_empty; trivial.

  rewrite undo_fourx_S.
  rewrite undo_fourx_S in Hcheck.
  destruct (eq_nat_dec (4 * S (length args)) slot); subst.

  rewrite NatMap.sel_eq.
  bool_elim; trivial.

  rewrite NatMap.sel_neq; trivial.
  bool_elim; eauto.
Qed.

Lemma enter_checkArgs : forall ss abs st args b c d ret slot ctx,
  aget_reg abs ESP = Stack ss
  -> conc_regs abs st ctx
  -> 4 * length args <= ss
  -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
  -> 4 <= slot <= 4 * length args
  -> S (S (S (S ss))) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> conc_slots abs st ctx
  -> hasType
  (Build_context (stack_slot ctx (ss - 4 * length args)) b c d (Some ret))
  (get_mem32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret)
    (stack_slot
      (Build_context (stack_slot ctx (ss - 4 * length args)) b c d
        (Some ret)) slot)) (NatMap.sel (argsStack args) slot).
  intros ss abs st args b c d ret slot ctx Hesp Hregs Hlen Hargs Hslot Hsize Hsptr Hslots.
  add Hbtn (btn_upper (sptr ctx)).

  add Hr (Hregs ESP).
  rewrite Hesp in Hr.
  inversion Hr; subst.
  
  rewrite invert_set_mem32_neq.

  replace (stack_slot
    (Build_context (stack_slot ctx (ss - 4 * length args)) b c d
      (Some ret)) slot)
    with (stack_slot ctx (ss - 4 * length args + slot)).
  apply enter_hasType with ctx.

  clear Hesp Hr H2 H.
  generalize dependent ss.
  induction args; intros ss Hss Hcheck Hsize; simpl argsStack.

  rewrite NatMap.sel_empty.
  myauto.

  simpl length.
  rewrite undo_fourx_S.

  simpl in Hcheck.
  rewrite undo_fourx_S in Hcheck.
  bool_elim.

  destruct (eq_nat_dec (4 * S (length args)) slot); subst.

  rewrite NatMap.sel_eq.
  apply subType_sound
    with (NatMap.sel (aStackSlots abs)
      (ss - 4 * S (length args) + 4 * S (length args))); trivial.
  red in Hslots; unfold aget_stack in Hslots.
  apply Hslots.
  omega.

  rewrite NatMap.sel_neq; trivial.
  destruct (eq_nat_dec (4 * S (length args) - 1) slot);
    [subst; rewrite argsStack_aligned; [myauto | omega]
      | idtac].
  destruct (eq_nat_dec (4 * S (length args) - 2) slot);
    [subst; rewrite argsStack_aligned; [myauto | omega]
      | idtac].
  destruct (eq_nat_dec (4 * S (length args) - 3) slot);
    [subst; rewrite argsStack_aligned; [myauto | omega]
      | idtac].
  assert (Heq : ss - 4 * S (length args)
    = (ss - 4) - 4 * length args).
  omega.
  rewrite Heq.
  simpl length in Hslot, Hss.
  apply IHargs; try (rewrite <- Heq; trivial; fail); omega.

  eapply argsStack_robust; eauto.

  unfold stack_slot.
  unfold sptr at 2.
  rewrite ntb_invert.
  apply ntb_inj.
  omega.
  omega.

  split.
  unfold badd, barith.
  rewrite plus_comm.
  unfold neg_four.
  assert (Heq : nat_to_bvec 32 (bvec_to_nat (get_reg32 st ESP) + bvec_to_nat (bneg four))
    = nat_to_bvec 32 (bvec_to_nat (get_reg32 st ESP) - bvec_to_nat four)).
  apply ntb_mod.
  apply eqMod_sym; apply bneg_btn_sub.
  rewrite <- H.
  unfold stack_slot.
  unfold four; rewrite btn_ntb_four.
  rewrite ntb_invert; omega.
  rewrite <- H in Heq.
  rewrite Heq.
  assert (Hlt : bvec_to_nat (stack_slot ctx ss) - bvec_to_nat four < pow2 32 - 3).
  unfold stack_slot.
  rewrite ntb_invert; omega.
  rewrite ntb_invert; omega.

  split.
  unfold stack_slot.
  unfold sptr at 1.
  repeat (rewrite ntb_invert); omega.

  apply apart_higher.
  unfold stack_slot.
  unfold sptr at 1.
  rewrite ntb_invert.
  
  replace (bvec_to_nat
    (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - (ss - 4 * length args))) - slot)
  with (bvec_to_nat (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - (ss - 4 * length args) - slot))).
  replace (bvec_to_nat (badd neg_four (nat_to_bvec 32 (bvec_to_nat (sptr ctx) - ss))))
    with (bvec_to_nat (nat_to_bvec 32 (bvec_to_nat neg_four + bvec_to_nat (stack_slot ctx ss)))); trivial.
  eapply enter_ret_neq; eauto.
  rewrite ntb_invert.
  rewrite ntb_invert.
  trivial.
  omega.
  omega.
  rewrite ntb_invert; omega.
Qed.

Lemma argsStack_cleared : forall args slot,
  4 * length args < slot
  -> NatMap.sel (argsStack args) slot = Top.
  induction args; simpl; intros slot Hslot.
  apply NatMap.sel_empty.
  rewrite undo_fourx_S in Hslot.
  rewrite undo_fourx_S.
  destruct (eq_nat_dec (4 * S (length args)) slot); subst.
  wrong; omega.
  rewrite NatMap.sel_neq; trivial.
  apply IHargs.
  omega.
Qed.

Lemma enter_conc_slots' : forall abs pc rs ss st ret b c d ctx args,
  pcInRange ret
  -> aget_reg abs ESP = Stack ss
  -> conc_regs abs st ctx
  -> S (S (S (S ss))) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> 4 * length args <= ss
  -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
  -> conc_slots abs st ctx
  -> conc_slots
  (Build_astate pc rs (aStackSize abs - (ss - 4 * length args))
    (NatMap.upd (argsStack args) (S (S (S (S (4 * length args))))) RetPtr))
  (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret)
  (Build_context (stack_slot ctx (ss - 4 * length args)) b c d (Some ret)).
  intros until args; intros Hret Hesp Hregs Hss Hsize Hargs HcheckArgs Hslots slot Hslot.
  simpl in Hslot.
  rewrite undo_fourx in Hslot.
  unfold aget_stack; simpl.
  destruct (eq_nat_dec slot (S (S (S (S (4 * length args)))))); subst.
  rewrite NatMap.sel_eq.

  rewrite undo_fourx.
  replace (stack_slot (Build_context (stack_slot ctx (ss - 4 * length args))
    b c d (Some ret)) (S (S (S (S (4 * length args))))))
    with (stack_slot ctx (S (S (S (S ss))))).
  add Hr (Hregs ESP).
  rewrite Hesp in Hr.
  inversion Hr; subst.
  rewrite (stack_slot_neg_four abs ctx ss); trivial; try omega.
  rewrite invert_set_mem32.
  myauto.

  unfold stack_slot.
  unfold sptr at 2.
  rewrite ntb_invert.
  assert (bvec_to_nat (sptr ctx) - S (S (S (S ss)))
    = bvec_to_nat (sptr ctx) - (ss - 4 * length args) - S (S (S (S (4 * length args))))).
  omega.
  congruence.
  generalize (btn_upper (sptr ctx)); omega.
  rewrite NatMap.sel_neq; auto with arith.
  rewrite undo_fourx.

  destruct (eq_nat_dec (4 * S (length args) - 1) slot);
    [subst; rewrite argsStack_aligned; try omega; myauto
      | idtac].
  destruct (eq_nat_dec (4 * S (length args) - 2) slot);
    [subst; rewrite argsStack_aligned; try omega; myauto
      | idtac].
  destruct (eq_nat_dec (4 * S (length args) - 3) slot);
    [subst; rewrite argsStack_aligned; try omega; myauto
      | idtac].
  destruct (le_lt_dec slot (4 * length args)).
  eapply enter_checkArgs; eauto.
  tauto.
  rewrite argsStack_cleared; myauto.
Qed.

Lemma get_mem32_pc : forall st pc addr,
  get_mem32 (set_pc st pc) addr = get_mem32 st addr.
  intros.
  unfold get_mem32.
  unfold get_mem.
  trivial.
Qed.

Lemma get_mem32_reg32 : forall st r v addr,
  get_mem32 (set_reg32 st r v) addr = get_mem32 st addr.
  intros.
  unfold get_mem32.
  unfold get_mem.
  trivial.
Qed.

Hint Rewrite get_mem32_pc get_mem32_reg32 : asm.

Lemma conc_slots_spc : forall abs st ctx pc,
  conc_slots abs st ctx
  -> conc_slots abs (set_pc st pc) ctx.
  intuition.
Qed.

Lemma conc_slots_sreg : forall abs st ctx r v,
  conc_slots abs st ctx
  -> conc_slots abs (set_reg32 st r v) ctx.
  intuition.
Qed.

Hint Resolve conc_slots_spc conc_slots_sreg : asm.

Lemma enter_conc_slots : forall abs pc rs ss st ret b c d ctx args,
  pcInRange ret
  -> aget_reg abs ESP = Stack ss
  -> conc_regs abs st ctx
  -> S (S (S (S ss))) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> 4 * length args <= ss
  -> conc_slots abs st ctx
  -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
  -> conc_slots
  (Build_astate (apcConst pc) rs (aStackSize abs - (ss - 4 * length args))
    (NatMap.upd (argsStack args) (S (S (S (S (4 * length args))))) RetPtr))
  (set_pc
    (set_reg32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret) ESP
      (bsub (get_reg32 st ESP) four)) pc)
  (Build_context (stack_slot ctx (ss - 4 * length args)) b c d (Some ret)).
  intros.
  apply conc_slots_spc; apply conc_slots_sreg;
    eapply enter_conc_slots'; eauto.
Qed.

Hint Resolve enter_conc_slots : asm.

Lemma enter_belowMax : forall ctx ss a b c d (args : list ty), 
  belowMax (sptr ctx) 4
  -> belowMax (sptr (Build_context (stack_slot ctx (ss - 4 * length args)) a b c d)) 4.
  simpl; myeauto.
Qed.

Hint Resolve enter_belowMax : asm.

Lemma enter_memIn' : forall ss abs st ctx ret args,
  aget_reg abs ESP = Stack ss
  -> S (S (S (S ss))) < aStackSize abs
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 4
  -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
  -> args <= ss
  -> agree_range'
  (stMem (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret))
  (stMem st) (stack_slot ctx (ss - args)) stackEnd.
  intros ss abs st ctx ret args Hss Hsize Hregs Hsptr Hbelow Hagree Hargs addr Hbound.
  add Hesp (Hregs ESP).
  rewrite Hss in Hesp.
  inversion Hesp; subst.
  assert (Heq : bvec_to_nat (stack_slot ctx (ss - args)) = bvec_to_nat (sptr ctx) - ss + args).
  unfold stack_slot.
  rewrite ntb_invert; trivial.
  generalize (btn_upper (sptr ctx)); omega.
  generalize (btn_upper (sptr ctx)); omega.
  rewrite Heq in Hbound.
  unfold set_mem32, set_mem.
  do 4 (unfold stMem at 1).
  rewrite (stack_slot_neg_four abs ctx ss); trivial; try omega.
  assert (Heq2 : bvec_to_nat (stack_slot ctx (S (S (S (S ss))))) = bvec_to_nat (sptr ctx) - S (S (S (S ss)))).
  unfold stack_slot.
  rewrite ntb_invert; trivial.
  generalize (btn_upper (sptr ctx)); omega.
  repeat (rewrite Int32Map.sel_neq); trivial; apply (btn_inj' 32).
  omega.
  rewrite binc_btn_belowMax; [omega | apply belowMax_dec with 4; myauto].
  rewrite binc2_btn_belowMax; [omega | apply belowMax_dec with 4; myauto].
  rewrite binc3_btn_belowMax; [omega | apply belowMax_dec with 4; myauto].
Qed.

Lemma enter_memIn : forall ss abs st ctx pc a b c d e f g h ret args,
  aget_reg abs ESP = Stack ss
  -> 4 <= S (S (S (S ss))) < aStackSize abs
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 4
  -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
  -> args <= ss
  -> agree_range'
  (stMem
    (set_pc
      (set_reg32 (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret) ESP
        (bsub (get_reg32 st ESP) four)) pc))
  (memIn
    (Build_context a b c (stMem st) d))
  (sptr (Build_context (stack_slot ctx (ss - args)) e f g h)) stackEnd.
  intros.
  add H' (enter_memIn' ss abs st ctx ret).
  simpl memIn; simpl sptr.
  apply H'; auto; omega.
Qed.

Hint Resolve enter_memIn : asm.

Lemma enter_esp : forall ctx ss a b c d size args,
  S (S (S (S ss))) < size
  -> bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
  -> bvec_to_nat stackStart + size < bvec_to_nat (sptr ctx)
  -> bvec_to_nat
  (sptr
    (Build_context (stack_slot ctx (ss - args)) a b c d)) <= bvec_to_nat initialESP.
  intros ctx ss a b c d size args Hss Hesp Hsptr.
  simpl sptr.
  unfold stack_slot.
  rewrite ntb_invert.
  omega.
  generalize (btn_upper (sptr ctx)); omega.
Qed.

Hint Resolve enter_esp : asm.

Definition leave_context st ctx ss (args : list ty) :=
  Build_context (stack_slot ctx (ss - 4 * length args))
  (4 * length args) (get_reg32 st EBP) (stMem st) (Some (stPc st)).

Definition enter : forall abs ret pc ss args,
  pcInRange pc
  -> pcInRange ret
  -> aget_reg abs ESP = Stack ss
  -> S (S (S (S ss))) < aStackSize abs
  -> 4 * length args <= ss
  -> checkArgs args (ss - 4 * length args) (aStackSlots abs) = true
  -> sig_option (fun abs' =>
    forall st ctx, conc' abs st ctx
      -> conc' abs' (set_pc
	(set_reg32
	  (set_mem32 st (badd neg_four (get_reg32 st ESP)) ret)
	  ESP (bsub (get_reg32 st ESP) four))
	pc)
      (leave_context (set_pc st ret) ctx ss args)).
  intros abs ret pc ss args Hpc Hret Hesp Hsize Hargs Hcheck.
  apply SSome with (Build_astate (apcConst pc)
    (Reg32Map.upd
      (Reg32Map.upd
	(Reg32Map.empty Top)
	ESP (Stack (S (S (S (S (4 * length args)))))))
      EBP OldEbp) (aStackSize abs - (ss - 4 * length args))
    (NatMap.upd
      (argsStack args)
       (S (S (S (S (4 * length args))))) RetPtr)).
  intros st ctx Hconc.
  unfold leave_context.
  simpl.
  replace (get_reg32 (set_pc st ret) EBP)
    with (get_reg32 st EBP); trivial.
  assert (Hsize' : 4 <= S (S (S (S ss))) < aStackSize abs).
  omega.
  inversion Hconc; unfold leave_context; split; try rewrite undo_fourx; myeauto.
  red; simpl; trivial.
Qed.


(** * Switch point of view from callee to caller.
  Use this on the state right before a RET instruction.
  *)

Definition leave_abs retType args := Build_astate
  apcRet
  (Reg32Map.upd
    (Reg32Map.upd
      (Reg32Map.upd
	(Reg32Map.empty Top)
	EAX retType)
      ESP (Stack args))
    EBP OldEbp) 0
  (NatMap.empty Top).

Import Datatypes.

Fixpoint keepValidSlots (n : nat) (stk : astack) {struct n} : astack :=
  match n with
    | O => NatMap.empty Top
    | S n => NatMap.upd (keepValidSlots n stk) n (NatMap.sel stk n)
  end.

(** Forget information across a function call *)
Definition ret_abs retType abs ss args := Build_astate
  (aPc abs)
  (Reg32Map.upd
    (Reg32Map.upd
      (Reg32Map.upd
	(Reg32Map.empty Top)
	EAX retType)
      ESP (aget_reg abs ESP))
    EBP (aget_reg abs EBP))
  (aStackSize abs)
  (keepValidSlots (S (ss - args)) (aStackSlots abs)).

Lemma ret_regs : forall abs st ctx rt ss args,
  conc_regs abs st ctx
  -> hasType ctx (get_reg32 st EAX) rt
  -> conc_regs (ret_abs rt abs ss args) st ctx.
  intros abs st ctx rt ss args Hregs Heax r.
  unfold aget_reg; simpl.

  destruct (reg32_eq_dec EBP r); subst.
  rewrite Reg32Map.sel_eq.
  auto.
  rewrite Reg32Map.sel_neq; auto.

  destruct (reg32_eq_dec ESP r); subst.
  rewrite Reg32Map.sel_eq.
  auto.
  rewrite Reg32Map.sel_neq; auto.
  
  destruct (reg32_eq_dec EAX r); subst.
  rewrite Reg32Map.sel_eq.
  trivial.
  rewrite Reg32Map.sel_neq; auto.

  rewrite Reg32Map.sel_empty.
  myauto.
Qed.

Hint Resolve ret_regs : asm.

Lemma ret_slots' : forall abs n st ctx pc a,
  conc_slots abs st ctx
  -> conc_slots (Build_astate pc a (aStackSize abs) (keepValidSlots n (aStackSlots abs))) st ctx.
  induction n; intros st ctx pc a Hslots sl Hsl; unfold aget_stack; simpl.
  rewrite NatMap.sel_empty.
  myauto.

  simpl in Hsl.
  destruct (eq_nat_dec n sl); subst.
  rewrite NatMap.sel_eq.
  red in Hslots.
  unfold aget_stack in Hslots.
  myauto.

  rewrite NatMap.sel_neq; trivial.
  apply (IHn st ctx pc a Hslots sl); trivial.
Qed.

Lemma ret_slots : forall abs st ctx rt ss args,
  conc_slots abs st ctx
  -> conc_slots (ret_abs rt abs ss args) st ctx.
  intros.
  unfold ret_abs.
  apply ret_slots'; trivial.
Qed.

Hint Resolve ret_slots : asm.

Lemma leave_pc : forall abs st ctx rt ss st' pc args,
  aPc abs = apcConst pc
  -> conc' abs st ctx
  -> conc' (leave_abs rt (4 * length args)) st' (leave_context st ctx ss args)
  -> Defs.conc_pc (ret_abs rt abs ss (4 * length args)) st' ctx.
  intros abs st ctx rt ss st' pc args Hpc Hconc Hconc'.
  add Hcpc (concPc abs st ctx Hconc).
  add Hcpc' (concPc (leave_abs rt (4 * length args)) st' (leave_context st ctx ss args) Hconc').
  red in Hcpc, Hcpc'.
  simpl in Hcpc'.
  rewrite Hpc in Hcpc.
  red.
  simpl.
  rewrite Hpc.
  congruence.
Qed.

Hint Resolve leave_pc : asm.

Lemma leave_hasType : forall rt st ctx ss v args,
  hasType (leave_context st ctx ss args) v rt
  -> robust_type rt = true
  -> hasType ctx v rt.
  intros rt st ctx ss v args Hty Hrobust.
  destruct rt; try discriminate.
  myauto.
  inversion Hty; myauto.
Qed.

Hint Resolve leave_hasType : asm.

Lemma leave_conc_regs : forall abs st ctx rt ss st' args,
  aget_reg abs ESP = Stack ss
  -> ss < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> conc' abs st ctx
  -> conc' (leave_abs rt (4 * length args)) st' (leave_context st ctx ss args)
  -> conc_regs (leave_abs rt (4 * length args)) st' (leave_context st ctx ss args)
  -> robust_type rt = true
  -> 4 * length args <= ss
  -> conc_regs (ret_abs rt abs ss (4 * length args)) st' ctx.
  intros abs st ctx rt ss st' args Hesp Hss Hsize Hconc Hconc' Hregs Hrobust Hargs r.
  unfold aget_reg; simpl.

  destruct (reg32_eq_dec EBP r); subst.
  rewrite Reg32Map.sel_eq.
  add Hebp' (Hregs EBP).
  unfold aget_reg in Hebp'; simpl in Hebp'.
  rewrite Reg32Map.sel_eq in Hebp'.
  inversion Hebp'; subst.
  myauto.
  
  rewrite Reg32Map.sel_neq; trivial.
  destruct (reg32_eq_dec ESP r); subst.
  rewrite Reg32Map.sel_eq.
  add Hesp' (Hregs ESP).
  unfold aget_reg in Hesp'; simpl in Hesp'.
  rewrite Reg32Map.sel_neq in Hesp'; trivial.
  rewrite Reg32Map.sel_eq in Hesp'.
  inversion Hesp'; subst.
  rewrite undo_fourx.
  replace (stack_slot (leave_context st ctx ss args) (4 * length args))
    with (stack_slot ctx ss).
  rewrite Hesp; myauto.
  apply HasType_Stack; omega.
  unfold stack_slot; simpl sptr.
  unfold stack_slot.
  rewrite ntb_invert.
  apply ntb_inj.
  omega.
  generalize (btn_upper (sptr ctx)); omega.

  rewrite Reg32Map.sel_neq; trivial.
  destruct (reg32_eq_dec EAX r); subst.
  rewrite Reg32Map.sel_eq.
  assert (rt = aget_reg (leave_abs rt (4 * length args)) EAX).
  unfold aget_reg; simpl.
  do 2 (rewrite Reg32Map.sel_neq; try discriminate);
    rewrite Reg32Map.sel_eq; trivial.
  rewrite H.
  rewrite H in Hrobust.
  myeauto.
  
  rewrite Reg32Map.sel_neq; trivial.
  rewrite Reg32Map.sel_empty; myauto.
Qed.

Hint Resolve leave_conc_regs : asm.

Lemma leave_stackSize : forall abs st ctx rt ss args,
  conc' abs st ctx
  -> bvec_to_nat stackStart + aStackSize (ret_abs rt abs ss args) < bvec_to_nat (sptr ctx).
  intros abs st ctx rt ss args Hconc.
  inversion Hconc.
  trivial.
Qed.

Hint Resolve leave_stackSize : asm.

Lemma agree_range'_read32 : forall st1 st2 start finish addr,
  agree_range' (stMem st1) (stMem st2) start finish
  -> bvec_to_nat start <= bvec_to_nat addr < bvec_to_nat finish - 4
  -> belowMax finish 0
  -> get_mem32 st1 addr = get_mem32 st2 addr.
  intros st1 st2 start finish addr Hrange Haddr Hbelow.
  assert (Hbelow' : belowMax addr 4).
  destruct Hbelow; split; omega.
  clear Hbelow; rename Hbelow' into Hbelow.
  red in Hrange.
  unfold get_mem32, get_mem.
  repeat (rewrite Hrange); trivial.
  replace (binc (binc (binc addr))) with (binc3 addr); trivial;
  rewrite binc3_btn_belowMax;
    [omega
      | apply belowMax_dec with 4; trivial; omega].
  replace (binc (binc addr)) with (binc2 addr); trivial;
  rewrite binc2_btn_belowMax;
    [omega
      | apply belowMax_dec with 4; trivial; omega].
  rewrite binc_btn_belowMax;
    [omega
      | apply belowMax_dec with 4; trivial; omega].
  omega.
Qed.

Lemma initialESP_stackEnd :
  bvec_to_nat initialESP + 5 = bvec_to_nat stackEnd.
  unfold initialESP, stackEnd.
  repeat (rewrite ntb_invert).
  omega.
  generalize stack_fits; omega.
  generalize stack_fits; omega.
Qed.

Lemma ret_abs_stack' : forall ss sl,
  sl < ss
  -> forall slots, NatMap.sel slots sl = NatMap.sel (keepValidSlots ss slots) sl.
  induction ss; intros sl Hle slots; simpl.
  rewrite NatMap.sel_empty; wrong; omega.

  destruct (eq_nat_dec ss sl); subst.
  rewrite NatMap.sel_eq; trivial.
  rewrite NatMap.sel_neq; trivial.
  simpl in IHss.
  apply IHss.
  omega.
Qed.

Lemma ret_abs_stack : forall rt ss sl args,
  sl <= ss - args
  -> forall abs, aget_stack abs sl = aget_stack (ret_abs rt abs ss args) sl.
  intros rt ss sl args Hss abs.
  unfold aget_stack.
  replace (aStackSlots (ret_abs rt abs ss args))
    with (keepValidSlots (S (ss - args)) (aStackSlots abs)); trivial.
  apply ret_abs_stack'; omega.
Qed.

Lemma ret_abs_stack_not' : forall ss sl,
  sl >= ss
  -> forall slots, NatMap.sel (keepValidSlots ss slots) sl = Top.
  induction ss; intros sl Hle slots; simpl.
  apply NatMap.sel_empty.
  rewrite NatMap.sel_neq; trivial.
  apply IHss; omega.
  omega.
Qed.

Lemma ret_abs_stack_not : forall rt ss sl args,
  sl > ss - args
  -> forall abs, aget_stack (ret_abs rt abs ss args) sl = Top.
  intros rt ss sl args Hss abs.
  unfold aget_stack.
  replace (aStackSlots (ret_abs rt abs ss args))
    with (keepValidSlots (S (ss - args)) (aStackSlots abs)); trivial.
  rewrite ret_abs_stack_not'; auto with arith.
Qed.

Lemma leave_slots : forall abs st ctx rt ss st' args,
  ss < aStackSize abs
  -> conc' abs st ctx
  -> conc' (leave_abs rt (4 * length args)) st' (leave_context st ctx ss args)
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> bvec_to_nat stackStart + aStackSize (leave_abs rt (4 * length args)) <
  bvec_to_nat (sptr (leave_context st ctx ss args))
  -> conc_slots abs st ctx
  -> conc_slots (leave_abs rt (4 * length args)) st' (leave_context st ctx ss args)
  -> agree_range' (stMem st') (memIn (leave_context st ctx ss args))
  (sptr (leave_context st ctx ss args)) stackEnd
  -> bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
  -> conc_slots (ret_abs rt abs ss (4 * length args)) st' ctx.
  intros abs st ctx rt ss st' args Hss Hconc Hconc' Hsize Hsize' Hslots Hslots' Hagree Hbelow sl Hsl.
  simpl in Hsl.
  destruct (dec_le sl (ss - 4 * length args)).

  replace (aget_stack (ret_abs rt abs ss (4 * length args)) sl)
    with (aget_stack abs sl).
  simpl sptr in Hagree; simpl memIn in Hagree.
  rewrite (agree_range'_read32 st' st (stack_slot ctx (ss - 4 * length args)) stackEnd);
    myauto.
  unfold stack_slot.
  repeat (rewrite ntb_invert); try (generalize (btn_upper (sptr ctx)); omega).
  generalize initialESP_stackEnd; omega.
  split.
  unfold stackEnd.
  rewrite ntb_invert;
    generalize stack_fits; omega.
  apply ret_abs_stack; omega.
  
  rewrite ret_abs_stack_not; myauto.
  omega.
Qed.

Hint Resolve leave_slots : asm.

Lemma leave_memIn : forall st st' ctx ss args,
  bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
  -> agree_range' (stMem st') (memIn (leave_context st ctx ss args))
  (sptr (leave_context st ctx ss args)) stackEnd
  -> agree_range' (stMem st) (memIn ctx) 
  (sptr ctx) stackEnd
  -> agree_range' (stMem st') (memIn ctx) (sptr ctx) stackEnd.
  intros st st' ctx ss args Hsptr Hagree Hagree'.
  simpl in Hagree.
  intros addr Haddr.
  transitivity (Int32Map.sel (stMem st) addr).

  apply Hagree.
  unfold stack_slot.
  rewrite ntb_invert.
  omega.
  generalize (btn_upper (sptr ctx)); omega.
  
  myauto.
Qed.

Hint Resolve leave_memIn : asm.

Definition leave' : forall (abs : astate) (rt : ty) ss args,
  ss < aStackSize abs
  -> aget_reg abs ESP = Stack ss
  -> 4 * length args <= ss
  -> sig_option (fun abss =>
    forall st ctx, conc' abs st ctx
      -> menter conc' (leave_context st ctx ss args) (snd abss) ctx (fst abss)).
  intros abs rt ss args Hss Hesp Hargs.

  caseEq (aPc abs); [idtac | type_fail].
  intros pc Hpc.
  destrEq Hrobust (robust_type rt); [idtac | type_fail].
  apply SSome with (ret_abs rt abs ss (4 * length args), leave_abs rt (4 * length args)).
  intros st ctx Hconc st' Hconc'.
  simpl; rewrite undo_fourx.
  inversion Hconc; inversion Hconc'; split;
    try (eapply leave_slots; eauto); myeauto.
Qed.

Definition leave : forall abs pc (rt : ty) ss args,
  aget_reg abs ESP = Stack ss
  -> S (S (S (S ss))) < aStackSize abs
  -> 4 * length args <= ss
  -> sig_option (fun abss =>
    forall st ctx, conc' abs st ctx
      -> menter conc'
      (leave_context
	(set_pc st pc) ctx ss args) (snd abss) ctx (fst abss)).
  intros abs pc rt ss args Hesp Hsize Hargs.
  destruct (pcInRange_dec pc); [idtac | type_fail].
  destruct (leave' (aset_pc abs pc) rt ss args); trivial.
  simpl; omega.

  apply SSome with x.
  intros st ctx Hconc.
  apply m; myauto.

  type_fail.
Qed.


(** * Creating the successor state of a RET instruction *)

Import ProofOS.Kernel.Defs.

Lemma invert_RetPtr: forall ctx v,
  hasType ctx v RetPtr
  -> Some v = ret ctx.
  intros ctx v Hret.
  inversion Hret; subst.
  auto.
Qed.

Lemma invert_RetPtr': forall ctx v,
  hasType ctx v RetPtr
  -> pcInRange v.
  intros ctx v Hret.
  inversion Hret; subst.
  auto.
Qed.

Hint Resolve invert_RetPtr invert_RetPtr' : asm.

Lemma ret_conc_pc : forall abs res st' st ctx,
  conc' abs st ctx
  -> (forall (st : state) (ctx : context),
    conc' abs st ctx ->
    hasType ctx
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
      st) RetPtr)
  -> conc_pc
  (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
    (aStackSize abs) (aStackSlots abs))
  (set_pc st'
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
  ctx.
  intros abs res st' st ctx Hconc Hret.
  red; simpl.
  add Hrp (Hret st ctx Hconc).
  replace (ret ctx)
    with (Some (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st));
    myauto.
Qed.

Hint Resolve ret_conc_pc : asm.

Lemma ret_pcInRange : forall abs st ctx st',
  conc' abs st ctx
  -> (forall (st : state) (ctx : context),
    conc' abs st ctx ->
    hasType ctx
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
      st) RetPtr)
  -> pcInRange
  (stPc
    (set_pc st'
      (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
        st))).
  intros abs st ctx st' Hconc Hret.
  add Hrp (Hret st ctx Hconc).
  myeauto.
Qed.

Hint Resolve ret_pcInRange : asm.

Lemma ret_agree : forall e st st' pc,
  write_genop32 (Reg ESP) e st st'
  -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
  -> agree_range
  (stMem
    (set_pc st' pc)) initialMem initialPc (actualCodeLen + longestInstr).
  intros e st st' pc Hwrite Hagree.
  inversion Hwrite; subst.
  trivial.
Qed.

Hint Resolve ret_agree : asm.

Lemma ret_conc_regs : forall abs st ctx res st' esp,
  conc' abs st ctx
  -> (forall (st : state) (ctx : context),
    conc' abs st ctx -> hasType ctx (eval_genop32 (Reg ESP) st) esp)
  -> (forall (st : state) (ctx : context) (v1 v2 : int32),
    conc' abs st ctx ->
    hasType ctx v1 (Const four) ->
    hasType ctx v2 esp -> hasType ctx (eval_arith Add v1 v2) res)
  -> write_genop32 (Reg ESP)
  (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
  -> conc_regs abs st ctx
  -> conc_regs
  (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
    (aStackSize abs) (aStackSlots abs))
  (set_pc st'
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
  ctx.
  intros abs st ctx res st' esp Hconc Hesp Hres Hwrite Hregs.
  intro r.
  simpl.
  unfold get_reg32, aget_reg; simpl.

  destruct (reg32_eq_dec ESP r); subst.
  rewrite Reg32Map.sel_eq.
  inversion Hwrite; subst.
  simpl.
  rewrite Reg32Map.sel_eq.
  unfold eval_arith in Hres.
  replace (badd32 (get_reg32 st ESP) four)
    with (badd32 four (get_reg32 st ESP)).
  apply Hres with st; trivial.
  myauto.
  unfold eval_genop32 in Hesp; auto.
  unfold badd32, badd, barith.
  assert (bvec_to_nat four + bvec_to_nat (get_reg32 st ESP) =
    bvec_to_nat (get_reg32 st ESP) + bvec_to_nat four).
  omega.
  congruence.
  
  rewrite Reg32Map.sel_neq; trivial.
  inversion Hwrite; subst.
  simpl.
  rewrite Reg32Map.sel_neq; trivial.
  red in Hregs; unfold get_reg32, aget_reg in Hregs; auto.
Qed.

Hint Resolve ret_conc_regs : asm.

Lemma ret_conc_slots : forall abs st ctx st' res,
  write_genop32 (Reg ESP)
  (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
  -> conc_slots abs st ctx
  -> conc_slots
  (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
    (aStackSize abs) (aStackSlots abs))
  (set_pc st'
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
  ctx.
  intros abs st ctx st' res Hwrite Hslots sl Hsl.
  simpl.
  unfold aget_stack.
  simpl.
  inversion Hwrite; subst.
  replace (get_mem32
    (set_pc
      (set_reg32 st ESP
        (eval_arith Add four (eval_genop32 (Reg ESP) st)))
      (get_mem32 st (eval_addr (Build_address bzero32 (Some ESP) None) st)))
    (stack_slot ctx sl))
  with (get_mem32 st (stack_slot ctx sl)); trivial.
  red in Hslots.
  unfold aget_stack in Hslots.
  auto.
Qed.

Hint Resolve ret_conc_slots : asm.

Lemma ret_memIn : forall st st' ctx,
  write_genop32 (Reg ESP)
  (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
  -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
  -> agree_range'
  (stMem
    (set_pc st'
      (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
        st))) (memIn ctx) (sptr ctx) stackEnd.
  intros st st' ctx Hwrite Hagree.
  inversion Hwrite; subst.
  trivial.
Qed.

Hint Resolve ret_memIn : asm.

Theorem ret_conc : forall abs st ctx esp st' res nextPc abs',
  (forall (st : state) (ctx : context),
    conc' abs st ctx -> hasType ctx (eval_genop32 (Reg ESP) st) esp)
  -> (forall (st : state) (ctx : context),
    conc' abs st ctx ->
    hasType ctx
    (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None))
      st) RetPtr)
  -> (forall (st : state) (ctx : context) (v1 v2 : int32),
    conc' abs st ctx ->
    hasType ctx v1 (Const four) ->
    hasType ctx v2 esp -> hasType ctx (eval_arith Add v1 v2) res)
  -> (forall (st st' : state) (ctx : context) (v : int32),
    conc' abs st ctx ->
    hasType ctx v res ->
    write_genop32 (Reg ESP) v st st' -> conc' abs' st' ctx)
  -> conc' abs st ctx
  -> write_genop32 (Reg ESP)
         (eval_arith Add four (eval_genop32 (Reg ESP) st)) st st'
   -> exec (Ret, nextPc) st
   (set_pc st'
     (eval_genop32
       (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
   -> conc'
   (Build_astate apcRet (Reg32Map.upd (aRegs abs) ESP res) 
     (aStackSize abs) (aStackSlots abs))
   (set_pc st'
     (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
   ctx.
  intros abs st ctx esp st' res nextPc abs' Hesp Hret Hadd Hwrite Hconc Hwrite32 Hexec.
  inversion Hexec; subst.
  inversion Hconc; split;
    try (eapply ret_conc_slots; eauto); myeauto.
Qed.

Hint Resolve ret_conc : asm.

