(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Proving that the soundness relation holds *)

Require Import Arith.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Tactics.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Env.

Require Import ProofOS.Kernel.Defs.
Require Import ProofOS.Kernel.Misc.

Set Implicit Arguments.

(** * Tactics *)

Ltac unfolder := unfold get_reg32, aget_reg, set_reg32, aset_reg, stack_slot, aset_stack, aget_stack.
Ltac sel_eq_replace e1 e2 lemma :=
  replace e1 with e2;
    [idtac | symmetry; unfolder; simpl; apply lemma; trivial].
Ltac sel_neq_replace e1 e2 lemma :=
  replace e1 with e2;
    [idtac | symmetry; unfolder; simpl; apply lemma; trivial].

Ltac ntb_conv :=
  match goal with
    | [ |- nat_to_bvec _ ?N = nat_to_bvec _ ?M ] =>
      assert (N = M); [idtac | congruence]
  end.

Hint Constructors exec write_genop32 : asm.
Hint Constructors hasType conc' : asm.

(** * Lemmas to be used in the following sections *)

Lemma get_set_stack_apart : forall abs s1 s2 t,
  nat_apart s1 s2
  -> aget_stack (aset_stack abs s1 t) s2 = aget_stack abs s2.
  intros abs s1 s2 t Hapart.
  unfolder; simpl;
    repeat (rewrite NatMap.sel_neq); trivial; inversion Hapart; omega.
Qed.

Lemma stack_slot_below : forall ctx slot n,
  belowMax (sptr ctx) n
  -> bvec_to_nat (stack_slot ctx slot) <= bvec_to_nat (sptr ctx).
  intros.
  unfold stack_slot.
  rewrite ntb_invert;
    [omega
      | inversion H; omega].
Qed.

Lemma stack_slot_belowMax : forall ctx slot n,
  belowMax (sptr ctx) n
  -> belowMax (stack_slot ctx slot) n.
  intros ctx slot n Hbelow.
  inversion Hbelow.
  add Hbelow2 (stack_slot_below ctx slot Hbelow).
  split; omega.
Qed.

Lemma stack_slot_belowMax' : forall ctx,
  bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP
  -> belowMax (sptr ctx) 4.
  intros.
  split.
  generalize initialESP_upper; omega.
Qed.

Hint Resolve stack_slot_belowMax stack_slot_belowMax' : asm.

Lemma slot_not_too_high : forall ctx n abs st,
  conc' abs st ctx
  -> not_too_high (stack_slot ctx n).
  intros.
  red.
  destruct H.
  myeauto.
Qed.

Import Datatypes.

Lemma plus4_apart : forall n m,
  nat_apart n m
  -> nat_apart (S (S (S (S n)))) (S (S (S (S m)))).
  intros n m Hapart.
  inversion Hapart; red; omega.
Qed.

Hint Resolve plus4_apart : asm.

Lemma slots_apart : forall ctx n m,
  n < bvec_to_nat (sptr ctx)
  -> m < bvec_to_nat (sptr ctx)
  -> nat_apart n m
  -> apart (stack_slot ctx n) (stack_slot ctx m).
  intros ctx n m Hn Hm Hapart.
  unfold stack_slot; add Hbound (btn_upper (sptr ctx));
    inversion Hapart; [right | left];
      repeat (rewrite ntb_invert); omega.
Qed.

Hint Rewrite get_set_stack_apart : asm.

Hint Rewrite invert_set_mem32 : asm.
Hint Rewrite invert_set_mem32_neq using (myeauto; fail) : asm.

Import Datatypes.

Lemma apart_hasType : forall st ctx abs t n n0,
  conc' abs st ctx
  -> nat_apart n n0
  -> 4 <= n0 < aStackSize (aset_stack abs n t)
  -> hasType ctx (get_mem32 st (stack_slot ctx n0))
  (aget_stack (aset_stack abs n t) n0).
  intros st ctx abs t n n0 Hconc Hapart Hlt.
  unfolder; unfold aStackSlots at 1;
    repeat (rewrite NatMap.sel_neq; try (inversion Hapart; omega));
      inversion Hconc; red in concStackSlots;
	unfold stack_slot, aget_stack in concStackSlots;
	  apply concStackSlots; trivial.
Qed.

Lemma stackSize_lt : forall abs st ctx n,
  conc' abs st ctx
  -> n < aStackSize abs
  -> n < bvec_to_nat (sptr ctx).
  intros abs st ctx n Hconc Hlt.
  inversion Hconc; omega.
Qed.

Hint Resolve slot_not_too_high apart_hasType slots_apart stackSize_lt : asm.


(** * Generic lemmas about aspects of [conc] *)

(** ** [conc'] *)

Lemma conc_pc_set_pc : forall abs st ctx pc,
  conc_pc abs st ctx
  -> pcInRange pc
  -> conc_pc (aset_pc abs pc) (set_pc st pc) ctx.
  intros abs st ctx pc Hconc Hpc.
  red; simpl; trivial.
Qed.

Hint Resolve conc_pc_set_pc : asm.

Lemma conc'_pc : forall abs st ctx pc,
  conc' abs st ctx
  -> pcInRange pc
  -> conc' (aset_pc abs pc) (set_pc st pc) ctx.
  intros abs st ctx pc Hconc Hpc.
  inversion Hconc; split; intuition.
Qed.

Hint Resolve conc'_pc : asm.

Lemma conc_pc : forall abs st pc,
  conc abs st
  -> pcInRange pc
  -> conc (aset_pc abs pc) (set_pc st pc).
  intros abs st pc Hconc Hpc.
  destruct Hconc as [ctx Hconc].
  exists ctx.
  myauto.
Qed.

Hint Resolve conc_pc : asm.

Lemma conc'_arith_flags : forall abs st ctx aop v1 v2,
  conc' abs st ctx
  -> conc' abs (arith_flags st aop v1 v2) ctx.
  intros abs st ctx aop v1 v2 Hconc.
  inversion Hconc.
  split; intuition.
Qed.

Hint Resolve conc'_arith_flags : asm.

Lemma conc'_arith_flags_no_CF : forall abs st ctx aop v1 v2,
  conc' abs st ctx
  -> conc' abs (arith_flags_no_CF st aop v1 v2) ctx.
  intros abs st ctx aop v1 v2 Hconc.
  inversion Hconc.
  split; intuition.
Qed.

Hint Resolve conc'_arith_flags_no_CF : asm.
  

(** ** [conc_regs] *)

Lemma conc_regs_upd : forall abs st ctx r v t,
  conc_regs abs st ctx
  -> hasType ctx v t
  -> conc_regs (aset_reg abs r t) (set_reg32 st r v) ctx.
  intros abs st ctx r v t Hconc Hty r0.
  destruct (reg32_eq_dec r r0); subst;
    [sel_eq_replace (get_reg32 (set_reg32 st r0 v) r0) v Reg32Map.sel_eq;
      sel_eq_replace (aget_reg (aset_reg abs r0 t) r0) t Reg32Map.sel_eq
      | sel_neq_replace (get_reg32 (set_reg32 st r v) r0) (get_reg32 st r0) Reg32Map.sel_neq;
	sel_neq_replace (aget_reg (aset_reg abs r t) r0) (aget_reg abs r0) Reg32Map.sel_neq];
    auto.
Qed.

Lemma conc_regs_pc : forall abs st ctx pc pc',
  conc_regs abs st ctx
  -> conc_regs (aset_pc abs pc) (set_pc st pc') ctx.
  intuition.
Qed.

Lemma conc_regs_stack : forall abs st ctx r t v a v' n t',
  conc_regs (aset_reg abs r t) (set_reg32 st r v) ctx
  -> conc_regs (aset_reg (aset_stack abs n t') r t)
  (set_reg32 (set_mem32 st a v') r v) ctx.
  intuition.
Qed.

Lemma reg_hasType : forall abs st ctx r,
  conc_regs abs st ctx
  -> hasType ctx (eval_genop32 (Reg r) st) (aget_reg abs r).
  intros abs st ctx r Hconc.
  simpl.
  apply Hconc.
Qed.

Lemma reg_hasType' : forall abs st ctx r,
  conc' abs st ctx
  -> hasType ctx (get_reg32 st r) (aget_reg abs r).
  intros abs st ctx r Hconc.
  destruct Hconc.
  simpl.
  apply concRegs.
Qed.

Hint Resolve conc_regs_upd conc_regs_pc conc_regs_stack reg_hasType reg_hasType' : asm.

(** ** [conc_slots] *)

Lemma conc_slots_upd : forall abs st ctx n v t,
  conc' abs st ctx
  -> hasType ctx v t
  -> 4 <= n < aStackSize abs
  -> conc_slots (aset_stack abs n t)
  (set_mem32 st (stack_slot ctx n) v) ctx.
  intros abs st ctx n v t Hconc Hty Hlt n0 Hbound.
  destruct (nat_apart_dec n n0);
    [rewrite invert_set_mem32_neq; myeauto
      | idtac].
  simpl in Hbound; inversion Hconc; apply slots_apart; myauto; omega.

  (* Reading the slot we just set *)
  assert (n0 = n-3 \/ n0 = n-2 \/ n0 = n-1 \/ n0 = n \/ n0 = n+1 \/ n0 = n+2 \/ n0 = n+3);
    [unfold nat_apart in n1; omega | idtac].
  unfold aset_stack, aget_stack; simpl.

  Hint Rewrite NatMap.sel_eq using omega : conc_slots_upd.
  Hint Rewrite NatMap.sel_neq using omega : conc_slots_upd.
  Hint Rewrite invert_set_mem32 : conc_slots_upd.

  Ltac doit :=
    autorewrite with conc_slots_upd; try apply HasType_Top; trivial;
    match goal with
      [ |- context[NatMap.sel (NatMap.upd _ ?N1 _) ?N2] ] =>
      let He := fresh "compare" in
	(destruct (eq_nat_dec N1 N2) as [He | He];
	  try rewrite <- He; autorewrite with conc_slots_upd;
	    try apply HasType_Top; trivial)
    end.

  intuition; subst; repeat doit.
Qed.

Lemma conc_slots_pc : forall abs st ctx pc pc',
  conc_slots abs st ctx
  -> conc_slots (aset_pc abs pc) (set_pc st pc') ctx.
  intuition.
Qed.

Lemma conc_slots_reg : forall abs st ctx r r' t v,
  conc_slots abs st ctx
  -> conc_slots (aset_reg abs r t) (set_reg32 st r' v) ctx.
  intuition.
Qed.

Hint Resolve conc_slots_upd conc_slots_pc conc_slots_reg.

(** ** [agree_range] *)

Lemma agree_range_write_above : forall st start len addr v,
  agree_range (stMem st) initialMem start len
  -> belowMax addr 3
  -> bvec_to_nat addr > bvec_to_nat start + len
  -> agree_range (stMem (set_mem32 st addr v)) initialMem start len.
  intros st start len addr v Hrange Hupper Hgt addr' Hrange'.
  simpl.
  repeat (rewrite Int32Map.sel_neq); auto.
  assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
  assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
    [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
    [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
    [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
Qed.

Lemma agree_range_write_below : forall st start len addr v,
  agree_range (stMem st) initialMem start len
  -> belowMax addr 3
  -> bvec_to_nat addr + 3 < bvec_to_nat start
  -> agree_range (stMem (set_mem32 st addr v)) initialMem start len.
  intros st start len addr v Hrange Hupper Hgt addr' Hrange'.
  simpl.
  repeat (rewrite Int32Map.sel_neq); auto.
  assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
  assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
    [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
    [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
    [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
Qed.

Lemma agree_range_write'_above : forall st memIn start finish addr v,
  agree_range' (stMem st) memIn start finish
  -> belowMax addr 3
  -> bvec_to_nat addr > bvec_to_nat finish
  -> agree_range' (stMem (set_mem32 st addr v)) memIn start finish.
  intros st memIn start len addr v Hrange Hupper Hgt addr' Hrange'.
  simpl.
  repeat (rewrite Int32Map.sel_neq); auto.
  assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
  assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
    [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
    [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
    [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
Qed.

Lemma agree_range_write'_below : forall st memIn start finish addr v,
  agree_range' (stMem st) memIn start finish
  -> belowMax addr 3
  -> bvec_to_nat addr + 3 < bvec_to_nat start
  -> agree_range' (stMem (set_mem32 st addr v)) memIn start finish.
  intros st memIn start len addr v Hrange Hupper Hgt addr' Hrange'.
  simpl.
  repeat (rewrite Int32Map.sel_neq); auto.
  assert (bvec_to_nat addr <> bvec_to_nat addr'); [omega | congruence].
  assert (bvec_to_nat (binc addr) <> bvec_to_nat addr');
    [rewrite binc_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc2 addr) <> bvec_to_nat addr');
    [rewrite binc2_btn_belowMax; [omega | myeauto] | congruence].
  assert (bvec_to_nat (binc3 addr) <> bvec_to_nat addr');
    [rewrite binc3_btn_belowMax; [omega | myeauto] | congruence].
Qed.

Lemma agree_range_set_pc : forall pc st m1 m2 m3,
  agree_range (stMem st) m1 m2 m3
  -> agree_range (stMem (set_pc st pc)) m1 m2 m3.
  intuition.
Qed.

Lemma agree_range_set_reg : forall r v st m1 m2 m3,
  agree_range (stMem st) m1 m2 m3
  -> agree_range (stMem (set_reg32 st r v)) m1 m2 m3.
  intuition.
Qed.

Hint Resolve agree_range_set_pc agree_range_set_reg : asm.
