(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Basic definitions *)

Require Import List Bvector.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Map.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Env.

Require Import ProofOS.AbsInt.Bounded.

Import Datatypes.

(** * Memory regions *)

Parameter stackStart : int32.
(** Lowest address of stack region *)

Parameter stackSize : nat.
(** Guaranteed among of stack space left above [stackStart] *)

Axiom stack_fits : bvec_to_nat stackStart + stackSize + 9 < pow2 32 - 4.

Definition stackEnd := nat_to_bvec 32 (bvec_to_nat stackStart + stackSize + 9).
(** The very last memory cell considered to be in the stack *)

Definition initialESP := nat_to_bvec 32 (bvec_to_nat stackStart + stackSize + 4).
(** Starting stack pointer *)

Parameter codeLen : nat.
(** Amount of space reserved for code *)

Axiom code_high : bvec_to_nat initialPc > 3.
(** The code segment starts high enough to avoid underflow. *)

Axiom code_space : bvec_to_nat initialPc + (codeLen + longestInstr) + 3 < bvec_to_nat stackStart.
(** The stack starts sufficiently after the code. *)

Axiom initialESP_location : bvec_to_nat stackStart + stackSize < bvec_to_nat initialESP.
(** The stack pointer starts sufficiently after the start of the stack. *)

Axiom initialESP_upper : bvec_to_nat initialESP < pow2 32 - 4.
(** The stack pointer starts sufficiently low to avoid overflow. *)

(** * Types *)

Inductive ty : Set :=
  | Top : ty
  | Const : int32 -> ty
  | Stack : nat -> ty
  | RetPtr : ty
  | OldEbp : ty.

(** * Setting up some maps to use in the abstract state type *)

Module NatFinite.
  Definition t := nat.
  Definition to_nat (n : nat) := n.
  Theorem inj : forall t1 t2, t1 = t2 <-> to_nat t1 = to_nat t2.
    intuition.
  Qed.
End NatFinite.
Module NatOrd := FiniteOrderedSet(NatFinite).
Module NatMap := TotalMap(NatOrd).

Definition aregs := Reg32Map.map ty.
Definition astack := NatMap.map ty.

(** * Abstract states *)

Inductive apc : Set :=
  | apcConst : int32 -> apc
  | apcRet : apc.

Record astate : Set :=
  {aPc : apc;            (** Current program counter *)
    aRegs : aregs;       (** Types of registers *)
    aStackSize : nat;    (** Lower bound on stack space left below initial
			     ESP *)
    aStackSlots : astack (** Types of stack slots *)
  }.

(** ** Accessor and mutator functions *)

Definition aset_pc abs pc :=
  Build_astate (apcConst pc) (aRegs abs) (aStackSize abs) (aStackSlots abs).

Definition aget_reg abs r := Reg32Map.sel (aRegs abs) r.
Definition aset_reg abs r ty :=
  Build_astate (aPc abs) (Reg32Map.upd (aRegs abs) r ty)
  (aStackSize abs) (aStackSlots abs).

Definition aget_stack abs n := NatMap.sel (aStackSlots abs) n.
Definition aset_stack abs n ty :=
  Build_astate (aPc abs) (aRegs abs) (aStackSize abs)
  (NatMap.upd
    (NatMap.upd
      (NatMap.upd
	(NatMap.upd
	  (NatMap.upd
	    (NatMap.upd
	      (NatMap.upd (aStackSlots abs) (n+3) Top)
	      (n+2) Top)
	    (n+1) Top)
	  (n-3) Top)
	(n-2) Top)
      (n-1) Top)
    n ty).
(** All this extra cruft is here to make sure we invalidate any overlapping
  stack slots. *)

(** * Characterizing valid program counters *)

Parameter actualCodeLen : nat.
(** The real length of the code segment *)

Axiom actual_code_bound : actualCodeLen < codeLen.

Theorem actual_code_space : bvec_to_nat initialPc + (actualCodeLen + longestInstr) + 3
  < bvec_to_nat stackStart.
  generalize code_space.
  generalize actual_code_bound.
  omega.
Qed.

Definition pcInRange (pc : int32) := bvec_to_nat initialPc <= bvec_to_nat pc
  <= bvec_to_nat initialPc + actualCodeLen.

(** * Typing judgment *)

(** ** Typing contexts *)
Record context : Set := {
  sptr : int32;      (** Function initial ESP *)
  oldesp : nat;      (** ESP value expected by the caller on return (as a stack slot number) *)
  oldebp : int32;    (** EBP value expected by the caller on return *)
  memIn : mem;       (** Memory at the start of this function call *)
  ret : option int32 (** Return pointer *)
  }.

Definition cset_sptr ctx sp := Build_context sp (oldesp ctx) (oldebp ctx) (memIn ctx) (ret ctx).
Definition cset_oldesp ctx sp := Build_context (sptr ctx) sp (oldebp ctx) (memIn ctx) (ret ctx).
Definition cset_oldebp ctx sp := Build_context (sptr ctx) (oldesp ctx) sp (memIn ctx) (ret ctx).
Definition cset_memIn ctx mi := Build_context (sptr ctx) (oldesp ctx) (oldebp ctx) mi (ret ctx).
Definition cset_ret ctx rp := Build_context (sptr ctx) (oldesp ctx) (oldebp ctx) (memIn ctx) rp.

(** ** Stack slot shorthand *)

Definition stack_slot ctx n :=
  nat_to_bvec 32 (bvec_to_nat (sptr ctx) - n).

(** ** The typing judgment *)

Definition cstep := cstep exec get_instr.

Inductive hasType : context -> int32 -> ty -> Prop :=
  | HasType_Top : forall ctx e,
    hasType ctx e Top
  | HasType_Const : forall ctx i,
    hasType ctx i (Const i)
  | HasType_Stack : forall ctx n,
    n <= bvec_to_nat (sptr ctx)
    -> hasType ctx (stack_slot ctx n) (Stack n)
  | HasType_RetPtr : forall ctx pc,
    ret ctx = Some pc
    -> pcInRange pc
    -> hasType ctx pc RetPtr
  | HasType_OldEbp : forall ctx,
    hasType ctx (oldebp ctx) OldEbp.

(** * The soundness relation *)

Definition conc_pc abs st ctx :=
  match aPc abs with
    | apcConst pc => stPc st = pc
    | apcRet => Some (stPc st) = ret ctx
  end.

Definition conc_regs abs st ctx :=
  forall r, hasType ctx (get_reg32 st r) (aget_reg abs r).

Definition conc_slots abs st ctx :=
  forall n, 4 <= n < aStackSize abs
    -> hasType ctx (get_mem32 st (stack_slot ctx n)) (aget_stack abs n).

Definition agree_range' (mem1 mem2 : mem) (addr1 addr2 : int32) :=
  forall addr', bvec_to_nat addr1 <= bvec_to_nat addr' < bvec_to_nat addr2
    -> Int32Map.sel mem1 addr' = Int32Map.sel mem2 addr'.

Record conc' (abs : astate) (st : state) (ctx : context) : Prop :=
  {concPc : conc_pc abs st ctx;
    concPcRange : pcInRange (stPc st);
    concCodeMem : agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr);
    concRegs : conc_regs abs st ctx;
    concStackSize : bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx);
    concStackSlots : conc_slots abs st ctx;
    concSptr : bvec_to_nat (sptr ctx) <= bvec_to_nat initialESP;
    concMemIn : agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd}.

Definition conc abs st := exists ctx : context, conc' abs st ctx.
