(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Choosing the set of root states *)

Require Import Arith List.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Env.

Require Import ProofOS.Kernel.Defs.
Require Import ProofOS.Kernel.Metadata.

Import Datatypes.

Definition initial_cp := Build_cutpoint_info (apcConst initialPc) stackSize nil nil.

Definition initial := (@nil cutpoint_info, initial_cp).
(** A [cutpoint] for the initial state *)

Definition initials := cons initial cutpoints.
(** All root states, represented as [cutpoint]s *)

Fixpoint make_init_regs (rs : list (reg32 * ty)) : aregs :=
  match rs with
    | nil => Reg32Map.empty Top
    | cons (r, t) rs => Reg32Map.upd (make_init_regs rs) r t
  end.

Fixpoint make_init_slots (sls : list (nat * ty)) : astack :=
  match sls with
    | nil => NatMap.empty Top
    | cons (sl, t) sls => NatMap.upd (make_init_slots sls) sl t
  end.

Definition make_init_state cpt :=
  Build_astate (ciPc cpt)
  (make_init_regs (ciRegs cpt))
  (ciStackSize cpt)
  (make_init_slots (ciSlots cpt)).

Lemma init_pcInRange : forall st,
  stPc st = initialPc
  -> pcInRange (stPc st).
  intros; red; rewrite H; omega.
Qed.

Lemma init_agree_range : forall st,
  stMem st = initialMem
  -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr).
  red; intros; congruence.
Qed.

Hint Constructors hasType : asm.

Definition init_ctx := Build_context initialESP 0 initialESP initialMem None.

Lemma init_conc_regs : forall st,
  initState st
  -> conc_regs (make_init_state initial_cp) st init_ctx.
  unfold init_ctx; intros st Hinit; intro.
  unfold aget_reg; simpl.
  rewrite Reg32Map.sel_empty.
  myauto.
Qed.

Lemma init_stack :
  bvec_to_nat stackStart + aStackSize (make_init_state initial_cp) <
  bvec_to_nat (sptr init_ctx).
  unfold init_ctx; simpl aStackSize; simpl sptr; apply initialESP_location.
Qed.

Lemma init_conc_slots : forall st,
  conc_slots (make_init_state initial_cp) st init_ctx.
  unfold init_ctx; intros; intro.
  unfold aget_stack; simpl.
  rewrite NatMap.sel_empty.
  myauto.
Qed.

Lemma init_sptr : bvec_to_nat (sptr init_ctx) < pow2 32 - 4.
  unfold init_ctx; unfold sptr; apply initialESP_upper.
Qed.

Lemma init_memIn : forall st, stMem st = initialMem
  -> agree_range' (stMem st) (memIn init_ctx) (sptr init_ctx) stackEnd.
  unfold init_ctx; intros; red; intros.
  simpl.
  congruence.
Qed.

Hint Resolve init_pcInRange init_agree_range init_conc_regs
  init_stack init_conc_slots init_sptr init_memIn : asm.

Definition make_init_hyps_state cpt :=
  (map make_init_state (fst cpt), make_init_state (snd cpt)).

Definition init : {abss : list (list astate * astate) |
  exists ctx, forall st, initState st ->
    exists abs, In (nil, abs) abss /\ conc' abs st (ctx st)}.
  exists (map make_init_hyps_state initials).
  exists (fun (_ : state) => init_ctx).
  intros st Hinit.
  inversion Hinit.
  exists (make_init_state initial_cp).
  simpl; intuition.
  split; myauto.
Defined.

