(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Analyzing assembly operands *)

Require Import Arith Bool Bvector.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Env.

Require Import ProofOS.Kernel.Defs.
Require Import ProofOS.Kernel.Conc.

Hint Constructors hasType : asm.

Set Implicit Arguments.

(** * Reading *)

Lemma double_sub : forall n v1 v2,
  bvec_to_nat v1 <= n
  -> bvec_to_nat (bsub32 v1 v2) < n
  -> eqMod (pow2 32)
  (n - bvec_to_nat (bsub32 v1 v2))
  (bvec_to_nat v2 + n - bvec_to_nat v1).
  intros n v1 v2 Hle Hlt.
  apply eqMod_trans
    with (n + bvec_to_nat (bneg (bsub32 v1 v2))).
  apply bneg_btn_sub; trivial.
  apply eqMod_trans
    with (n + bvec_to_nat (bsub32 v2 v1)).
  apply eqMod_plus.
  apply eqMod_refl.
  unfold bsub32; apply bsub_btn_neg.
  apply eqMod_trans
    with (n + (pow2 32 + bvec_to_nat v2 - bvec_to_nat v1)).
  apply eqMod_plus.
  apply eqMod_refl.
  unfold bsub32; apply bsub_btn.
  add Hup (btn_upper v1).
  replace (n + (pow2 32 + bvec_to_nat v2 - bvec_to_nat v1))
    with (pow2 32 + (bvec_to_nat v2 + n - bvec_to_nat v1)).
  apply eqMod_sym; apply eqMod_mod.
  apply eqMod_refl.
  omega.
Qed.

Import Datatypes.

Lemma slot_addr : forall d r st ss ctx abs,
  conc_regs abs st ctx
  -> aget_reg abs r = Stack ss
  -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) d) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> eval_addr (Build_address d (Some r) None) st
  = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) d)).
  intros d r st ss ctx abs Hconc Hr Hlt Hsptr.
  add Hrconc (Hconc r).
  rewrite Hr in Hrconc.
  inversion Hrconc as [| |Ha Hb Hle Hd Heq Hf| |]; subst.
  unfold eval_addr, addrIndex, addrBase, addrDisp, stack_slot, badd32, badd, barith.
  apply (ntb_mod 32).
  rewrite <- Heq.
  unfold stack_slot.
  rewrite ntb_invert; try (add Hup (btn_upper (sptr ctx)); omega).
  rewrite plus_minus; trivial.
  replace (bvec_to_nat d + bvec_to_nat (sptr ctx) - ss)
    with (bvec_to_nat d + bvec_to_nat (sptr ctx) - bvec_to_nat (nat_to_bvec 32 ss)).
  apply eqMod_sym; apply double_sub.
  rewrite ntb_invert; add Hup (btn_upper (sptr ctx)); omega.
  omega.
  rewrite ntb_invert; add Hup (btn_upper (sptr ctx)); omega.
Qed.

Ltac type_fail := intros; apply SNone.

Definition typeof_genop32 : forall abs gop,
  sig_option (fun ty =>
    forall st ctx,
      conc' abs st ctx
      -> hasType ctx (eval_genop32 gop st) ty).
  intros abs gop.
  destruct gop.

  (* Imm *)
  apply SSome with (Const i); myeauto.

  (* Reg *)
  apply SSome with (aget_reg abs r); intros st ctx Hconc; simpl; destruct Hconc; auto.

  (* Address *)
  destruct a.
  destruct addrBase; [idtac | apply SNone].
  destruct addrIndex; [apply SNone | idtac].
  caseEq (aget_reg abs r); [type_fail | type_fail | idtac | type_fail | type_fail].
  intros ss Hr.
  destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
    [idtac | apply SNone].
  destruct (le_lt_dec (aStackSize abs) (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
    [apply SNone | idtac].
  apply SSome with (aget_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
  intros st ctx Hconc.
  inversion Hconc.
  unfold eval_genop32.
  assert (eval_addr (Build_address addrDisp (Some r) None) st
    = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
  eapply slot_addr with abs; eauto.
  rewrite H.
  myauto.
Qed.


(** * Writing *)

(** ** Inversion lemmas, needed to be able to use inversion on [write_genop32]
     in sort [Set] *)

Lemma invert_write_reg : forall r n st st',
  write_genop32 (Reg r) n st st'
  -> st' = set_reg32 st r n.
  inversion 1; trivial.
Qed.

Lemma invert_write_mem : forall a n st st',
  write_genop32 (Address reg32 a) n st st'
  -> st' = set_mem32 st (eval_addr a st) n.
  inversion 1; trivial.
Qed.

Lemma invert_write_imm : forall i n st st',
  write_genop32 (Imm reg32 i) n st st'
  -> False.
  inversion 1.
Qed.

(** ** Writing to the stack *)

Lemma awrite_stack_high : forall abs st ctx r ss addrDisp,
  aget_reg abs r = Stack ss
  -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
  -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 3
  -> bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st)
  >= bvec_to_nat stackStart.
  intros abs st ctx r ss addrDisp Hr Hss Hagree Hconc Hsptr Hbelow.
  assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
    = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
  eapply slot_addr; eauto.
  rewrite Hslot.
  unfold stack_slot.
  add Hup (btn_upper (sptr ctx)).
  rewrite ntb_invert; omega.
Qed.

Lemma btn_conv : forall len (v1 v2 : Bvector len),
  bvec_to_nat v1 <> bvec_to_nat v2
  -> v1 <> v2.
  intros; congruence.
Qed.

Lemma stack_slot_belowMax : forall abs st ctx r ss addrDisp,
  aget_reg abs r = Stack ss
  -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 3
  -> belowMax (eval_addr (Build_address addrDisp (Some r) None) st) 3.
  intros abs st ctx r ss addrDisp Hr Hsize Hconc Hsptr Hbelow.
  assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
    = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
  eapply slot_addr; eauto.
  rewrite Hslot.
  unfold stack_slot.
  split.
  inversion Hbelow.
  rewrite ntb_invert; omega.
Qed.

Hint Resolve stack_slot_belowMax : asm.

Lemma awrite_stack_agree_range : forall abs st ctx r ss addrDisp v,
  aget_reg abs r = Stack ss
  -> bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
  -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 3
  -> agree_range
  (stMem
    (set_mem32 st (eval_addr (Build_address addrDisp (Some r) None) st) v))
  initialMem initialPc (actualCodeLen + longestInstr).
  intros abs st ctx r ss addrDisp v Hr Hss Hagree Hconc Hsptr Hbelow addr Hrange.
  unfold set_mem32, set_mem, stMem.
  assert (Hhigh : bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st) >=
    bvec_to_nat stackStart).
  eapply awrite_stack_high; eauto.
  add Hspace actual_code_space.
  repeat (rewrite Int32Map.sel_neq).
  auto.
  apply (@btn_conv 32); omega.
  apply (@btn_conv 32); rewrite binc_btn_belowMax; [omega | myeauto].
  apply (@btn_conv 32); rewrite binc2_btn_belowMax; [omega | myeauto].
  apply (@btn_conv 32); rewrite binc3_btn_belowMax; [omega | myeauto].
Qed.

Hint Resolve awrite_stack_agree_range : asm.

Lemma ignore_four : forall k n m,
  k <= n < m
  -> n < m.
  intros; omega.
Qed.

Hint Resolve ignore_four : asm.

Lemma stackSlot_below_sptr : forall r ss addrDisp abs ctx st,
  aget_reg abs r = Stack ss
  -> 4 <= bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> conc_regs abs st ctx
  -> bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st) + 3
  < bvec_to_nat (sptr ctx).
  intros r ss addrDisp abs ctx st Hr Hss Hsptr Hconc.
  assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
    = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
  eapply slot_addr; eauto.
  omega.
  rewrite Hslot.
  unfold stack_slot.
  rewrite ntb_invert.
  omega.
  add Hup (btn_upper (sptr ctx)); omega.
Qed.

Lemma awrite_stack_memIn : forall abs st ctx r ss addrDisp v,
  aget_reg abs r = Stack ss
  -> 4 <= bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
  -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 4
  -> agree_range'
  (stMem
    (set_mem32 st (eval_addr (Build_address addrDisp (Some r) None) st) v))
  (memIn ctx) (sptr ctx) stackEnd.
  intros abs st ctx r ss addrDisp v Hr Hss Hagree Hconc Hsptr Hbelow.
  assert (Hlow : bvec_to_nat (eval_addr (Build_address addrDisp (Some r) None) st) + 3
    < bvec_to_nat (sptr ctx)).
  eapply stackSlot_below_sptr; eauto.
  apply agree_range_write'_below; myeauto.
Qed.

Hint Resolve awrite_stack_memIn : asm.

Lemma awrite_stack_conc_slots : forall abs st ctx v t r ss addrDisp,
  conc' abs st ctx
  -> hasType ctx v t
  -> aget_reg abs r = Stack ss
  -> 4 <= bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp) < aStackSize abs
  -> conc_regs abs st ctx
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> conc_slots abs st ctx
  -> conc_slots
  (aset_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)) t)
  (set_mem32 st (eval_addr (Build_address addrDisp (Some r) None) st) v)
  ctx.
  intros abs st ctx v t r ss addrDisp Hconc Ht Hr Hsize Hregs Hsptr Hslots.
  assert (Hslot : eval_addr (Build_address addrDisp (Some r) None) st
    = stack_slot ctx (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp))).
  eapply slot_addr; eauto.
  omega.
  rewrite Hslot.
  apply conc_slots_upd; trivial.
Qed.

Hint Resolve awrite_stack_conc_slots : asm.


(** ** Concrete write function *)

Definition writable_genop32 (gop : genop32) :=
  match gop with
    | Imm _ => false
    | _ => true
  end.

Definition write_concrete : forall (gop : genop32) (st : state) (v : int32)
  (writable : writable_genop32 gop = true), state.
  intros.
  destruct gop.

  discriminate.
  exact (set_reg32 st r v).
  exact (set_mem32 st (eval_addr a st) v).
Defined.

Lemma belowMax_stack_slot : forall ctx n,
  belowMax (sptr ctx) 3
  -> belowMax (stack_slot ctx n) 3.
  intros.
  inversion H.
  split.
  unfold stack_slot.
  rewrite ntb_invert; omega.
Qed.

Hint Resolve belowMax_stack_slot : asm.

Lemma heap_stack_apart : forall ctx abs d n,
  bvec_to_nat initialPc + (actualCodeLen + longestInstr) < bvec_to_nat d
  -> bvec_to_nat d + 3 < bvec_to_nat stackStart
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> 4 <= n < aStackSize abs
  -> apart d (stack_slot ctx n).
  intros ctx abs d n Hlt1 Hlt2 Hsptr Hn.
  apply apart_higher.
  unfold stack_slot.
  rewrite ntb_invert.
  omega.
  add Hup (btn_upper (sptr ctx)); omega.
Qed.

Hint Resolve heap_stack_apart : asm.

Lemma agree_range_heapWrite : forall d st abs ctx v,
  bvec_to_nat initialPc + (actualCodeLen + longestInstr) < bvec_to_nat d
  -> bvec_to_nat d + 3 < bvec_to_nat stackStart
  -> agree_range (stMem st) initialMem initialPc (actualCodeLen + longestInstr)
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 3
  -> belowMax d 3
  -> agree_range (stMem (set_mem32 st d v)) initialMem initialPc
  (actualCodeLen + longestInstr).
  intros.
  apply agree_range_write_above; trivial.
Qed.

Hint Resolve agree_range_heapWrite : asm.

Lemma conc_slots_heapWrite : forall d abs ctx st v,
  bvec_to_nat initialPc + (actualCodeLen + longestInstr) < bvec_to_nat d
  -> bvec_to_nat d + 3 < bvec_to_nat stackStart
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> conc_slots abs st ctx
  -> belowMax (sptr ctx) 3
  -> belowMax d 3
  -> conc_slots abs (set_mem32 st d v) ctx.
  intros; intros n Hn.
  rewrite invert_set_mem32_neq; myeauto.
  red; myeauto.
Qed.

Hint Resolve conc_slots_heapWrite : asm.

Lemma memIn_heapWrite : forall d st abs ctx v,
  bvec_to_nat d + 3 < bvec_to_nat stackStart
  -> agree_range' (stMem st) (memIn ctx) (sptr ctx) stackEnd
  -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (sptr ctx)
  -> belowMax (sptr ctx) 4
  -> belowMax d 3
  -> agree_range' (stMem (set_mem32 st d v)) (memIn ctx) (sptr ctx) stackEnd.
  intros.
  apply agree_range_write'_below; trivial.
  omega.
Qed.

Hint Resolve memIn_heapWrite : asm.

Lemma write_heap' : forall abs st ctx v d,
  conc' abs st ctx
  -> bvec_to_nat initialPc + (actualCodeLen + longestInstr) < bvec_to_nat d
  -> bvec_to_nat d + 3 < bvec_to_nat stackStart
  -> conc' abs (set_mem32 st (eval_addr (Build_address d None None) st) v) ctx.
  intros abs st ctx v d Hconc Hle Hlt.
  unfold eval_addr, addrIndex, addrBase, addrDisp.
  inversion Hconc.
  assert (Hbelow : belowMax d 3).
  add Hactual actual_code_space.
  split; assert (Hbel : belowMax (sptr ctx) 3);
    [myeauto | inversion Hbel; omega].
  split; try (intros n Hn; rewrite invert_set_mem32_neq; myeauto); myeauto.
Qed.

Lemma write_heap : forall abs st ctx v d st',
  conc' abs st ctx  
  -> bvec_to_nat initialPc + (actualCodeLen + longestInstr) < bvec_to_nat d
  -> bvec_to_nat d + 3 < bvec_to_nat stackStart
  -> write_genop32 (Address reg32 (Build_address d None None)) v st st'
  -> conc' abs st' ctx.
  intros abs st ctx v d st' Hconc Hle Hlt Hwrite.
  inversion Hwrite.
  apply write_heap'; trivial.
Qed.

Hint Resolve write_heap : asm.

(** ** The main write function *)

Definition awrite_genop32 : forall abs gop ty,
  sig_option (fun abs' =>
    forall st st' ctx v,
      conc' abs st ctx
      -> hasType ctx v ty
      -> write_genop32 gop v st st'
      -> conc' abs' st' ctx).
  intros abs gop ty.
  destruct gop.

  (* Imm *)
  apply SNone.

  (* Reg *)
  apply SSome with (aset_reg abs r ty).
  intros st st' ctx v Hconc Ht Hwrite.
  add Hst' (invert_write_reg Hwrite); subst; clear Hwrite.
  inversion Hconc; split; intuition.

  (* Address *)
  destruct a.
  destruct addrBase.

  (* Base register *)
  destruct addrIndex; [apply SNone | idtac].
  caseEq (aget_reg abs r); [type_fail | type_fail | idtac | type_fail | type_fail].
  intros ss Hr.
  destruct (le_lt_dec 4 (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
    [idtac | apply SNone].
  destruct (le_lt_dec (aStackSize abs) (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)));
    [apply SNone | idtac].
  apply SSome with (aset_stack abs (bvec_to_nat (bsub32 (nat_to_bvec 32 ss) addrDisp)) ty).
  intros st st' ctx v Hconc Ht Hwrite.
  add Hst' (invert_write_mem Hwrite); subst; clear Hwrite.
  inversion Hconc; split; myeauto.

  (* No base register *)
  destruct addrIndex; [apply SNone | idtac].
  destruct (le_lt_dec (bvec_to_nat addrDisp) (bvec_to_nat initialPc + (actualCodeLen + longestInstr)));
    [apply SNone | idtac].
  destruct (le_lt_dec (bvec_to_nat stackStart) (bvec_to_nat addrDisp + 3));
    [apply SNone | idtac].
  apply SSome with abs.
  myeauto.
Qed.
