(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Analyzing functional assembly operations *)

Require Import Arith Bool Bvector.

Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Kernel.Defs.
Require Import ProofOS.Kernel.Conc.

Set Implicit Arguments.

Hint Constructors hasType : asm.

Lemma stack_slot_add : forall i n ctx,
  bvec_to_nat i <= n
  -> n <= bvec_to_nat (sptr ctx)
  -> badd32 i (stack_slot ctx n) = stack_slot ctx (n - bvec_to_nat i).
  intros i n ctx Hi Hsptr.
  simpl.
  apply btn_inj.
  unfold stack_slot, badd32, badd, barith.
  add Hup (btn_upper (sptr ctx)).
  repeat (rewrite ntb_invert); omega.
Qed.

Lemma stack_slot_sub : forall i n ctx,
  n + bvec_to_nat i <= bvec_to_nat (sptr ctx)
  -> bsub32 (stack_slot ctx n) i = stack_slot ctx (n + bvec_to_nat i).
  intros i n ctx Hsptr.
  unfold bsub32, bsub, barith, stack_slot.
  apply ntb_mod.
  add Hup (btn_upper (sptr ctx)).
  rewrite ntb_invert; try omega.
  apply eqMod_trans
    with (bvec_to_nat (sptr ctx) - n + (pow2 32 - bvec_to_nat i)).
  apply eqMod_plus; [apply eqMod_refl | apply bneg_btn].
  replace (bvec_to_nat (sptr ctx) - n + (pow2 32 - bvec_to_nat i))
    with (pow2 32 + (bvec_to_nat (sptr ctx) - n - bvec_to_nat i)).
  apply eqMod_mod'.
  rewrite rewrite_minus_plus.
  apply eqMod_refl.
  omega.
Qed.

Lemma badd_comm : forall len (v1 v2 : Bvector len),
  badd v1 v2 = badd v2 v1.
  intros len v1 v2.
  unfold badd, barith.
  rewrite plus_comm.
  trivial.
Qed.

Ltac give_up := exists Top; intuition.

Lemma badd32_swap_neg : forall v1 v2,
  badd32 v2 v1 = bsub32 v1 (bneg v2).
  intros.
  unfold bsub32, badd32, bsub, badd, barith.
  rewrite bneg_btn_double.
  rewrite plus_comm.
  trivial.
Qed.

Definition aeval_arith : forall abs aop t1 t2,
  {t : ty |
    forall st ctx v1 v2,
      conc' abs st ctx 
      -> hasType ctx v1 t1
      -> hasType ctx v2 t2
      -> hasType ctx (eval_arith aop v1 v2) t}.
  intros abs aop t1 t2.
  destruct t1.

  (* Top vs. anything *)
  give_up.

  destruct t2; [give_up | idtac | idtac | give_up | give_up].

  (* Const vs. Const *)
  exists (Const (eval_arith aop i i0)).
  intros st ctx v1 v2 Hconc Hv1 Hv2.
  inversion Hv1; subst.
  inversion Hv2; subst.
  intuition.

  (* Const vs. Stack *)
  destruct aop; [idtac | give_up | give_up | give_up].

  (* Addition *)
  destruct (le_lt_dec (bvec_to_nat i) n).

  (* Add positive *)
  exists (Stack (n - bvec_to_nat i)).
  intros st ctx v1 v2 Hconc Hv1 Hv2.
  inversion Hv1; subst.
  inversion Hv2; subst.
  simpl.
  rewrite stack_slot_add; intuition.

  clear l.
  destruct (le_lt_dec (n + bvec_to_nat (bneg i)) (aStackSize abs)); [idtac | give_up].

  (* Add negative *)
  exists (Stack (n + bvec_to_nat (bneg i))).
  intros st ctx v1 v2 Hconc Hv1 Hv2.
  inversion Hv1; subst.
  inversion Hv2; subst.
  simpl.
  assert (Hsptr : aStackSize abs <= bvec_to_nat (sptr ctx)).
  inversion Hconc; omega.
  rewrite badd32_swap_neg.
  rewrite stack_slot_sub; intuition.
  
  destruct t2; [give_up | idtac | idtac | give_up | give_up].

  (* Stack vs. Const *)
  destruct aop; [idtac | give_up | give_up | idtac].

  (* Addition *)
  destruct (le_lt_dec (bvec_to_nat i) n); [idtac | give_up].
  exists (Stack (n - bvec_to_nat i)).
  intros st ctx v1 v2 Hconc Hv1 Hv2.
  inversion Hv1; subst.
  inversion Hv2; subst.
  simpl.
  unfold badd32; rewrite badd_comm.
  rewrite stack_slot_add; intuition.

  (* Subtraction *)
  destruct (le_lt_dec (n + bvec_to_nat i) (aStackSize abs)); [idtac | give_up].
  exists (Stack (n + bvec_to_nat i)).
  intros st ctx v1 v2 Hconc Hv1 Hv2.
  inversion Hv1; subst.
  inversion Hv2; subst.
  simpl.
  assert (Hsptr : aStackSize abs <= bvec_to_nat (sptr ctx));
    [inversion Hconc; omega
      | rewrite stack_slot_sub; intuition].

  (* Stack vs. Stack *)
  give_up.

  (* RetPtr vs. Anything *)
  give_up.

  (* OldEbp vs. Anything *)
  give_up.
Qed.

Definition aeval_addr : forall abs a,
  {t : ty |
    forall st ctx,
      conc' abs st ctx 
      -> hasType ctx (eval_addr a st) t}.
  intros abs a.

  destruct a.
  destruct addrBase.
  destruct addrIndex.

  (* disp + base + scale * idx *)
  give_up.

  (* disp + base *)
  destruct (aeval_arith abs Add (Const addrDisp) (aget_reg abs r)) as [t Ht].
  exists t.
  intuition.
  replace (eval_addr (Build_address addrDisp (Some r) None) st)
    with (eval_arith Add addrDisp (get_reg32 st r)).
  myeauto.
  simpl.
  unfold eval_addr, addrIndex, addrBase, Defs.addrDisp.
  trivial.

  destruct addrIndex.

  (* disp + scale * idx *)
  give_up.

  (* disp *)
  exists (Const addrDisp).
  myeauto.
Qed.
