(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Basic definitions *)

Require Import Arith List Bvector TheoryList.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Map.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.

Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Env.

Require Import ProofOS.AbsInt.Bounded.

Require Import ProofOS.Kernel.Defs.
Require Import ProofOS.Kernel.Metadata.
Require Import ProofOS.Kernel.Abs.
Require Import ProofOS.Kernel.Init.
Require Import ProofOS.Kernel.Conc.
Require Import ProofOS.Kernel.Misc.
Require Import ProofOS.Kernel.Call.
Require Import ProofOS.Kernel.Operands.
Require Import ProofOS.Kernel.Operations.

Set Implicit Arguments.

Import Datatypes.

Hint Constructors hasType exec write_genop32 : asm.

Definition isInitializingStack : forall (g1 g2 : genop32),
  sig_option (fun esp0 => g1 = Reg ESP /\ g2 = Imm reg32 esp0).
  intros.
  destruct g1; [apply SNone | idtac | apply SNone].
  destruct (reg32_eq_dec r ESP); [subst | apply SNone].
  destruct g2; [idtac | apply SNone | apply SNone].
  apply SSome with i.
  eauto.
Qed.

Lemma setESP_conc_regs : forall st nextPc mi,
  conc_regs
  (Build_astate (apcConst nextPc) (Reg32Map.upd (Reg32Map.empty Top) ESP (Stack 0))
    stackSize (NatMap.empty Top))
  (set_pc (set_reg32 st ESP (eval_genop32 (Imm reg32 initialESP) st))
    nextPc)
  (Build_context initialESP 0 initialESP mi None).
  intros st nextPc mi r.
  unfold get_reg32, aget_reg; simpl.
  destruct (reg32_eq_dec r ESP); subst.

  repeat (rewrite Reg32Map.sel_eq).
  assert (hasType (Build_context initialESP 0 initialESP mi None)
    (stack_slot (Build_context initialESP 0 initialESP mi None) 0)
    (Stack 0)).
  apply HasType_Stack.
  omega.
  replace (stack_slot (Build_context initialESP 0 initialESP mi None) 0) with initialESP in H.
  trivial.
  unfold stack_slot.
  rewrite <- minus_n_O.
  rewrite btn_invert.
  trivial.
  repeat (rewrite Reg32Map.sel_neq); auto with arith.
  repeat (rewrite Reg32Map.sel_empty).
  myauto.
Qed.

Hint Resolve setESP_conc_regs : asm.

Lemma setESP_conc_slots : forall nextPc rs ss st ctx,
  conc_slots
  (Build_astate nextPc rs ss (NatMap.empty Top))
  st ctx.
  intros; intros slot Hbound.
  unfold aget_stack; simpl.
  rewrite NatMap.sel_empty.
  myauto.
Qed.

Hint Resolve setESP_conc_slots : asm.

Lemma setESP_belowMax : belowMax initialESP 4.
  split; apply initialESP_upper.
Qed.

Hint Resolve setESP_belowMax : asm.

Lemma setESP_pc : forall nextPc r ss sl st ctx,
   Defs.conc_pc
     (Build_astate (apcConst nextPc) r ss sl)
     (set_pc st nextPc)
     ctx.
  intros.
  red; simpl.
  trivial.
Qed.

Hint Resolve setESP_pc : asm.

Lemma setESP_memIn : forall v1 v2 m1 m2,
  bvec_to_nat v1 <= bvec_to_nat v2
  -> agree_range' m1 m2 v1 stackEnd
  -> agree_range' m1 m2 v2 stackEnd.
  intros v1 v2 m1 m2 Hr Hm addr Haddr.
  apply Hm.
  omega.
Qed.
  
Hint Resolve setESP_memIn : asm.

Lemma setESP' : forall ctx st st' abs nextPc,
  pcInRange nextPc
  -> conc' abs st ctx
  -> write_genop32 (Reg ESP) (eval_genop32 (Imm reg32 initialESP) st) st st'
  -> exec (Mov (Reg ESP) (Imm reg32 initialESP), nextPc) st (set_pc st' nextPc)
  -> agree_range' (stMem st) (memIn ctx) initialESP stackEnd
  -> conc' (Build_astate (apcConst nextPc) (Reg32Map.upd (Reg32Map.empty Top) ESP (Stack 0))
    stackSize (NatMap.empty Top))
  (set_pc st' nextPc) (Build_context initialESP 0 initialESP (memIn ctx) None).
  intros ctx st st' abs nextPc Hpc Hconc Hwrite Hexec Hagree.
  inversion Hwrite; subst.
  inversion Hexec; subst.
  inversion Hconc.
  split; myeauto.
Qed.

Hint Resolve setESP_memIn : asm.

Lemma setESP : forall ctx st st' abs nextPc abs',
  pcInRange nextPc
  -> conc' abs st ctx
  -> write_genop32 (Reg ESP) (eval_genop32 (Imm reg32 initialESP) st) st st'
  -> exec (Mov (Reg ESP) (Imm reg32 initialESP), nextPc) st (set_pc st' nextPc)
  -> proj1_sig (subAbs
    (Build_astate (apcConst nextPc)
      (Reg32Map.upd (Reg32Map.empty Top) ESP (Stack 0)) stackSize
      (NatMap.empty Top)) abs') = true
  -> conc' abs'
  (set_pc st' nextPc) (Build_context initialESP 0 initialESP (memIn ctx) None).
  intros ctx st st' abs nextPc abs' Hpc Hconc Hwrite Hexec Hsub.
  destruct (subAbs
    (Build_astate (apcConst nextPc)
      (Reg32Map.upd (Reg32Map.empty Top) ESP (Stack 0)) stackSize
      (NatMap.empty Top)) abs').
  simpl in Hsub; subst; intuition.
  apply H.
  inversion Hconc.
  eapply setESP'; myeauto.
Qed.

Hint Resolve setESP : asm.

Lemma AllS_trivial : forall (T : Set) (f : T -> Prop) ls,
  (forall x, In x ls -> f x)
  -> AllS f ls.
  induction ls; intuition.
Qed.

Lemma AllS_imp : forall (T : Set) (f g : T -> Prop),
  (forall x, f x -> g x)
  -> forall ls, AllS f ls
    -> AllS g ls.
  induction ls; intuition.
  inversion H0; subst.
  apply allS_cons; auto.
Qed.

Lemma AllS_goodHyps : forall hyps' hyps ctx ls,
  goodHyps hyps hyps'
  -> AllS
  (fun need : astate =>
    exists have : astate,
      (In have ls \/ In have hyps) /\ menter conc' ctx need ctx have)
  hyps'.
  induction hyps'; intros hyps ctx ls Hhyps;
    destruct hyps; simpl in Hhyps; intuition.
  apply allS_cons.
  exists a0.
  intuition.
  destruct (subAbs a a0).
  simpl in H.
  unfold menter.
  auto.
  apply AllS_imp with
    (fun need : astate =>
      exists have : astate,
        (In have nil \/ In have hyps) /\
        menter conc' ctx need ctx have).
  simpl.
  intros x Hx.
  destruct Hx as [have Hx].
  exists have.
  intuition.
  auto.
Qed.

Hint Resolve AllS_goodHyps : asm.

Lemma menter_refl : forall ctx abs,
  menter conc' ctx abs ctx abs.
  unfold menter; intuition.
Qed.

Definition step : forall hyps abs insPc, {result : option (list astate)
    | match result with
	| None => True
	| Some abss =>
	  forall st ctx, conc' abs st ctx
	  -> (exists st' : state, exec insPc st st')
	  /\ forall st' (Hexec : exec insPc st st'),
	    exists abs', ((In abs' abss \/ In abs' hyps) /\ conc' abs' st' ctx)
	      \/ exists ctx', exists hyps',
		In (hyps', abs') (proj1_sig init)
		/\ AllS (fun need => exists have,
		  (In have abss \/ In have hyps) /\ menter conc' ctx' need ctx have) hyps'
		/\ conc' abs' st' ctx'
      end}.
  intros hyps abs insPc.
  destruct insPc as [ins nextPc].

  Ltac failure := exists (@None (list astate)); trivial.
  Ltac assertOpt hyp e := destrEq hyp e; [idtac | failure].
  Ltac assertOptEq v H e := caseEq e; [intros v H | failure].
  Ltac sig_opt v H e := case e; [intros v H | failure].

  Ltac succ' abs' :=
    match goal with
      | [ nextPc : int32 |- _ ] =>
	case (pcInRange_dec nextPc);
	  [intro HpcRange; exists (Some (cons abs' nil))
	    | failure]
    end.
  Ltac succ f abs :=
    succ' abs; intros st ctx Hconc; split;
      [exists (f st); myauto
	| exists abs; inversion Hexec; subst; left; split; [left; intu | idtac]].
  (*Ltac succSame f abs := succ f abs (fun (ctx : context) => ctx).*)
  Ltac destrConc' abs' :=
    match goal with
      [ H : conc' _ _ |- _ ] => destruct H as [abs']
    end.

  Ltac concrete g := destruct g; simpl; myauto; discriminate.
  Ltac concrete_arith rule g := destruct g; myauto; try discriminate;
    apply rule; simpl; myeauto.

  Ltac jmp :=
    match goal with
      [ hyps : list astate, abs : astate, i : int32, nextPc : int32 |- _] =>
      let Hx' := fresh "Hx'" in let c := fresh "c" in
 	(case (pcInRange_dec i); [intro Hpc | failure];
	  case (abs_covered (proj1_sig init) hyps (aset_pc abs i));
	    intros x Hx; destruct x; [idtac | failure];
	      add Hx' (Hx (refl_equal true)); clear Hx; rename Hx' into Hx;
		exists (Some (@nil astate)); intros st ctx Hconc; split;
		  [exists (set_pc st i); myauto
		    | intros st' Hexec; inversion Hexec;
		      subst; destruct Hx as [abs' Hx];
			exists abs'; destruct (subAbs (aset_pc abs i) abs') as [b c];
			  right; exists ctx; exists hyps; intuition;
			    apply AllS_trivial; intros x Hx; exists x;
			      unfold menter; intuition])
    end.

  (* Preservation by case analysis on ins *)
  destruct ins.

  (* Arith *)
  sig_opt t Ht (typeof_genop32 abs g).
  sig_opt t0 Ht0 (typeof_genop32 abs g0).
  destruct (aeval_arith abs a t t0) as [res Hres].
  assertOpt writable (writable_genop32 g).
  sig_opt abs' Habs' (awrite_genop32 abs g res).

  succ
  (fun st => set_pc
    (arith_flags (write_concrete g st (eval_arith a (eval_genop32 g st) (eval_genop32 g0 st)) writable)
      a (eval_genop32 g st) (eval_genop32 g0 st)) nextPc)
  (aset_pc abs' nextPc); eauto 7 with asm; concrete_arith XArith g.

  (* Arithb *)
  failure.

  (* Call *)
  destruct g; [idtac | failure | failure].
  destruct (pcInRange_dec i); [idtac | failure].
  destruct (pcInRange_dec nextPc); [idtac | failure].
  caseEq (aget_reg abs ESP); [failure | failure | idtac | failure | failure].
  intros ss Hesp.
  destruct (le_lt_dec (aStackSize abs) (S (S (S (S ss))))); [failure | idtac].
  destruct (functions i); [idtac | failure].
  destruct (le_lt_dec (4 * length (fiArgs f)) ss); [idtac | failure].
  destrEq Hcheck (checkArgs (fiArgs f) (ss - 4 * length (fiArgs f)) (aStackSlots abs)); [idtac | failure].
  destruct (enter abs nextPc i ss (fiArgs f)) as [absEnter |]; trivial; [idtac | failure].
  destruct (leave abs nextPc (fiRet f) ss (fiArgs f)) as [absLeave |]; trivial; [idtac | failure].
  destruct (abs_covered (proj1_sig init) (cons (snd absLeave) nil) absEnter).
  destruct x; [intuition | failure].

  exists (Some (cons (fst absLeave) nil)).
  destruct H as [hyps' H].
  destruct H as [abs' Habs'].
  destruct Habs' as [Hin Hsub].
  destruct Hsub as [Hsub Hhyps].
  destruct hyps' as [| hyp' hyps']; try (simpl in Hhyps; tauto).
  destruct hyps'; simpl in Hhyps; try tauto.

  intros st ctx Hconc.
  split.
  myeauto.
  intros st' Hexec.
  inversion Hexec; subst.
  exists abs'.
  right.
  exists (leave_context (set_pc st nextPc) ctx ss (fiArgs f)).
  exists (cons hyp' nil).
  intuition.
  apply allS_cons; [idtac | apply allS_nil].
  exists (fst absLeave).
  intuition.
  inversion H4; subst.
  inversion H1; subst.
  assert (Henter : menter conc' (leave_context (set_pc st nextPc) ctx ss (fiArgs f))
    (snd absLeave) ctx (fst absLeave)).
  apply m; trivial.
  intros st' Hconc'.
  destruct (subAbs hyp' (snd absLeave)).
  simpl in H.
  apply Henter.
  auto.

  inversion H4; subst.
  inversion H1; subst.
  destruct (subAbs absEnter abs').
  simpl in Hsub.
  replace (eval_addr (Build_address neg_four (Some ESP) None) st)
    with (badd neg_four (get_reg32 st ESP)); trivial.
  replace (eval_arith Sub (eval_genop32 (Reg ESP) st) four)
    with (bsub (get_reg32 st ESP) four); trivial.
  replace (eval_genop32 (Imm reg32 i)
    (set_reg32 (set_mem32 st (get_reg32 st ESP) nextPc) ESP
      (bsub (get_reg32 st ESP) four)))
  with i; trivial.
  auto.

  (* Cmp *)
  sig_opt t Ht (typeof_genop32 abs g).
  sig_opt t0 Ht0 (typeof_genop32 abs g0).
  destruct (aeval_arith abs Sub t t0) as [res Hres].

  succ
  (fun st => set_pc
    (arith_flags st Sub (eval_genop32 g st) (eval_genop32 g0 st)) nextPc)
  (aset_pc abs nextPc); eauto 7 with asm; concrete_arith XArith g.

  (* Inc *)
  sig_opt t Ht (typeof_genop32 abs g).
  destruct (aeval_arith abs Add t (Const one)) as [res Hres].
  assertOpt writable (writable_genop32 g).
  sig_opt abs' Habs' (awrite_genop32 abs g res).

  succ
  (fun st => set_pc
    (arith_flags_no_CF (write_concrete g st (eval_arith Add (eval_genop32 g st) one) writable)
      Add (eval_genop32 g st) one) nextPc)
  (aset_pc abs' nextPc); eauto 7 with asm; concrete_arith XInc g.

  (* Jcc *)
  destruct (pcInRange_dec i); [idtac | failure].
  destruct (pcInRange_dec nextPc); [idtac | failure].
  case (abs_covered (proj1_sig init) hyps (aset_pc abs i)).
  intros cov Hcov.
  destruct cov; [idtac | failure].
  intuition; rename H into Hcov.
  exists (Some (cons (aset_pc abs nextPc) nil)).
  intros st ctx Hconc.
  destruct Hcov as [hyps' Hcov].
  destruct Hcov as [abs' Hcov].
  split.

  (* Progress *)
  destrEq Hc (check_cc st c).
  exists (set_pc st i); myauto.
  exists (set_pc st nextPc); myauto.

  (* Preservation *)
  intros st' Hexec.
  inversion Hexec; subst.

  (* Taking the jump *)
  exists abs'.
  right; intuition.
  exists ctx.
  exists hyps'.
  intuition.

  destruct (subAbs (aset_pc abs i) abs').
  destruct x; simpl in H1; try discriminate.
  apply c0; myauto.

  (* Falling through *)
  exists (aset_pc abs nextPc).
  intuition; myauto.
  
  (* Jmp *)
  case (pcInRange_dec i); [intro Hpc | failure].
  case (abs_covered (proj1_sig init) hyps (aset_pc abs i)).
  intros x Hx; destruct x; [idtac | failure].
  add Hx' (Hx (refl_equal true)); clear Hx; rename Hx' into Hx.
  exists (Some (@nil astate)); intros st ctx Hconc; split.
  myeauto.
  intros st' Hexec; inversion Hexec; subst.
  destruct Hx as [hyps' Hx].
  destruct Hx as [abs' Hx].
  exists abs'.
  destruct (subAbs (aset_pc abs i) abs') as [b c].
  right; exists ctx; exists hyps'; intuition.

  (* Lea *)
  destruct (aeval_addr abs a) as [res Hres].

  succ
  (fun st => set_pc (set_reg32 st r (eval_addr a st)) nextPc)
  (aset_pc (aset_reg abs r res) nextPc); eauto 6 with asm;
    apply conc'_pc; trivial; inversion Hconc; split; intuition.

  (* Leave *)
  failure.

  (* Mov *)
  destruct (isInitializingStack g g0) as [esp0 Hinit | ].

  (* We are initializing ESP from a constant. *)
  destruct Hinit as [Hg Hg0]; subst.
  destruct (pcInRange_dec nextPc); [idtac | failure].
  destruct (bvec_eq_dec esp0 initialESP); [subst | failure].
  destruct (abs_covered (proj1_sig init) nil 
    (Build_astate (apcConst nextPc) (Reg32Map.upd (Reg32Map.empty Top) ESP (Stack 0))
      stackSize (NatMap.empty Top))).
  destruct x; [subst; intuition | failure].

  exists (Some (@nil astate)).
  destruct H as [hyps' H].
  destruct H as [abs' Habs'].
  destruct Habs' as [Hin Hproj].
  intros st ctx Hconc.
  split.
  exists (set_pc (set_reg32 st ESP initialESP) nextPc); myauto.
  intros st' Hexec.
  inversion Hexec; subst.
  exists abs'.
  right.
  exists (Build_context initialESP 0 initialESP (memIn ctx) None).
  exists (@nil astate).
  intuition.
  destruct hyps'; simpl in H0; tauto.
  eapply setESP; eauto.

  (* Other moves. *)
  sig_opt t0 Ht0 (typeof_genop32 abs g0).
  assertOpt writable (writable_genop32 g).
  sig_opt abs' Habs' (awrite_genop32 abs g t0).

  succ
  (fun st => set_pc (write_concrete g st (eval_genop32 g0 st) writable) nextPc)
  (aset_pc abs' nextPc); myeauto; concrete g.

  (* Movb *)
  failure.

  (* Pop *)
  sig_opt esp Hesp (typeof_genop32 abs (Reg ESP)).
  sig_opt t Ht (typeof_genop32 abs (Address reg32 (Build_address bzero32 (Some ESP) None))).
  destruct (aeval_arith abs Add (Const four) esp) as [res Hres].
  assertOpt writable (writable_genop32 g).
  sig_opt abs' Habs' (awrite_genop32 abs g t).
  sig_opt abs'' Habs'' (awrite_genop32 abs' (Reg ESP) res).

  destruct g; simpl in writable; try discriminate.

  succ
  (fun st => set_pc
    (set_reg32
      (set_reg32 st r (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
      ESP (eval_arith Add four (eval_genop32 (Reg ESP) st)))
    nextPc)
  (aset_pc abs'' nextPc); try inversion H1; try inversion H4; subst; eauto 7 with asm.

  succ
  (fun st => set_pc
    (set_reg32
      (set_mem32 st (eval_addr a st) (eval_genop32 (Address reg32 (Build_address bzero32 (Some ESP) None)) st))
      ESP (eval_arith Add four (eval_genop32 (Reg ESP) st)))
    nextPc)
  (aset_pc abs'' nextPc); try inversion H1; try inversion H4; subst; eauto 7 with asm.

  (* Push *)
  sig_opt esp Hesp (typeof_genop32 abs (Reg ESP)).
  sig_opt t Ht (typeof_genop32 abs g).
  destruct (aeval_arith abs Add (Const neg_four) esp) as [res Hres].
  assertOpt writable (writable_genop32 (Address reg32 (Build_address neg_four (Some ESP) None))).
  sig_opt abs' Habs' (awrite_genop32 abs (Address reg32 (Build_address neg_four (Some ESP) None)) t).
  sig_opt abs'' Habs'' (awrite_genop32 abs' (Reg ESP) res).

  succ
  (fun st => set_pc
    (set_reg32
      (set_mem32 st (eval_addr (Build_address neg_four (Some ESP) None) st) (eval_genop32 g st))
      ESP (eval_arith Add neg_four (eval_genop32 (Reg ESP) st)))
    nextPc)
  (aset_pc abs'' nextPc); try inversion H1; subst; eauto 7 with asm.

  (* Ret *)
  destruct hyps as [| hyp hyps]; [failure | destruct hyps; [idtac | failure]].
  sig_opt esp Hesp (typeof_genop32 abs (Reg ESP)).
  sig_opt t Ht (typeof_genop32 abs (Address reg32 (Build_address bzero32 (Some ESP) None))).
  destruct t; [failure | failure | failure | idtac | failure].
  destruct (aeval_arith abs Add (Const four) esp) as [res Hres].
  sig_opt abs' Habs' (awrite_genop32 abs (Reg ESP) res).
  pose (Build_astate
    apcRet
    (Reg32Map.upd (aRegs abs) ESP res)
    (aStackSize abs)
    (aStackSlots abs)).
  rename a into abs''.
  destruct (subAbs abs'' hyp); destruct x; [intuition | failure].

  exists (Some (@nil astate)).
  intros st ctx Hconc.
  split.
  myeauto.
  intros st' Hexec.
  inversion Hexec; subst.
  exists hyp.
  left.
  intuition.
  apply H.
  unfold abs''.
  eapply ret_conc; eauto.

  (* Shift *)
  failure.
Qed.

