(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** X86 driver policy *)

Require Import List TheoryList.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Env.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.Machine.Stateless.
Require Import ProofOS.Machine.Arches.

Module Type DRIVER_PERMISSIONS.
  Parameter memReadable : int32 -> Prop.
  Parameter memWritable : int32 -> Prop.
End DRIVER_PERMISSIONS.

Definition memGenop32 st (gop : genop32) :=
  match gop with
    | Address a => eval_addr a st :: nil
    | _ => nil
  end.

Definition memRead32 st ins :=
  match ins with
    | Arith _ dst src => memGenop32 st dst ++ memGenop32 st src
    | Call tgt => memGenop32 st tgt
    | Inc src => memGenop32 st src
    | Mov _ src => memGenop32 st src
    | Pop _ => get_reg32 st ESP :: nil
    | Push src => memGenop32 st src
    | Ret => get_reg32 st ESP :: nil
    | Shift _ dst _ => memGenop32 st dst
    | _ => nil
  end.

Definition memWrite32 st ins :=
  match ins with
    | Arith _ dst _ => memGenop32 st dst
    | Inc dst => memGenop32 st dst
    | Mov dst _ => memGenop32 st dst
    | Push dst => badd neg_four (get_reg32 st ESP) :: nil
    | Shift _ dst _ => memGenop32 st dst
    | _ => nil
  end.

Module M86_Arg <: M86_PARAM.
  Definition minitState := initState.
End M86_Arg.
Module M86_with_init := M86(M86_Arg).

Module DriverStateless (Perms : DRIVER_PERMISSIONS) <: STATELESS.
  Import Perms.

  Module Mac := M86_with_init.

  Definition memReadable32 addr :=
    memReadable addr
    /\ memReadable (binc addr)
    /\ memReadable (binc2 addr)
    /\ memReadable (binc3 addr).

  Definition memWritable32 addr :=
    memWritable addr
    /\ memWritable (binc addr)
    /\ memWritable (binc2 addr)
    /\ memWritable (binc3 addr).

  Definition isSafe (ins : instr) (st : state) :=
    AllS memReadable32 (memRead32 st ins)
    /\ AllS memWritable32 (memWrite32 st ins).
End DriverStateless.
