(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Building abstractions that depend on a fixed code segment *)

Require Import Arith List TheoryList.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.ListUtil.
Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Monad.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Machine.
Require Import ProofOS.Trusted.Arches.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.AbsInt.Sal.

Lemma apart_or : forall a a',
  bvec_to_nat a' + 4 <= bvec_to_nat a
  \/ bvec_to_nat a' >= bvec_to_nat a + 4
  -> apart a a'.
  intuition.
Qed.

Module Type M86_PARAM_FIXED_CODE.
  Parameter minitState : state -> Prop.

  Parameter minitMem : mem.
  Parameter codeStart : int32.
  Parameter codeLength : nat.
  Axiom codeHigh : bvec_to_nat codeStart >= 4.
  Axiom codeFits : bvec_to_nat codeStart + codeLength < pow2 32 - 3.

  Axiom pcInRange : forall st, minitState st
    -> bvec_to_nat codeStart
    <= bvec_to_nat (stPc st)
    < bvec_to_nat codeStart + codeLength - longestInstr.
  Axiom codeUniform : forall st, minitState st
    -> agree_range (stMem st) minitMem codeStart codeLength.
End M86_PARAM_FIXED_CODE.

Module Make (Param : M86_PARAM_FIXED_CODE).
  Import Param.
  Module MySal := Sal.Make(Param).
  Import MySal.

  Definition validPc (pc : int32) :=
    bvec_to_nat codeStart
    <= bvec_to_nat pc
    < bvec_to_nat codeStart + codeLength - longestInstr.

  Definition apartFromCode (pc : int32) :=
    (bvec_to_nat pc <= bvec_to_nat codeStart - 4
      \/ bvec_to_nat pc >= bvec_to_nat codeStart + codeLength + 4).

  (** * Does an instruction avoid writing over code memory? *)

  Definition cmdOk st cmd :=
    match cmd with
      | SStore dst src => apartFromCode (evalExp st dst)
	/\ not_too_high (evalExp st dst)
      | _ => True
    end.

  Fixpoint cmdsOk (st : salState) (cmds : list salCmd) {struct cmds} : Prop :=
    match cmds with
      | nil => True
      | cons cmd cmds => cmdOk st cmd /\ cmdsOk (salExecCmd st cmd) cmds
    end.

  Definition instrOk st (ins : salInstr) := cmdsOk st (fst ins).
  Definition instructionOk st (ins : salInstruction) := instrOk st (snd ins).
  
  Lemma agree_range_refl : forall a low lim,
    agree_range a a low lim.
    unfold agree_range; trivial.
  Qed.

  Hint Resolve agree_range_refl.

  Lemma agree_range_trans : forall a1 a2 a3 low lim,
    agree_range a1 a2 low lim
    -> agree_range a2 a3 low lim
    -> agree_range a1 a3 low lim.
    unfold agree_range; intuition.
    rewrite (H addr'); trivial.
    rewrite (H0 addr'); trivial.
    omega.
    omega.
  Qed.

  Lemma cmdOk_sound : forall st cmd,
    cmdOk st cmd
    -> agree_range (SstMem (salExecCmd st cmd)) (SstMem st) codeStart codeLength.
    intros st cmd Hcmd.
    unfold cmdOk, validPc in Hcmd.
    destruct cmd; simpl; trivial.
    intros a Ha.
    apply invert_set_mem32'_neq'.
    tauto.
    apply apart_or.
    intuition.
    red in H.
    generalize codeHigh; omega.

    unfold set_arith_flags.
    destruct s; trivial.
  Qed.

  Lemma jmpMem : forall jmp pc st st',
    salExecJmp jmp pc st = Some st'
    -> SstMem st = SstMem st'.
    destruct jmp; simpl; intuition; try discriminate;
      try destruct (check_cc st c); injection H; intros; subst; trivial.
  Qed.

  Lemma cmdsOk_sound : forall cmds st st',
    cmdsOk st' cmds ->
    agree_range (SstMem st') 
    (SstMem st) codeStart codeLength
    -> agree_range (SstMem (fold_left salExecCmd cmds st'))
    (SstMem st) codeStart codeLength.
    induction cmds; intuition.
    simpl in *; intuition.
    apply IHcmds; intuition.
    generalize (cmdOk_sound _ _ H1); intro.
    eapply agree_range_trans; eauto.
  Qed.

  Theorem instructionsOk_sound : forall st insPc st',
    instructionOk st (fst insPc)
    -> M.mexec insPc st st'
    -> agree_range (SstMem st') (SstMem st) codeStart codeLength.
    intros st insPc st' Hins Hexec.
    do 2 red in Hins.
    destruct insPc as [[x86 [cmds jmp]] pc]; simpl in *; idtac.
    do 2 red in Hexec; unfold salExec' in Hexec; simpl in Hexec.
    rewrite <- (jmpMem _ _ _ _ Hexec).
    clear Hexec.
    apply cmdsOk_sound; trivial.
  Qed.


  (** * An abstraction of a concrete machine whose code segment is never touched *)
  Module Type FIXED_CODE_ABSTRACTION.
    Declare Module Mac := M.

    Parameter mastate : Set.
    Parameter print_mastate : mastate -> unit.
    Parameter mcontext : Set.
    Parameter mconc : mcontext -> Mac.mstate -> mastate -> Prop.

    Parameter maPc : forall (abs : mastate), sig_option (fun pc =>
      forall ctx st, mconc ctx st abs -> SstPc st = pc).

    Definition absValidPc abs :=
      match maPc abs with
	| SNone => True
	| SSome pc _ => validPc pc
      end.

    Parameter minit : {init : list (list mastate * mastate)
      | (forall abs hyps, In (hyps, abs) init -> absValidPc abs)
	/\ forall st, Mac.minitState st -> exists ctx, exists abs,
	  In (nil, abs) init /\ mconc ctx st abs}.

    Definition progress (insPc : Mac.minstr * int32) (st : Mac.mstate) :=
      instructionOk st (fst insPc)
      /\ exists st' : Mac.mstate,
	Mac.mexec insPc st st'.

    Definition localStep (hyps abss : list mastate)
      (ctx : mcontext) (st' : Mac.mstate) (abs' : mastate) :=
      ((In abs' abss \/ In abs' hyps) /\ mconc ctx st' abs').

    Definition callStep (hyps abss : list mastate) (ctx : mcontext)
      (st' : Mac.mstate)
      (abs' : mastate) (ctx' : mcontext) (hyps' : list mastate) :=
      In (hyps', abs') (proj1_sig minit)
      /\ AllS (fun need => exists have,
	(In have abss \/ In have hyps)
	/\ subsumes mconc ctx' need ctx have) hyps'
      /\ mconc ctx' st' abs'.

    Definition preservation (hyps : list mastate) (abs : mastate)
      (insPc : Mac.minstr * int32) (abss : list mastate)
      (st : Mac.mstate) (ctx : mcontext) :=
      forall (st' : Mac.mstate) (Hexec : Mac.mexec insPc st st'),
	exists abs', localStep hyps abss ctx st' abs'
	  \/ exists ctx', exists hyps', callStep hyps abss ctx st' abs' ctx' hyps'.

    Parameter mstep : forall hyps abs insPc, sig_option (fun abss : list mastate =>
      forall st ctx, mconc ctx st abs
	-> progress insPc st
	/\ preservation hyps abs insPc abss st ctx).
  End FIXED_CODE_ABSTRACTION.

  Module Fix (Abs : FIXED_CODE_ABSTRACTION) : ABSTRACTION with Module Mac := M.
    Module Mac := M.
    Import Abs.

    Definition mastate := mastate.
    Definition print_mastate := print_mastate.
    Definition mcontext := mcontext.

    Definition absValidPc abs :=
      match maPc abs with
	| SNone => True
	| SSome pc _ => validPc pc
      end.

    Record _mconc (ctx : mcontext) (st : Mac.mstate) (abs : mastate) : Prop := {
      mcPc : absValidPc abs;
      mcCode : agree_range (SstMem st) minitMem codeStart codeLength;
      mcCust : mconc ctx st abs
    }.

    Definition get_instr pc :=
      let (ins, pc) := decode_instr minitMem pc in
	((ins, salCompile (ins, pc)), pc).

    Definition maget_instr : forall (abs : mastate),
      sig_option (fun insPc : Mac.minstr * int32 =>
	forall st ctx, _mconc ctx st abs
	  -> insPc = Mac.mget_instr st).
      intros.
      caseEq (maPc abs); [idtac | intros; apply SNone].
      intros pc Hpc HpcEq.
      apply SSome with (get_instr pc).
      intros st ctx Hconc.
      unfold get_instr, Mac.mget_instr, get_sal_instr, Sal.get_instr.
      destruct Hconc.
      rewrite (Hpc ctx); trivial.
      replace (decode_instr minitMem pc)
	with (decode_instr (SstMem st) pc); trivial.
      apply decode_same.
      do 2 intro.
      apply mcCode0.
      red in mcPc0.
      rewrite HpcEq in mcPc0.
      red in mcPc0.
      omega.
    Qed.

    Definition minit : {init : list (list mastate * mastate)
      | forall st, Mac.minitState st -> exists ctx, exists abs,
	In (nil, abs) init /\ _mconc ctx st abs}.
      destruct minit as [init [Hpcs Hinit]].
      exists init.
      intuition.
      destruct (Hinit _ H) as [ctx [abs [Hin Hconc]]].
      exists ctx.
      exists abs.
      intuition.
      destruct H as [st' [HR Hminit]].
      destruct HR.
      split; intuition.

      red.
      destruct (maPc abs); trivial.
      red.
      rewrite <- (e ctx st); trivial.
      rewrite Rpc.
      apply pcInRange; tauto.

      rewrite Rmem.
      apply codeUniform; trivial.
    Defined.

    (** ** On to the home stretch *)

    Definition progress (insPc : Mac.minstr * int32) (st : Mac.mstate) :=
      exists st' : Mac.mstate,
	Mac.mexec insPc st st'.

    Definition localStep (hyps abss : list mastate)
      (ctx : mcontext) (st' : Mac.mstate) (abs' : mastate) :=
      ((In abs' abss \/ In abs' hyps) /\ _mconc ctx st' abs').

    Definition callStep (hyps abss : list mastate) (ctx : mcontext)
      (st' : Mac.mstate)
      (abs' : mastate) (ctx' : mcontext) (hyps' : list mastate) :=
      In (hyps', abs') (proj1_sig minit)
      /\ AllS (fun need => exists have,
	(In have abss \/ In have hyps)
	/\ subsumes _mconc ctx' need ctx have) hyps'
      /\ _mconc ctx' st' abs'.

    Definition preservation (hyps : list mastate) (abs : mastate)
      (insPc : Mac.minstr * int32) (abss : list mastate)
      (st : Mac.mstate) (ctx : mcontext) :=
      forall (st' : Mac.mstate) (Hexec : Mac.mexec insPc st st'),
	exists abs', localStep hyps abss ctx st' abs'
	  \/ exists ctx', exists hyps', callStep hyps abss ctx st' abs' ctx' hyps'.

    Definition absValidPc_dec : forall (abs : mastate), pred_option (absValidPc abs).
      intros.
      refine (match maPc abs as pc return (maPc abs = pc -> pred_option (absValidPc abs)) with
		| SNone => fun pf => PSome _
		| SSome pc Hpc => fun pf =>
		  if le_lt_dec (bvec_to_nat codeStart) (bvec_to_nat pc)
		    then if le_lt_dec (bvec_to_nat codeStart + codeLength - longestInstr) (bvec_to_nat pc)
		      then PNone _
		      else PSome _
		    else
		      PNone _
	      end (refl_equal _));
      unfold absValidPc, validPc; rewrite pf; trivial; omega.
    Qed.

    Definition mstep : forall hyps abs insPc, sig_option (fun abss : list mastate =>
      forall st ctx, _mconc ctx st abs
	-> progress insPc st
	/\ preservation hyps abs insPc abss st ctx).
      intros.
      destruct (AllS_pdec _ absValidPc_dec hyps) as [HallHyps |]; [idtac | apply SNone].
      destruct (mstep hyps abs insPc) as [abss Habss |]; [idtac | apply SNone].
      destruct (AllS_pdec _ absValidPc_dec abss) as [HallAbss |]; [idtac | apply SNone].
      apply SSome with abss.
      intros st ctx Hconc.
      destruct Hconc.
      destruct (Habss _ _ mcCust0) as [Hprog Hpres]; clear Habss.
      intuition.

      destruct Hprog.
      red; eauto.
      
      intros st' Hst'.
      destruct (Hpres _ Hst') as [abs' Habs'].
      exists abs'.

      assert ((In abs' abss \/ In abs' hyps) /\ mconc ctx st' abs'
	-> _mconc ctx st' abs').
      clear Habs'.
      intro H.
      split; try tauto.

      assert (popt_Prop (absValidPc_dec abs')).
      intuition;
	[apply (AllS_In HallAbss)
	  | apply (AllS_In HallHyps)]; trivial.
      destruct (absValidPc_dec abs'); simpl in H0; tauto.

      apply agree_range_trans with (SstMem st); trivial.
      destruct Hprog as [HinsOk Hprog].
      eapply instructionsOk_sound; eauto.
      
      intuition; [left | right].

      red.
      red in H0.
      intuition.

      destruct H0 as [ctx' [hyps' Hcall]].
      exists ctx'; exists hyps'.
      red.
      red in Hcall.
      unfold minit.
      destruct Abs.minit as [init [Hpcs Hinit]]; simpl in *; idtac.
      intuition.

      eapply AllS_imp; [idtac | apply H3].
      intuition.
      destruct H1 as [have [Hin Hsub]].
      exists have.
      split; [tauto | idtac].
      intros st'' Hst''.
      destruct Hst''.
      split; [idtac | intuition | intuition].
      
      assert (popt_Prop (absValidPc_dec have)).
      intuition;
	[apply (AllS_In HallAbss)
	  | apply (AllS_In HallHyps)]; trivial.
      destruct (absValidPc_dec have); simpl in H1; tauto.

      split; intuition.
      eapply Hpcs; eauto.

      apply agree_range_trans with (SstMem st); trivial.
      destruct Hprog as [HinsOk Hprog].
      eapply instructionsOk_sound; eauto.
    Qed.

    Definition mconc := _mconc.
  End Fix.

End Make.


