(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Verifying a machine language by compiling it to another (probably simpler)
    language *)

Require Import List TheoryList ZArith.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.ListUtil.
Require Import Asm.Util.Monad.

Require Import Asm.Bitvector.Bitvector.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Machine.

Require Import ProofOS.AbsInt.Bounded.

Set Implicit Arguments.

(** * A reduction from one machine to another *)
Module Type REDUCTION.
  Declare Module Conc : MACHINE.
  (** Concrete machine whose programs we want to verify *)

  Declare Module Abs : ABSTRACTION.
  (** An abstraction, probably over a _different_ machine *)

  Parameter R : Conc.mstate -> Abs.Mac.mstate -> Prop.
  (** Concretization relation *)

  Parameter cameFrom : forall (ins : Abs.Mac.minstr),
    {ins' : Conc.minstr
      | forall abs nextPc, Abs.Mac.mget_instr abs = (ins, nextPc)
	-> forall st, R st abs -> Conc.mget_instr st = (ins', nextPc)}.
  (** For simplicity in verification, we want to know which instruction
      each abstract instruction was compiled from. *)

  Axiom init : forall conc, Conc.minitState conc
    -> exists abs, R conc abs /\ Abs.Mac.minitState abs.
  (** Every concrete start state is covered by an abstract start state. *)

  Parameter compile : forall (insPc : Conc.minstr * int32),
    {ins' : Abs.Mac.minstr
      | forall conc abs,
	R conc abs
	-> (forall abs', Abs.Mac.mexec (ins', snd insPc) abs abs'
	  -> exists conc', Conc.mexec insPc conc conc')
	/\ (forall conc', Conc.mexec insPc conc conc'
	  -> exists abs', Abs.Mac.mexec (ins', snd insPc) abs abs'
	    /\ R conc' abs')}.
  (** Translate concrete into abstract instructions that are safe in the usual
      progress and preservation sense. *)
End REDUCTION.

(** * Functor to build a verifier based on a reduction *)

Module Make(Red : REDUCTION) : CHECKER with Module Mac := Red.Conc.
  Import Red.

  Module ReduceAbs <: ABSTRACTION.
    Module Mac := Conc.

    Definition mastate := Abs.mastate.
    Definition print_mastate := Abs.print_mastate.
    Definition mcontext := Abs.mcontext.

    Definition mconc (ctx : mcontext) (st : Mac.mstate) (abs : mastate) :=
      exists ab, Abs.mconc ctx ab abs /\ R st ab.

    Definition maget_instr : forall (abs : mastate),
      sig_option (fun insPc : Conc.minstr * int32 =>
	forall st ctx, mconc ctx st abs
	  -> insPc = Conc.mget_instr st).
      intros.
      refine (insPc :: Hins <- Abs.maget_instr abs;
	  ins' ::: Hins' <- cameFrom (fst insPc);
	  SSome _ (ins', snd insPc) _).

      intros.
      destruct H as [abs' Habs'].
      rewrite Hins' with abs' (snd insPc) st; intuition.
      rewrite Hins with abs' ctx; trivial.
      destruct (Abs.Mac.mget_instr abs'); trivial.
    Qed.

    Definition minit : {init : list (list mastate * mastate)
      | forall st, Mac.minitState st -> exists ctx, exists abs,
	In (nil, abs) init /\ mconc ctx st abs}.
      destruct Abs.minit as [minit Hinit].
      exists minit.
      intros st Hst.
      destruct (init Hst) as [abs [HR Habs]].
      destruct (Hinit abs Habs) as [ctx [abs' [Hin Hconc]]].
      exists ctx.
      exists abs'.
      intuition.
      red; eauto.
    Defined.

    Definition progress (insPc : Mac.minstr * int32) (st : Mac.mstate) :=
      exists st' : Mac.mstate,
	Mac.mexec insPc st st'.

    Definition localStep (hyps abss : list mastate)
      (ctx : mcontext) (st' : Mac.mstate) (abs' : mastate) :=
      ((In abs' abss \/ In abs' hyps) /\ mconc ctx st' abs').

    Definition callStep (hyps abss : list mastate) (ctx : mcontext)
      (st' : Mac.mstate)
      (abs' : mastate) (ctx' : mcontext) (hyps' : list mastate) :=
      In (hyps', abs') (proj1_sig minit)
      /\ AllS (fun need => exists have,
	(In have abss \/ In have hyps)
	/\ subsumes mconc ctx' need ctx have) hyps'
      /\ mconc ctx' st' abs'.

    Definition preservation (hyps : list mastate) (abs : mastate)
      (insPc : Mac.minstr * int32) (abss : list mastate)
      (st : Mac.mstate) (ctx : mcontext) :=
      forall (st' : Mac.mstate) (Hexec : Mac.mexec insPc st st'),
	exists abs', localStep hyps abss ctx st' abs'
	  \/ exists ctx', exists hyps', callStep hyps abss ctx st' abs' ctx' hyps'.

    Definition mstep : forall hyps abs insPc, sig_option (fun abss : list mastate =>
      forall st ctx, mconc ctx st abs
	-> progress insPc st
	/\ preservation hyps abs insPc abss st ctx).
      intros.
      refine (ins' ::: Hins' <- compile insPc;
	abs' :: Habs' <- Abs.mstep hyps abs (ins', snd insPc);
	SSome _ abs' _); intros st ctx Hconc.
      red in Hconc.
      destruct Hconc as [ab [Hconc HR]].
      destruct (Hins' _ _ HR) as [Hprog Hpres].
      destruct (Habs' _ _ Hconc) as [[next Hprog'] Hpres'].
      red in Hpres'.
      intuition; red.
      destruct insPc; eauto.
    
      intros st' Hst'.
      destruct insPc; simpl in *; idtac.
      destruct (Hpres _ Hst') as [abs'' [Habs'' HR']].
      destruct (Hpres' _ Habs'') as [abs'0 [Hlocal | [ctx' [hyps' Hcall]]]].

      exists abs'0.
      left.
      red.
      red in Hlocal.
      split; [intuition | idtac].
      red.
      exists abs''; intuition.

      exists abs'0.
      right.
      exists ctx'; exists hyps'.
      red.
      red in Hcall.
      intuition.
      unfold minit.
      destruct Abs.minit.
      trivial.

      eapply AllS_imp; try apply H1.
      intuition.
      destruct H0 as [have Hhave].
      exists have.
      split; [intuition | idtac].
      unfold subsumes, mconc in *; idtac.
      destruct Hhave; clear H0.
      intros st0 [ab0 Hab0].
      exists ab0.
      intuition.

      red.
      eauto.
    Qed.

  End ReduceAbs.

  Module Ch := Checker(ReduceAbs).
  Definition verify := Ch.verify.

  Module Mac := Red.Conc.
End Make.
