(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Adding simple conditional branch consideration support to a type system *)

Require Import Arith List TheoryList ZArith.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.ListUtil.
Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Monad.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Maps.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Machine.
Require Import ProofOS.Trusted.Arches.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.AbsInt.Sal.
Require Import ProofOS.AbsInt.FixedCode.
Require Import ProofOS.AbsInt.TypeSystem.
Require Import ProofOS.AbsInt.StackTypes.

Set Implicit Arguments.

Definition eval_test fg :=
  match fg with
    | CF => Some eval_CF
    | PF => Some eval_PF
    | AF => Some eval_AF
    | ZF => Some eval_ZF
    | SF => Some eval_SF
    | OF => Some eval_OF
    | _ => None
  end.

Definition eval_condition' c aop e1 e2 :=
  match c with
    | O => eval_OF aop e1 e2
    | B => eval_CF aop e1 e2
    | Z => eval_ZF aop e1 e2
    | BE => eval_CF aop e1 e2 || eval_ZF aop e1 e2
    | S => eval_SF aop e1 e2
    | P => eval_PF aop e1 e2
    | L => negb (eqb (eval_SF aop e1 e2) (eval_OF aop e1 e2))
    | LE => eval_ZF aop e1 e2 || negb (eqb (eval_SF aop e1 e2) (eval_OF aop e1 e2))
  end.

Definition eval_cc c aop e1 e2 :=
  eqb (eval_condition' (snd c) aop e1 e2) (fst c).

Section condTy.
  Variable mastate : Set.
  
  Record condState : Set := {
    csFlags : FlagMap.t (option (arith_op * salExp * salExp));
    csCust : mastate
  }.
  
  Parameter print_condState : (mastate -> unit)
    -> ((flag * option (arith_op * salExp * salExp) -> unit) -> unit)
    -> mastate -> unit.
End condTy.

Module Make (Param : M86_PARAM_FIXED_CODE_AND_STACK).
  Module Stack := StackTypes.Make(Param).
  Import Stack.
  Import TySys.
  Import Param.
  Import Fixed.
  Import MySal.

  Section condTy2.
    Variables mastate mcontext : Set.
    Variable mconc : mcontext -> salState -> mastate -> Prop.
    
    Record condConc (ctx : mcontext) (st : salState) (abs : condState mastate) : Prop := {
      ccFlags : forall fg, match FlagMap.get fg (csFlags abs) with
			     | None => True
			     | Some (aop, e1, e2) =>
			       match eval_test fg with
				 | None => False
				 | Some eval_flag =>
				   FlagMap.get fg (SstFlags st)
				   = eval_flag aop (evalExp st e1) (evalExp st e2)
			       end
			   end;
      ccCust : mconc ctx st (csCust abs)
    }.
  End condTy2.

  Module Type TYPE_SYSTEM_WITH_TESTING.
    Declare Module Sys : STACK_READY_TYPE_SYSTEM.

    Parameter considerTest : forall (abs : stateDescription (stackTy Sys.ty)
      (stackState Sys.ty Sys.mastate))
      (co : cc) (aop : arith_op) (e1 e2 : salExp) (b : bool),
      {abs' : _
	| forall ctx st,
	  typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs
	  -> eval_cc co aop (evalExp st e1) (evalExp st e2) = b
	  -> typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs'}.
  End TYPE_SYSTEM_WITH_TESTING.

  (** * Add conditional testing support to a type system *)
  Module AddCond (SysPlus : TYPE_SYSTEM_WITH_TESTING) : STACK_READY_TYPE_SYSTEM
    with Definition ty := SysPlus.Sys.ty
      with Definition mastate := condState SysPlus.Sys.mastate
	with Definition mcontext := SysPlus.Sys.mcontext
	  with Definition hasTy := SysPlus.Sys.hasTy
	    with Definition mconc := condConc SysPlus.Sys.mconc.
    Module Mac := M.

    Import SysPlus.
    Import Sys.

    Definition ty := ty.
    Definition print_ty := print_ty.
    Definition mastate := condState mastate.
    Definition print_mastate ss :=
      let (default, elems) := FlagMap.elements (csFlags ss) in
	print_condState print_mastate
	(fun f => iter f elems) (csCust ss).

    Definition mcontext := mcontext.
    Definition hasTy := hasTy.
    Hint Unfold hasTy.

    Definition mconc := condConc mconc.
    Hint Unfold mconc.
    Hint Constructors condConc.

    Definition subTy := subTy.

    Definition arith_op_eq_dec : forall (x y : arith_op), {x = y} + {x <> y}.
      decide equality.
    Qed.

    Definition shift_op_eq_dec : forall (x y : shift_op), {x = y} + {x <> y}.
      decide equality.
    Qed.

    Definition salExp_eq_dec : forall (x y : salExp), {x = y} + {x <> y}.
      decide equality.
      apply int32_eq_dec.
      apply salReg_eq_dec.
      apply arith_op_eq_dec.
      apply eq_nat_dec.
      apply shift_op_eq_dec.
    Qed.

    Definition eq_dec_triple : forall (x y : arith_op * salExp * salExp), {x = y} + {x <> y}.
      decide equality.
      apply salExp_eq_dec.
      decide equality a p.
      apply salExp_eq_dec.
      apply arith_op_eq_dec.
    Qed.

    Definition eq_dec_tripleOpt : forall (x y : option (arith_op * salExp * salExp)), {x = y} + {x <> y}.
      decide equality; apply eq_dec_triple.
    Qed.

    Definition tripleOpt_ok (x y : option (arith_op * salExp * salExp)) :=
      x = y \/ y = None.

    Definition check_tripleOpt : forall (x y : option (arith_op * salExp * salExp)),
      pred_option (tripleOpt_ok x y).
      intros.
      refine (match y as y' return (y' = y -> _) with
		| None => fun pfEq => PSome _
		| _ => fun pfEq =>
		  if eq_dec_tripleOpt x y
		    then PSome _
		    else PNone _
	      end (refl_equal _)); red; intuition.
    Qed.

    Definition subMastate : forall (abs1 abs2 : mastate),
      pred_option (forall ctx st,
	mconc ctx st abs1 -> mconc ctx st abs2).
      intros.
      refine (Hflags <-- FlagMap.subMap tripleOpt_ok check_tripleOpt (csFlags abs1) (csFlags abs2);;
	      Hcust <-- subMastate (csCust abs1) (csCust abs2);;
	      PSome _).
      intros ctx st Hconc.
      destruct Hconc.
      split; intuition.
      add Hfg (Hflags fg).
      red in Hfg.
      intuition.

      rewrite <- H.
      apply ccFlags0.

      rewrite H.
      trivial.
    Qed.
    
    Definition tyTop := tyTop.
    Definition tyTopSound := tyTopSound.

    Definition tyConst := tyConst.
    Definition isTyConst := isTyConst.
    Definition tyConstSound1 := tyConstSound1.
    Definition tyConstSound2 := tyConstSound2.

    Definition typeof : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (evalExp st e) t}.
      intros.
      refine (t ::: Ht <- typeof (csCust abs) rs e;
	exist _ t _); intuition.
      destruct H.
      intuition.
    Qed.

    Definition typeofLoad : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (get_mem32 st (evalExp st e)) t}.
      intros.
      refine (t ::: Ht <- typeofLoad (csCust abs) rs e;
	exist _ t _); intuition.
      destruct H.
      intuition.
    Qed.

    Definition checkStore : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      pred_option (forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> apartFromCode (evalExp st e)
	/\ not_too_high (evalExp st e)
	/\ apartFromStack (evalExp st e)).
      intros.
      refine (Hok <-- checkStore (csCust abs) rs e;;
	PSome _); intros.
      destruct H.
      generalize (Hok ctx st).
      intuition.
    Qed.

    Definition expUnaffected : forall (cmd : salCmd) (e : salExp),
      pred_option (forall st, evalExp st e = evalExp (salExecCmd st cmd) e).
      refine (fix expUnaffected (cmd : salCmd) (e : salExp) {struct e}
	: pred_option (forall st, evalExp st e = evalExp (salExecCmd st cmd) e) :=
	match e as e
	  return (pred_option (forall st, evalExp st e = evalExp (salExecCmd st cmd) e)) with
	  | SConst _ => PSome _
	  | SReg r =>
	    match cmd as cmd' return (cmd' = cmd -> _) with
	      | SSet r' _ => fun _ =>
		if salReg_eq_dec r r'
		  then PNone _
		  else PSome _
	      | SLoad r' _ => fun _ =>
		if salReg_eq_dec r r'
		  then PNone _
		  else PSome _
	      | _ => fun pfEq => PSome _	    
	    end (refl_equal _)
	  | SArith _ e1 e2 =>
	    He1 <-- expUnaffected cmd e1;;
	    He2 <-- expUnaffected cmd e2;;
 	    PSome _
	  | SShift _ e1 _ =>
	    He1 <-- expUnaffected cmd e1;;
 	    PSome _
	end); subst; simpl; intros.

      intuition.

      unfold get_reg, set_reg; simpl.
      rewrite SalRegMap.gso; trivial.

      unfold get_reg, set_reg; simpl.
      rewrite SalRegMap.gso; trivial.

      trivial.

      unfold set_arith_flags.
      destruct s; trivial.

      rewrite He1.
      rewrite He2.
      trivial.

      rewrite He1.
      trivial.
    Qed.

    Lemma evalExp_flags : forall e st fls,
      evalExp (Build_salState (SstRegs st) fls (SstMem st) (SstPc st)) e
      = evalExp st e.
      induction e; simpl; intros.

      trivial.

      trivial.

      rewrite IHe1.
      rewrite IHe2.
      trivial.

      rewrite IHe.
      trivial.
    Qed.

    Hint Rewrite FlagMap.gss : Maps.
    Hint Rewrite FlagMap.gso using (auto; discriminate) : Maps.

    Definition effectOfCmd : forall (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> mconc ctx (salExecCmd st cmd) abs').
      intros.
      refine (cust :: Hcust <- effectOfCmd (csCust abs) rs cmd;
	match cmd as cmd' return (cmd' = cmd -> _) with
	  | SCompare cmp aop e1 e2 => fun pfEq =>
	    SSome _ (Build_condState
	      (FlagMap.set OF (Some (aop, e1, e2))
		(FlagMap.set SF (Some (aop, e1, e2))
		  (FlagMap.set ZF (Some (aop, e1, e2))
		    (FlagMap.set AF (Some (aop, e1, e2))
		      (FlagMap.set PF (Some (aop, e1, e2))
			(FlagMap.init None))))))
	      cust) _
	  | _ => fun pfEq =>
	    SSome _ (Build_condState
	      (FlagMap.map (fun topt : option (arith_op * salExp * salExp) =>
		match topt with
		  | None => None
		  | Some triple =>
		    match triple with
		      | (_, e1, e2) =>
			if expUnaffected cmd e1
			  then if expUnaffected cmd e2
			    then topt
			    else None
			  else None
		    end
		end) (csFlags abs))
	      cust) _
	end (refl_equal _)); simpl; intros; subst; simpl in *; intuition; destruct H;
      try (split; simpl; intuition;
	add Hfg (ccFlags0 fg);
	rewrite FlagMap.gmap; destruct (FlagMap.get fg (csFlags abs)); trivial;
	  destruct p; destruct p;
	    destruct (eval_test fg); try tauto;
	      match goal with
		| [ |- context[expUnaffected ?cmd _]] =>
		  destruct (expUnaffected cmd s2); trivial;
		    destruct (expUnaffected cmd s1); trivial;
		      simpl in e, e0; rewrite <- e; rewrite <- e0; trivial
	      end).

      split; simpl; intuition.

      repeat (rewrite FlagMap.gsspec;
	match goal with
	  | [ |- context[FlagIndexed.eq _ ?F] ] =>
	    destruct (FlagIndexed.eq fg F);
	      [subst; simpl;
		unfold set_arith_flags, arith_flags, arith_flags_no_CF, arith_flags', arith_flags_no_CF';
		  destruct cmp; simpl;
		    autorewrite with Maps;
		      repeat rewrite evalExp_flags; trivial
		| idtac]
	end).
    
      rewrite FlagMap.gi; trivial.
    Qed.

    Lemma evalExp_mem : forall e st pc a v,
      evalExp (Build_salState (SstRegs st) (SstFlags st) (set_mem32' (SstMem st) a v) pc) e
      = evalExp st e.
      induction e; simpl; intros.

      trivial.

      trivial.

      rewrite IHe1.
      rewrite IHe2.
      trivial.

      rewrite IHe.
      trivial.
    Qed.

    Definition doStackWrite : forall (abs : mastate),
      sig_option (fun abs' => forall ctx st dst src,
	mconc ctx st abs
	-> inStack dst
	-> mconc ctx (set_mem32 st dst src) abs').
      intros.
      refine (cust :: Hcust <- doStackWrite (csCust abs);
	SSome _ (Build_condState (csFlags abs) cust) _); intros.
      destruct H.
      split; intuition.
      add Hfg (ccFlags0 fg).
      simpl.
      destruct (FlagMap.get fg (csFlags abs)); trivial.
      do 2 destruct p.
      destruct (eval_test fg); trivial.
      unfold set_mem32; simpl.
      repeat rewrite evalExp_mem; trivial.
    Qed.

    Lemma evalExp_pc : forall e st pc,
      evalExp (Build_salState (SstRegs st) (SstFlags st) (SstMem st) pc) e
      = evalExp st e.
      induction e; simpl; intros.

      trivial.

      trivial.

      rewrite IHe1.
      rewrite IHe2.
      trivial.

      rewrite IHe.
      trivial.
    Qed.

    Definition effectOfJump : forall (abs : mastate) (rs : SalRegMap.t ty)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> match salExecJmp jmp nextPc st with
	     | None => False
	     | Some st' => mconc ctx st' abs'
	   end).
      intros.
      refine (cust :: Hcust <- effectOfJump (csCust abs) rs jmp nextPc;
	SSome _ (Build_condState (csFlags abs) cust) _); intuition.
      generalize (Hcust ctx st).
      destruct jmp; simpl; try destruct (check_cc st c); destruct H; intuition;
	split; simpl; intuition;
	  generalize (ccFlags0 fg);
	    destruct (FlagMap.get fg (csFlags abs)); trivial;
	      do 2 destruct p;
		destruct (eval_test fg); try tauto;
		  unfold set_pc;
		    repeat rewrite evalExp_pc;
		      trivial.
    Qed.

    Definition absEval_cc : forall (abs : mastate) (co : cc),
      sig_option (fun triple => forall ctx st,
	mconc ctx st abs
	-> check_cc st co = eval_cc co (fst (fst triple))
	(evalExp st (snd (fst triple)))
	(evalExp st (snd triple))).
      intros.
      destruct co.
      idtac.
      refine (match c as c
		return (sig_option
		  (fun triple =>
		    forall ctx st,
		      mconc ctx st abs ->
		      check_cc st (b, c) =
		      eval_cc (b, c) (fst (fst triple)) (evalExp st (snd (fst triple)))
		      (evalExp st (snd triple)))) with
		| O =>
		  match FlagMap.get OF (csFlags abs)
		    as b return (b = FlagMap.get OF (csFlags abs) -> _) with
		    | None => fun _ => SNone _
		    | Some triple => fun pfEq => SSome _ triple _
		  end (refl_equal _)
		| B =>
		  match FlagMap.get CF (csFlags abs)
		    as b return (b = FlagMap.get CF (csFlags abs) -> _) with
		    | None => fun _ => SNone _
		    | Some triple => fun pfEq => SSome _ triple _
		  end (refl_equal _)
		| Z =>
		  match FlagMap.get ZF (csFlags abs)
		    as b return (b = FlagMap.get ZF (csFlags abs) -> _) with
		    | None => fun _ => SNone _
		    | Some triple => fun pfEq => SSome _ triple _
		  end (refl_equal _)
		| BE =>
		  match (FlagMap.get CF (csFlags abs), FlagMap.get ZF (csFlags abs))
		    as b return (b = (FlagMap.get CF (csFlags abs), FlagMap.get ZF (csFlags abs)) -> _) with
		    | (Some triple, Some triple') => fun pfEq =>
		      if eq_dec_triple triple' triple
			then SSome _ triple _
			else SNone _
		    | _ => fun _ => SNone _
		  end (refl_equal _)
		| S =>
		  match FlagMap.get SF (csFlags abs)
		    as b return (b = FlagMap.get SF (csFlags abs) -> _) with
		    | None => fun _ => SNone _
		    | Some triple => fun pfEq => SSome _ triple _
		  end (refl_equal _)
		| P =>
		  match FlagMap.get PF (csFlags abs)
		    as b return (b = FlagMap.get PF (csFlags abs) -> _) with
		    | None => fun _ => SNone _
		    | Some triple => fun pfEq => SSome _ triple _
		  end (refl_equal _)
		| L =>
		  match (FlagMap.get SF (csFlags abs), FlagMap.get OF (csFlags abs))
		    as b return (b = (FlagMap.get SF (csFlags abs), FlagMap.get OF (csFlags abs)) -> _) with
		    | (Some triple, Some triple') => fun pfEq =>
		      if eq_dec_triple triple' triple
			then SSome _ triple _
			else SNone _
		    | _ => fun _ => SNone _
		  end (refl_equal _)
		| LE =>
		  match (FlagMap.get ZF (csFlags abs), FlagMap.get SF (csFlags abs), FlagMap.get OF (csFlags abs))
		    as b return (b = (FlagMap.get ZF (csFlags abs),
		      FlagMap.get SF (csFlags abs),
		      FlagMap.get OF (csFlags abs)) -> _) with
		    | (Some triple, Some triple', Some triple'') => fun pfEq =>
		      if eq_dec_triple triple' triple
			then if eq_dec_triple triple'' triple
			  then SSome _ triple _
			  else SNone _
			else SNone _
		    | _ => fun _ => SNone _
		  end (refl_equal _)
	      end);
      subst; try (intuition;
	unfold check_cc, check_cc', eval_cc, check_condition', eval_condition'; simpl;
	  match goal with
	    | [ |- context[FlagMap.get ?F _] ] =>
	      add Hf (ccFlags H F);
	      rewrite <- pfEq in Hf;
		destruct triple;
		  destruct p;
		    simpl in *; congruence
	  end).
      
      intuition.
      unfold check_cc, check_cc', eval_cc, check_condition', eval_condition'; simpl.
      add Hcf (ccFlags H CF).
      add Hzf (ccFlags H ZF).
      injection pfEq.
      intros Heqz Heqc.
      rewrite <- Heqz in Hzf.
      rewrite <- Heqc in Hcf.
      destruct triple.
      destruct p.
      simpl in *; congruence.

      intuition.
      unfold check_cc, check_cc', eval_cc, check_condition', eval_condition'; simpl.
      add Hsf (ccFlags H SF).
      add Hof (ccFlags H OF).
      injection pfEq.
      intros Heqo Heqs.
      rewrite <- Heqo in Hof.
      rewrite <- Heqs in Hsf.
      destruct triple.
      destruct p.
      simpl in *; congruence.

      intuition.
      unfold check_cc, check_cc', eval_cc, check_condition', eval_condition'; simpl.
      add Hzf (ccFlags H ZF).
      add Hsf (ccFlags H SF).
      add Hof (ccFlags H OF).
      injection pfEq.
      intros Heqo Heqs Heqz.
      rewrite <- Heqz in Hzf.
      rewrite <- Heqo in Hof.
      rewrite <- Heqs in Hsf.
      destruct triple.
      destruct p.
      simpl in *; congruence.
    Qed.

    Definition effectOfConditionalJump : forall (abs : stateDescription (stackTy ty)
      (stackState ty mastate))
      (co : cc) (pc nextPc : int32) (b : bool),
      sig_option (fun abs' => forall ctx st,
	typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st abs
	-> check_cc st co = b
	-> match salExecJmp (SConditionalJump co pc) nextPc st with
	     | None => False
	     | Some st' => typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st' abs'
	   end).
      intros.
      refine (abs' :: Habs' <- effectOfConditionalJump
	(Build_stateDescription
	  (sdPc abs)
	  (sdRegs abs)
	  (Build_stackState
	    (aIsRet (sdCust abs))
	    (aStackSize (sdCust abs))
	    (aStackSlots (sdCust abs))
	    (csCust (aCustom (sdCust abs))))) co pc nextPc b;
	match absEval_cc (aCustom (sdCust abs)) co with
	    | SNone =>
	      SSome _ (Build_stateDescription
		(sdPc abs')
		(sdRegs abs')
		(Build_stackState
		  (aIsRet (sdCust abs'))
		  (aStackSize (sdCust abs'))
		  (aStackSlots (sdCust abs'))
		  (Build_condState
		    (csFlags (aCustom (sdCust abs)))
		    (aCustom (sdCust abs'))))) _
	    | SSome triple Htriple =>
	      abs'' ::: Habs'' <- considerTest abs' co
	      (fst (fst triple)) (snd (fst triple)) (snd triple) b;
	      SSome _ (Build_stateDescription
		(sdPc abs'')
		(sdRegs abs'')
		(Build_stackState
		  (aIsRet (sdCust abs''))
		  (aStackSize (sdCust abs''))
		  (aStackSlots (sdCust abs''))
		  (Build_condState
		    (csFlags (aCustom (sdCust abs)))
		    (aCustom (sdCust abs''))))) _
	end); simpl in *; intuition; rewrite H0.

      assert (typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc)
	ctx (if b then set_pc st pc else set_pc st nextPc) abs'').
      apply Habs''.
      rewrite <- H0.
      apply Habs'; trivial.
      destruct H; destruct mcCust0; destruct mcCust0; split; simpl; intuition;
	split; simpl; intuition.

      rewrite <- H0.
      rewrite Htriple with (ctxCust ctx) st.
      replace (evalExp
        (if eval_cc co (fst (fst triple)) (evalExp st (snd (fst triple)))
          (evalExp st (snd triple))
          then set_pc st pc
          else set_pc st nextPc) (snd (fst triple)))
      with (evalExp st (snd (fst triple))).
      replace (evalExp
        (if eval_cc co (fst (fst triple)) (evalExp st (snd (fst triple)))
          (evalExp st (snd triple))
          then set_pc st pc
          else set_pc st nextPc) (snd triple))
      with (evalExp st (snd triple)); trivial.
      destruct (eval_cc co (fst (fst triple)) (evalExp st (snd (fst triple)))
        (evalExp st (snd triple)));
      unfold set_pc; rewrite evalExp_pc; trivial.
      destruct (eval_cc co (fst (fst triple)) (evalExp st (snd (fst triple)))
        (evalExp st (snd triple)));
      unfold set_pc; rewrite evalExp_pc; trivial.

      destruct H; destruct mcCust0; intuition.

      destruct H1.
      split; simpl in *; intuition.
      destruct mcCust0.
      split; simpl in *; intuition.
      split; simpl in *; intuition.

      replace (SstFlags (if b then set_pc st pc else set_pc st nextPc))
	with (SstFlags st); [idtac | destruct b; trivial].

      destruct H.
      destruct mcCust1; destruct mcCust1.
      generalize (Habs' ctx st); generalize (Habs'' ctx st).
      generalize (ccFlags0 fg).
      caseEq (FlagMap.get fg (csFlags (aCustom (sdCust abs)))); trivial.
      intros p Heq.
      unfold ty, mastate in Heq.
      rewrite Heq.

      do 2 destruct p.
      destruct (eval_test fg); try tauto.
      unfold set_pc.

      destruct b;
	repeat rewrite evalExp_pc;
	  trivial.

      assert (typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc)
	ctx (if b then set_pc st pc else set_pc st nextPc) abs').
      rewrite <- H0.
      apply Habs'; trivial.

      destruct H.
      destruct mcCust0.
      destruct mcCust0.
      split; simpl; intuition.
      split; simpl; intuition.

      destruct H.
      destruct mcCust0.
      destruct mcCust0.
      destruct H1.
      destruct mcCust0.
      split; simpl; intuition.
      split; simpl; intuition.
      split; simpl; intuition.

      generalize (ccFlags0 fg).
      unfold ty, mastate.
      match goal with
	| [ |- context[FlagMap.get ?fg ?fls]] =>
	  destruct (FlagMap.get fg fls); trivial
      end.
      do 2 destruct p.
      destruct (eval_test fg); trivial.
      destruct b;
	unfold set_pc; simpl;
	  repeat rewrite evalExp_pc; trivial.
    Qed.

    Definition viewShiftOfJump : forall (hyps : list (stateDescription ty mastate))
      (abs : stateDescription ty mastate)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun vs => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> exists ctx', typeMconc hasTy mconc ctx' st (vsAbs vs)
	  /\ AllS (fun need => exists have, (In have (vsSuccs vs) \/ In have hyps)
	    /\ subsumes (typeMconc hasTy mconc) ctx' need ctx have) (vsHyps vs)).
      intros; apply SNone.
    Qed.
  End AddCond.

End Make.
