(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Verification driver *)

Require Import List TheoryList.

Require Import Asm.Util.ArithUtil.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.Sal.
Require Import ProofOS.AbsInt.TypeSystem.
Require Import ProofOS.AbsInt.StackTypes.
Require Import ProofOS.AbsInt.SimpleTypes.

Set Implicit Arguments.

Module Type DRIVER_REGIONS.
  Parameter minitMem : mem.
  Parameter codeStart : int32.
  Parameter codeLength : nat.
  Axiom codeHigh : bvec_to_nat codeStart >= 4.
  Axiom codeFits : bvec_to_nat codeStart + codeLength < pow2 32 - 3.
  Axiom codeBigEnough : longestInstr + 1 <= codeLength.

  Parameter stackStart : int32.
  Parameter stackLength : nat.
  Axiom stackHigh : bvec_to_nat stackStart >= 4.
  Axiom stackBig : stackLength > 8.
  Axiom stackFits : bvec_to_nat stackStart + stackLength < pow2 32 - 3.
  Axiom stackCodeSeparate : disjointRegions codeStart codeLength stackStart stackLength.
  Axiom stackLongEnough : stackLength >= 4.
End DRIVER_REGIONS.

Module Type DRIVER_PARAM.
  Declare Module Regions : DRIVER_REGIONS.
  Declare Module Sys : SIMPLE_TYPE_SYSTEM.
End DRIVER_PARAM.

Module Driver (Param : DRIVER_PARAM).
  Import Param.
  Import Regions.

  Module StackParam <: M86_PARAM_FIXED_CODE_AND_STACK.
    Definition minitState st := stPc st = codeStart /\ stMem st = minitMem.

    Definition minitMem := minitMem.
    Definition codeStart := codeStart.
    Definition codeLength := codeLength.
    Definition codeHigh := codeHigh.
    Definition codeFits := codeFits.

    Theorem pcInRange : forall st, minitState st
      -> bvec_to_nat codeStart
      <= bvec_to_nat (stPc st)
      < bvec_to_nat codeStart + codeLength - longestInstr.
      intros.
      destruct H.
      rewrite H.
      unfold codeStart, codeLength.
      generalize codeBigEnough; omega.
    Qed.

    Theorem codeUniform : forall st, minitState st
      -> agree_range (stMem st) minitMem codeStart codeLength.
      intros.
      destruct H.
      rewrite H0.
      unfold minitMem.
      red; trivial.
    Qed.

    Definition stackStart := stackStart.
    Definition stackLength := stackLength.
    Definition stackHigh := stackHigh.
    Definition stackFits := stackFits.
    Definition stackCodeSeparate := stackCodeSeparate.
  End StackParam.

  Module Simple := SimpleTypes.Make(StackParam).
  Import Simple.
  Import Stack.
  Import TySys.
  Import Fixed.
  Import MySal.

  Module MySys := Create(Sys).

  (** * Specify the root states for a verification *)
  Module Type ROOT_STATES.
    Parameter roots : {rs : list (list (stateDescription (stackTy MySys.ty) (stackState MySys.ty unit))
      * stateDescription (stackTy MySys.ty) (stackState MySys.ty unit))
      | AllS (fun sd => match sdPc (snd sd) with
			  | ApcUnknown => False
			  | ApcConst pc => validPc pc
			end) rs}.

    Parameters functions : list (function_info (stackTy MySys.ty)).
  End ROOT_STATES.

  Module Make (Roots : ROOT_STATES).
    Import Roots.

    Module MyTypeSystemPlus <: TYPE_SYSTEM_WITH_FUNCTIONS.
      Module Sys := MySys.
      Definition functions := functions.
    End MyTypeSystemPlus.

    Module TypeSystem := Stack.AddStackTypes(MyTypeSystemPlus).
    
    (** Produce the first input to the chain of functors to come *)
    Module User <: USE_TYPE_SYSTEM.
      Module Sys := TypeSystem.

      Definition minit : {init : list (list (stateDescription Sys.ty Sys.mastate)
	* stateDescription Sys.ty Sys.mastate)
	| (AllS (fun sd => match sdPc (snd sd) with
			   | ApcUnknown => False
			   | ApcConst pc => validPc pc
			 end) init)
	  /\ forall st, M.minitState st -> exists ctx, exists sd,
	    In (nil, sd) init
	    /\ ApcConst (SstPc st) = sdPc sd
	    /\ (forall r, Sys.hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs sd)))
	    /\ Sys.mconc ctx st (sdCust sd)}.
	destruct roots as [rs Hrs].
	exists (cons (nil, Build_stateDescription (ApcConst codeStart) (SalRegMap.init TypeSystem.tyTop)
	  (Build_stackState false 0 (NatMap.init (Top _)) tt))
	rs); intuition.

	constructor; simpl; trivial.
	red.
	generalize codeBigEnough; unfold StackParam.codeStart, StackParam.codeLength;
	  omega.

	exists (Build_stackContext (nat_to_bvec _ (bvec_to_nat stackStart + stackLength - 8))
	  minitMem bzero32 None tt).
	exists (Build_stateDescription (ApcConst codeStart) (SalRegMap.init TypeSystem.tyTop)
          (Build_stackState false 0 (NatMap.init (Top MySys.ty)) tt)).
	destruct H as [st' [HR Hinit]].
	destruct Hinit.
	intuition.

	simpl; congruence.

	simpl sdRegs.
	rewrite SalRegMap.gi.
	red.
	apply TypeSystem.tyTopSound.

	simpl sdCust.
	red.
	split.

	intros; discriminate.

	unfold ctxSptr.
	rewrite ntb_invert; auto.
	unfold StackParam.stackStart, StackParam.stackLength; omega.
	generalize stackFits; omega.

	unfold ctxSptr.
	rewrite ntb_invert; auto.
	simpl.
	generalize stackLongEnough; intuition.
	generalize stackBig.
	unfold StackParam.stackStart.
	omega.
	generalize stackFits; omega.

	intros slot Hslot.
	unfold aStackSlots.
	rewrite NatMap.gi.
	constructor.

	rewrite Rmem.
	rewrite H0.
	unfold ctxMemIn.
	red; trivial.

	split.
      Qed.
    End User.

    Module Abs1 := Abstraction(User).
    Module Abs2 := Fix(Abs1).
    Module Ch := MySal.Checker(Abs2).

    Module Mac := MC.
    Definition verify := Ch.verify.

  End Make.

  Definition fold_right := fold_right.
End Driver.

