(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Simple type system construction *)

Require Import Arith List TheoryList.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.ListUtil.
Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Monad.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Machine.
Require Import ProofOS.Trusted.Arches.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.AbsInt.Sal.
Require Import ProofOS.AbsInt.FixedCode.
Require Import ProofOS.AbsInt.TypeSystem.
Require Import ProofOS.AbsInt.StackTypes.

Set Implicit Arguments.

(** * Definition of a simple type system *)
Module Type SIMPLE_TYPE_SYSTEM.
  Parameter ty : Set.
  Parameter print_ty : ty -> unit.
  
  Parameter hasTy : int32 -> ty -> Prop.
  
  Parameter tyTop : ty.
  Axiom tyTopSound : forall v, hasTy v tyTop.

  Parameter tyConst : int32 -> ty.
  Axiom tyConstSound1 : forall v, hasTy v (tyConst v).
  Axiom tyConstSound2 : forall v1 v2, hasTy v1 (tyConst v2) -> v1 = v2.
  Parameter isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).
  
  Parameter subTy : forall (t1 t2 : ty),
    pred_option (forall v,
      hasTy v t1 -> hasTy v t2).
  
  Parameter typeofArith : forall (aop : arith_op) (t1 t2 : ty),
    {t : ty
      | forall v1 v2,
	hasTy v1 t1
	-> hasTy v2 t2
	-> hasTy (eval_arith aop v1 v2) t}.
End SIMPLE_TYPE_SYSTEM.

Module Make (Param : M86_PARAM_FIXED_CODE_AND_STACK).
  Module Stack := StackTypes.Make(Param).
  Import Stack.
  Import TySys.
  Import Fixed.
  Import MySal.

  Definition dummyConc (_ : unit) (_ : M.mstate) (_ : unit) := True.

  Section dummies.
    Variable ty : Set.
    Variable hasTy : int32 -> ty -> Prop.

    Definition dummyHasTy (_ : unit) := hasTy.
  End dummies.

  Module Create (Sys : SIMPLE_TYPE_SYSTEM) : STACK_READY_TYPE_SYSTEM
    with Definition ty := Sys.ty
	with Definition mastate := unit
	  with Definition mcontext := unit
	    with Definition hasTy := dummyHasTy Sys.hasTy
	      with Definition mconc := dummyConc.
    Import Sys.
    Module Mac := M.

    Definition ty := ty.
    Definition print_ty := print_ty.

    Definition mastate := unit.
    Definition print_mastate (u : unit) := tt.

    Definition mcontext := unit.

    Definition mconc := dummyConc.

    Definition subMastate : forall (abs1 abs2 : mastate),
      pred_option (forall ctx st,
	mconc ctx st abs1 -> mconc ctx st abs2).
      intros.
      apply PSome.
      intuition.
    Qed.
    
    Definition hasTy := dummyHasTy hasTy.
    Hint Unfold hasTy.

    Definition tyTop := tyTop.
    Lemma tyTopSound : forall ctx v, hasTy ctx v tyTop.
      intros.
      apply tyTopSound.
    Qed.

    Hint Resolve tyTopSound.

    Definition tyConst := tyConst.
    Theorem tyConstSound1 : forall ctx v, hasTy ctx v (tyConst v).
      intros; apply tyConstSound1.
    Qed.
    Theorem tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
      intros; apply tyConstSound2; trivial.
    Qed.
    Definition isTyConst := isTyConst.

    Definition subTy : forall (t1 t2 : ty),
      pred_option (forall ctx v,
	hasTy ctx v t1 -> hasTy ctx v t2).
      intros.
      refine (Hsub <-- subTy t1 t2;; PSome _);
	intuition.
    Qed.

    Hint Resolve tyConstSound1.

    Definition typeof : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (evalExp st e) t}.
      refine (fix typeof (abs : mastate) (rs : SalRegMap.t ty) (e : salExp) {struct e}
	: {t : ty
	  | forall ctx st,
	    mconc ctx st abs
	    -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	    -> hasTy ctx (evalExp st e) t} :=
	match e as e' return (e' = e -> _) with
	  | SConst n => fun pfEq => exist _ (tyConst n) _
	  | SReg r => fun pfEq =>
	    exist _ (SalRegMap.get r rs) _
	  | SArith aop e1 e2 => fun pfEq =>
	    t1 ::: Ht1 <- typeof abs rs e1;
	    t2 ::: Ht2 <- typeof abs rs e2;
	    t ::: Ht <- typeofArith aop t1 t2;
	    exist _ t _
	  | _ => fun _ => exist _ tyTop _
	end (refl_equal _)); subst; simpl; intuition.
      
      unfold hasTy in *; apply Ht.
      apply Ht1 with tt; auto.
      apply Ht2 with tt; auto.
    Qed.

    Definition typeofLoad : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (get_mem32 st (evalExp st e)) t}.
      intros; exists tyTop; auto.
    Qed.

    Definition checkStore : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      pred_option (forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> apartFromCode (evalExp st e)
	/\ not_too_high (evalExp st e)
	/\ apartFromStack (evalExp st e)).
      intros; apply PNone.
    Qed.

    Definition doStackWrite : forall (abs : mastate),
      sig_option (fun abs' => forall ctx st dst src,
	mconc ctx st abs
	-> inStack dst
	-> mconc ctx (set_mem32 st dst src) abs').
      intros.
      apply SSome with abs.
      intuition.
    Qed.

    Definition effectOfCmd : forall (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> mconc ctx (salExecCmd st cmd) abs').
      intros; apply SSome with tt; intuition.
    Qed.

    Definition effectOfJump : forall (abs : mastate) (rs : SalRegMap.t ty)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> match salExecJmp jmp nextPc st with
	     | None => False
	     | Some st' => mconc ctx st' abs'
	   end).
      intros.
      refine (match jmp as jmp' return (jmp' = jmp -> _) with
		| SFail => fun _ => SNone _
		| _ => fun pfEq => SSome _ tt _
	      end (refl_equal _)); subst; intuition.
    Qed.

    Definition effectOfConditionalJump : forall (abs : stateDescription (stackTy ty)
      (stackState ty mastate))
      (co : cc) (pc nextPc : int32) (b : bool),
      sig_option (fun abs' => forall ctx st,
	typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st abs
	-> check_cc st co = b
	-> match salExecJmp (SConditionalJump co pc) nextPc st with
	     | None => False
	     | Some st' => typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st' abs'
	   end).
      intros.
      apply SSome with (Build_stateDescription
	(ApcConst (if b then pc else nextPc))
	(sdRegs abs)
	(Build_stackState
	  false
	  (aStackSize (sdCust abs))
	  (aStackSlots (sdCust abs))
	  tt)); subst; intuition.
      subst.
      simpl.
      
      destruct H.
      split; simpl; intuition.
      destruct (check_cc st co); trivial.

      replace (get_reg (if check_cc st co then set_pc st pc else set_pc st nextPc) r)
	with (get_reg st r); auto.
      destruct (check_cc st co); trivial.
    
      simpl in mcCust0.
      destruct mcCust0.
      destruct (check_cc st co); split; simpl; intuition; discriminate.
    Qed.
      
    Definition viewShiftOfJump : forall (hyps : list (stateDescription ty mastate))
      (abs : stateDescription ty mastate)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun vs => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> exists ctx', typeMconc hasTy mconc ctx' st (vsAbs vs)
	  /\ AllS (fun need => exists have, (In have (vsSuccs vs) \/ In have hyps)
	    /\ subsumes (typeMconc hasTy mconc) ctx' need ctx have) (vsHyps vs)).
      intros; apply SNone.
    Qed.
  End Create.

End Make.
