(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Adding simple stack support to a type system *)

Require Import Arith List TheoryList ZArith.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.ListUtil.
Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Monad.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Maps.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Machine.
Require Import ProofOS.Trusted.Arches.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.AbsInt.Sal.
Require Import ProofOS.AbsInt.FixedCode.
Require Import ProofOS.AbsInt.TypeSystem.

Set Implicit Arguments.

(** * Maps used in abstract state *)

Module NatIndexed <: INDEXED_TYPE.
  Definition t := nat.

  Definition index := P_of_succ_nat.

  Theorem index_inj: forall (x y: t), index x = index y -> x = y.
    induction x; destruct y; simpl in *; intuition.
    destruct (index y); discriminate.
    destruct (index x); discriminate.
    rewrite (IHx y); trivial.
    apply Psucc_inj; trivial.
  Qed.

  Definition eq := eq_nat_dec.

  Definition invert n := pred (nat_of_P n).

  Theorem invert_index : forall x, invert (index x) = x.
    intro.
    unfold invert, index.
    rewrite nat_of_P_o_P_of_succ_nat_eq_succ.
    auto.
  Qed.
End NatIndexed.

Module NatMap := IMap(NatIndexed).


Definition disjointRegions (start1 : int32) len1 (start2 : int32) len2 :=
  bvec_to_nat start1 + len1 + 4 <= bvec_to_nat start2
  \/ bvec_to_nat start1 >= bvec_to_nat start2 + len2 + 4.

Module Type M86_PARAM_FIXED_CODE_AND_STACK.
  Parameter minitState : state -> Prop.

  Parameter minitMem : mem.
  Parameter codeStart : int32.
  Parameter codeLength : nat.
  Axiom codeHigh : bvec_to_nat codeStart >= 4.
  Axiom codeFits : bvec_to_nat codeStart + codeLength < pow2 32 - 3.

  Axiom pcInRange : forall st, minitState st
    -> bvec_to_nat codeStart
    <= bvec_to_nat (stPc st)
    < bvec_to_nat codeStart + codeLength - longestInstr.
  Axiom codeUniform : forall st, minitState st
    -> agree_range (stMem st) minitMem codeStart codeLength.

  Parameter stackStart : int32.
  Parameter stackLength : nat.
  Axiom stackHigh : bvec_to_nat stackStart >= 4.
  Axiom stackFits : bvec_to_nat stackStart + stackLength < pow2 32 - 3.
  Axiom stackCodeSeparate : disjointRegions codeStart codeLength stackStart stackLength.
End M86_PARAM_FIXED_CODE_AND_STACK.

(** * Metadata types *)

Section metadata.
  Variable ty : Set.

  Record function_info : Set := {
    fiPc : int32;               (** First program counter *)
    fiRet : ty;                 (** Return type *)
    fiArgs : list ty            (** How many arguments? *)
  }.
End metadata.

Section stackTy.
  Variables ty mastate : Set.
  
  Inductive stackTy : Set :=
    | Top 
    | Const : int32 -> stackTy
    | Stack : nat -> stackTy
    | OldEbp : stackTy
    | RetPtr : stackTy
    | Cust : ty -> stackTy.

  Parameter print_stackTy : (ty -> unit) -> stackTy -> unit.
  
  Record stackState : Set := {
    aIsRet : bool;
    aStackSize : nat;
    aStackSlots : NatMap.t stackTy;
    aCustom : mastate
  }.
  
  Parameter print_stackState : (ty -> unit) -> (mastate -> unit)
    -> nat -> stackTy -> ((nat * stackTy -> unit) -> unit) -> mastate -> unit.
End stackTy.

Module Make (Param : M86_PARAM_FIXED_CODE_AND_STACK).
  Module TySys := TypeSystem.Make(Param).
  Import TySys.
  Import Param.
  Import Fixed.
  Import MySal.

  Definition stackEnd := nat_to_bvec 32 (bvec_to_nat stackStart + stackLength).

  Definition inStack (a : int32) :=
    bvec_to_nat stackStart <= bvec_to_nat a <= bvec_to_nat stackStart + stackLength.

  Definition apartFromStack (a : int32) :=
    bvec_to_nat a <= bvec_to_nat stackStart - 4
    \/ bvec_to_nat a >= bvec_to_nat stackStart + stackLength + 4.

  (** * The language of types used in basic block preconditions for outputs of
      this functor *)
  Section stackTy2.
    Variables ty mastate mcontext : Set.
    Variable hasTy : mcontext -> int32 -> ty -> Prop.
    Variable mconc : mcontext -> M.mstate -> mastate -> Prop.

    Record stackContext : Set := {
      ctxSptr : int32;
      ctxMemIn : mem;
      ctxOldEbp : int32;
      ctxRetPtr : option int32;
      ctxCust : mcontext
    }.

    Definition stack_slot (ctx : stackContext) n :=
      nat_to_bvec 32 (bvec_to_nat (ctxSptr ctx) - n).

    Inductive stackHasTy : stackContext -> int32 -> stackTy ty -> Prop :=
      | HT_Top : forall ctx v, stackHasTy ctx v (Top _)
      | HT_Const : forall ctx v, stackHasTy ctx v (Const _ v)
      | HT_Stack : forall ctx n,
	n <= bvec_to_nat (ctxSptr ctx)
	-> stackHasTy ctx (stack_slot ctx n) (Stack _ n)
      | HT_OldEbp : forall ctx, stackHasTy ctx (ctxOldEbp ctx) (OldEbp _)
      | HT_RetPtr : forall ctx r, ctxRetPtr ctx = Some r
	-> stackHasTy ctx r (RetPtr _)
      | HT_Cust : forall ctx v t, hasTy (ctxCust ctx) v t
	-> stackHasTy ctx v (Cust t).

    Definition conc_slots ctx st (abs : stackState ty mastate) :=
      forall n, 4 <= n < aStackSize abs
	-> stackHasTy ctx (get_mem32 st (stack_slot ctx n)) (NatMap.get n (aStackSlots abs)).

    Definition agree_range' (mem1 mem2 : mem) (addr1 addr2 : int32) :=
      forall addr', bvec_to_nat addr1 <= bvec_to_nat addr' < bvec_to_nat addr2
	-> Int32Map.get addr' mem1 = Int32Map.get addr' mem2.

    Record stackConc (ctx : stackContext) (st : M.mstate) (abs : stackState ty mastate) : Prop := {
      mcPc : aIsRet abs = true -> ctxRetPtr ctx = Some (SstPc st);
      mcSptr : bvec_to_nat (ctxSptr ctx) <= bvec_to_nat stackStart + stackLength - 4;
      mcStackSize : bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (ctxSptr ctx);
      mcStackSlots : conc_slots ctx st abs;
      mcMemIn : agree_range' (SstMem st) (ctxMemIn ctx) (ctxSptr ctx) stackEnd;
      mcCust : mconc (ctxCust ctx) st (aCustom abs)
    }.

    Lemma ntb_cong : forall len n1 n2,
      n1 = n2
      -> nat_to_bvec len n1 = nat_to_bvec len n2.
      congruence.
    Qed.

    Definition stackSlotOf : forall (abs : stateDescription (stackTy ty) (stackState ty mastate)) (e : salExp),
      sig_option (fun ss => forall ctx st,
	typeMconc stackHasTy stackConc ctx st abs
	-> evalExp st e = stack_slot ctx ss).
      intros.
      refine (match e as e' return (e' = e -> _) with
		| SReg r => fun pfEq =>
		  match SalRegMap.get r (sdRegs abs)
		    as t return (t = SalRegMap.get r (sdRegs abs) -> _) with
		    | Stack ss => fun pfEq' =>
		      SSome _ ss _
		    | _ => fun _ => SNone _
		  end (refl_equal _)
		| SArith Add (SConst n) (SReg r) => fun pfEq =>
		  match SalRegMap.get r (sdRegs abs)
		    as t return (t = SalRegMap.get r (sdRegs abs) -> _) with
		    | Stack ss => fun pfEq' =>
		      let n' := bvec_to_nat n in
			if le_lt_dec n' ss
			  then SSome _ (ss - n') _
			  else
			    let n'' := ss + bvec_to_nat (bneg n) in
			      if le_lt_dec n'' (aStackSize (sdCust abs))
				then SSome _ n'' _
				else SNone _
		    | _ => fun _ => SNone _
		  end (refl_equal _)
		| _ => fun _ => SNone _
	      end (refl_equal _)); subst; simpl; intuition; destruct H;
      add Hr (mcRegs r);
      rewrite <- pfEq' in Hr;
	inversion Hr; subst; trivial.

      subst n'.
      unfold stack_slot, badd32, badd, barith.
      rewrite ntb_invert.
      match goal with
	| [ |- nat_to_bvec _ ?X = nat_to_bvec _ ?Y ] =>
	  replace Y with X; trivial
      end.
      omega.
      destruct mcCust0; generalize stackFits; omega.

      subst n'.
      subst n''.
      unfold stack_slot, badd32, badd, barith.
      rewrite ntb_invert.
      rewrite plus_comm.
      replace (bvec_to_nat (ctxSptr ctx) - (ss + bvec_to_nat (bneg n)))
	with (bvec_to_nat (ctxSptr ctx) - ss - bvec_to_nat (bneg n)).
      apply (@ntb_mod 32).
      apply eqMod_sym.
      replace (bvec_to_nat (ctxSptr ctx) - ss + bvec_to_nat n)
	with (bvec_to_nat (ctxSptr ctx) - ss + bvec_to_nat (bneg (bneg n))).
      apply bneg_btn_sub; destruct mcCust0; omega.
      rewrite bneg_btn_double; trivial.
      omega.
      destruct mcCust0; generalize stackFits; omega.
    Qed.

  End stackTy2.

  Module Type STACK_READY_TYPE_SYSTEM.
    Parameter ty : Set.
    Parameter print_ty : ty -> unit.
    (** A language of types describing machine words *)

    Parameter mastate : Set.
    Parameter print_mastate : mastate -> unit.
    (** Custom abstract state information *)

    Parameter mcontext : Set.
    (** A summary of aspects of a machine state that don't need to be represented
	explicitly at verification time, but are important in soundness proofs *)

    Parameter mconc : mcontext -> M.mstate -> mastate -> Prop.
    (** Concretization relation *)

    Parameter subMastate : forall (abs1 abs2 : mastate),
      pred_option (forall ctx st,
	mconc ctx st abs1 -> mconc ctx st abs2).
    (** A way to check compatibility of abstract states *)
    
    Parameter hasTy : mcontext -> int32 -> ty -> Prop.
    (** What it means for a word to be of a type *)

    Parameter tyTop : ty.
    Axiom tyTopSound : forall ctx v, hasTy ctx v tyTop.

    Parameter tyConst : int32 -> ty.
    Axiom tyConstSound1 : forall ctx v, hasTy ctx v (tyConst v).
    Axiom tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
    Parameter isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).

    Parameter subTy : forall (t1 t2 : ty),
      pred_option (forall ctx v,
	hasTy ctx v t1 -> hasTy ctx v t2).
    (** A way to check compatibility of types *)

    Parameter typeof : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (evalExp st e) t}.
    (** A type inference procedure for SAL expressions *)

    Parameter typeofLoad : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (get_mem32 st (evalExp st e)) t}.
    (** Type inference for memory loads *)

    Parameter checkStore : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      pred_option (forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> apartFromCode (evalExp st e)
	/\ not_too_high (evalExp st e)
	/\ apartFromStack (evalExp st e)).
    (** Type inference for memory loads *)

    Parameter effectOfCmd : forall (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> mconc ctx (salExecCmd st cmd) abs').
    (** Update the custom state to reflect the result of a command. *)

    Parameter doStackWrite : forall (abs : mastate),
      sig_option (fun abs' => forall ctx st dst src,
	mconc ctx st abs
	-> inStack dst
	-> mconc ctx (set_mem32 st dst src) abs').
    (** The custom type system should be unaffected by stack writes. *)

    Parameter effectOfJump : forall (abs : mastate) (rs : SalRegMap.t ty)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> match salExecJmp jmp nextPc st with
	     | None => False
	     | Some st' => mconc ctx st' abs'
	   end).
    (** Update the custom state to reflect the result of a jump. *)

    Parameter effectOfConditionalJump : forall (abs : stateDescription (stackTy ty)
      (stackState ty mastate))
      (co : cc) (pc nextPc : int32) (b : bool),
      sig_option (fun abs' => forall ctx st,
	typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st abs
	-> check_cc st co = b
	-> match salExecJmp (SConditionalJump co pc) nextPc st with
	     | None => False
	     | Some st' => typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st' abs'
	   end).
    (** Specialized version of [effectOfJump] that allows different states in
        different cases of a conditional jump test *)

    Parameter viewShiftOfJump : forall (hyps : list (stateDescription ty mastate))
      (abs : stateDescription ty mastate)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun vs => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> exists ctx', typeMconc hasTy mconc ctx' st (vsAbs vs)
	  /\ AllS (fun need => exists have, (In have (vsSuccs vs) \/ In have hyps)
	    /\ subsumes (typeMconc hasTy mconc) ctx' need ctx have) (vsHyps vs)).
    (** Massage the form of an abstract state to match what a jump target
        expects; for instance, at a function call. *)
  End STACK_READY_TYPE_SYSTEM.

  Module Type TYPE_SYSTEM_WITH_FUNCTIONS.
    Declare Module Sys : STACK_READY_TYPE_SYSTEM.

    Parameter functions : list (function_info (stackTy Sys.ty)).
    (** Information on function types *)
  End TYPE_SYSTEM_WITH_FUNCTIONS.

  (** * Add stack support to a type system *)
  Module AddStackTypes (SysPlus : TYPE_SYSTEM_WITH_FUNCTIONS) : TYPE_SYSTEM
    with Definition ty := stackTy SysPlus.Sys.ty
      with Definition mastate := stackState SysPlus.Sys.ty SysPlus.Sys.mastate
	with Definition mcontext := stackContext SysPlus.Sys.mcontext
	  with Definition hasTy := stackHasTy SysPlus.Sys.hasTy
	    with Definition mconc := stackConc SysPlus.Sys.hasTy SysPlus.Sys.mconc.
    Module Mac := M.

    Import SysPlus.
    Import Sys.

    Definition ty := stackTy ty.
    Definition print_ty := print_stackTy print_ty.
    Definition mastate := stackState Sys.ty mastate.
    Definition print_mastate ss :=
      let (default, elems) := NatMap.elements (aStackSlots ss) in
	print_stackState Sys.print_ty print_mastate
	(aStackSize ss) default (fun f => iter f elems) (aCustom ss).

    Definition mcontext := stackContext mcontext.

    Definition hasTy := stackHasTy hasTy.
    Hint Unfold hasTy.
    Hint Constructors stackHasTy.

    Definition mconc := stackConc Sys.hasTy mconc.
    Hint Unfold mconc.
    Hint Constructors stackConc.

    Hint Resolve tyConstSound1 tyConstSound2.

    Lemma HT_Const' : forall ctx v1 v2,
      v1 = v2
      -> hasTy ctx v1 (Const _ v2).
      intros; subst; auto.
    Qed.

    Hint Resolve HT_Const'.

    Definition subTy : forall (t1 t2 : ty),
      pred_option (forall ctx v,
	hasTy ctx v t1 -> hasTy ctx v t2).
      intros.
      refine (match (t1, t2) as ts return (ts = (t1, t2) -> _) with
		| (_, Top) => fun pfEq => PSome _
		| (Const n1, Const n2) => fun pfEq =>
		  if int32_eq_dec n1 n2
		    then PSome _
		    else PNone _
		| (Stack n1, Stack n2) => fun pfEq =>
		  if eq_nat_dec n1 n2
		    then PSome _
		    else PNone _
		| (OldEbp, OldEbp) => fun pfEq => PSome _
		| (RetPtr, RetPtr) => fun pfEq => PSome _
		| (Cust t1, Cust t2) => fun pfEq =>
		  if subTy t1 t2
		    then PSome _
		    else PNone _
		| (Const n1, Cust t2) => fun pfEq =>
		  if subTy (tyConst n1) t2
		    then PSome _
		    else PNone _
		| (Cust t1, Const n2) => fun pfEq =>
		  if subTy t1 (tyConst n2)
		    then PSome _
		    else PNone _
		| _ => fun _ => PNone _
	      end (refl_equal _)); injection pfEq; intros; subst; intuition;
      inversion H1; intuition eauto.
    Qed.

    Definition bool_eq_dec : forall (x y : bool), {x = y} + {x <> y}.
      decide equality.
    Qed.

    Definition subMastate : forall (abs1 abs2 : mastate),
      pred_option (forall ctx st,
	mconc ctx st abs1 -> mconc ctx st abs2).
      intros.
      refine (if le_lt_dec (aStackSize abs2) (aStackSize abs1)
	then if bool_eq_dec (aIsRet abs2) (aIsRet abs1)
	  then (Hslots <-- NatMap.subMap (fun t1 t2 => forall ctx v,
	    hasTy ctx v t1 -> hasTy ctx v t2)
	      subTy (aStackSlots abs1) (aStackSlots abs2);;
	    Hcust <-- subMastate (aCustom abs1) (aCustom abs2);;
	    PSome _)
	  else PNone _
	else PNone _); intros.
      destruct H.
      split; intuition.
      apply mcPc0; congruence.

      unfold hasTy in Hslots.
      red; intuition.
    Qed.

    Lemma mconc_sptr_not_too_high : forall ctx st abs,
      mconc ctx st abs
      -> not_too_high (ctxSptr ctx).
      destruct 1.
      red; split.
      generalize stackFits; omega.
    Qed.

    Hint Resolve mconc_sptr_not_too_high.

    Lemma not_too_high_minus : forall n1 n2,
      not_too_high n1
      -> bvec_to_nat n1 - n2 < pow2 32.
      destruct 1.
      omega.
    Qed.

    Hint Resolve not_too_high_minus.

    Lemma stack_slot_rewrite : forall ctx st abs n1 n2,
      mconc ctx st abs
      -> n1 <= bvec_to_nat (ctxSptr ctx)
      -> bvec_to_nat n2 <= n1
      -> badd32 (stack_slot ctx n1) n2
      = stack_slot ctx (n1 - bvec_to_nat n2).
      intros.
      unfold stack_slot, badd32, badd, barith.
      rewrite ntb_invert; eauto.
      apply ntb_cong.
      omega.
    Qed.

    Lemma stack_slot_rewrite2 : forall ctx st abs n1 n2,
      mconc ctx st abs
      -> n1 <= bvec_to_nat (ctxSptr ctx)
      -> n1 + bvec_to_nat n2 <= aStackSize abs
      -> bsub32 (stack_slot ctx n1) n2
      = stack_slot ctx (n1 + bvec_to_nat n2).
      intros.
      unfold stack_slot, bsub32, bsub, barith.
      rewrite ntb_invert; eauto.
      apply ntb_mod.
      rewrite rewrite_minus_plus.
      apply eqMod_sym.
      apply bneg_btn_sub.
      destruct H.
      omega.
    Qed.

    Lemma stack_slot_rewrite3 : forall (ctx : mcontext) st (abs : mastate) n1 n2,
      mconc ctx st abs
      -> n1 < bvec_to_nat n2
      -> n1 + bvec_to_nat (bneg n2) <= aStackSize abs
      -> n1 <= bvec_to_nat (ctxSptr ctx)
      -> badd32 n2 (stack_slot ctx n1)
      = stack_slot ctx (n1 + bvec_to_nat (bneg n2)).
      intros until n2; intros Hconc Hlt Hle Hsptr.
      destruct Hconc.
      unfold stack_slot, badd32, badd, barith.
      rewrite ntb_invert.
      rewrite plus_comm.
      replace (bvec_to_nat (ctxSptr ctx) - (n1 + bvec_to_nat (bneg n2)))
	with (bvec_to_nat (ctxSptr ctx) - n1 - bvec_to_nat (bneg n2)).
      apply ntb_mod.
      apply eqMod_sym.
      replace (bvec_to_nat (ctxSptr ctx) - n1 + bvec_to_nat n2)
	with (bvec_to_nat (ctxSptr ctx) - n1 + bvec_to_nat (bneg (bneg n2))).
      apply bneg_btn_sub; omega.
      rewrite bneg_btn_double; trivial.
      omega.
      generalize stackFits; omega.
    Qed.

    Definition lowerTy t :=
      match t with
	| Const n => tyConst n
	| Cust t' => t'
	| _ => tyTop
      end.

    Hint Resolve tyTopSound.

    Lemma lowerTy_sound : forall st r ctx t,
      hasTy ctx (get_reg st r) t
      -> Sys.hasTy (ctxCust ctx) (get_reg st r) (lowerTy t).
      destruct t; inversion 1; subst; simpl; intuition.
    Qed.

    Hint Resolve lowerTy_sound.

    Definition typeofR (abs : mastate) (rs : SalRegMap.t ty) (e : salExp) :=
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (evalExp st e) t}.

    Definition raiseTy : forall (t : Sys.ty), {t' : ty |
      forall ctx v, Sys.hasTy (ctxCust ctx) v t -> hasTy ctx v t'}.
      intros.
      refine (match Sys.isTyConst t with
		| SNone =>
		  if Sys.subTy Sys.tyTop t
		    then exist _ (Top _) _
		    else exist _ (Cust t) _
		| SSome n Hn => exist _ (Const _ n) _
	      end); subst; intuition.
      generalize (Sys.tyConstSound2 H); intro; subst.
      auto.
    Qed.

    Lemma lowerTy_map_sound : forall ctx st rs,
      (forall r : SalRegIndexed.t,
	hasTy ctx (get_reg st r) (SalRegMap.get r rs))
      -> (forall r : SalRegIndexed.t,
	Sys.hasTy (ctxCust ctx) (get_reg st r)
	(SalRegMap.get r (SalRegMap.map lowerTy rs))).
      intros.
      rewrite SalRegMap.gmap.
      auto.
    Qed.

    Hint Resolve lowerTy_map_sound.

    Definition typeof : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      typeofR abs rs e.
      refine (fix typeof (abs : mastate) (rs : SalRegMap.t ty) (e : salExp) {struct e}
	: typeofR abs rs e :=
	t ::: Ht <- Sys.typeof (aCustom abs) (SalRegMap.map lowerTy rs) e;
	if Sys.subTy Sys.tyTop t
	then match e return typeofR abs rs e with
	  | SConst n => exist _ (Const _ n) _
	  | SReg r => exist _ (SalRegMap.get r rs) _
	  | SArith aop e1 e2 =>
	    t1 ::: Ht1 <- typeof abs rs e1;
	    t2 ::: Ht2 <- typeof abs rs e2;
	    match (t1, t2) as ts
	      return (ts = (t1, t2) -> typeofR abs rs (SArith aop e1 e2)) with
	      | (Const n1, Const n2) => fun pfEq => exist _ (Const _ (eval_arith aop n1 n2)) _
	      | (Stack n1, Const n2) => fun pfEq =>
		match aop return typeofR abs rs (SArith aop e1 e2) with
		  | Add =>
		    let n2' := bvec_to_nat n2 in
		      if le_lt_dec n2' n1
			then exist _ (Stack _ (n1 - n2')) _
			else exist _ (Top _) _
		  | Sub =>
		    let n2' := bvec_to_nat n2 in
		      if le_lt_dec (n1 + n2') (aStackSize abs)
			then exist _ (Stack _ (n1 + n2')) _
			else exist _ (Top _) _
		  | _ => exist _ (Top _) _
		end
	      | (Const n2, Stack n1) => fun pfEq =>
		match aop return typeofR abs rs (SArith aop e1 e2) with
		  | Add =>
		    let n2' := bvec_to_nat n2 in
		      if le_lt_dec n2' n1
			then exist _ (Stack _ (n1 - n2')) _
			else
			  let n2'' := n1 + bvec_to_nat (bneg n2) in
			    if le_lt_dec n2'' (aStackSize abs)
			      then exist _ (Stack _ n2'') _
			      else exist _ (Top _) _
		  | _ => exist _ (Top _) _
		end
	      | _ => fun _ => exist _ (Top _) _
	    end (refl_equal _)
	  | SShift _ _ _ => exist _ (Top _) _
	end
	  else t' ::: Ht' <- raiseTy t;
	       exist _ t' _); try (injection pfEq; clear pfEq; intros; subst); simpl evalExp; intuition;
      try (try subst n2'; try subst n2'';
	add Ht1' (Ht1 _ _ H1 H2); inversion Ht1'; subst;
	add Ht2' (Ht2 _ _ H1 H2); inversion Ht2'; subst;
	match goal with
	  | [ |- context[badd32 (stack_slot ?ctx ?n1) ?n2] ] =>
	    rewrite stack_slot_rewrite with ctx st abs n1 n2; trivial;
	  red; apply HT_Stack; omega
	  | [ |- context[bsub32 (stack_slot ?ctx ?n1) ?n2] ] =>
	    rewrite stack_slot_rewrite2 with ctx st abs n1 n2; trivial; destruct H1;
	  red; apply HT_Stack; omega
	  | [ |- context[badd32 ?n2 (stack_slot ?ctx ?n1)] ] =>
	    try (rewrite badd32_comm;
	      rewrite stack_slot_rewrite with ctx st abs n1 n2; trivial;
	      red; apply HT_Stack; omega);
	    rewrite stack_slot_rewrite3 with ctx st abs n1 n2; trivial;
	  red; apply HT_Stack; destruct H1; omega
	  | _ => intuition
	end);
      red; apply Ht';
	destruct H; auto.
    Qed.

    Definition typeofLoadR (abs : mastate) (rs : SalRegMap.t ty) (e : salExp) :=
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (get_mem32 st (evalExp st e)) t}.

    Definition typeofLoad : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      typeofLoadR abs rs e.
      intros.
      refine (t ::: Ht <- typeof abs rs e;
	match t as t' return (t' = t -> typeofLoadR abs rs e) with
	  | Stack n => fun pfEq =>
	    if le_lt_dec 4 n
	      then if le_lt_dec (aStackSize abs) n
		then exist _ (Top _) _
		else exist _ (NatMap.get n (aStackSlots abs)) _
	      else exist _ (Top _) _
	  | _ => fun _ =>
	    t'' ::: Ht' <- typeofLoad (aCustom abs) (SalRegMap.map lowerTy rs) e;
	    t''' ::: Ht''' <- raiseTy t'';
	    exist _ t''' _
	end (refl_equal _)); subst; intuition;
      try (destruct H; auto; fail).

      add Ht' (Ht _ _ H H0); inversion Ht'; subst.
      destruct H.
      intuition.
    Qed.

    Lemma apartFromCode_slot : forall (ctx : mcontext) (abs : mastate) n,
      bvec_to_nat (ctxSptr ctx) <= bvec_to_nat stackStart + stackLength - 4
      -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (ctxSptr ctx)
      -> n <= bvec_to_nat (ctxSptr ctx)
      -> n < aStackSize abs
      -> apartFromCode (stack_slot ctx n).
      intuition.
      red.
      unfold stack_slot.
      rewrite ntb_invert.
      generalize stackCodeSeparate.
      unfold disjointRegions.
      intuition.
      generalize stackFits; omega.
    Qed.

    Hint Resolve apartFromCode_slot.

    Lemma not_too_high_slot : forall n (ctx : mcontext),
      bvec_to_nat (ctxSptr ctx) <= bvec_to_nat stackStart + stackLength - 4
      -> n <= bvec_to_nat (ctxSptr ctx)
      -> not_too_high (stack_slot ctx n).
      intros.
      unfold stack_slot.
      split.
      rewrite ntb_invert; generalize stackFits; omega.
    Qed.

    Hint Resolve not_too_high_slot.

    Definition checkStoreR (abs : mastate) (rs : SalRegMap.t ty) (e : salExp) :=
      pred_option (forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> apartFromCode (evalExp st e)
	/\ not_too_high (evalExp st e)).

    Definition checkStore : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      checkStoreR abs rs e.
       intros.
       refine (t ::: Ht <- typeof abs rs e;
	match t as t' return (t' = t -> checkStoreR abs rs e) with
	  | Stack n => fun pfEq =>
	    if le_lt_dec 4 n
	      then if le_lt_dec (aStackSize abs) n
		then PNone _
		else PSome _
	      else PNone _
	  | _ => fun _ =>
	    Hcheck <-- checkStore (aCustom abs) (SalRegMap.map lowerTy rs) e;;
	    PSome _
	end (refl_equal _)); subst; intuition;
       add Ht' (Ht _ _ H H0); inversion Ht'; subst;
	 destruct H; intuition eauto;
	   try (assert (apartFromCode (evalExp st e) /\
             not_too_high (evalExp st e) /\ apartFromStack (evalExp st e));
	   [eauto | intuition]);
	   rewrite H3; trivial.
    Qed.

    Definition set_stack32 (m : NatMap.t ty) n t :=
      NatMap.set (n-3) (Top _)
	(NatMap.set (n-2) (Top _)
	  (NatMap.set (n-1) (Top _)
	    (NatMap.set (n+1) (Top _)
	      (NatMap.set (n+2) (Top _)
		(NatMap.set (n+3) (Top _)
		  (NatMap.set n t m)))))).

    Hint Rewrite NatMap.gss : Maps.
    Hint Rewrite NatMap.gso using (trivial; fail) : Maps.

    Lemma set_stack32_sound : forall (ctx : mcontext) st (abs : mastate) src tSrc n (cust : Sys.mastate),
      bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (ctxSptr ctx)
      -> bvec_to_nat (ctxSptr ctx) <= bvec_to_nat stackStart + stackLength - 4
      -> conc_slots Sys.hasTy ctx st abs
      -> hasTy ctx src tSrc
      -> n < aStackSize abs
      -> conc_slots Sys.hasTy ctx (set_mem32 st (stack_slot ctx n) src)
      (Build_stackState (aIsRet abs) (aStackSize abs)
        (set_stack32 (aStackSlots abs) n tSrc) cust).
      intros.
      intros slot Hslot; simpl in Hslot.
      simpl.
      unfold set_stack32; simpl.

      repeat match goal with
	       | [ |- context[NatMap.get ?slot (NatMap.set ?n _ _)] ] =>
		 destruct (eq_nat_dec slot n); subst; autorewrite with Maps; intuition
	     end.
      rewrite invert_set_mem32; trivial.
      rewrite invert_set_mem32_neq; auto.

      intuition.
      intuition.
      apply apart_or.
      unfold stack_slot.
      rewrite ntb_invert; [rewrite ntb_invert | idtac].
      omega.
      generalize stackFits; omega.
      generalize stackFits; omega.
    Qed.

    Hint Resolve set_stack32_sound.

    Lemma agree_range'_write : forall m1 m2 a1 a2 a v,
      agree_range' m1 m2 a1 a2
      -> bvec_to_nat a + 4 <= bvec_to_nat a1
      -> not_too_high a
      -> agree_range' (set_mem32' m1 a v) m2 a1 a2.
      intros until v; intros Hagree Hlt Hhigh a' Ha'.
      rewrite invert_set_mem32'_neq'; auto.
      apply apart_or; omega.
    Qed.

    Hint Resolve agree_range'_write.

    Ltac use_gmap :=
      match goal with
	| [ |- context[SalRegMap.get _ (SalRegMap.map _ _)] ] =>
	  rewrite SalRegMap.gmap
      end.
    Hint Extern 1 (Sys.hasTy _ _ _) => use_gmap.

    Lemma apartFromStack_heap_write : forall dst ctx st (abs : mastate) src (cust : Sys.mastate),
      apartFromStack dst
      -> not_too_high dst
      -> bvec_to_nat (ctxSptr ctx) <= bvec_to_nat stackStart + stackLength - 4
      -> bvec_to_nat stackStart + aStackSize abs < bvec_to_nat (ctxSptr ctx)
      -> conc_slots Sys.hasTy ctx st abs
      -> conc_slots Sys.hasTy ctx (set_mem32 st dst src)
      (Build_stackState (aIsRet abs) (aStackSize abs) (aStackSlots abs) cust).
      unfold conc_slots; simpl.
      intros.
      rewrite invert_set_mem32_neq; eauto.
      apply not_too_high_slot; eauto; omega.

      red in H.
      apply apart_or.
      unfold stack_slot.
      rewrite ntb_invert.

      generalize stackHigh; omega.

      generalize stackFits; omega.
    Qed.

    Lemma apartFromStack_agree_range' : forall dst m1 m2 src sptr ssize,
      not_too_high dst
      -> apartFromStack dst
      -> bvec_to_nat sptr <= bvec_to_nat stackStart + stackLength - 4
      -> bvec_to_nat stackStart + ssize < bvec_to_nat sptr
      -> agree_range' m1 m2 sptr stackEnd
      -> agree_range' (set_mem32' m1 dst src) m2 sptr stackEnd.
      unfold agree_range', apartFromStack, not_too_high.
      intros.
      rewrite invert_set_mem32'_neq'; auto.
      apply apart_or.

      unfold stackEnd in H4.
      rewrite ntb_invert in H4.
      generalize stackHigh; omega.
      generalize stackFits; omega.
    Qed.

    Definition effectOfCmdR (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd) :=
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> mconc ctx (salExecCmd st cmd) abs').

    Definition effectOfCmd : forall (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd),
      effectOfCmdR abs rs cmd.
      intros.
      refine (match cmd as cmd' return (cmd' = cmd -> effectOfCmdR abs rs cmd) with
		| SStore dst src => fun pfEq =>
		  tDst ::: HtDst <- typeof abs rs dst;
		  match tDst as tDst' return (tDst' = tDst -> effectOfCmdR abs rs cmd) with
		    | Stack n => fun pfEq =>
		      if le_lt_dec 4 n
			then if le_lt_dec (aStackSize abs) n
			  then SNone _
			  else (tSrc ::: HtSrc <- typeof abs rs src;
			    cust :: Hcust <- doStackWrite (aCustom abs);
			    SSome _ (Build_stackState
			      (aIsRet abs) (aStackSize abs)
			      (set_stack32 (aStackSlots abs) n tSrc)
			      cust) _)
			else SNone _
		    | _ => fun pfEq =>
		      cust :: Hcust <- effectOfCmd (aCustom abs) (SalRegMap.map lowerTy rs) cmd;
		      Hcheck <-- Sys.checkStore (aCustom abs) (SalRegMap.map lowerTy rs) dst;
		      SSome _ (Build_stackState (aIsRet abs) (aStackSize abs)
			(aStackSlots abs) cust) _
		  end (refl_equal _)
		| _ => fun pfEq =>
		  cust :: Hcust <- effectOfCmd (aCustom abs) (SalRegMap.map lowerTy rs) cmd;
		  SSome _ (Build_stackState (aIsRet abs) (aStackSize abs) (aStackSlots abs) cust) _
	      end (refl_equal _)); subst; simpl in *; intuition;
      try match goal with
	| [ |- context[Build_stackState (aIsRet ?abs) (aStackSize ?abs)
	  (aStackSlots ?abs) ?cust] ] =>
	destruct H; split; simpl;
	  [intuition | intuition | intuition
	    | assert (apartFromCode (evalExp st dst) /\
              not_too_high (evalExp st dst) /\ apartFromStack (evalExp st dst));
	    [eauto
	      | destruct H as [Ha [Hb Hc]];
		apply apartFromStack_heap_write; eauto]
	    | assert (apartFromCode (evalExp st dst) /\
              not_too_high (evalExp st dst) /\ apartFromStack (evalExp st dst));
	    [eauto
	      | destruct H as [Ha [Hb Hc]];
		eapply apartFromStack_agree_range'; eauto]
	    | intuition]
      end;
      try (add Ht' (HtDst _ _ H H0); inversion Ht'; subst);
	try (unfold set_arith_flags; destruct s);
	  destruct H; intuition (auto 6).

      split; simpl; intuition.
      simpl; apply agree_range'_write; auto.
      unfold stack_slot.
      rewrite ntb_invert; generalize stackFits; omega.

      apply Hcust; auto.
      red.
      unfold stack_slot.
      rewrite ntb_invert.
      omega.
      generalize stackFits; omega.
    Qed.

    Definition effectOfJumpR (abs : mastate) (rs : SalRegMap.t ty)
      (jmp : salJump) (nextPc : int32) :=
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> match salExecJmp jmp nextPc st with
	     | None => False
	     | Some st' => mconc ctx st' abs'
	   end).

    Definition effectOfJump : forall (abs : mastate) (rs : SalRegMap.t ty)
      (jmp : salJump) (nextPc : int32),
      effectOfJumpR abs rs jmp nextPc.
      intros.
      refine (cust :: Hcust <- effectOfJump (aCustom abs) (SalRegMap.map lowerTy rs)
	jmp nextPc;
	match jmp as jmp' return (jmp' = jmp -> effectOfJumpR abs rs jmp nextPc) with
	  | SFail => fun _ => SNone _
	  | SIndirectJump SRet e => fun pfEq =>
	    t ::: Ht <- typeof abs rs e;
	    match t as t' return (t' = t -> _) with
	      | RetPtr => fun pfEq' =>
		SSome _ (Build_stackState true (aStackSize abs) (aStackSlots abs) cust) _
	      | _ => fun _ => SNone _
	    end (refl_equal _)
	  | _ => fun pfEq =>
	    SSome _ (Build_stackState false (aStackSize abs) (aStackSlots abs) cust) _
	end (refl_equal _)); subst; intros; simpl in *; destruct H;
      generalize (Hcust (ctxCust ctx) st); try destruct (check_cc st c);
	split; intuition; simpl; try discriminate.
    
      assert (He : hasTy ctx (evalExp st e) (RetPtr Sys.ty)); intuition.
      inversion He; trivial.
    Qed.

    Definition effectOfConditionalJump : forall (abs : stateDescription ty mastate)
      (co : cc) (pc nextPc : int32) (b : bool),
      sig_option (fun abs' => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> check_cc st co = b
	-> match salExecJmp (SConditionalJump co pc) nextPc st with
	     | None => False
	     | Some st' => typeMconc hasTy mconc ctx st' abs'
	   end).
      intros.

      refine (abs' :: Habs' <- effectOfConditionalJump abs co pc nextPc b;
	SSome _ abs' _); intuition.
    Qed.

    Definition tyTop : ty := Top _.
    Theorem tyTopSound : forall ctx v, hasTy ctx v tyTop.
      unfold tyTop; intuition.
    Qed.

    Definition tyConst (n : int32) : ty := Const _ n.
    Hint Unfold tyConst.
    Theorem tyConstSound1 : forall ctx v, hasTy ctx v (tyConst v).
      unfold tyConst; intuition.
    Qed.
    Theorem tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
      unfold tyConst.
      inversion 1; trivial.
    Qed.
    Definition isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).
      intros.
      refine (match t as t return sig_option (fun n => t = tyConst n) with
		| Const n => SSome _ n _
		| _ => SNone _
	      end); trivial.
    Qed.

    Lemma agree_range'_refl : forall m a1 a2,
      agree_range' m m a1 a2.
      red; intuition.
    Qed.

    Hint Resolve agree_range'_refl.

    Definition shiftContext (args : list ty) (ctx : mcontext) (st : salState) ss oldebp retptr :=
      Build_stackContext
      (stack_slot ctx (ss - 4 - 4 * length args))
      (SstMem st)
      oldebp
      retptr
      (ctxCust ctx).

    Fixpoint argsStack (args : list ty) : NatMap.t ty :=
      match args with
	| nil => NatMap.init (Top _)
	| cons t rest => NatMap.set (4 * length args) t (argsStack rest)
      end.

    Lemma undo_rewrite_fourx : forall n,
      n + (n + (n + (n + 0))) = 4 * n.
      intros; omega.
    Qed.

    Import Datatypes.

    Lemma undo_rewrite_fourx_plus : forall n,
      S (n + (S (n + S (n + (S (n + 0)))))) = 4 * (S n).
      intros; omega.
    Qed.

    Hint Rewrite undo_rewrite_fourx undo_rewrite_fourx_plus : fixup.
    Ltac mysimpl := simpl; autorewrite with fixup.

    Ltac clear_one :=
      match goal with
	[ H : _ |- _ ] => clear H
      end.
    Ltac clear_all := repeat clear_one.

    Definition robustType : forall (t : ty),
      pred_option (forall ctx v, hasTy ctx v t
	-> forall args st ss oldebp retptr, hasTy (shiftContext args ctx st ss oldebp retptr) v t).
      intros.
      refine (match t as t return (pred_option (forall ctx v, hasTy ctx v t
	-> forall args st ss oldebp retptr, hasTy (shiftContext args ctx st ss oldebp retptr) v t)) with
		| Stack n => PNone _
		| OldEbp => PNone _
		| RetPtr => PNone _
		| _ => PSome _
	      end); intuition; inversion H; intuition.
    Qed.

    Lemma checkArgs_add : forall n1 n2 n3 n4 n5,
      n1 <= n2
      -> n3 + n2 <= n4
      -> n2 <= n4 - n3
      -> n5 - (n4 - n3 - n1) = n5 - (n4 - n3 - n2) - (n2 - n1).
      intros.
      repeat rewrite <- rewrite_minus_plus.
      replace (n4 - (n3 + n2) + (n2 - n1))
	with (n4 - n3 - n1).
      repeat rewrite rewrite_minus_plus.
      trivial.
      clear H0; omega.
    Qed.

    Definition checkArgsR (args args' : list ty) (slots : NatMap.t ty) ssize ss :=
      sig_option (fun slots' => forall ctx st,
	bvec_to_nat (ctxSptr ctx) <= bvec_to_nat stackStart + stackLength
	-> (forall n, 4 <= n < ssize
	  -> hasTy ctx (get_mem32 st (stack_slot ctx n)) (NatMap.get n slots))
	-> forall oldebp retptr,
	  forall n, 4 <= n < ssize - (ss - 4 - 4 * length args)
	    -> hasTy (shiftContext args ctx st ss oldebp retptr) (get_mem32 st
	      (stack_slot (shiftContext args ctx st ss oldebp retptr) n)) (NatMap.get n slots')).

    Definition checkArgs' : forall (args args' : list ty) (Hargs : length args' <= length args)
      (slots : NatMap.t ty) ssize ss,
      4 <= ss - 4 * length args' /\ ss < ssize /\ 4 + 4 * length args <= ss
      -> checkArgsR args args' slots ssize ss.
      refine (fix checkArgs (args args' : list ty) (Hargs' : length args' <= length args)
	(slots : NatMap.t ty) (ssize ss : nat)
	(Hss : 4 <= ss - 4 * length args' /\ ss < ssize /\ 4 + 4 * length args <= ss)
	{struct args'} : checkArgsR args args' slots ssize ss :=
	match args' as args'' return (args'' = args' -> _) with
	  | nil => fun pfEq => SSome _ (NatMap.init (Top _)) _
	  | cons t rest => fun pfEq =>
	    Hrobust <-- robustType t;
	    Hsub <-- subTy (NatMap.get (ss - 4 - 4 * length rest) slots) t;
	    slots' :: Hslots' <- checkArgs args rest _ slots ssize ss _;
	    SSome _ (NatMap.set (4 * length args - 4 * length rest) t slots') _
	end (refl_equal _)); subst; intuition.
      rewrite NatMap.gi; auto.

      simpl length in H; omega.

      rewrite NatMap.gsspec.
      destruct (NatIndexed.eq n (4 * length args - 4 * length rest)); subst; intuition.
      replace (stack_slot (shiftContext args ctx st ss oldebp retptr)
        (4 * length args - 4 * length rest))
      with (stack_slot ctx
        (ss - 4 - 4 * length rest)).
      apply Hrobust; apply Hsub; apply H3.
      simpl length in H; generalize H H1; clear_all; intros; omega.

      unfold stack_slot.
      simpl ctxSptr; autorewrite with fixup.
      unfold stack_slot.
      rewrite ntb_invert.
      apply ntb_cong.
      simpl length in Hargs'.
      apply checkArgs_add; trivial.
      generalize Hargs'; clear_all; intros; omega.
      generalize H H2; clear_all; intros; omega.

      generalize stackFits H0; clear_all; intros; omega.
    Qed.

    Definition checkArgs : forall (args : list ty) (slots : NatMap.t ty) ssize ss,
      4 <= ss - 4 * length args /\ ss < ssize /\ 4 + 4 * length args <= ss
      -> checkArgsR args args slots ssize ss.
      intros.
      refine (checkArgs' args args _ slots (ssize := ssize) (ss := ss) _); subst; intuition.
    Qed.

    Lemma shiftAbs_lt' : forall n1 n2 ss,
      n1 < n2
      -> ss < n1
      -> ss < n2
      -> n1 - ss < n2 - ss.
      intros; omega.
    Qed.

    Lemma shiftAbs_lt'' : forall n1 n2 n3,
      n3 < n2
      -> n1 + n2 - n3 = n1 + (n2 - n3).
      intros; omega.
    Qed.

    Lemma shiftAbs_lt : forall start size sptr ss n1 n2,
      start + size < sptr
      -> ss < size
      -> start +
      (size - (ss - n1 - n2)) <
      sptr - (ss - n1 - n2).
      intros.
      replace (start + (size - (ss - n1 - n2)))
	with (start + size - (ss - n1 - n2)).
      apply shiftAbs_lt'; omega.
      apply shiftAbs_lt''; omega.
    Qed.

    Lemma rewrite_plus_four : forall n,
      S (S (S (S n))) = 4 + n.
      trivial.
    Qed.

    Lemma HT_Stack' : forall ctx n v,
      n <= bvec_to_nat (ctxSptr ctx)
      -> v = stack_slot ctx n
      -> hasTy ctx v (Stack _ n).
      intros; subst; auto.
    Qed.

    Lemma Some_get_mem32_cong : forall st e e',
      e = e'
      -> Some (get_mem32 st e) = Some (get_mem32 st e').
      congruence.
    Qed.

    Lemma viewShift_eq : forall ss size sptr n1 n2,
      ss <= size
      -> n1 + n2 <= ss
      -> sptr - (ss - n1 - n2) - (n1 + n2) = sptr - ss.
      intros.
      rewrite <- rewrite_minus_plus.
      omega.
    Qed.

    Definition shiftAbs : forall (fi : function_info ty) ss (abs : mastate),
      sig_option (fun abs' => forall ctx st oldebp,
	mconc ctx st abs
	-> mconc (shiftContext (fiArgs fi) ctx st ss oldebp
	  (Some (get_mem32 st (stack_slot ctx ss)))) st abs').
      intros.
      refine (if le_lt_dec (aStackSize abs) ss
	then SNone _
	else if le_lt_dec 4 (ss - 4 * length (fiArgs fi))
	  then if le_lt_dec (4 + 4 * length (fiArgs fi)) ss
	    then slots :: Hslots <- checkArgs (fiArgs fi) (aStackSlots abs)
	      (ssize := aStackSize abs) (ss := ss) _;
	      SSome _ (Build_stackState false
		(aStackSize abs - (ss - 4 - 4 * length (fiArgs fi)))
		(NatMap.set (4 + 4 * length (fiArgs fi)) (RetPtr _) slots)
		(aCustom abs)) _
	    else SNone _
	  else SNone _); intuition.

      destruct H.
      split; intuition; try discriminate.

      simpl ctxSptr; autorewrite with fixup.
      unfold stack_slot.
      rewrite ntb_invert;
	generalize mcSptr0 stackFits; clear_all; intros; omega.

      simpl aStackSize; simpl ctxSptr; autorewrite with fixup.
      unfold stack_slot.
      rewrite ntb_invert.
      assert (Hss : ss < aStackSize abs); trivial.
      apply shiftAbs_lt; trivial.
      generalize mcSptr0 stackFits; clear_all; intros; omega.

      intros sl Hsl.
      simpl aStackSlots; autorewrite with fixup.
      rewrite rewrite_plus_four.
      rewrite NatMap.gsspec.
      destruct (NatIndexed.eq sl (4 + 4 * length (fiArgs fi))); subst.

      apply HT_RetPtr.
      simpl; autorewrite with fixup; rewrite rewrite_plus_four.
      apply Some_get_mem32_cong.
      unfold stack_slot.
      simpl ctxSptr; autorewrite with fixup.
      unfold stack_slot.
      rewrite ntb_invert.
      match goal with
	| [ |- nat_to_bvec 32 ?X = nat_to_bvec 32 ?Y] =>
	  assert (X = Y); try congruence
      end.
      symmetry.
      eapply viewShift_eq; eauto.

      generalize mcSptr0 stackFits; clear_all; intros; omega.

      apply Hslots; trivial.
      generalize mcSptr0; clear_all; intros; omega.
    Qed.

    (** * We build up a fast look-up structure to find function information
          from program counters. *)
    Definition pcToFi_map' (fis : list (function_info ty)) m :=
      fold_left (fun m fi => Int32Map.set (fiPc fi) (Some fi) m) fis m.

    Definition pcToFi_map := pcToFi_map' functions (Int32Map.init None).
    Definition pcToFi pc := Int32Map.get pc pcToFi_map.

    Lemma viewShift_boring : forall abs hyps,
      forall (ctx : stackContext Sys.mcontext) (st : salState),
	typeMconc hasTy mconc ctx st abs ->
	exists ctx' : stackContext Sys.mcontext,
	  typeMconc hasTy mconc ctx' st abs /\
	  AllS
	  (fun
            need : stateDescription (stackTy Sys.ty)
            (stackState Sys.ty Sys.mastate) =>
            exists
              have : stateDescription (stackTy Sys.ty)
              (stackState Sys.ty Sys.mastate),
              (False \/ In have hyps) /\
              subsumes (typeMconc hasTy mconc) ctx' need ctx have) hyps.
      intuition.
      exists ctx.
      intuition.
      clear_all.
      induction hyps; simpl; intuition.
      apply allS_cons.

      exists (Build_stateDescription sdPc sdRegs
        (Build_stackState aIsRet0 aStackSize0 aStackSlots0 aCustom0)); intuition.
      red; trivial.

      eapply AllS_imp; [idtac | apply IHhyps].
      intuition.
      destruct H as [have Hhave].
      exists have.
      intuition.
    Qed.

    Hint Resolve viewShift_boring.

    Fixpoint preserveSlots (sl : nat) (slots : NatMap.t ty) {struct sl} : NatMap.t ty :=
      match sl with
	| O => NatMap.init (Top _)
	| S sl' => NatMap.set sl' (NatMap.get sl' slots) (preserveSlots sl' slots)
      end.

    Definition reverseRobustType : forall (t : ty),
      pred_option (forall ctx v args st ss oldebp retptr, hasTy (shiftContext args ctx st ss oldebp retptr) v t
	  -> hasTy ctx v t).
      intros.
      refine (match t as t return (pred_option (forall ctx v args st ss oldebp retptr,
	hasTy (shiftContext args ctx st ss oldebp retptr) v t
	-> hasTy ctx v t)) with
		| Stack n => PNone _
		| OldEbp => PNone _
		| RetPtr => PNone _
		| _ => PSome _
	      end); intuition; inversion H; intuition.
    Qed.

    Lemma viewShift_ret_eq : forall n1 n2 n3 n4,
      n2 <= n1
      -> n3 + n4 <= n2
      -> n1 - (n2 - n3 - n4) - n4 = n1 - (n2 - n3).
      intros.
      rewrite <- rewrite_minus_plus.
      omega.
    Qed.

    Lemma agree_range'_slot : forall m1 m2 bound finish a,
      belowMax finish 3
      -> agree_range' m1 m2 bound finish
      -> bvec_to_nat bound <= bvec_to_nat a < bvec_to_nat finish - 3
      -> get_mem32' m2 a = get_mem32' m1 a.
      intros until a; intros Hfinish Hagree Hlt.
      unfold get_mem32'.
      repeat rewrite Hagree; trivial.

      replace (binc (binc (binc a))) with (binc3 a); trivial.
      rewrite binc3_btn_belowMax.
      omega.
      split; destruct Hfinish; omega.

      replace (binc (binc a)) with (binc2 a); trivial.
      rewrite binc2_btn_belowMax.
      omega.
      split; destruct Hfinish; omega.

      rewrite binc_btn_belowMax.
      omega.
      split; destruct Hfinish; omega.

      omega.
    Qed.

    Lemma minus_lt : forall n m o,
      n < m - 0
      -> n - o < m.
      intros; omega.
    Qed.

    Lemma minus_lt2 : forall n m o p,
      n - m <= o < p
      -> n - S m <= o < p.
      intros; omega.
    Qed.

    Lemma agree_range'_S : forall m1 m2 (ctx : mcontext) bound finish,
      belowMax (ctxSptr ctx) 0
      -> agree_range' m1 m2 (stack_slot ctx (S bound)) finish
      -> agree_range' m1 m2 (stack_slot ctx bound) finish.
      intros until finish; intros Hsptr Hagree addr Haddr.
      apply Hagree.
      unfold stack_slot.
      rewrite ntb_invert.
      unfold stack_slot in Haddr.
      rewrite ntb_invert in Haddr.
      apply minus_lt2; trivial.
      apply minus_lt; destruct Hsptr; trivial.
      apply minus_lt; destruct Hsptr; trivial.
    Qed.

    Lemma belowMax_stackEnd : forall (n : int32),
      bvec_to_nat n < bvec_to_nat stackEnd - 3
      -> belowMax n 0.
      intros.
      unfold stackEnd in H.
      rewrite ntb_invert in H.
      split.
      generalize stackFits; omega.
      generalize stackFits; omega.
    Qed.

    Hint Resolve belowMax_stackEnd.

    Lemma bound_sptr : forall sptr bound,
      sptr < bvec_to_nat stackEnd - 3
      -> sptr - bound < pow2 32.
      intros.
      unfold stackEnd in H.
      rewrite ntb_invert in H.
      generalize stackFits; omega.
      generalize stackFits; omega.
    Qed.

    Hint Resolve bound_sptr.

    
    Lemma agree_range'_weaken : forall m1 m2 s1 s2 e,
      agree_range' m1 m2 s1 e
      -> bvec_to_nat s1 <= bvec_to_nat s2
      -> agree_range' m1 m2 s2 e.
      intros; intros a Ha.
      apply H.
      omega.
    Qed.

    Lemma viewShift_concSlots : forall ctx ssize st st' bound slots,
      bvec_to_nat (ctxSptr ctx) < bvec_to_nat stackEnd - 3
      -> agree_range' (SstMem st') (SstMem st) (stack_slot ctx (bound - 1)) stackEnd
      -> (forall n,
	4 <= n < ssize ->
	stackHasTy Sys.hasTy ctx (get_mem32 st (stack_slot ctx n))
	(NatMap.get n slots))
      -> forall n,
	4 <= n < ssize ->
	stackHasTy Sys.hasTy ctx (get_mem32 st' (stack_slot ctx n))
	(NatMap.get n
          (preserveSlots bound slots)).
      induction bound; intros.
      
      simpl; rewrite NatMap.gi; auto.

      simpl; rewrite NatMap.gsspec.
      destruct (NatIndexed.eq n bound); subst.
      clear IHbound.
      replace (get_mem32 st' (stack_slot ctx bound))
	with (get_mem32 st (stack_slot ctx bound)); [auto | idtac].
      unfold get_mem32.
      apply agree_range'_slot with (stack_slot ctx bound) stackEnd.
      split.
      unfold stackEnd.
      rewrite ntb_invert.
      apply stackFits.
      generalize stackFits; omega.
      intros addr Haddr.
      assert (agree_range' (SstMem st') (SstMem st) (stack_slot ctx bound)
        stackEnd).
      replace (S bound - 1) with bound in H0; auto.
      omega.
      auto.

      split; try omega.
      unfold stack_slot.
      repeat rewrite ntb_invert.
      omega.
      auto.

      apply IHbound; auto.
      apply agree_range'_weaken with (stack_slot ctx (S bound - 1)); trivial.
      unfold stack_slot.
      repeat rewrite ntb_invert; eauto; omega.
    Qed.

    Lemma agree_range'_trans : forall m1 m2 m3 s e,
      agree_range' m1 m2 s e
      -> agree_range' m2 m3 s e
      -> agree_range' m1 m3 s e.
      intros; intros addr Haddr.
      rewrite H; auto.
    Qed.

    Definition viewShift_ret : forall (fi : function_info ty)
      (abs : stateDescription ty mastate) ss rp,
      ss <= aStackSize (sdCust abs)
      -> 4 + 4 * length (fiArgs fi) <= ss
      -> sig_option (fun retAbss => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> subsumes (typeMconc hasTy mconc)
	(shiftContext (fiArgs fi) ctx st ss (get_reg st (SX86 EBP))
	  (Some rp)) (fst retAbss)
	ctx (snd retAbss)).
      intros fi abs ss rp Hss Hargs.
      refine (Hrobust <-- reverseRobustType (fiRet fi);
	SSome _ (Build_stateDescription
	  ApcUnknown
	  (SalRegMap.set (SX86 EAX) (fiRet fi)
	    (SalRegMap.set (SX86 EBP) (OldEbp _)
	      (SalRegMap.set (SX86 ESP) (Stack _ (4 * length (fiArgs fi)))
		(SalRegMap.init (Top _)))))
	  (Build_stackState (ty := Sys.ty)
	    true
	    0
	    (NatMap.init (Top _))
	    (aCustom (sdCust abs))),
	  Build_stateDescription
	  (ApcConst rp)
	  (SalRegMap.set (SX86 EAX) (fiRet fi)
	    (SalRegMap.set (SX86 EBP) (SalRegMap.get (SX86 EBP) (sdRegs abs))
	      (SalRegMap.set (SX86 ESP) (Stack _ (ss - 4))
		(SalRegMap.init (Top _)))))
	  (Build_stackState
	    false
	    (aStackSize (sdCust abs))
	    (preserveSlots (ss - 3 - 4 * length (fiArgs fi)) (aStackSlots (sdCust abs)))
	    (aCustom (sdCust abs)))) _).
      simpl; autorewrite with fixup; intros ctx st Hconc st' Hconc'; intuition.

      destruct Hconc.
      destruct mcCust0.
      destruct Hconc'.
      destruct mcCust1.
      split; simpl; autorewrite with fixup; intuition.

      simpl in H; congruence.

      rewrite SalRegMap.gsspec.
      destruct (SalRegIndexed.eq r (SX86 EAX)); subst.
      add Heax (mcRegs0 (SX86 EAX)).
      simpl in Heax.
      rewrite SalRegMap.gss in Heax.
      eauto.

      rewrite SalRegMap.gsspec.
      destruct (SalRegIndexed.eq r (SX86 EBP)); subst.
      add Hebp (mcRegs0 (SX86 EBP)).
      simpl in Hebp.
      rewrite SalRegMap.gso in Hebp; try discriminate.
      rewrite SalRegMap.gss in Hebp.
      inversion Hebp; subst.
      eauto.

      rewrite SalRegMap.gsspec.
      destruct (SalRegIndexed.eq r (SX86 ESP)); subst.
      add Hesp (mcRegs0 (SX86 ESP)).
      simpl in Hesp.
      rewrite SalRegMap.gso in Hesp; try discriminate.
      rewrite SalRegMap.gso in Hesp; try discriminate.
      rewrite SalRegMap.gss in Hesp.
      apply HT_Stack'.
      generalize Hss mcStackSize0; clear_all.
      unfold ty, mastate.
      intros; omega.
      rewrite undo_rewrite_fourx in Hesp.
      inversion Hesp; subst.
      rewrite undo_rewrite_fourx.

      unfold stack_slot; simpl ctxSptr; autorewrite with fixup.
      unfold stack_slot.
      rewrite ntb_invert.
      apply ntb_cong.
      apply viewShift_ret_eq; trivial.
      generalize Hss mcStackSize0; clear_all.
      unfold ty, mastate.
      intros; omega.

      generalize stackFits mcSptr0; clear_all; intros; omega.

      rewrite SalRegMap.gi; auto.

      split; intuition; try discriminate.

      red; simpl; autorewrite with fixup.
      intros.
      simpl in mcMemIn1, mcStackSlots1.
      rewrite undo_rewrite_fourx in mcMemIn1.
      apply viewShift_concSlots with (aStackSize (sdCust abs)) st; eauto.

      unfold stackEnd.
      rewrite ntb_invert.
      generalize mcSptr0 stackHigh; clear_all; intros; omega.
      generalize stackFits; clear_all; intros; omega.

      replace (ss - 3 - 4 * length (fiArgs fi) - 1)
	with (ss - 4 - 4 * length (fiArgs fi)); [trivial | clear_all; omega].

      simpl in mcMemIn1.
      rewrite undo_rewrite_fourx in mcMemIn1.
      apply agree_range'_trans with (SstMem st); trivial.
      apply agree_range'_weaken with (stack_slot ctx (ss - 4 - 4 * length (fiArgs fi))); trivial.

      unfold stack_slot.
      rewrite ntb_invert.
      clear_all; omega.
      generalize mcSptr0 stackFits; clear_all; intros; omega.
    Qed.

    Definition viewShiftOfJump : forall (hyps : list (stateDescription ty mastate))
      (abs : stateDescription ty mastate)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun vs => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> exists ctx', typeMconc hasTy mconc ctx' st (vsAbs vs)
	  /\ AllS (fun need => exists have, (In have (vsSuccs vs) \/ In have hyps)
	    /\ subsumes (typeMconc hasTy mconc) ctx' need ctx have) (vsHyps vs)).
      intros hyps abs jmp nextPc.
      refine (match jmp with
		| SFallthru =>
		  match SalRegMap.get (SX86 ESP) (sdRegs abs) as t
		    return (t = SalRegMap.get (SX86 ESP) (sdRegs abs) -> _) with
		    | Const n => fun pfEq =>
		      if le_lt_dec (bvec_to_nat n) (bvec_to_nat stackStart)
			then SNone _
			else if le_lt_dec (bvec_to_nat stackStart + stackLength - 4) (bvec_to_nat n)
			  then SNone _
			  else SSome _
			    (Build_viewShift
			      nil
			      nil
			      (Build_stateDescription
				(sdPc abs)
				(SalRegMap.set (SX86 ESP) (Stack _ 0) (SalRegMap.init (Top _)))
				(Build_stackState (aIsRet (sdCust abs))
				  (pred (bvec_to_nat n - bvec_to_nat stackStart))
				  (NatMap.init (Top _)) (aCustom (sdCust abs)))))
			    _
		    | _ => fun _ => SNone _
		  end (refl_equal _)
		| SIndirectJump SCall (SConst tgt) =>
		  match pcToFi tgt with
		    | None => SNone _
		    | Some fi =>
		      match SalRegMap.get (SX86 ESP) (sdRegs abs) as tEsp
			return (tEsp = SalRegMap.get (SX86 ESP) (sdRegs abs) -> _) with
			| Stack ss => fun pfEq =>
			  match NatMap.get ss (aStackSlots (sdCust abs)) as tSlot
			    return (tSlot = NatMap.get ss (aStackSlots (sdCust abs)) -> _) with
			    | Const rp => fun pfEq' =>
			      if le_lt_dec (aStackSize (sdCust abs)) ss
				then SNone _
				else if le_lt_dec (4 + 4 * length (fiArgs fi)) ss
				  then abs' :: Habs' <- shiftAbs fi ss (sdCust abs);
				    rets :: Hrets <- viewShift_ret fi abs (ss := ss) rp _ _;
				    SSome _
				    (Build_viewShift
				      (cons (snd rets) nil)
				      (cons (fst rets) nil)
				      (Build_stateDescription
					(sdPc abs)
					(SalRegMap.set (SX86 EBP) (OldEbp _)
					  (SalRegMap.set (SX86 ESP)
					    (Stack _ (4 + 4 * length (fiArgs fi)))
					    (SalRegMap.init (Top _))))
					abs'))
				    _
				  else SNone _
			    | _ => fun _ => SNone _
			  end (refl_equal _)
			| _ => fun _ => SNone _
		      end (refl_equal _)
		  end
		| _ => SSome _ (Build_viewShift nil hyps abs) _
	      end); simpl; eauto.

      (* Setting ESP for the first time *)
      intros ctx st Hconc; simpl.
      exists (Build_stackContext n (SstMem st) (ctxOldEbp ctx) (ctxRetPtr ctx) (ctxCust ctx)).
      destruct Hconc.
      intuition.
      split; simpl; intuition.

      rewrite SalRegMap.gsspec.
      destruct (SalRegIndexed.eq r (SX86 ESP)); subst.
      unfold ty, mastate in pfEq.
      add Hr (mcRegs (SX86 ESP)).
      rewrite <- pfEq in Hr.
      inversion Hr; subst.
      apply HT_Stack'; intuition.
      unfold stack_slot.
      rewrite <- minus_n_O.
      rewrite btn_invert; trivial.

      rewrite SalRegMap.gi; intuition.

      destruct mcCust0.
      split; simpl; intuition.
      intro.
      simpl.
      rewrite NatMap.gi.
      auto.
      intuition.

      (* Function call *)
      intuition.
      exists (shiftContext (fiArgs fi) ctx st ss
	(get_reg st (SX86 EBP))
	(Some (get_mem32 st (stack_slot ctx ss)))).
      simpl; intuition; autorewrite with fixup.

      destruct H.
      destruct mcCust0.
      split; simpl; autorewrite with fixup; intuition.

      rewrite SalRegMap.gsspec.

      destruct (SalRegIndexed.eq r (SX86 EBP)); subst.
      replace (get_reg st (SX86 EBP))
	with (ctxOldEbp (shiftContext (fiArgs fi) ctx st ss (get_reg st (SX86 EBP))
          (Some (get_mem32 st (stack_slot ctx ss))))); auto.
      
      rewrite SalRegMap.gsspec.
      destruct (SalRegIndexed.eq r (SX86 ESP)); subst.
      add Hesp (mcRegs (SX86 ESP)).
      unfold ty, mastate in pfEq.
      rewrite <- pfEq in Hesp.
      inversion Hesp; subst.
      apply HT_Stack'.

      simpl ctxSptr; autorewrite with fixup.
      unfold stack_slot.
      rewrite ntb_invert.
      destruct H.
      omega.
      destruct H; generalize stackFits; omega.

      unfold stack_slot.
      simpl ctxSptr; autorewrite with fixup.
      unfold stack_slot.
      rewrite ntb_invert.
      apply ntb_cong.
      symmetry.
      rewrite (rewrite_plus_four (4 * length (fiArgs fi))).
      eapply viewShift_eq; eauto.
      destruct H; generalize stackFits; omega.

      rewrite SalRegMap.gi; intuition.
    
      apply allS_cons; trivial.
      exists (snd rets).
      intuition.
      replace (get_mem32 st (stack_slot ctx ss))
	with rp.
      auto.
      add Hrp (mcStackSlots (TypeSystem.mcCust H) (n := ss)).
      assert (Hrange : 4 <= ss < aStackSize (sdCust abs)).
      omega.
      add Hrp' (Hrp Hrange).
      unfold ty, mastate in pfEq'.
      rewrite <- pfEq' in Hrp'.
      generalize Hrp'; clear_all; intros.
      generalize dependent (get_mem32 st (stack_slot ctx ss)).
      inversion 1; trivial.
    Qed.
  End AddStackTypes.

End Make.
