(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** A mock-up type system with only a Top type *)

Require Import List TheoryList.

Require Import Asm.Util.Specif.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.AbsInt.Sal.
Require Import ProofOS.AbsInt.TypeSystem.
Require Import ProofOS.AbsInt.StackTypes.

Set Implicit Arguments.

Parameter print_int32_option : option int32 -> unit.
Parameter print_nostate : unit -> unit.

Module Make (Param : M86_PARAM_FIXED_CODE_AND_STACK).
  Module Stack := StackTypes.Make(Param).
  Import Stack.
  Import TySys.
  Import Fixed.
  Import MySal.

  Module MyTypeSystem <: STACK_READY_TYPE_SYSTEM.
    Module Mac := M.

    Definition ty := option int32.
    Definition print_ty := print_int32_option.
    Definition mastate := unit.
    Definition print_mastate := print_nostate.
    Definition mcontext := unit.
    Definition mconc (ctx : mcontext) (st : Mac.mstate) (abs : mastate) := True.

    Definition hasTy (ctx : mcontext) (v : int32) (t : ty) :=
      match t with
	| None => True
	| Some n => v = n
      end.
    Definition tyTop : ty := None.
    Hint Unfold hasTy.

    Lemma tyTopSound' : forall ctx v, hasTy ctx v None.
      intros; red; trivial.
    Qed.

    Hint Resolve tyTopSound'.

    Theorem tyTopSound : forall ctx v, hasTy ctx v tyTop.
      intuition.
    Qed.

    Definition tyConst (n : int32) := Some n.
    Hint Unfold tyConst.

    Theorem tyConstSound1 : forall ctx v, hasTy ctx v (tyConst v).
      unfold tyConst; intuition.
    Qed.
    Theorem tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
      unfold tyConst; intuition.
    Qed.

    Definition isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).
      intros.
      refine (match t as t return (sig_option (fun n => t = tyConst n)) with
		| None => SNone _
		| Some n => SSome _ n _
	      end); trivial.
    Qed.

    Definition subTy : forall (t1 t2 : ty),
      pred_option (forall ctx v,
	hasTy ctx v t1 -> hasTy ctx v t2).
      intros.
      refine (match (t1, t2) as ts return (ts = (t1, t2) -> _) with
		| (_, None) => fun pfEq => PSome _
		| (Some n1, Some n2) => fun pfEq =>
		  if int32_eq_dec n1 n2
		    then PSome _
		    else PNone _
		| _ => fun _ => PNone _
	      end (refl_equal _)); injection pfEq; intros; subst; intuition.
    Qed.

    Definition subMastate : forall (abs1 abs2 : mastate),
      pred_option (forall ctx st,
	mconc ctx st abs1 -> mconc ctx st abs2).
      intros; apply PSome; trivial.
    Qed.

    Definition typeof : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (evalExp st e) t}.
      intros; exists (None : ty); intuition.
    Qed.

    Definition typeofLoad : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (get_mem32 st (evalExp st e)) t}.
      intros; exists (None : ty); intuition.
    Qed.

    Definition checkStore : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      pred_option (forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> apartFromCode (evalExp st e)
	/\ not_too_high (evalExp st e)
	/\ apartFromStack (evalExp st e)).
      intros; apply PNone.
    Qed.

    Definition doStackWrite : forall (abs : mastate),
      sig_option (fun abs' => forall ctx st dst src,
	mconc ctx st abs
	-> inStack dst
	-> mconc ctx (set_mem32 st dst src) abs').
      intros.
      apply SSome with abs.
      intuition.
    Qed.

    Definition effectOfCmd : forall (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> mconc ctx (salExecCmd st cmd) abs').
      intros; apply SSome with tt; intuition.
    Qed.

    Definition effectOfJump : forall (abs : mastate) (rs : SalRegMap.t ty)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> match salExecJmp jmp nextPc st with
	     | None => False
	     | Some st' => mconc ctx st' abs'
	   end).
      intros.
      destruct jmp; [apply SNone
	| apply SSome with tt
	| apply SSome with tt
	| apply SSome with tt
	| apply SSome with tt]; intuition.
    Qed.

    Definition effectOfConditionalJump : forall (abs : stateDescription (stackTy ty)
      (stackState ty mastate))
      (co : cc) (pc nextPc : int32) (b : bool),
      sig_option (fun abs' => forall ctx st,
	typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st abs
	-> check_cc st co = b
	-> match salExecJmp (SConditionalJump co pc) nextPc st with
	     | None => False
	     | Some st' => typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st' abs'
	   end).
      intros; apply SNone.
    Qed.

    Definition viewShiftOfJump : forall (hyps : list (stateDescription ty mastate))
      (abs : stateDescription ty mastate)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun vs => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> exists ctx', typeMconc hasTy mconc ctx' st (vsAbs vs)
	  /\ AllS (fun need => exists have, (In have (vsSuccs vs) \/ In have hyps)
	    /\ subsumes (typeMconc hasTy mconc) ctx' need ctx have) (vsHyps vs)).
      intros; apply SNone.
    Qed.
  End MyTypeSystem.

End Make.
