(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Building abstractions from word-level type systems *)

Require Import Arith List TheoryList.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.ListUtil.
Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Monad.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Machine.
Require Import ProofOS.Trusted.Arches.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.AbsInt.Sal.
Require Import ProofOS.AbsInt.FixedCode.

Set Implicit Arguments.

Inductive apc : Set :=
  | ApcConst : int32 -> apc
  | ApcUnknown.

Section stateDescription.
  Variables ty mastate mcontext : Set.
  Variable hasTy : mcontext -> int32 -> ty -> Prop.
  Variable mconc : mcontext -> salState -> mastate -> Prop.

  Record stateDescription : Set := {
    sdPc : apc;
    sdRegs : SalRegMap.t ty;
    sdCust : mastate
  }.

  Parameter print_stateDescription : (ty -> unit) -> (mastate -> unit)
    -> apc -> ty -> ((salReg * ty -> unit) -> unit) -> mastate -> unit.

  Definition pcOk pc st :=
    match pc with
      | ApcUnknown => True
      | ApcConst pc => pc = SstPc st
    end.

  Record typeMconc (ctx : mcontext) (st : salState) (abs : stateDescription) : Prop := {
    mcPc : pcOk (sdPc abs) st;
    mcRegs : forall r, hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs abs));
    mcCust : mconc ctx st (sdCust abs)
  }.
End stateDescription.

Module Make (Param : M86_PARAM_FIXED_CODE).
  Module Fixed := FixedCode.Make(Param).
  Import Fixed.
  Import MySal.

  Section viewShift.
    Variables ty mastate : Set.

    (** Describe how to change perspective at a jump *)
    Record viewShift : Set := {
      vsSuccs : list (stateDescription ty mastate);
      (** Additional states to queue for visitation *)
      vsHyps : list (stateDescription ty mastate);
      (** Continuation hypotheses known to be covered by the successors and the
	  current hypotheses together *)
      vsAbs : stateDescription ty mastate
      (** Main abstract state to switch to *)
    }.
  End viewShift.

  (** * Definition of a type system *)
  Module Type TYPE_SYSTEM.
    Parameter ty : Set.
    Parameter print_ty : ty -> unit.
    (** A language of types describing machine words *)

    Parameter mastate : Set.
    Parameter print_mastate : mastate -> unit.
    (** Custom abstract state information *)

    Parameter mcontext : Set.
    (** A summary of aspects of a machine state that don't need to be represented
	explicitly at verification time, but are important in soundness proofs *)

    Parameter mconc : mcontext -> M.mstate -> mastate -> Prop.
    (** Concretization relation *)

    Parameter subMastate : forall (abs1 abs2 : mastate),
      pred_option (forall ctx st,
	mconc ctx st abs1 -> mconc ctx st abs2).
    (** A way to check compatibility of abstract states *)
    
    Parameter hasTy : mcontext -> int32 -> ty -> Prop.
    (** What it means for a word to be of a type *)

    Parameter tyTop : ty.
    Axiom tyTopSound : forall ctx v, hasTy ctx v tyTop.

    Parameter tyConst : int32 -> ty.
    Axiom tyConstSound1 : forall ctx v, hasTy ctx v (tyConst v).
    Axiom tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
    Parameter isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).

    Parameter subTy : forall (t1 t2 : ty),
      pred_option (forall ctx v,
	hasTy ctx v t1 -> hasTy ctx v t2).
    (** A way to check compatibility of types *)

    Parameter typeof : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (evalExp st e) t}.
    (** A type inference procedure for SAL expressions *)

    Parameter typeofLoad : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      {t : ty
	| forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> hasTy ctx (get_mem32 st (evalExp st e)) t}.
    (** Type inference for memory loads *)

    Parameter checkStore : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
      pred_option (forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> apartFromCode (evalExp st e)
	/\ not_too_high (evalExp st e)).
    (** Type inference for memory loads *)

    Parameter effectOfCmd : forall (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> mconc ctx (salExecCmd st cmd) abs').
    (** Update the custom state to reflect the result of a command. *)

    Parameter effectOfJump : forall (abs : mastate) (rs : SalRegMap.t ty)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun abs' => forall ctx st,
	mconc ctx st abs
	-> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	-> match salExecJmp jmp nextPc st with
	     | None => False
	     | Some st' => mconc ctx st' abs'
	   end).
    (** Update the custom state to reflect the result of a jump. *)

    Parameter effectOfConditionalJump : forall (abs : stateDescription ty mastate)
      (co : cc) (pc nextPc : int32) (b : bool),
      sig_option (fun abs' => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> check_cc st co = b
	-> match salExecJmp (SConditionalJump co pc) nextPc st with
	     | None => False
	     | Some st' => typeMconc hasTy mconc ctx st' abs'
	   end).
    (** Specialized version of [effectOfJump] that allows different states in
        different cases of a conditional jump test *)

    Parameter viewShiftOfJump : forall (hyps : list (stateDescription ty mastate))
      (abs : stateDescription ty mastate)
      (jmp : salJump) (nextPc : int32),
      sig_option (fun vs => forall ctx st,
	typeMconc hasTy mconc ctx st abs
	-> exists ctx', typeMconc hasTy mconc ctx' st (vsAbs vs)
	  /\ AllS (fun need => exists have, (In have (vsSuccs vs) \/ In have hyps)
	    /\ subsumes (typeMconc hasTy mconc) ctx' need ctx have) (vsHyps vs)).
    (** Massage the form of an abstract state to match what a jump target
        expects; for instance, at a function call. *)
  End TYPE_SYSTEM.

  (** * Instantiation of a type system to a particular verification problem *)
  Module Type USE_TYPE_SYSTEM.
    Declare Module Sys : TYPE_SYSTEM.
    
    Parameter minit : {init : list (list (stateDescription Sys.ty Sys.mastate)
      * stateDescription Sys.ty Sys.mastate)
      | (AllS (fun sd => match sdPc (snd sd) with
			   | ApcUnknown => False
			   | ApcConst pc => validPc pc
			 end) init)
	/\ forall st, M.minitState st -> exists ctx, exists sd,
	  In (nil, sd) init
	  /\ ApcConst (SstPc st) = sdPc sd
	  /\ (forall r, Sys.hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs sd)))
	  /\ Sys.mconc ctx st (sdCust sd)}.
    (** An initial set of states, given by program counter and register types *)
  End USE_TYPE_SYSTEM.

  Module Abstraction (Use : USE_TYPE_SYSTEM) : FIXED_CODE_ABSTRACTION with Module Mac := M.
    Import Use.
    Module Mac := M.
    Import Sys.

    Definition mastate := stateDescription ty mastate.
    Definition print_mastate abs :=
      let (default, elems) := SalRegMap.elements (sdRegs abs) in
	print_stateDescription print_ty print_mastate
	(sdPc abs)
	default
	(fun f => iter f elems)
	(sdCust abs).

    Definition mcontext := mcontext.

    Definition mconc := typeMconc hasTy mconc.
    Hint Unfold mconc.
    Hint Constructors typeMconc.
    
    Hint Unfold pcOk.

    Definition apc_eq_dec : forall (pc1 pc2 : apc), {pc1 = pc2} + {pc1 <> pc2}.
      decide equality.
      apply int32_eq_dec.
    Qed.

    Definition subMastate : forall (abs1 abs2 : mastate),
      pred_option (forall ctx st,
	mconc ctx st abs1 -> mconc ctx st abs2).
      intros.
      refine (if apc_eq_dec (sdPc abs1) (sdPc abs2)
	then
	  Hregs <-- SalRegMap.subMap (fun t1 t2 =>
	    forall (ctx : Sys.mcontext) (v : int32),
	      hasTy ctx v t1 -> hasTy ctx v t2) subTy
	  (sdRegs abs1) (sdRegs abs2);;
	  Hcust <-- subMastate (sdCust abs1) (sdCust abs2);;
	  PSome _
	else PNone _).
      intuition.
      destruct H; split; intuition.
      red.
      destruct (sdPc abs2); trivial.
      red in mcPc0.
      assert (Heq : sdPc abs1 = ApcConst i); trivial.
      rewrite Heq in mcPc0.
      trivial.
    Qed.

    Definition maPc' : forall (abs : mastate), sig_option (fun pc =>
      sdPc abs = ApcConst pc /\ forall ctx st, mconc ctx st abs -> SstPc st = pc).
      intros.
      refine (match sdPc abs as pc return (pc = sdPc abs -> _) with
		| ApcUnknown => fun pfEq => SNone _
		| ApcConst n => fun pfEq => SSome _ n _
	      end (refl_equal _)); intuition.
      destruct H.
      red in mcPc0.
      rewrite <- pfEq in mcPc0.
      auto.
    Defined.

    Definition maPc : forall (abs : mastate), sig_option (fun pc =>
      forall ctx st, mconc ctx st abs -> SstPc st = pc).
      intros.
      refine (match maPc' abs with
		| SNone => SNone _
		| SSome pc _ => SSome _ pc _
	      end); intuition.
    Defined.

    Definition absValidPc' (abs : mastate) :=
      match maPc' abs with
	| SNone => True
	| SSome pc _ => validPc pc
      end.

    Definition absValidPc abs :=
      match maPc abs with
	| SNone => True
	| SSome pc _ => validPc pc
      end.

    Lemma absValidPc_imp : forall abs, absValidPc' abs -> absValidPc abs.
      intros abs Hvalid.
      red.
      red in Hvalid.
      unfold maPc.
      destruct (maPc' abs); trivial.
    Qed.

    Definition minit : {init : list (list mastate * mastate)
      | (forall abs hyps, In (hyps, abs) init -> absValidPc abs)
	/\ forall st, Mac.minitState st -> exists ctx, exists abs,
	  In (nil, abs) init /\ mconc ctx st abs}.
      destruct minit as [init [Hall Hcov]].
      exists init.
      intuition.

      red.
      unfold maPc.
      caseEq (maPc' abs); trivial.
      intros pc [Heq Hrest] Hpc.
      assert (match sdPc (snd (hyps, abs)) with
		| ApcConst pc => validPc pc
		| ApcUnknown => False
	      end).
      apply (AllS_In Hall); trivial.
      simpl in H0; rewrite Heq in H0; trivial.

      destruct (Hcov _ H) as [ctx [sd [Hin [Hpc Hregs]]]].
      exists ctx; exists sd.
      intuition.
      split; intuition.
      red.
      rewrite <- Hpc.
      trivial.
    Qed.


    (** * We build up a fast look-up structure to find preconditions of code
          addresses represented in [minit]. *)
    Definition pcToAbs_map' (states : list (list mastate * mastate)) m :=
      fold_left (fun m abs =>
	match sdPc (snd abs) with
	  | ApcUnknown => m
	  | ApcConst pc => Int32Map.set pc (Some abs) m
	end) states m.

    Definition pcToAbs_map := pcToAbs_map' (proj1_sig minit) (Int32Map.init None).

    Lemma pcToAbs_map'_sound : forall init states m,
      incl states init
      -> (forall r ha, Int32Map.get r m = Some ha -> In ha init)
      -> (forall r ha, Int32Map.get r (pcToAbs_map' states m) = Some ha -> In ha init).
      induction states; simpl; intuition.
      simpl in H1.

      caseEq (sdPc b).

      intros pc Hpc.
      rewrite Hpc in H1.

      apply IHstates with (Int32Map.set pc (Some (a0, b)) m) r; eauto.
      red in H; red; simpl in *; eauto.

      intros.
      rewrite Int32Map.gsspec in H2.
      destruct (Int32Indexed.eq r0 pc); subst.
      injection H2; intros; subst; intuition.
      eauto.
    
      intro Hpc.
      rewrite Hpc in H1.
      eapply IHstates; eauto.
      red in H; red; simpl in *; eauto.
    Qed.

    Definition pcToAbs : forall (pc : int32), sig_option (fun hypsAbs =>
      In hypsAbs (proj1_sig minit)).
      intro pc.
      refine (match Int32Map.get pc pcToAbs_map as lookup
		return (Int32Map.get pc pcToAbs_map = lookup -> _) with
		| None => fun _ => SNone _
		| Some hypsAbs => fun pfEq => SSome _ hypsAbs _
	      end (refl_equal _)).
      apply pcToAbs_map'_sound
	with (proj1_sig minit)
	(Int32Map.init (@None (list mastate * mastate)))
	pc; auto.
      apply incl_refl.
      intros.
      rewrite Int32Map.gi in H; discriminate.
    Qed.
      
    Definition progress (insPc : Mac.minstr * int32) (st : Mac.mstate) :=
      instructionOk st (fst insPc)
      /\ exists st' : Mac.mstate,
	Mac.mexec insPc st st'.

    Definition localStep (hyps abss : list mastate)
      (ctx : mcontext) (st' : Mac.mstate) (abs' : mastate) :=
      ((In abs' abss \/ In abs' hyps) /\ mconc ctx st' abs').

    Definition callStep (hyps abss : list mastate) (ctx : mcontext)
      (st' : Mac.mstate)
      (abs' : mastate) (ctx' : mcontext) (hyps' : list mastate) :=
      In (hyps', abs') (proj1_sig minit)
      /\ AllS (fun need => exists have,
	(In have abss \/ In have hyps)
	/\ subsumes mconc ctx' need ctx have) hyps'
      /\ mconc ctx' st' abs'.

    Definition preservation (hyps : list mastate) (abs : mastate)
      (insPc : Mac.minstr * int32) (abss : list mastate)
      (st : Mac.mstate) (ctx : mcontext) :=
      forall (st' : Mac.mstate) (Hexec : Mac.mexec insPc st st'),
	exists abs', localStep hyps abss ctx st' abs'
	  \/ exists ctx', exists hyps', callStep hyps abss ctx st' abs' ctx' hyps'.

    Lemma execCmd_pc : forall st cmd,
      SstPc (salExecCmd st cmd) = SstPc st.
      destruct cmd; simpl; trivial.
      unfold set_arith_flags; destruct s; trivial.
    Qed.

    Hint Rewrite SalRegMap.gss : Maps.
    Hint Rewrite SalRegMap.gso using (trivial; fail) : Maps.

    Definition checkCmd : forall abs cmd, sig_option (fun abs' =>
      forall ctx st, mconc ctx st abs
	-> mconc ctx (salExecCmd st cmd) abs'
	/\ cmdOk st cmd).
      intros.
      refine (abs' :: Habs' <- effectOfCmd (sdCust abs) (sdRegs abs) cmd;
	match cmd as cmd'
	  return (cmd = cmd' -> sig_option (fun abs' => forall ctx st, mconc ctx st abs
	    -> mconc ctx (salExecCmd st cmd) abs' /\ cmdOk st cmd)) with
	  | SSet r e => fun pfEq =>
	    t ::: Ht <- typeof (sdCust abs) (sdRegs abs) e;
	    SSome _
	    (Build_stateDescription
	      (sdPc abs)
	      (SalRegMap.set r t (sdRegs abs))
	      abs')
	    _
	  | SLoad r e => fun pfEq =>
	    t ::: Ht <- typeofLoad (sdCust abs) (sdRegs abs) e;
	    SSome _
	    (Build_stateDescription
	      (sdPc abs)
	      (SalRegMap.set r t (sdRegs abs))
	      abs')
	    _
	  | SStore dst src => fun pfEq =>
	    Hload <-- checkStore (sdCust abs) (sdRegs abs) dst;
	    SSome _
	    (Build_stateDescription
	      (sdPc abs)
	      (sdRegs abs)
	      abs')
	    _
	  | SCompare cmp aop e1 e2 => fun pfEq =>
	    SSome _
	    (Build_stateDescription
	      (sdPc abs)
	      (sdRegs abs)
	      abs')
	    _
	end (refl_equal _)); simpl; intuition; subst; simpl in *; intuition;
      destruct H; try destruct (Hload _ _ mcCust0); intuition;
	split; simpl in *; unfold get_reg, set_reg, set_arith_flags, arith_flags in *;
	  try destruct cmp; simpl in *; intuition;
	    try destruct (salReg_eq_dec r0 r); subst; autorewrite with Maps; auto.
    Qed.

    Definition checkCmds : forall abs cmds, sig_option (fun abs' =>
      forall ctx st, mconc ctx st abs
	-> mconc ctx (fold_left salExecCmd cmds st) abs'
	/\ cmdsOk st cmds).
      refine (fix checkCmds (abs : mastate) (cmds : list salCmd) {struct cmds}
	: sig_option (fun abs' : mastate =>
	  forall (ctx : mcontext) (st : Mac.mstate),
	    mconc ctx st abs
	    -> mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds) :=
	match cmds as cmds
	  return (sig_option (fun abs' : mastate =>
	    forall (ctx : mcontext) (st : Mac.mstate),
	      mconc ctx st abs ->
	      mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)) with
	  | nil => SSome _ abs _
	  | cons cmd cmds =>
	    abs' :: Habs' <- checkCmd abs cmd;
   	    abs'' :: Habs'' <- checkCmds abs' cmds;
            SSome _ abs'' _
	end); simpl; intuition;
      destruct (Habs' _ _ H) as [Hconc Hok];
	generalize (Habs'' _ _ Hconc);
	intuition.
    Qed.

    Definition hypCovered : forall (hyp : mastate) (hyps : list mastate),
      pred_option (exists have, In have hyps
	/\ forall st ctx, mconc ctx st hyp
	  -> mconc ctx st have).
      refine (fix hypCovered (hyp : mastate) (hyps : list mastate) {struct hyps}
	: pred_option (exists have, In have hyps
	  /\ forall st ctx, mconc ctx st hyp
	    -> mconc ctx st have) :=
	match hyps return (pred_option (exists have, In have hyps
	  /\ forall st ctx, mconc ctx st hyp
	    -> mconc ctx st have)) with
	  | nil => PNone _
	  | cons have rest =>
	    if subMastate hyp have
	      then PSome _
	      else Hrest <-- hypCovered hyp rest;; PSome _
	end).

      exists have; intuition.
      destruct Hrest as [have' Hhave'].
      exists have'; intuition.
    Qed.

    Definition subHyps : forall (hyps1 hyps2 : list mastate),
      pred_option (AllS (fun need => exists have, In have hyps1
	/\ forall st ctx, mconc ctx st need
	  -> mconc ctx st have) hyps2).
      refine (fix subHyps (hyps1 hyps2 : list mastate) {struct hyps2}
	: pred_option (AllS (fun need => exists have, In have hyps1
	  /\ forall st ctx, mconc ctx st need
	    -> mconc ctx st have) hyps2) :=
	match hyps2 return (pred_option (AllS (fun need => exists have, In have hyps1
	  /\ forall st ctx, mconc ctx st need
	    -> mconc ctx st have) hyps2)) with
	  | nil => PSome _
	  | cons need rest =>
	    Hneed <-- hypCovered need hyps1;;
	    Hrest <-- subHyps hyps1 rest;;
	    PSome _
	end); intuition.
    Qed.

    Definition checkJump : forall hyps abs (jmp : salJump) (nextPc : int32),
      sig_option (fun succs =>
	(exists abs', (In abs' hyps /\ forall ctx st, mconc ctx st abs -> mconc ctx st abs')
	  \/ exists hyps', In (hyps', abs') (proj1_sig minit)
	    /\ forall ctx st, mconc ctx st abs
	      -> exists ctx', mconc ctx' st abs'
		/\ AllS (fun need => exists have,
		  (In have succs \/ In have hyps)
		  /\ subsumes mconc ctx' need ctx have) hyps')).
      intros hyps abs jmp nextPc.
      refine (match hypCovered abs hyps with
		| PSome Hcovered => SSome _ nil _
		| PNone =>
		  match sdPc abs as pc return (pc = sdPc abs -> _) with
		    | ApcUnknown => fun pfEq => SNone _
		    | ApcConst pc => fun pfEq =>
		      hypsAbs :: Habs' <- pcToAbs pc;
		      vs :: Hvs <- viewShiftOfJump hyps abs jmp nextPc;
		      Hsub <-- subMastate (vsAbs vs) (snd hypsAbs);
		      Hhyps <-- subHyps (vsHyps vs) (fst hypsAbs);
		      SSome _ (vsSuccs vs) _
		  end (refl_equal _)
	      end).

      destruct Hcovered as [have [Hin Hsubsume]].
      exists have.
      left.
      intuition.

      destruct hypsAbs as [hyps' abs'];	simpl in *; subst.
      exists abs'; right; exists hyps'.
      intuition.

      destruct (Hvs _ _ H) as [ctx' [Hconc Hall]].
      destruct H.
      exists ctx'.
      intuition.
      unfold subsumes.

      eapply AllS_imp; [idtac | apply Hhyps].
      intuition.
      destruct H as [have [Hin Hsubsume]].
      destruct (AllS_In Hall _ Hin) as [have' [Hin' Hsubsume']].
      exists have'.
      split; trivial.
      clear Hin'.
      intros st' Hst'.
      red in Hsubsume'.
      unfold mconc in Hsubsume.
      auto.
    Qed.

    Lemma progress_fallthru : forall x86 cmds nextPc st ctx abs abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
	mconc ctx st abs ->
	mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> mconc ctx st abs
      -> progress (x86, (cmds, SFallthru), nextPc) st.
      intros until abs'; intros Habs' Hconc.
      red; simpl; intuition.
      do 2 red; simpl.
      generalize (Habs' ctx st); intuition.
      unfold Mac.mexec, salExec; simpl.
      unfold salExec'; simpl.
      eauto.
    Qed.

    Hint Resolve progress_fallthru.

    Lemma preservation_fallthru_covered : forall ctx st abs succs x86 cmds nextPc hyps cust abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
        mconc ctx st abs ->
        mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> (exists abs'0 : mastate,
	(In abs'0 hyps /\ forall ctx st, mconc ctx st
	  (Build_stateDescription (ApcConst nextPc) (sdRegs abs') cust)
	  -> mconc ctx st abs'0)
        \/ (exists hyps' : list mastate,
          In (hyps', abs'0) (proj1_sig minit) /\
          (forall (ctx : mcontext) (st : Mac.mstate),
            mconc ctx st
            (Build_stateDescription (ApcConst nextPc) (sdRegs abs') cust) ->
            exists ctx' : mcontext,
              mconc ctx' st abs'0 /\
              AllS
              (fun need : mastate =>
                exists have : mastate,
                  (In have succs \/ In have hyps) /\
                  subsumes mconc ctx' need ctx have) hyps')))
      -> (forall (ctx : Sys.mcontext) (st : Mac.mstate),
        Sys.mconc ctx st (sdCust abs') ->
        (forall r : SalRegIndexed.t,
          hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs abs'))) ->
        Sys.mconc ctx (set_pc st nextPc) cust)
      -> mconc ctx st abs
      -> preservation hyps abs (x86, (cmds, SFallthru), nextPc) succs st ctx.
      intros until abs'; intros Habs' [abs'' [[Hin Hcover] | [hyps' [Hin Hcover]]]]
	Hcust Hconc; red; intros st' Hst'; exists abs''.

      left; red; intuition.
      inversion Hst'; subst.
      apply Hcover.
      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; simpl; intuition.

      right.
      destruct (Hcover ctx (set_pc (fold_left salExecCmd cmds st) nextPc)) as [ctx' Hctx'].

      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; intuition.
      red; simpl; trivial.
      
      exists ctx'; exists hyps'.
      red.
      intuition.

      inversion Hst'; trivial.
    Qed.

    Hint Resolve preservation_fallthru_covered.

    Lemma preservation_fallthru : forall ctx st cust abs' nextPc x86 cmds abs hyps,
      (forall (ctx : mcontext) (st : Mac.mstate),
        mconc ctx st abs ->
        mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> (forall (ctx : Sys.mcontext) (st : Mac.mstate),
        Sys.mconc ctx st (sdCust abs') ->
        (forall r : SalRegIndexed.t,
          hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs abs'))) ->
        Sys.mconc ctx (set_pc st nextPc) cust)
      -> mconc ctx st abs
      -> preservation hyps abs (x86, (cmds, SFallthru), nextPc)
      (cons (Build_stateDescription (ApcConst nextPc) (sdRegs abs') cust) nil) st ctx.
      intros.
      red.
      intuition.
      exists (Build_stateDescription (ApcConst nextPc) (sdRegs abs') cust).
      left.
      red; intuition.
      inversion Hexec; subst.
      generalize (H _ _ H1); intuition.
      destruct H3.
      split; intuition.
      red; simpl; trivial.
    Qed.

    Hint Resolve preservation_fallthru.

    Lemma progress_jump : forall x86 cmds nextPc st ctx abs pc abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
	mconc ctx st abs ->
	mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> mconc ctx st abs
      -> progress (x86, (cmds, SJump pc), nextPc) st.
      intros until abs'; intros Habs' Hconc.
      red; simpl; intuition.
      do 2 red; simpl.
      generalize (Habs' ctx st); intuition.
      unfold Mac.mexec, salExec; simpl.
      unfold salExec'; simpl.
      eauto.
    Qed.

    Hint Resolve progress_jump.

    Lemma preservation_jump_covered : forall pc ctx st abs succs x86 cmds nextPc hyps cust abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
        mconc ctx st abs ->
        mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> (exists abs'0 : mastate,
	(In abs'0 hyps /\ forall ctx st, mconc ctx st
	  (Build_stateDescription (ApcConst pc) (sdRegs abs') cust)
	  -> mconc ctx st abs'0)
        \/ (exists hyps' : list mastate,
          In (hyps', abs'0) (proj1_sig minit) /\
          (forall (ctx : mcontext) (st : Mac.mstate),
            mconc ctx st
            (Build_stateDescription (ApcConst pc) (sdRegs abs') cust) ->
            exists ctx' : mcontext,
              mconc ctx' st abs'0 /\
              AllS
              (fun need : mastate =>
                exists have : mastate,
                  (In have succs \/ In have hyps) /\
                  subsumes mconc ctx' need ctx have) hyps')))
      -> (forall (ctx : Sys.mcontext) (st : Mac.mstate),
        Sys.mconc ctx st (sdCust abs') ->
        (forall r : SalRegIndexed.t,
          hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs abs'))) ->
        Sys.mconc ctx (set_pc st pc) cust)
      -> mconc ctx st abs
      -> preservation hyps abs (x86, (cmds, SJump pc), nextPc) succs st ctx.
      intros until abs'; intros Habs' [abs'' [[Hin Hcover] | [hyps' [Hin Hcover]]]]
	Hcust Hconc; red; intros st' Hst'; exists abs''.

      left; red; intuition.
      inversion Hst'; subst.
      apply Hcover.
      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; simpl; intuition.

      right.
      destruct (Hcover ctx (set_pc (fold_left salExecCmd cmds st) pc)) as [ctx' Hctx'].

      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; intuition.
      red; simpl; trivial.
      
      exists ctx'; exists hyps'.
      red.
      intuition.

      inversion Hst'; trivial.
    Qed.

    Hint Resolve preservation_jump_covered.

    Lemma progress_call : forall x86 cmds nextPc st ctx abs pc abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
	mconc ctx st abs ->
	mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> mconc ctx st abs
      -> progress (x86, (cmds, SIndirectJump SCall (SConst pc)), nextPc) st.
      intros until abs'; intros Habs' Hconc.
      red; simpl; intuition.
      do 2 red; simpl.
      generalize (Habs' ctx st); intuition.
      unfold Mac.mexec, salExec; simpl.
      unfold salExec'; simpl.
      eauto.
    Qed.

    Hint Resolve progress_call.

    Lemma preservation_call_covered : forall pc ctx st abs succs x86 cmds nextPc hyps cust abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
        mconc ctx st abs ->
        mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> (exists abs'0 : mastate,
	(In abs'0 hyps /\ forall ctx st, mconc ctx st
	  (Build_stateDescription (ApcConst pc) (sdRegs abs') cust)
	  -> mconc ctx st abs'0)
        \/ (exists hyps' : list mastate,
          In (hyps', abs'0) (proj1_sig minit) /\
          (forall (ctx : mcontext) (st : Mac.mstate),
            mconc ctx st
            (Build_stateDescription (ApcConst pc) (sdRegs abs') cust) ->
            exists ctx' : mcontext,
              mconc ctx' st abs'0 /\
              AllS
              (fun need : mastate =>
                exists have : mastate,
                  (In have succs \/ In have hyps) /\
                  subsumes mconc ctx' need ctx have) hyps')))
      -> (forall (ctx : Sys.mcontext) (st : Mac.mstate),
        Sys.mconc ctx st (sdCust abs') ->
        (forall r : SalRegIndexed.t,
          hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs abs'))) ->
        Sys.mconc ctx (set_pc st pc) cust)
      -> mconc ctx st abs
      -> preservation hyps abs (x86, (cmds, SIndirectJump SCall (SConst pc)), nextPc) succs st ctx.
      intros until abs'; intros Habs' [abs'' [[Hin Hcover] | [hyps' [Hin Hcover]]]]
	Hcust Hconc; red; intros st' Hst'; exists abs''.

      left; red; intuition.
      inversion Hst'; subst.
      apply Hcover.
      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; simpl; intuition.

      right.
      destruct (Hcover ctx (set_pc (fold_left salExecCmd cmds st) pc)) as [ctx' Hctx'].

      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; intuition.
      red; simpl; trivial.
      
      exists ctx'; exists hyps'.
      red.
      intuition.

      inversion Hst'; trivial.
    Qed.

    Hint Resolve preservation_call_covered.

    Lemma progress_ret : forall x86 cmds nextPc st ctx abs e abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
	mconc ctx st abs ->
	mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> mconc ctx st abs
      -> progress (x86, (cmds, SIndirectJump SRet e), nextPc) st.
      intros until abs'; intros Habs' Hconc.
      red; simpl; intuition.
      do 2 red; simpl.
      generalize (Habs' ctx st); intuition.
      unfold Mac.mexec, salExec; simpl.
      unfold salExec'; simpl.
      eauto.
    Qed.

    Hint Resolve progress_ret.

    Lemma preservation_ret_covered : forall e ctx st abs succs x86 cmds nextPc hyps cust abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
        mconc ctx st abs ->
        mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> (exists abs'0 : mastate,
	(In abs'0 hyps /\ forall ctx st, mconc ctx st
	  (Build_stateDescription ApcUnknown (sdRegs abs') cust)
	  -> mconc ctx st abs'0)
        \/ (exists hyps' : list mastate,
          In (hyps', abs'0) (proj1_sig minit) /\
          (forall (ctx : mcontext) (st : Mac.mstate),
            mconc ctx st
            (Build_stateDescription ApcUnknown (sdRegs abs') cust) ->
            exists ctx' : mcontext,
              mconc ctx' st abs'0 /\
              AllS
              (fun need : mastate =>
                exists have : mastate,
                  (In have succs \/ In have hyps) /\
                  subsumes mconc ctx' need ctx have) hyps')))
      -> (forall (ctx : Sys.mcontext) (st : Mac.mstate),
        Sys.mconc ctx st (sdCust abs') ->
        (forall r : SalRegIndexed.t,
          hasTy ctx (get_reg st r) (SalRegMap.get r (sdRegs abs'))) ->
        Sys.mconc ctx (set_pc st (evalExp st e)) cust)
      -> mconc ctx st abs
      -> preservation hyps abs (x86, (cmds, SIndirectJump SRet e), nextPc) succs st ctx.
      intros until abs'; intros Habs' [abs'' [[Hin Hcover] | [hyps' [Hin Hcover]]]]
	Hcust Hconc; red; intros st' Hst'; exists abs''.

      left; red; intuition.
      inversion Hst'; subst.
      apply Hcover.
      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; simpl; intuition.

      right.
      destruct (Hcover ctx (set_pc (fold_left salExecCmd cmds st)
	(evalExp (fold_left salExecCmd cmds st) e))) as [ctx' Hctx'].

      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      split; intuition.
      red; simpl; trivial.

      exists ctx'; exists hyps'.
      red.
      intuition.

      inversion Hst'; trivial.
    Qed.

    Hint Resolve preservation_ret_covered.

    Lemma progress_conditional : forall co pc x86 cmds nextPc st ctx abs abs',
      (forall (ctx : mcontext) (st : Mac.mstate),
	mconc ctx st abs ->
	mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> mconc ctx st abs
      -> progress (x86, (cmds, SConditionalJump co pc), nextPc) st.
      intros until abs'; intros Habs' Hconc.
      red; simpl; intuition.
      do 2 red; simpl.
      generalize (Habs' ctx st); intuition.
      unfold Mac.mexec, salExec; simpl.
      unfold salExec'; simpl.
      destruct (check_cc (fold_left salExecCmd cmds st) co); eauto.
    Qed.

    Hint Resolve progress_conditional.

    Lemma preservation_conditional_covered : forall ctx st custFalse custTrue succs abs' pc co
      x86 cmds nextPc abs hyps,
      (forall (ctx : Sys.mcontext) (st : salState),
        mconc ctx st abs ->
        mconc ctx (fold_left salExecCmd cmds st) abs' /\ cmdsOk st cmds)
      -> (forall (ctx : Sys.mcontext) (st : Mac.mstate),
        typeMconc hasTy Sys.mconc ctx st abs' ->
        check_cc st co = true ->
        typeMconc hasTy Sys.mconc ctx
        (if check_cc st co then set_pc st pc else set_pc st nextPc)
        custTrue)
      -> (forall (ctx : Sys.mcontext) (st : Mac.mstate),
        typeMconc hasTy Sys.mconc ctx st abs' ->
        check_cc st co = false ->
        typeMconc hasTy Sys.mconc ctx
        (if check_cc st co then set_pc st pc else set_pc st nextPc)
        custFalse)
      -> (exists abs'0 : stateDescription ty Sys.mastate,
        In abs'0 hyps /\
        (forall (ctx : Sys.mcontext) (st : salState),
          mconc ctx st custTrue ->
          mconc ctx st abs'0) \/
        (exists hyps' : list mastate,
          In (hyps', abs'0) (proj1_sig minit) /\
          (forall (ctx : Sys.mcontext) (st : salState),
            mconc ctx st
            custTrue ->
            exists ctx' : Sys.mcontext,
              mconc ctx' st abs'0 /\
              AllS
              (fun need : stateDescription ty Sys.mastate =>
                exists have : stateDescription ty Sys.mastate,
                  (In have succs \/ In have hyps) /\
                  subsumes mconc ctx' need ctx have) hyps')))
      -> mconc ctx st abs
      -> preservation hyps abs (x86, (cmds, SConditionalJump co pc), nextPc)
      (cons custFalse succs)
      st ctx.
      intros until hyps; intros Habs' HcustTrue HcustFalse Habs'' Hconc; red; intros st' Hst'.

      inversion Hst'; subst.
      destrEq Hcc (check_cc (fold_left salExecCmd cmds st) co); rewrite Hcc in Hst'.

      destruct Habs'' as [abs'' [[Hin Hcover] | [hyps' [Hin Hcover]]]].
      exists abs''; left; red; intuition.

      apply Hcover.
      generalize (Habs' _ _ Hconc); intuition.
      inversion H0; intuition.
      generalize (HcustTrue ctx (fold_left salExecCmd cmds st)).
      rewrite Hcc.
      intuition.

      destruct (Hcover ctx (set_pc (fold_left salExecCmd cmds st) pc)) as [ctx' Hctx'].

      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      generalize (HcustTrue ctx (fold_left salExecCmd cmds st)).
      rewrite Hcc.
      intuition.

      exists abs''.
      right.
      exists ctx'; exists hyps'.
      red.
      intuition.
      eapply AllS_imp; [idtac | apply H0].
      intuition.
      destruct H1 as [have Hhave].
      exists have.
      intuition.

      exists custFalse.
      left; red; intuition.
      clear Habs'' HcustTrue.
      generalize (Habs' _ _ Hconc); intuition.
      destruct H0; intuition.
      generalize (HcustFalse ctx (fold_left salExecCmd cmds st)).
      rewrite Hcc.
      intuition.
    Qed.

    Hint Resolve preservation_conditional_covered.

    Definition mstep : forall hyps abs insPc, sig_option (fun abss : list mastate =>
      forall st ctx, mconc ctx st abs
	-> progress insPc st
	/\ preservation hyps abs insPc abss st ctx).
      intros.
      destruct insPc as [[x86 [cmds jmp]] nextPc].
      refine (abs' :: Habs' <- checkCmds abs cmds;
	match jmp as jmp'
	  return (jmp' = jmp -> _) with
	  | SFail => fun _ => SNone _
	  | SFallthru => fun pfEq =>
	    cust :: Hcust <- effectOfJump (sdCust abs') (sdRegs abs') jmp nextPc;
	    match checkJump hyps (Build_stateDescription (ApcConst nextPc) (sdRegs abs') cust) jmp nextPc with
	      | SSome succs Hsuccs => SSome _ succs _
	      | _ => SSome _ (cons (Build_stateDescription (ApcConst nextPc) (sdRegs abs') cust) nil) _
	    end
	  | SJump pc => fun pfEq =>
	    cust :: Hcust <- effectOfJump (sdCust abs') (sdRegs abs') jmp nextPc;
	    succs :: Hsuccs <- checkJump hyps (Build_stateDescription (ApcConst pc) (sdRegs abs') cust) jmp nextPc;
	    SSome _ succs _
	  | SIndirectJump SCall (SConst pc) => fun pfEq =>
	    cust :: Hcust <- effectOfJump (sdCust abs') (sdRegs abs') jmp nextPc;
	    succs :: Hsuccs <- checkJump hyps (Build_stateDescription (ApcConst pc) (sdRegs abs') cust) jmp nextPc;
	    SSome _ succs _
	  | SIndirectJump SRet _ => fun pfEq =>
	    cust :: Hcust <- effectOfJump (sdCust abs') (sdRegs abs') jmp nextPc;
	    succs :: Hsuccs <- checkJump hyps (Build_stateDescription ApcUnknown (sdRegs abs') cust) jmp nextPc;
	    SSome _ succs _
	  | SConditionalJump co pc => fun pfEq =>
	    custTrue :: HcustTrue <- effectOfConditionalJump abs' co pc nextPc true;
	    custFalse :: HcustFalse <- effectOfConditionalJump abs' co pc nextPc false;
	    succs :: Hsuccs <- checkJump hyps custTrue jmp nextPc;
	    SSome _ (cons custFalse succs) _
	  | _ => fun _ => SNone _
	end (refl_equal _)); subst; simpl in *; intuition eauto.
    Qed.
  End Abstraction.

End Make.
