(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Simplified construction of a type system involving memory accesses *)

Require Import Arith List TheoryList ZArith.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Tactics.
Require Import Asm.Util.ListUtil.
Require Import Asm.Util.ArithUtil.
Require Import Asm.Util.Monad.
Require Import Asm.Util.Mod.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.Trusted.Safety.
Require Import ProofOS.Trusted.Machine.
Require Import ProofOS.Trusted.Arches.

Require Import ProofOS.AbsInt.Bounded.
Require Import ProofOS.AbsInt.Sal.
Require Import ProofOS.AbsInt.FixedCode.
Require Import ProofOS.AbsInt.TypeSystem.
Require Import ProofOS.AbsInt.StackTypes.
Require Import ProofOS.AbsInt.SimpleCond.

Set Implicit Arguments.

(** * Definition of a simple type system *)
Module Type WEAK_UPDATE_TYPE_SYSTEM.
  Parameter ty : Set.
  Parameter print_ty : ty -> unit.
  
  Parameter hasTy : (int32 -> option ty) -> int32 -> ty -> Prop.
  
  Parameter tyTop : ty.
  Axiom tyTopSound : forall ctx v, hasTy ctx v tyTop.

  Parameter tyConst : int32 -> ty.
  Axiom tyConstSound1 : forall ctx v, hasTy ctx v (tyConst v).
  Axiom tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
  Parameter isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).
  
  Parameter subTy : forall (t1 t2 : ty),
    pred_option (forall ctx v,
      hasTy ctx v t1 -> hasTy ctx v t2).
  
  Parameter typeofArith : forall (aop : arith_op) (t1 t2 : ty),
    {t : ty
      | forall ctx v1 v2,
	hasTy ctx v1 t1
	-> hasTy ctx v2 t2
	-> hasTy ctx (eval_arith aop v1 v2) t}.

  Parameter typeofCell : forall (t : ty),
    sig_option (fun t' =>
      forall ctx v,
	hasTy ctx v t
	-> ctx v = Some t').

  Parameter considerNeq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx v1,
	hasTy ctx v1 t
	-> v1 <> v
	-> hasTy ctx v1 t'}.

  Parameter considerDerefEq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx st v1,
	(forall a,
	  match ctx a with
	    | None => True
	    | Some t2 => hasTy ctx (get_mem32 st a) t2
	  end)
	-> hasTy ctx v1 t
	-> get_mem32 st v1 = v
	-> hasTy ctx v1 t'}.

  Parameter considerDerefNeq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx st v1,
	(forall a,
	  match ctx a with
	    | None => True
	    | Some t2 => hasTy ctx (get_mem32 st a) t2
	  end)
	-> hasTy ctx v1 t
	-> get_mem32 st v1 <> v
	-> hasTy ctx v1 t'}.
End WEAK_UPDATE_TYPE_SYSTEM.

Definition wordAligned (n : int32) := exists k, bvec_to_nat n = 4 * k.

Lemma wordAligned_neq : forall a1 a2,
  a1 <> a2
  -> wordAligned a1
  -> wordAligned a2
  -> apart a1 a2.
  intros a1 a2 Hneq [k1 Hk1] [k2 Hk2].
  apply apart_or.
  assert (bvec_to_nat a1 <> bvec_to_nat a2).
  intro.
  assert (nat_to_bvec 32 (bvec_to_nat a1) = nat_to_bvec 32 (bvec_to_nat a2)).
  congruence.
  repeat rewrite btn_invert in H0.
  tauto.
  omega.
Qed.

Hint Resolve wordAligned_neq.

Module Type M86_PARAM_FIXED_CODE_AND_STACK_AND_HEAP.
  Parameter minitState : state -> Prop.

  Parameter minitMem : mem.
  Parameter codeStart : int32.
  Parameter codeLength : nat.
  Axiom codeHigh : bvec_to_nat codeStart >= 4.
  Axiom codeFits : bvec_to_nat codeStart + codeLength < pow2 32 - 3.

  Axiom pcInRange : forall st, minitState st
    -> bvec_to_nat codeStart
    <= bvec_to_nat (stPc st)
    < bvec_to_nat codeStart + codeLength - longestInstr.
  Axiom codeUniform : forall st, minitState st
    -> agree_range (stMem st) minitMem codeStart codeLength.

  Parameter stackStart : int32.
  Parameter stackLength : nat.
  Axiom stackHigh : bvec_to_nat stackStart >= 4.
  Axiom stackFits : bvec_to_nat stackStart + stackLength < pow2 32 - 3.
  Axiom stackCodeSeparate : disjointRegions codeStart codeLength stackStart stackLength.

  Parameter heapStart : int32.
  Parameter heapLength : nat.
  Axiom heapHigh : bvec_to_nat heapStart >= 4.
  Axiom heapFits : bvec_to_nat heapStart + heapLength < pow2 32 - 3.
  Axiom heapCodeSeparate : disjointRegions codeStart codeLength heapStart heapLength.
  Axiom heapStackSeparate : disjointRegions stackStart stackLength heapStart heapLength.
End M86_PARAM_FIXED_CODE_AND_STACK_AND_HEAP.

Definition weakState := SalRegMap.t (option salExp).
(** Map each register to a possible expression whose dereference it is known to
    contain *)

Parameter print_weakState : ((salReg * option salExp -> unit) -> unit) -> unit.

Module Make (Param : M86_PARAM_FIXED_CODE_AND_STACK_AND_HEAP).
  Module Cond := SimpleCond.Make(Param).
  Import Cond.
  Import Stack.
  Import TySys.
  Import Fixed.
  Import MySal.

  Import Param.

  Definition inHeap (addr : int32) :=
    bvec_to_nat heapStart
    <= bvec_to_nat addr
    < bvec_to_nat heapStart + heapLength - 3.

  Section weakTy2.
    Variable ty : Set.
    Variable hasTy : (int32 -> option ty) -> int32 -> ty -> Prop.

    Definition concAddr (ctx : int32 -> option ty) (st : salState) (addr : int32) :=
      match ctx addr with
	| None => True
	| Some t =>
	  wordAligned addr
	  /\ inHeap addr
	  /\ hasTy ctx (get_mem32 st addr) t
      end.

    Definition concReg (st : salState) (ws : weakState) (r : salReg) :=
      match SalRegMap.get r ws with
	| None => True
	| Some e => get_reg st r = get_mem32 st (evalExp st e)
      end.

    Record weakConc (ctx : int32 -> option ty) (st : salState) (ws : weakState) : Prop := {
      wcAddrs : forall addr, concAddr ctx st addr;
      wcTmp1 : forall r, concReg st ws r
    }.
  End weakTy2.

  Module Create (_Sys : WEAK_UPDATE_TYPE_SYSTEM) : STACK_READY_TYPE_SYSTEM
    with Definition ty := _Sys.ty
	with Definition mastate := condState weakState
	  with Definition mcontext := int32 -> option _Sys.ty
	    with Definition hasTy := _Sys.hasTy
	      with Definition mconc := condConc (weakConc _Sys.hasTy).

    Module IntermediateTySys <: STACK_READY_TYPE_SYSTEM.
      Import _Sys.
      Module Mac := M.

      Definition ty := ty.
      Definition print_ty := print_ty.

      Definition mastate := weakState.
      Definition print_mastate ws :=
	let (default, elems) := SalRegMap.elements ws in
	print_weakState (fun f => iter f elems).

      Definition mcontext := int32 -> option ty.

      Definition mconc := weakConc hasTy.
      Hint Unfold mconc.

      Definition arith_op_eq_dec : forall (x y : arith_op), {x = y} + {x <> y}.
	decide equality.
      Qed.

      Definition shift_op_eq_dec : forall (x y : shift_op), {x = y} + {x <> y}.
	decide equality.
      Qed.

      Definition salExp_eq_dec : forall (x y : salExp), {x = y} + {x <> y}.
	decide equality.
	apply int32_eq_dec.
	apply salReg_eq_dec.
	apply arith_op_eq_dec.
	apply eq_nat_dec.
	apply shift_op_eq_dec.
      Qed.

      Definition salExpOpt_ok (x y : option salExp) := x = y \/ y = None.
      Hint Unfold salExpOpt_ok.

      Definition salExpOpt_ok_dec : forall (x y : option salExp), {salExpOpt_ok x y} + {~salExpOpt_ok x y}.
	intros.
	destruct y; destruct x.
	destruct (salExp_eq_dec s0 s); subst; eauto.

	right.
	unfold salExpOpt_ok.
	intro.
	destruct H; congruence.
	
	right.
	unfold salExpOpt_ok.
	intro.
	destruct H; discriminate.

	eauto.
	eauto.
      Qed.

      Definition salExpOpt_popt : forall (x y : option salExp), pred_option (salExpOpt_ok x y).
	intros.
	refine (if salExpOpt_ok_dec x y
	  then PSome _
	  else PNone _); trivial.
      Qed.

      Definition subMastate : forall (abs1 abs2 : mastate),
	pred_option (forall ctx st,
	  mconc ctx st abs1 -> mconc ctx st abs2).
	intros.
	refine (Hsub <-- SalRegMap.subMap salExpOpt_ok salExpOpt_popt
	  abs1 abs2;;
	  PSome _); intuition.
	destruct H.
	split; intuition.
	add Hsub' (Hsub r).
	destruct Hsub'.
	unfold concReg in *; rewrite <- H; apply wcTmp2.
	red; rewrite H; trivial.
      Qed.
      
      Definition hasTy := hasTy.

      Definition tyTop := tyTop.
      Definition tyTopSound := tyTopSound.

      Hint Resolve tyTopSound.

      Definition tyConst := tyConst.
      Definition tyConstSound1 := tyConstSound1.
      Definition tyConstSound2 := tyConstSound2.
      Definition isTyConst := isTyConst.

      Definition subTy := subTy.
      
      Hint Unfold hasTy.
      Hint Resolve tyConstSound1.

      Definition typeof : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
	{t : ty
	  | forall ctx st,
	    mconc ctx st abs
	    -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	    -> hasTy ctx (evalExp st e) t}.
	refine (fix typeof (abs : mastate) (rs : SalRegMap.t ty) (e : salExp) {struct e}
	  : {t : ty
	    | forall ctx st,
	      mconc ctx st abs
	      -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	      -> hasTy ctx (evalExp st e) t} :=
	  match e as e' return (e' = e -> _) with
	    | SConst n => fun pfEq => exist _ (tyConst n) _
	    | SReg r => fun pfEq =>
	      exist _ (SalRegMap.get r rs) _
	    | SArith aop e1 e2 => fun pfEq =>
	      t1 ::: Ht1 <- typeof abs rs e1;
	    t2 ::: Ht2 <- typeof abs rs e2;
	    t ::: Ht <- typeofArith aop t1 t2;
	    exist _ t _
	    | _ => fun _ => exist _ tyTop _
	  end (refl_equal _)); subst; simpl; unfold hasTy in *; intuition.
      Qed.

      Definition typeofLoad : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
	{t : ty
	  | forall ctx st,
	    mconc ctx st abs
	    -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	    -> hasTy ctx (get_mem32 st (evalExp st e)) t}.
	intros.
	refine (t ::: Ht <- typeof abs rs e;
	  match typeofCell t with
	    | SSome t' Ht' => exist _ t' _
	    | _ => exist _ tyTop _
	  end); intuition.

	add Hcell (wcAddrs H (evalExp st e)).
	red in Hcell.
	rewrite Ht' in Hcell; intuition.
	apply Ht; intuition.
      Qed.

      Lemma inHeap_apartFromCode : forall v,
	inHeap v
	-> apartFromCode v.
	generalize heapCodeSeparate.
	unfold inHeap, apartFromCode, disjointRegions.
	intros.
	omega.
      Qed.

      Hint Resolve inHeap_apartFromCode.

      Lemma inHeap_not_too_high : forall v,
	inHeap v
	-> not_too_high v.
	generalize heapFits.
	unfold inHeap, not_too_high.
	intros.
	split.
	omega.
      Qed.

      Hint Resolve inHeap_not_too_high.

      Lemma inHeap_apartFromStack : forall v,
	inHeap v
	-> apartFromStack v.
	generalize heapStackSeparate.
	unfold inHeap, apartFromStack, disjointRegions.
	intros.
	omega.
      Qed.

      Hint Resolve inHeap_apartFromStack.

      Definition checkStore : forall (abs : mastate) (rs : SalRegMap.t ty) (e : salExp),
	pred_option (forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> apartFromCode (evalExp st e)
	  /\ not_too_high (evalExp st e)
	  /\ apartFromStack (evalExp st e)).
        intros.
	refine (t ::: Ht <- typeof abs rs e;;
	  t' :: Ht' <- typeofCell t;;
	  PSome _); intros.

	add Hcell (wcAddrs H (evalExp st e)).
	red in Hcell.
	rewrite Ht' in Hcell; intuition.
	apply Ht; auto.
      Qed.

      Lemma evalExp_mem : forall e st pc a v,
	evalExp (Build_salState (SstRegs st) (SstFlags st) (set_mem32' (SstMem st) a v) pc) e
	= evalExp st e.
	induction e; simpl; intros.

	trivial.

	trivial.

	rewrite IHe1.
	rewrite IHe2.
	trivial.

	rewrite IHe.
	trivial.
      Qed.
      
      Theorem doStackWrite : forall (abs : mastate),
	sig_option (fun abs' => forall ctx st dst src,
	  mconc ctx st abs
	  -> inStack dst
	  -> mconc ctx (set_mem32 st dst src) abs').
	intros.
	apply SSome with (SalRegMap.init (@None salExp)).
	intros; split; intuition.

	add Haddr (wcAddrs H addr).
	red; red in Haddr.
	destruct (ctx addr); try (split; fail).

	rewrite invert_set_mem32_neq; intuition.
	red in H0.
	split; generalize stackFits; omega.

	red in H0.
	red in H3.
	apply apart_or.
	generalize heapStackSeparate.
	unfold disjointRegions.
	omega.
      
	red.
	rewrite SalRegMap.gi; trivial.
      Qed.

      Hint Unfold mconc.
      Hint Constructors weakConc.

      Definition regFree : forall (r : salReg) (e : salExp),
	pred_option (forall st v,
	  evalExp st e = evalExp (set_reg st r v) e).
	refine (fix regFree (r : salReg) (e : salExp) {struct e}
	  : pred_option (forall st v,
	    evalExp st e = evalExp (set_reg st r v) e) :=
	  match e as e return (pred_option (forall st v,
	    evalExp st e = evalExp (set_reg st r v) e)) with
	    | SConst _ => PSome _
	    | SReg r' =>
	      if salReg_eq_dec r' r
		then PNone _
		else PSome _
	    | SArith _ e1 e2 =>
	      He1 <-- regFree r e1;;
	      He2 <-- regFree r e2;;
	      PSome _
	    | SShift _ e1 _ =>
	      He1 <-- regFree r e1;;
	      PSome _
	  end); simpl; intros.

	trivial.

	unfold get_reg, set_reg; simpl.
	rewrite SalRegMap.gso; trivial.

	rewrite (He1 st v); rewrite (He2 st v); trivial.

	rewrite (He1 st v); trivial.
      Qed.

      Lemma get_mem32_flags : forall st fl a,
	get_mem32
	(Build_salState (SstRegs st) fl (SstMem st) (SstPc st)) a
	= get_mem32 st a.
	trivial.
      Qed.

      Lemma evalExp_flags : forall st fl e,
	evalExp
        (Build_salState (SstRegs st) fl (SstMem st) (SstPc st)) e
	= evalExp st e.
	induction e; simpl.

	trivial.

	trivial.

	rewrite IHe1.
	rewrite IHe2.
	trivial.

	rewrite IHe.
	trivial.
      Qed.

      Lemma concReg_easy : forall st r,
	concReg st (SalRegMap.init None) r.
	intros.
	red.
	rewrite SalRegMap.gi.
	trivial.
      Qed.

      Hint Resolve concReg_easy.

      Definition contains : forall (abs : mastate) (e : salExp),
	sig_option (fun e' => forall ctx st,
	  mconc ctx st abs
	  -> evalExp st e = get_mem32 st (evalExp st e')).
	intros.
	refine (match e as e' return (e' = e -> _) with
		  | SReg r => fun pfEq =>
		    match SalRegMap.get r abs as gr return (gr = SalRegMap.get r abs -> _) with
		      | Some e' => fun pfEq' => SSome _ e' _
		      | _ => fun _ => SNone _
		    end (refl_equal _)
		  | _ => fun _ => SNone _
		end (refl_equal _)); subst; intuition.

	add Hr (wcTmp1 H r).
	red in Hr.
	rewrite <- pfEq' in Hr.
	trivial.
      Qed.

      Definition effectOfCmd : forall (abs : mastate) (rs : SalRegMap.t ty) (cmd : salCmd),
	sig_option (fun abs' => forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> mconc ctx (salExecCmd st cmd) abs').
	intros.
	refine (match cmd as cmd' return (cmd' = cmd -> _) with
		  | SStore dst src => fun pfEq =>
		    tDst ::: HtDst <- typeof abs rs dst;
		    tDst' :: HtDst' <- typeofCell tDst;
		    tSrc ::: HtSrc <- typeof abs rs src;
		    match subTy tSrc tDst' with
		      | PSome Hsub => SSome _ (SalRegMap.init (@None salExp)) _
		      | PNone => SNone _
		    end
		  | SLoad r e => fun pfEq =>
		    SSome _ (SalRegMap.map
		      (fun v =>
			match v with
			  | None => None
			  | Some e' =>
			    match regFree r e' with
			      | PNone => None
			      | PSome Hfree => v
			    end
			end)
		      (SalRegMap.set r (Some e) abs)) _
		  | SCompare _ _ _ _ => fun pfEq =>
		    SSome _ abs _
		  | SSet r e => fun pfEq =>
		    let e' := match contains abs e with
				| SNone => None
				| SSome e' He' => Some e'
			      end in
		    SSome _ (SalRegMap.map
		      (fun v =>
			match v with
			  | None => None
			  | Some e' =>
			    match regFree r e' with
			      | PNone => None
			      | PSome Hfree => v
			    end
			end)
		      (SalRegMap.set r e' abs)) _
		end (refl_equal _)); subst; simpl; intros.

	destruct H; split.
	intuition.
	intro.
	red.
	rewrite SalRegMap.gmap.
	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r); subst; trivial.

	subst e'.
	destruct (contains abs e); trivial.
	destruct (regFree r x); trivial.
	rewrite <- e1.
	replace (get_mem32 (set_reg st r (evalExp st e)) (evalExp st x))
	  with (get_mem32 st (evalExp st x)); trivial.
	unfold get_reg, set_reg; simpl.
	rewrite SalRegMap.gss.
	eauto.

	add Hr0 (wcTmp2 r0).
	red in Hr0.
	destruct (SalRegMap.get r0 abs); trivial.
	destruct (regFree r s); try (split; fail).
	rewrite <- e0.
	unfold get_reg, set_reg; simpl.
	rewrite SalRegMap.gso; trivial.

	destruct H; split.
	intuition.
	intro.
	red.
	rewrite SalRegMap.gmap.
	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r); subst.

	destruct (regFree r e); trivial.
	rewrite <- e0.
	unfold get_reg, set_reg; simpl.
	rewrite SalRegMap.gss; trivial.

	add Hr0 (wcTmp2 r0).
	red in Hr0.
	destruct (SalRegMap.get r0 abs); trivial.
	destruct (regFree r s); try (split; fail).
	rewrite <- e0.
	unfold get_reg, set_reg; simpl.
	rewrite SalRegMap.gso; trivial.

	destruct H; split; try (split; fail).

	intro.

	assert (Heq : ctx (evalExp st dst) = Some tDst').
	apply HtDst'.
	apply HtDst; auto.

	add Hdst (wcAddrs0 (evalExp st dst)).
	red in Hdst.
	rewrite Heq in Hdst.

	destruct (int32_eq_dec addr (evalExp st dst)); subst.
	red.
	rewrite Heq.
	unfold get_mem32, set_mem32; simpl.
	rewrite invert_set_mem32'.
	intuition.
	apply Hsub.
	apply HtSrc; auto.

	add Haddr (wcAddrs0 addr).
	red; red in Haddr.
	destruct (ctx addr); try (split; fail).
	unfold get_mem32, set_mem32; simpl.
	rewrite invert_set_mem32'_neq; intuition.

	intro; auto.

	destruct H.
	unfold set_arith_flags, arith_flags, arith_flags_no_CF; destruct s;
	  (split;
	    [intuition
	      | intro; red;
		add Hr (wcTmp2 r);
		red in Hr;
		  destruct (SalRegMap.get r abs); try (split; fail);
		    rewrite get_mem32_flags;
		      rewrite evalExp_flags;
			trivial]).
      Qed.

      Lemma evalExp_pc : forall e st pc,
	evalExp (Build_salState (SstRegs st) (SstFlags st) (SstMem st) pc) e
	= evalExp st e.
	induction e; simpl; intros.
	
	trivial.
	
	trivial.
	
	rewrite IHe1.
	rewrite IHe2.
	trivial.
	
	rewrite IHe.
	trivial.
      Qed.

      Definition effectOfJump : forall (abs : mastate) (rs : SalRegMap.t ty)
	(jmp : salJump) (nextPc : int32),
	sig_option (fun abs' => forall ctx st,
	  mconc ctx st abs
	  -> (forall r, hasTy ctx (get_reg st r) (SalRegMap.get r rs))
	  -> match salExecJmp jmp nextPc st with
	       | None => False
	       | Some st' => mconc ctx st' abs'
	     end).
	intros.
	refine (match jmp as jmp' return (jmp' = jmp -> _) with
		  | SFail => fun _ => SNone _
		  | _ => fun pfEq => SSome _ abs _
		end (refl_equal _)); subst; simpl; intros; destruct H;
	try destruct (check_cc st c);
	  (split;
	    [trivial
	      | intro; red;
		add Hr (wcTmp2 r); red in Hr;
		  destruct abs; try (split; fail);
		    destruct (SalRegMap.get r (o, t)); try (split; fail);
		      unfold set_pc;
			rewrite evalExp_pc; trivial]).
      Qed.

      Definition effectOfConditionalJump : forall (abs : stateDescription (stackTy ty)
	(stackState ty mastate))
      (co : cc) (pc nextPc : int32) (b : bool),
      sig_option (fun abs' => forall ctx st,
	typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st abs
	-> check_cc st co = b
	-> match salExecJmp (SConditionalJump co pc) nextPc st with
	     | None => False
	     | Some st' => typeMconc (stackHasTy hasTy) (stackConc hasTy mconc) ctx st' abs'
	   end).
	intros.
	apply SSome with (Build_stateDescription
	  (ApcConst (if b then pc else nextPc))
	  (sdRegs abs)
	  (Build_stackState
	    false
	    (aStackSize (sdCust abs))
	    (aStackSlots (sdCust abs))
	    (aCustom (sdCust abs)))); subst; intuition.
	subst.
	simpl.
	
	destruct H.
	destruct mcCust0.
	destruct (check_cc st co); split; simpl; intuition.
	split; simpl; intuition; try discriminate.
	destruct mcCust0.
	split.
	intuition.
	intro.
	red.
	add Hr (wcTmp2 r); red in Hr; simpl in Hr.
	destruct (SalRegMap.get r aCustom).
	unfold set_pc.
	rewrite evalExp_pc.
	trivial.
	trivial.

	split; simpl; intuition; try discriminate.
	destruct mcCust0.
	split.
	auto.
	intro.
	add Hr (wcTmp2 r); red in Hr; simpl in Hr.
	red.
	destruct (SalRegMap.get r aCustom).
	unfold set_pc.
	rewrite evalExp_pc.
	trivial.
	trivial.
      Qed.
      
      Definition viewShiftOfJump : forall (hyps : list (stateDescription ty mastate))
	(abs : stateDescription ty mastate)
	(jmp : salJump) (nextPc : int32),
	sig_option (fun vs => forall ctx st,
	  typeMconc hasTy mconc ctx st abs
	  -> exists ctx', typeMconc hasTy mconc ctx' st (vsAbs vs)
	    /\ AllS (fun need => exists have, (In have (vsSuccs vs) \/ In have hyps)
	      /\ subsumes (typeMconc hasTy mconc) ctx' need ctx have) (vsHyps vs)).
	intros; apply SNone.
      Qed.

    End IntermediateTySys.

    Lemma bsub32_eq : forall n,
      bsub32 n n = bzero32.
      intros.
      unfold bsub32, bsub, barith.
      rewrite (ntb_mod 32 (n1 := bvec_to_nat n + bvec_to_nat (bneg n))
	(n2 := pow2 32 + bvec_to_nat n - bvec_to_nat n)).

      replace (pow2 32 + bvec_to_nat n - bvec_to_nat n)
	with (pow2 32).
      rewrite (ntb_mod 32 (n1 := pow2 32) (n2 := 0)).
      unfold bzero32.
      apply zero_ntb.
      red; exists 1%Z.
      generalize (Z_of_nat (pow2 32)).
      intro.
      replace (Z_of_nat 0) with 0%Z; trivial.
      omega.
      omega.
      apply eqMod_trans with (bvec_to_nat (bsub n n)); [idtac | apply bsub_btn].

      unfold bsub, barith.
      apply eqMod_sym.
      apply ntb_invert_mod.
    Qed.
    
    Ltac clear_one :=
      match goal with
	| [ H : _ |- _ ] => clear H
      end.
    Ltac clear_all := repeat clear_one.

    Lemma mult_inc : forall k n,
      k > 1
      -> n > 1
      -> k * n >= 2 * n.
      intros.
      destruct k; try omega.
      destruct k; try omega.
      simpl.
      generalize (k * n).
      intros; omega.
    Qed.

    Import Datatypes.

    Lemma pow2_gt_1 : forall len, pow2 (S len) > 1.
      simpl; intros.
      generalize (pow2_pos len).
      omega.
    Qed.

    Hint Resolve pow2_gt_1.

    Lemma bsub32_eq' : forall n1 n2,
      bsub32 n1 n2 = bzero32
      -> n1 = n2.
      intros.
      unfold bsub32, bsub, barith in H.
      apply btn_inj.
      rewrite (ntb_mod 32 (n1 := bvec_to_nat n1 + bvec_to_nat (bneg n2))
	(n2 := pow2 32 + bvec_to_nat n1 - bvec_to_nat n2)) in H.
      unfold bzero32 in H.
      rewrite <- zero_ntb in H.
      add Hmod (ntb_mod' _ _ H).
      destruct (eqMod_nat Hmod) as [k Hk].
      
      assert (k = 1).

      assert (k = 0 -> False).
      intro.
      subst.
      rewrite mult_0_l in Hk.
      rewrite plus_0_l in Hk.
      rewrite plus_0_r in Hk.
      generalize (pow2_pos 32); generalize (btn_upper n1); generalize (btn_upper n2); omega.
      
      assert (k > 1 -> False).
      intro.
      assert (k * pow2 32 >= 2 * pow2 32).
      apply mult_inc; auto.
      generalize dependent (k * pow2 32); intros.
      generalize (pow2_pos 32); generalize (btn_upper n1); generalize (btn_upper n2); omega.
      omega.

      subst.
      generalize (pow2_pos 32); generalize (btn_upper n1); generalize (btn_upper n2); omega.

      rewrite (plus_comm (pow2 32) (bvec_to_nat n1)).
      apply eqMod_trans
	with (bvec_to_nat n1 + pow2 32 - bvec_to_nat n2).
      replace (bvec_to_nat n1 + pow2 32 - bvec_to_nat n2)
	with (bvec_to_nat n1 + (pow2 32 - bvec_to_nat n2)).
      apply eqMod_plus.
      apply eqMod_refl.
      apply bneg_btn.
      generalize (btn_upper n2); omega.
      apply eqMod_refl.
    Qed.

    Module WithTesting <: TYPE_SYSTEM_WITH_TESTING.
      Module Sys := IntermediateTySys.
      Import Sys.

      Definition lowerTy t :=
	match t with
	  | Const n => tyConst n
	  | Cust t' => t'
	  | _ => tyTop
	end.
      
      Hint Resolve tyTopSound.
      
      Lemma lowerTy_sound : forall v ctx t,
	stackHasTy _Sys.hasTy ctx v t
	-> _Sys.hasTy (ctxCust ctx) v (lowerTy t).
	destruct t; inversion 1; subst; simpl; intuition.
      Qed.
      
      Hint Resolve lowerTy_sound.

      Definition raiseTy t :=
	match _Sys.isTyConst t with
	  | SNone =>
	    if _Sys.subTy _Sys.tyTop t
	      then Top _
	      else Cust t
	  | SSome n _ => Const _ n
	end.

      Lemma raiseTy_sound : forall v ctx t,
	_Sys.hasTy (ctxCust ctx) v t
	-> stackHasTy _Sys.hasTy ctx v (raiseTy t).
	intros.
	unfold raiseTy.
	destruct (_Sys.isTyConst t); subst.

	add Hconst (tyConstSound2 H).
	subst; constructor.

	destruct (_Sys.subTy _Sys.tyTop t);
	  constructor; trivial.
      Qed.

      Definition doCastEq : forall r n (abs : stateDescription (stackTy Sys.ty)
	(stackState Sys.ty Sys.mastate)),
      {abs' : stateDescription (stackTy Sys.ty) (stackState Sys.ty Sys.mastate)
	| forall ctx st,
	  typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs
	  -> get_reg st r = n
	  -> typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs'}.
	intros.
	refine (match SalRegMap.get r (aCustom (sdCust abs)) as gr
		  return (gr = SalRegMap.get r (aCustom (sdCust abs)) -> _) with
		  | Some e => fun pfEq =>
		    match e as e' return (e' = e -> _) with
		      | SArith Add (SConst z) (SReg r') => fun pfEq' =>
			if int32_eq_dec z bzero32
			  then match SalRegMap.get r' (aCustom (sdCust abs)) as gr'
				 return (gr' = SalRegMap.get r' (aCustom (sdCust abs)) -> _) with
				 | Some e'' => fun pfEq'' =>
				   t' ::: Ht' <- _Sys.considerDerefEq (lowerTy (SalRegMap.get r' (sdRegs abs))) n;
				   match stackSlotOf Sys.hasTy Sys.mconc abs e'' with
				     | SSome ss Hss =>
				       exist _ (Build_stateDescription
					 (sdPc abs)
					 (SalRegMap.set r' (raiseTy t') (sdRegs abs))
					 (Build_stackState
					   false
					   (aStackSize (sdCust abs))
					   (NatMap.set ss (raiseTy t') (aStackSlots (sdCust abs)))
					   (SalRegMap.init None))) _
				     | _ =>
				       exist _ (Build_stateDescription
					 (sdPc abs)
					 (SalRegMap.set r' (raiseTy t') (sdRegs abs))
					 (sdCust abs)) _
				   end
				 | _ => fun pfEq'' =>
				   exist _ (Build_stateDescription
				     (sdPc abs)
				     (sdRegs abs)
				     (sdCust abs)) _
			       end (refl_equal _)
			  else
			    exist _ (Build_stateDescription
			      (sdPc abs)
			      (sdRegs abs)
			      (sdCust abs)) _
		      | _ => fun pfEq' =>
			exist _ (Build_stateDescription
			  (sdPc abs)
			  (sdRegs abs)
			  (sdCust abs)) _
		    end (refl_equal _)
		  | _ => fun pfEq =>
		    exist _ (Build_stateDescription
		      (sdPc abs)
		      (sdRegs abs)
		      (sdCust abs)) _
		end (refl_equal _)); intuition;
	destruct H;
	  split; simpl; intuition.

	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r'); subst.

	apply raiseTy_sound.
	inversion mcCust0.
	inversion mcCust1.
	apply Ht' with st; auto.
	
	intro.
	generalize (wcAddrs0 a0).
	unfold concAddr.
	destruct (ctxCust ctx a0); intuition.

	add Hr (wcTmp2 r); red in Hr.
	unfold ty, mastate in pfEq.
	rewrite <- pfEq in Hr.
	rewrite Hr.
	simpl.
	unfold badd32, bzero32; rewrite badd_zero'; trivial.
	
	auto.

	destruct mcCust0.
	destruct mcCust0.
	split; intuition.

	intros sl Hsl; simpl in *; idtac.
	rewrite NatMap.gsspec.
	destruct (NatIndexed.eq sl ss); subst.

	apply raiseTy_sound.
	apply Ht' with st.

	intro.
	generalize (wcAddrs0 a0).
	unfold concAddr.
	destruct (ctxCust ctx a0); intuition.
	
	apply lowerTy_sound.
	replace (get_mem32 st (stack_slot ctx ss))
	  with (get_reg st r'); auto.
	add Hr' (wcTmp2 r'); red in Hr'.
	unfold ty, mastate in pfEq''.
	rewrite <- pfEq'' in Hr'.
	rewrite Hr'.
	rewrite Hss with ctx st; auto.

	split; intuition.
	split; intuition.

	add Hr (wcTmp2 r); red in Hr.
	unfold ty, mastate in pfEq.
	rewrite <- pfEq in Hr.
	rewrite Hr.
	simpl.
	rewrite <- Hss with ctx st.
	add Hr' (wcTmp2 r'); red in Hr'.
	unfold ty, mastate in pfEq''.
	rewrite <- pfEq'' in Hr'.
	rewrite Hr'.
	unfold badd32, bzero32; rewrite badd_zero'; clear_all; trivial.

	split; intuition.
	split; intuition.

	auto.

	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r'); subst.

	apply raiseTy_sound.
	inversion mcCust0.
	inversion mcCust1.
	apply Ht' with st; auto.
	
	intro.
	generalize (wcAddrs0 a0).
	unfold concAddr.
	destruct (ctxCust ctx a0); intuition.

	add Hr (wcTmp2 r); red in Hr.
	unfold ty, mastate in pfEq.
	rewrite <- pfEq in Hr.
	rewrite Hr.
	simpl.
	unfold badd32, bzero32; rewrite badd_zero'; trivial.
	
	auto.
      Qed.

      Definition doCastNeq : forall r n (abs : stateDescription (stackTy Sys.ty)
	(stackState Sys.ty Sys.mastate)),
      {abs' : stateDescription (stackTy Sys.ty) (stackState Sys.ty Sys.mastate)
	| forall ctx st,
	  typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs
	  -> get_reg st r <> n
	  -> typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs'}.
	intros.
	refine (t ::: Ht <- _Sys.considerNeq (lowerTy (SalRegMap.get r (sdRegs abs))) n;
	  match SalRegMap.get r (aCustom (sdCust abs)) as gr
	    return (gr = SalRegMap.get r (aCustom (sdCust abs)) -> _) with
	    | Some e => fun pfEq =>
	      match stackSlotOf Sys.hasTy Sys.mconc abs e with
		| SSome ss Hss =>
		  exist _ (Build_stateDescription
		    (sdPc abs)
		    (SalRegMap.set r (raiseTy t) (sdRegs abs))
		    (Build_stackState
		      false
		      (aStackSize (sdCust abs))
		      (NatMap.set ss (raiseTy t) (aStackSlots (sdCust abs)))
		      (SalRegMap.init None))) _
		| _ =>
		  match e as e' return (e' = e -> _) with
		    | SArith Add (SConst z) (SReg r') => fun pfEq' =>
		      if int32_eq_dec z bzero32
			then
			  t' ::: Ht' <- _Sys.considerDerefNeq (lowerTy (SalRegMap.get r' (sdRegs abs))) n;
			  match SalRegMap.get r' (aCustom (sdCust abs)) as gr'
			       return (gr' = SalRegMap.get r' (aCustom (sdCust abs)) -> _) with
			       | Some e'' => fun pfEq'' =>
				 match stackSlotOf Sys.hasTy Sys.mconc abs e'' with
				   | SSome ss Hss =>
				     exist _ (Build_stateDescription
				       (sdPc abs)
				       (SalRegMap.set r (raiseTy t)
					 (SalRegMap.set r' (raiseTy t') (sdRegs abs)))
				       (Build_stackState
					 false
					 (aStackSize (sdCust abs))
					 (NatMap.set ss (raiseTy t') (aStackSlots (sdCust abs)))
					 (SalRegMap.init None))) _
				   | _ =>
				     exist _ (Build_stateDescription
				       (sdPc abs)
				       (SalRegMap.set r (raiseTy t)
					 (SalRegMap.set r' (raiseTy t') (sdRegs abs)))
				       (sdCust abs)) _
				 end
			       | _ => fun pfEq'' =>
				 exist _ (Build_stateDescription
				   (sdPc abs)
				   (SalRegMap.set r (raiseTy t)
				     (SalRegMap.set r' (raiseTy t') (sdRegs abs)))
				   (sdCust abs)) _
			     end (refl_equal _)
			else
			  exist _ (Build_stateDescription
			    (sdPc abs)
			    (SalRegMap.set r (raiseTy t) (sdRegs abs))
			    (sdCust abs)) _
		    | _ => fun pfEq' =>
		      exist _ (Build_stateDescription
			(sdPc abs)
			(SalRegMap.set r (raiseTy t) (sdRegs abs))
			(sdCust abs)) _
		  end (refl_equal _)
	      end
	    | _ => fun pfEq =>
	      exist _ (Build_stateDescription
		(sdPc abs)
		(SalRegMap.set r (raiseTy t) (sdRegs abs))
		(sdCust abs)) _
	  end (refl_equal _)); intuition;
	try (destruct H;
	  split; simpl; intuition;
	    rewrite SalRegMap.gsspec;
	      destruct (SalRegIndexed.eq r0 r); subst; auto;
		apply raiseTy_sound;
		  apply Ht;
		    [apply lowerTy_sound; apply mcRegs
		      | trivial]).

	inversion H.
	split; simpl; intuition.
	rewrite SalRegMap.gsspec;
	  destruct (SalRegIndexed.eq r0 r); subst; auto;
	    apply raiseTy_sound;
	      apply Ht;
		[apply lowerTy_sound; apply mcRegs
		  | trivial].

	inversion mcCust0.
	split; simpl; intuition.
	intros sl Hsl; simpl in *; idtac.
	rewrite NatMap.gsspec.

	destruct (NatIndexed.eq sl ss); subst.
	add Hslt (mcStackSlots0 ss).
	apply raiseTy_sound.
	apply Ht.
	inversion mcCust0.
	inversion mcCust2.
	unfold ty, mastate in pfEq.
	add Hr (wcTmp2 r); red in Hr.
	rewrite <- pfEq in Hr.
	rewrite <- (Hss ctx st).
	rewrite <- Hr.
	apply lowerTy_sound; auto.
	intuition.

	inversion mcCust0.
	inversion mcCust2.
	unfold ty, mastate in pfEq.
	add Hr (wcTmp2 r); red in Hr.
	rewrite <- pfEq in Hr.
	rewrite <- (Hss ctx st).
	rewrite <- Hr.
	trivial.

	trivial.

	auto.

	destruct mcCust1.
	intuition.

	inversion H.
	split; simpl; intuition.
	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r); subst; auto.
	apply raiseTy_sound; apply Ht;
	  [apply lowerTy_sound; apply mcRegs
	    | trivial].
	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r'); subst; auto.
	apply raiseTy_sound; apply Ht' with st; auto.
	intro.
	inversion mcCust0.
	inversion mcCust1.
	generalize (wcAddrs0 a0).
	unfold concAddr.
	destruct (ctxCust ctx a0); intuition.
	replace (get_mem32 st (get_reg st r'))
	  with (get_reg st r); trivial.
	inversion mcCust0.
	inversion mcCust1.
	add Hr' (wcTmp2 r'); red in Hr'.
	unfold ty, mastate in pfEq''.
	rewrite <- pfEq'' in Hr'.
	add Hr (wcTmp2 r); red in Hr.
	unfold ty, mastate in pfEq.
	rewrite <- pfEq in Hr; simpl in Hr.
	rewrite Hr.
	unfold badd32, bzero32; rewrite badd_zero'; trivial.

	inversion mcCust0.
	
	split; simpl; intuition.
	intros sl Hsl; simpl in *; idtac.
	rewrite NatMap.gsspec.

	destruct (NatIndexed.eq sl ss); subst.
	add Hslt (mcStackSlots0 ss).
	apply raiseTy_sound.
	apply Ht' with st; auto.

	destruct mcCust1.
	intro.
	generalize (wcAddrs0 a0).
	unfold concAddr.
	destruct (ctxCust ctx a0); intuition.
	rewrite <- (Hss ctx st).
	inversion mcCust0.
	inversion mcCust2.
	add Hr' (wcTmp2 r').
	red in Hr'.
	unfold ty, mastate in pfEq''.
	rewrite <- pfEq'' in Hr'.
	rewrite <- Hr'.
	apply lowerTy_sound.
	auto.
	auto.

	replace (get_mem32 st (get_mem32 st (stack_slot ctx ss)))
	  with (get_reg st r); trivial.
	inversion mcCust0.
	inversion mcCust2.
	add Hr (wcTmp2 r); red in Hr.
	unfold ty, mastate in pfEq.
	rewrite <- pfEq in Hr.
	rewrite Hr; simpl.
	unfold badd32, bzero32; rewrite badd_zero'.
	add Hr' (wcTmp2 r'); red in Hr'.
	unfold ty, mastate in pfEq''.
	rewrite <- pfEq'' in Hr'.
	rewrite Hr'; simpl.
	rewrite (Hss ctx st).
	clear_all; trivial.
	auto.
	auto.

	destruct mcCust1.
	intuition.

	destruct H.
	split; simpl; intuition.
      
	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r); subst; auto.
	apply raiseTy_sound.
	apply Ht;
	  [apply lowerTy_sound; apply mcRegs
	    | trivial].

      	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r'); subst; auto.
	apply raiseTy_sound.
	destruct mcCust0; destruct mcCust0.
	apply Ht' with st; auto.

	intro.
	generalize (wcAddrs0 a0).
	unfold concAddr.
	destruct (ctxCust ctx a0); intuition.
	add Hr (wcTmp2 r); red in Hr.
	unfold ty, mastate in pfEq.
	rewrite <- pfEq in Hr.
	simpl in Hr.
	unfold badd32, bzero32 in Hr; rewrite badd_zero' in Hr.
	rewrite <- Hr.
	trivial.

	destruct H.
	split; simpl; intuition.
      
	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r); subst; auto.
	apply raiseTy_sound.
	apply Ht;
	  [apply lowerTy_sound; apply mcRegs
	    | trivial].

      	rewrite SalRegMap.gsspec.
	destruct (SalRegIndexed.eq r0 r'); subst; auto.
	apply raiseTy_sound.
	destruct mcCust0; destruct mcCust0.
	apply Ht' with st; auto.

	intro.
	generalize (wcAddrs0 a0).
	unfold concAddr.
	destruct (ctxCust ctx a0); intuition.
	add Hr (wcTmp2 r); red in Hr.
	unfold ty, mastate in pfEq.
	rewrite <- pfEq in Hr.
	simpl in Hr.
	unfold badd32, bzero32 in Hr; rewrite badd_zero' in Hr.
	rewrite <- Hr.
	trivial.
      Qed.

      Lemma bsub_comm : forall n1 n2,
	bsub32 n1 n2 = bzero32
	-> bsub32 n2 n1 = bzero32.
	intros.
	assert (bneg (bsub32 n1 n2) = bneg bzero32).
	congruence.
	unfold bzero32 in H0; rewrite bzero_neg in H0.
	apply btn_inj.
	unfold bzero32.

	rewrite <- H0.
	assert (eqMod (pow2 32) (bvec_to_nat (bsub32 n2 n1)) (bvec_to_nat (bneg (bsub32 n1 n2)))).
	unfold bsub32.
	apply eqMod_sym.
	apply bsub_btn_neg.

	apply eqMod_inj with (pow2 32); trivial.
	generalize (btn_upper (bsub32 n2 n1)); omega.
	generalize (btn_upper (bneg (bsub32 n1 n2))); omega.
      Qed.

      Hint Resolve bsub_comm.

      Definition considerTest : forall (abs : stateDescription (stackTy Sys.ty)
	(stackState Sys.ty Sys.mastate))
      (co : cc) (aop : arith_op) (e1 e2 : salExp) (b : bool),
      {abs' : _
	| forall ctx st,
	  typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs
	  -> eval_cc co aop (evalExp st e1) (evalExp st e2) = b
	  -> typeMconc (stackHasTy Sys.hasTy) (stackConc Sys.hasTy Sys.mconc) ctx st abs'}.
	intros.
	refine (match (co, b) as cob return (cob = (co, b) -> _) with
		  | ((false, Z), true) => fun pfEq =>
		    match (aop, e1, e2) as aope1e2 return (aope1e2 = (aop, e1, e2) -> _) with
		      | (Sub, SReg r, SConst n) => fun pfEq' =>
			abs' ::: Habs' <- doCastNeq r n abs;
			exist _ abs' _
		      | (Sub, SConst n, SReg r) => fun pfEq' =>
			abs' ::: Habs' <- doCastNeq r n abs;
			exist _ abs' _
		      | _ => fun _ => exist _ abs _
		    end (refl_equal _)
		  | ((true, Z), false) => fun pfEq =>
		    match (aop, e1, e2) as aope1e2 return (aope1e2 = (aop, e1, e2) -> _) with
		      | (Sub, SReg r, SConst n) => fun pfEq' =>
			abs' ::: Habs' <- doCastNeq r n abs;
			exist _ abs' _
		      | (Sub, SConst n, SReg r) => fun pfEq' =>
			abs' ::: Habs' <- doCastNeq r n abs;
			exist _ abs' _
		      | _ => fun _ => exist _ abs _
		    end (refl_equal _)
		  | ((false, Z), false) => fun pfEq =>
		    match (aop, e1, e2) as aope1e2 return (aope1e2 = (aop, e1, e2) -> _) with
		      | (Sub, SReg r, SConst n) => fun pfEq' =>
			abs' ::: Habs' <- doCastEq r n abs;
			exist _ abs' _
		      | (Sub, SConst n, SReg r) => fun pfEq' =>
			abs' ::: Habs' <- doCastEq r n abs;
			exist _ abs' _
		      | _ => fun _ => exist _ abs _
		    end (refl_equal _)
		  | ((true, Z), true) => fun pfEq =>
		    match (aop, e1, e2) as aope1e2 return (aope1e2 = (aop, e1, e2) -> _) with
		      | (Sub, SReg r, SConst n) => fun pfEq' =>
			abs' ::: Habs' <- doCastEq r n abs;
			exist _ abs' _
		      | (Sub, SConst n, SReg r) => fun pfEq' =>
			abs' ::: Habs' <- doCastEq r n abs;
			exist _ abs' _
		      | _ => fun _ => exist _ abs _
		    end (refl_equal _)
		  | _ => fun _ => exist _ abs _
		end (refl_equal _)); intros; try (injection pfEq; clear pfEq);
	try (injection pfEq'; clear pfEq');
	  intros; subst; intuition.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 n (get_reg st r) = bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 n (get_reg st r)) bzero32); trivial; discriminate.
	apply bsub32_eq'; auto.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 (get_reg st r) n = bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 (get_reg st r) n) bzero32); trivial; discriminate.
	apply bsub32_eq'; auto.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 n (get_reg st r) <> bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 n (get_reg st r)) bzero32); trivial; discriminate.
	intro Hwrong.
	rewrite Hwrong in H1.
	rewrite bsub32_eq in H1.
	tauto.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 (get_reg st r) n <> bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 (get_reg st r) n) bzero32); trivial; discriminate.
	intro Hwrong.
	rewrite Hwrong in H1.
	rewrite bsub32_eq in H1.
	tauto.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 n (get_reg st r) <> bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 n (get_reg st r)) bzero32); trivial; discriminate.
	intro Hwrong.
	rewrite Hwrong in H1.
	rewrite bsub32_eq in H1.
	tauto.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 (get_reg st r) n <> bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 (get_reg st r) n) bzero32); trivial; discriminate.
	intro Hwrong.
	rewrite Hwrong in H1.
	rewrite bsub32_eq in H1.
	tauto.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 n (get_reg st r) = bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 n (get_reg st r)) bzero32); trivial; discriminate.
	apply bsub32_eq'; auto.

	apply Habs'; auto.
	unfold eval_cc, eval_condition', eval_ZF in H0.
	simpl in H0.
	assert (bsub32 (get_reg st r) n = bzero32).
	unfold bvec_eq in H0.
	destruct (bvec_eq_dec (bsub32 (get_reg st r) n) bzero32); trivial; discriminate.
	apply bsub32_eq'; auto.
      Qed.
    End WithTesting.

    Module CondSys := AddCond(WithTesting).
    Import CondSys.

    Definition ty := ty.
    Definition print_ty := print_ty.
    Definition mastate := mastate.
    Definition print_mastate := print_mastate.
    Definition mcontext := mcontext.
    Definition mconc := mconc.
    Definition subMastate := subMastate.
    Definition hasTy := hasTy.
    Definition tyTop := tyTop.
    Definition tyTopSound := tyTopSound.
    Definition tyConst := tyConst.
    Definition tyConstSound1 := tyConstSound1.
    Definition tyConstSound2 := tyConstSound2.
    Definition isTyConst := isTyConst.
    Definition subTy := subTy.
    Definition typeof := typeof.
    Definition typeofLoad := typeofLoad.
    Definition checkStore := checkStore.
    Definition effectOfCmd := effectOfCmd.
    Definition doStackWrite := doStackWrite.
    Definition effectOfJump := effectOfJump.
    Definition effectOfConditionalJump := effectOfConditionalJump.
    Definition viewShiftOfJump := viewShiftOfJump.
  End Create.

End Make.
