(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Even/odd abstract interpretation *)

Require Import Bool Bvector Eqdep.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Vector.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Monad.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.SimpleTypes.
Require Import ProofOS.AbsInt.SimpleDriver.

Set Implicit Arguments.

Inductive evenOddTy : Set :=
  | Anything
  | Constant : int32 -> evenOddTy
  | Even
  | Odd.

Parameter print_evenOddTy : evenOddTy -> unit.

Definition isEven (v : int32) :=
  match v with
    | Vcons false _ _ => True
    | _ => False
  end.

Definition isOdd (v : int32) :=
  match v with
    | Vcons true _ _ => True
    | _ => False
  end.

Definition to_int32 : forall size, vector bool size
  -> size = 32 -> int32.
  intros; subst; trivial.
Defined.

Definition parity_dec : forall (v : int32), {isEven v} + {isOdd v}.
  intros.
  refine (match v as v' in (vector _ size) return (forall (pfSize : size = 32), to_int32 v' pfSize = v -> _) with
	    | Vcons false _ _ => fun pfSize pfEq => left (isOdd v) _
	    | Vcons true _ _ => fun pfSize pfEq => right (isEven v) _
	    | _ => fun pfSize pfEq => left (isOdd v) _
	  end (refl_equal _) _); subst; try discriminate;
  try (assert (n = 31); [omega | subst];
    rewrite (UIP_refl _ _ pfSize);
      simpl); trivial.
Qed.

Inductive hasEvenOddTy : int32 -> evenOddTy -> Prop :=
  | HT_Anything : forall v,
    hasEvenOddTy v Anything
  | HT_Constant : forall v,
    hasEvenOddTy v (Constant v)
  | HT_Even : forall v,
    isEven v
    -> hasEvenOddTy v Even
  | HT_Odd : forall v,
    isOdd v
    -> hasEvenOddTy v Odd.

Lemma rewrite_double : forall n,
  n + (n + 0) = 2 * n.
  intros; omega.
Qed.

Lemma merge_times : forall m1 m2 n,
  n * m1 + n * m2 = n * (m1 + m2).
  induction n; simpl; intuition.
Qed.

Import Datatypes.

Lemma merge_times_S : forall m1 m2 n,
  n * m1 + S (n * m2) = S (n * (m1 + m2)).
  induction n; simpl; intuition.
Qed.

Lemma merge_times_S2 : forall m1 m2 n,
  S (n * m1) + n * m2 = S (n * (m1 + m2)).
  induction n; simpl; intuition.
Qed.

Lemma merge_times_S3 : forall m1 m2,
  S (2 * m1) + S (2 * m2) = 2 * (m1 + m2 + 1).
  intros; omega.
Qed.

Hint Rewrite rewrite_double merge_times merge_times_S
  merge_times_S2 merge_times_S3 : fixup.

Hint Rewrite mod2_bool_2 mod2_bool_S2 : fixup.

Ltac int32_arith :=
  let v1 := fresh "v1" with v2 := fresh "v2"
    with Hv1 := fresh "Hv1" with Hv2 := fresh "Hv2"
    with Hv1Eq := fresh "Hv1Eq" with Hv2Eq := fresh "Hv2Eq" in
    (intros v1 v2 Hv1 Hv2;
      simpl; unfold int32, Bvector in *;
	VSntac Hv1Eq v1; VSntac Hv2Eq v2;
	rewrite Hv1Eq in Hv1; rewrite Hv2Eq in Hv2;
	  simpl in *;
	    destruct (Vhead v1); try tauto;
	      destruct (Vhead v2); try tauto;
		autorewrite with fixup; trivial).

Lemma add_even_even : forall v1 v2,
  isEven v1
  -> isEven v2
  -> isEven (eval_arith Add v1 v2).
  int32_arith.
Qed.

Lemma add_even_odd : forall v1 v2,
  isEven v1
  -> isOdd v2
  -> isOdd (eval_arith Add v1 v2).
  int32_arith.
Qed.

Lemma add_odd_even : forall v1 v2,
  isOdd v1
  -> isEven v2
  -> isOdd (eval_arith Add v1 v2).
  int32_arith.
Qed.

Lemma add_odd_odd : forall v1 v2,
  isOdd v1
  -> isOdd v2
  -> isEven (eval_arith Add v1 v2).
  int32_arith.
Qed.

Hint Resolve add_even_even add_even_odd add_odd_even add_odd_odd.

Module EvenOddTySys <: SIMPLE_TYPE_SYSTEM.
  Definition ty := evenOddTy.
  Definition print_ty := print_evenOddTy.
  
  Definition hasTy := hasEvenOddTy.
  Hint Unfold hasTy.
  Hint Constructors hasEvenOddTy.
  
  Definition tyTop := Anything.
  Theorem tyTopSound : forall v, hasTy v tyTop.
    intuition.
  Qed.
  
  Definition tyConst := Constant.
  Theorem tyConstSound1 : forall v, hasTy v (tyConst v).
    unfold tyConst; intuition.
  Qed.
  Theorem tyConstSound2 : forall v1 v2, hasTy v1 (tyConst v2) -> v1 = v2.
    unfold tyConst; inversion 1; trivial.
  Qed.
  Definition isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).
    intros.
    refine (match t as t return (sig_option (fun n => t = tyConst n)) with
	      | Constant n => SSome _ n _
	      | _ => SNone _
	    end); trivial.
  Qed.

  Definition subTy : forall (t1 t2 : ty),
    pred_option (forall v,
      hasTy v t1 -> hasTy v t2).
    intros.
    refine (match (t1, t2) as ts return (ts = (t1, t2) -> _) with
	      | (_, Anything) => fun pfEq => PSome _
	      | (Even, Even) => fun pfEq => PSome _
	      | (Odd, Odd) => fun pfEq => PSome _
	      | (Constant n, Even) => fun pfEq =>
		if parity_dec n
		  then PSome _
		  else PNone _
	      | (Constant n, Odd) => fun pfEq =>
		if parity_dec n
		  then PNone _
		  else PSome _
	      | _ => fun _ => PNone _
	    end (refl_equal _)); injection pfEq; intros; subst; try inversion H1; intuition.
  Qed.

  Definition relaxTy : forall (t : ty), {t' : ty | forall v, hasTy v t -> hasTy v t'}.
    intros.
    refine (match t as t' return (t' = t -> _) with
	      | Constant n => fun pfEq =>
		if parity_dec n
		  then exist _ Even _
		  else exist _ Odd _
	      | _ => fun pfEq => exist _ t _
	    end (refl_equal _)); subst; intuition;
    inversion H; eauto.
  Qed.

  Definition typeofArith : forall (aop : arith_op) (t1 t2 : ty),
    {t : ty
      | forall v1 v2,
	hasTy v1 t1
	-> hasTy v2 t2
	-> hasTy (eval_arith aop v1 v2) t}.
    intros.
    refine (match aop as aop' return (aop' = aop -> _) with
	      | Add => fun pfEq =>
		match (t1, t2) as ts return (ts = (t1, t2) -> _) with
		  | (Constant n1, Constant n2) => fun pfEq' => exist _ (Constant (badd32 n1 n2)) _
		  | _ => fun pfEq' =>
		    t1' ::: Ht1' <- relaxTy t1;
		    t2' ::: Ht2' <- relaxTy t2;
		    match (t1', t2') as ts return (ts = (t1', t2') -> _) with
		      | (Even, Even) => fun pfEq'' => exist _ Even _
		      | (Even, Odd) => fun pfEq'' => exist _ Odd _
		      | (Odd, Even) => fun pfEq'' => exist _ Odd _
		      | (Odd, Odd) => fun pfEq'' => exist _ Even _
		      | _ => fun _ => exist _ Anything _
		    end (refl_equal _)
		end (refl_equal _)
	      | _ => fun _ => exist _ Anything _
	    end (refl_equal _)); intuition; injection pfEq'; clear pfEq'; intros; subst;
    try (injection pfEq''; clear pfEq''; intros; subst);
      inversion H; inversion H0;
	try (add Ht1'' (Ht1' _ H); add Ht2'' (Ht2' _ H0);
	  inversion Ht1''; inversion Ht2''); subst; eauto.
  Qed.
End EvenOddTySys.

Module Verifier (Param : DRIVER_REGIONS).
  Module MyParam <: DRIVER_PARAM.
    Module Regions := Param.
    Module Sys := EvenOddTySys.
  End MyParam.

  Module Dr := Driver(MyParam).
End Verifier.
