(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Integer pointer type-checking *)

Require Import Bool Bvector Eqdep.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Vector.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Monad.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.WeakUpdateTypes.
Require Import ProofOS.AbsInt.WeakDriver.
Require Import ProofOS.AbsInt.Sal.

Set Implicit Arguments.

Inductive intPtrTy : Set :=
  | Anything
  | Constant : int32 -> intPtrTy
  | Ptr
  | NnPtr.

Parameter print_intPtrTy : intPtrTy -> unit.

Inductive hasIntPtrTy : (int32 -> option intPtrTy) -> int32 -> intPtrTy -> Prop :=
  | HT_Anything : forall ctx v,
    hasIntPtrTy ctx v Anything
  | HT_Constant : forall ctx v,
    hasIntPtrTy ctx v (Constant v)
  | HT_Ptr_Null : forall ctx,
    hasIntPtrTy ctx bzero32 Ptr
  | HT_Ptr_NonNull : forall ctx v,
    ctx v = Some Anything
    -> hasIntPtrTy ctx v Ptr
  | HT_NnPtr : forall ctx v,
    ctx v = Some Anything
    -> hasIntPtrTy ctx v NnPtr.

Module IntPtrTySys <: WEAK_UPDATE_TYPE_SYSTEM.
  Definition ty := intPtrTy.
  Definition print_ty := print_intPtrTy.
  
  Definition hasTy := hasIntPtrTy.
  Hint Unfold hasTy.
  Hint Constructors hasIntPtrTy.
  
  Definition tyTop := Anything.
  Definition tyTopSound := HT_Anything.

  Definition tyConst := Constant.
  Definition tyConstSound1 := HT_Constant.
  Theorem tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
    unfold tyConst; inversion 1; trivial.
  Qed.
  Definition isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).
    intros.
    refine (match t as t return (sig_option (fun n => t = tyConst n)) with
	      | Constant n => SSome _ n _
	      | _ => SNone _
	    end); trivial.
  Qed.
  
  Definition subTy : forall (t1 t2 : ty),
    pred_option (forall ctx v,
      hasTy ctx v t1 -> hasTy ctx v t2).
    intros.
    refine (match (t1, t2) as ts return (ts = (t1, t2) -> _) with
	      | (_, Anything) => fun pfEq => PSome _
	      | (Constant n1, Constant n2) => fun pfEq =>
		if int32_eq_dec n1 n2
		  then PSome _
		  else PNone _
	      | (Ptr, Ptr) => fun pfEq => PSome _
	      | (NnPtr, NnPtr) => fun pfEq => PSome _
	      | (NnPtr, Ptr) => fun pfEq => PSome _
	      | (Constant n, Ptr) => fun pfEq =>
		if int32_eq_dec n bzero32
		  then PSome _
		  else PNone _
	      | _ => fun _ => PNone _
	    end (refl_equal _)); injection pfEq; intros; subst; simpl; intuition;
    inversion H1; auto.
  Qed.

  Definition typeofArith : forall (aop : arith_op) (t1 t2 : ty),
    {t : ty
      | forall ctx v1 v2,
	hasTy ctx v1 t1
	-> hasTy ctx v2 t2
	-> hasTy ctx (eval_arith aop v1 v2) t}.
    intros.
    refine (match aop as aop' return (aop' = aop -> _) with
	      | Add => fun pfEq =>
		match (t1, t2) as ts return (ts = (t1, t2) -> _) with
		  | (Constant n, _) => fun pfEq' =>
		    if int32_eq_dec n bzero32
		      then exist _ t2 _
		      else exist _ Anything _
		  | (_, Constant n) => fun pfEq' =>
		    if int32_eq_dec n bzero32
		      then exist _ t1 _
		      else exist _ Anything _
		  | _ => fun _ => exist _ Anything _
		end (refl_equal _)
	      | _ => fun _ => exist _ Anything _
	    end (refl_equal _)); try injection pfEq'; intros; subst; simpl; intuition.

    inversion H1; subst; unfold badd32, bzero32; rewrite badd_zero'; trivial.
    inversion H2; subst; unfold badd32, bzero32; rewrite badd_comm; rewrite badd_zero'; trivial.
    inversion H2; subst; unfold badd32, bzero32; rewrite badd_comm; rewrite badd_zero'; trivial.
  Qed.

  Definition typeofCell : forall (t : ty),
    sig_option (fun t' =>
      forall ctx v,
	hasTy ctx v t
	-> ctx v = Some t').
    intros.
    refine (match t as t return (sig_option
      (fun t' : intPtrTy =>
	forall (ctx : int32 -> option intPtrTy) (v : int32),
	  hasTy ctx v t -> ctx v = Some t')) with
	      | NnPtr => SSome _ Anything _
	      | _ => SNone _
	    end); intuition.

    inversion H; trivial.
  Qed.

  Definition considerNeq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx v1,
	hasTy ctx v1 t
	-> v1 <> v
	-> hasTy ctx v1 t'}.
    intros.
    refine (match t as t' return (t' = t -> _) with
	      | Ptr => fun pfEq =>
		if int32_eq_dec v bzero32
		  then exist _ NnPtr _
		  else exist _ t _
	      | _ => fun _ => exist _ t _
	    end (refl_equal _)); subst; intuition.

    inversion H; subst; auto; tauto.
  Qed.

  Definition considerDerefEq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx st v1,
	(forall a,
	  match ctx a with
	    | None => True
	    | Some t2 => hasTy ctx (get_mem32 st a) t2
	  end)
	-> hasTy ctx v1 t
	-> get_mem32 st v1 = v
	-> hasTy ctx v1 t'}.
    intros; exists t; trivial.
  Qed.

  Definition considerDerefNeq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx st v1,
	(forall a,
	  match ctx a with
	    | None => True
	    | Some t2 => hasTy ctx (get_mem32 st a) t2
	  end)
	-> hasTy ctx v1 t
	-> get_mem32 st v1 <> v
	-> hasTy ctx v1 t'}.
    intros; exists t; trivial.
  Qed.
End IntPtrTySys.


Module Verifier (Param : DRIVER_REGIONS).
  Module MyParam <: DRIVER_PARAM.
    Module Regions := Param.
    Module Sys := IntPtrTySys.
  End MyParam.

  Module Dr := Driver(MyParam).
End Verifier.
