(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** The usual suspects in type systems describing heap-allocated values *)

Require Import Arith ArithUtil Bool Bvector Eqdep.

Require Import Asm.Util.Specif.
Require Import Asm.Util.Vector.
Require Import Asm.Util.Mod.
Require Import Asm.Util.Monad.
Require Import Asm.Util.Tactics.

Require Import Asm.Bitvector.Bitvector.
Require Import Asm.X86.X86.

Require Import ProofOS.AbsInt.WeakUpdateTypes.
Require Import ProofOS.AbsInt.WeakDriver.
Require Import ProofOS.AbsInt.Sal.

Set Implicit Arguments.

Definition var := nat.

Inductive memoryTy : Set :=
  | Constant : int32 -> memoryTy
  | Product : productTy -> memoryTy
  | Sum : memoryTy -> memoryTy -> memoryTy

  | Var : var -> memoryTy
  | Recursive : var -> memoryTy -> memoryTy

    with productTy : Set :=
  | PT_Nil
  | PT_Cons : memoryTy -> productTy -> productTy.

Parameter print_memoryTy : memoryTy -> unit.

Fixpoint subst (x : var) (t b : memoryTy) {struct b} : memoryTy :=
  match b with
    | Constant _ => b
    | Product ts => Product (substProduct x t ts)
    | Sum b1 b2 => Sum (subst x t b1) (subst x t b2)
    | Var x' =>
      if eq_nat_dec x x'
	then t
	else b
    | Recursive x' b =>
      if eq_nat_dec x x'
	then b
	else Recursive x' (subst x t b)
  end

  with substProduct (x : var) (t : memoryTy) (ts : productTy) {struct ts} : productTy :=
  match ts with
    | PT_Nil => ts
    | PT_Cons b ts => PT_Cons (subst x t b) (substProduct x t ts)
  end.

Inductive hasMemoryTy : (int32 -> option memoryTy) -> int32 -> memoryTy -> Prop :=
  | HT_Constant : forall ctx v,
    hasMemoryTy ctx v (Constant v)
  | HT_Unit : forall ctx v,
    hasMemoryTy ctx v (Product PT_Nil)
  | HT_Product : forall ctx v t ts,
    ctx v = Some t
    -> hasMemoryTy ctx (binc4 v) (Product ts)
    -> hasMemoryTy ctx v (Product (PT_Cons t ts))
  | HT_Suml : forall ctx v t1 t2,
    hasMemoryTy ctx v
    (Product (PT_Cons (Constant bzero32)
      (PT_Cons t1 PT_Nil)))
    -> hasMemoryTy ctx v (Sum t1 t2)
  | HT_Sumr : forall ctx v t1 t2,
    hasMemoryTy ctx v
    (Product (PT_Cons (Constant bone32)
      (PT_Cons t2 PT_Nil)))
    -> hasMemoryTy ctx v (Sum t1 t2)
  | HT_Recursive : forall ctx x t v,
    hasMemoryTy ctx v (subst x (Recursive x t) t)
    -> hasMemoryTy ctx v (Recursive x t).

Module MemoryTySys <: WEAK_UPDATE_TYPE_SYSTEM.
  Definition ty := memoryTy.
  Definition print_ty := print_memoryTy.
  
  Definition hasTy := hasMemoryTy.
  Hint Unfold hasTy.
  Hint Constructors hasMemoryTy.
  
  Definition tyTop := Product PT_Nil.
  Definition tyTopSound := HT_Unit.

  Definition tyConst := Constant.
  Definition tyConstSound1 := HT_Constant.
  Theorem tyConstSound2 : forall ctx v1 v2, hasTy ctx v1 (tyConst v2) -> v1 = v2.
    unfold tyConst; inversion 1; trivial.
  Qed.
  Definition isTyConst : forall (t : ty), sig_option (fun n => t = tyConst n).
    intros.
    refine (match t as t return (sig_option (fun n => t = tyConst n)) with
	      | Constant n => SSome _ n _
	      | _ => SNone _
	    end); trivial.
  Qed. 

  Fixpoint ty_eq_dec' (a b : ty) {struct b} : bool :=
    match (a, b) with
      | (Constant k1, Constant k2) => sum_to_bool (int32_eq_dec k1 k2)
      | (Product ts1, Product ts2) => product_eq_dec' ts1 ts2
      | (Sum t1 u1, Sum t2 u2) => ty_eq_dec' t1 t2 && ty_eq_dec' u1 u2
      | (Recursive t1 u1, Recursive t2 u2) => sum_to_bool (eq_nat_dec t1 t2) && ty_eq_dec' u1 u2
      | (Var x1, Var x2) => sum_to_bool (eq_nat_dec x1 x2)
      | _ => false
    end

    with product_eq_dec' (a b : productTy) {struct b} : bool :=
      match (a, b) with
	| (PT_Nil, PT_Nil) => true
	| (PT_Cons t1 ts1, PT_Cons t2 ts2) =>
	  ty_eq_dec' t1 t2 && product_eq_dec' ts1 ts2
	| _ => false
      end.

  Scheme ty_ind := Induction for memoryTy Sort Prop
    with productTy_ind := Induction for productTy Sort Prop.

  Lemma and_true : forall b1 b2,
    b1 && b2 = true -> b1 = true /\ b2 = true.
    destruct b1; destruct b2; simpl; intuition.
  Qed.

  Lemma and_false : forall b1 b2,
    b1 && b2 = false -> b1 = false \/ b2 = false.
    destruct b1; destruct b2; simpl; intuition.
  Qed.

  Lemma ty_eq_dec'_sound1 : forall (b a : ty),
    ty_eq_dec' a b = true
    -> a = b.
    apply (ty_ind
      (fun b => forall a, ty_eq_dec' a b = true -> a = b)
      (fun b => forall a, product_eq_dec' a b = true -> a = b)); simpl; intuition.
    destruct a; try discriminate;
      destruct (int32_eq_dec i0 i); try discriminate; congruence.
    destruct a; try discriminate.
    rewrite (H p0); trivial.

    destruct a; try discriminate.
    destruct (and_true _ _ H1).
    rewrite (H a1); trivial.
    rewrite (H0 a2); trivial.

    destruct a; try discriminate.
    destruct (eq_nat_dec v0 v); try discriminate; congruence.

    destruct a; try discriminate.
    destruct (and_true _ _ H0).
    destruct (eq_nat_dec v0 v); try discriminate.
    rewrite e.
    rewrite (H a); auto.

    destruct a; trivial; discriminate.
    destruct a; try discriminate.
    destruct (and_true _ _ H1).
    rewrite (H m0); trivial.
    rewrite (H0 a); trivial.
  Qed.

  Lemma ty_eq_dec'_sound2 : forall (b a : ty),
    ty_eq_dec' a b = false
    -> a <> b.
    apply (ty_ind
      (fun b => forall a, ty_eq_dec' a b = false -> a <> b)
      (fun b => forall a, product_eq_dec' a b = false -> a <> b)); simpl; intuition.
    destruct a; try discriminate; destruct (int32_eq_dec i0 i);
      try discriminate; congruence.
    destruct a; try discriminate.
    generalize (H _ H0); congruence.

    destruct a; try discriminate.
    destruct (and_false _ _ H1).
    generalize (H _ H3); congruence.
    generalize (H0 _ H3); congruence.

    destruct a; try discriminate.
    destruct (eq_nat_dec v0 v); try discriminate; congruence.

    destruct a; try discriminate.
    destruct (and_false _ _ H0).
    destruct (eq_nat_dec v0 v); try discriminate; congruence.
    apply (H a); trivial.
    congruence.

    destruct a; discriminate.
    destruct a; try discriminate.
    destruct (and_false _ _ H1).
    generalize (H _ H3); congruence.
    generalize (H0 _ H3); congruence.
  Qed.

  Definition ty_eq_dec : forall (a b : ty), {a = b} + {a <> b}.
    intros.
    refine ((if ty_eq_dec' a b as ted return (ty_eq_dec' a b = ted -> _)
      then fun pfEq => left (a <> b) _
      else fun pfEq => right (a = b) _) (refl_equal _));
    [apply ty_eq_dec'_sound1 | apply ty_eq_dec'_sound2]; trivial.
  Qed.

  Fixpoint subTy' (t1 t2 : ty) {struct t1} : bool :=
    match (t1, t2) with
      | (_, Product PT_Nil) => true
      | (Product ts1, Product ts2) => subProduct' ts1 ts2
      | (Sum t1 u1, Sum t2 u2) => ty_eq_dec' t1 t2 && ty_eq_dec' u1 u2
      | (Product (PT_Cons (Constant n) (PT_Cons t PT_Nil)), Sum t1 t2) =>
	(sum_to_bool (int32_eq_dec n bzero32)
	  && ty_eq_dec' t t1)
	|| (sum_to_bool (int32_eq_dec n bone32)
	  && ty_eq_dec' t t2)
      | (Constant n1, Constant n2) =>
	if int32_eq_dec n1 n2
	  then true
	  else false
      | (Var x1, Var x2) =>
	if eq_nat_dec x1 x2
	  then true
	  else false
      | (Recursive x1 b1, Recursive x2 b2) =>
	if eq_nat_dec x1 x2
	  then ty_eq_dec' b1 b2
	  else false
      | _ => false
    end

    with subProduct' (ts1 ts2 : productTy) {struct ts1} : bool :=
    match (ts1, ts2) with
      | (_, PT_Nil) => true
      | (PT_Cons t1 ts1, PT_Cons t2 ts2) =>
	ty_eq_dec' t1 t2 && subProduct' ts1 ts2
      | _ => false
    end.

  Lemma or_true : forall b1 b2,
    b1 || b2 = true -> b1 = true \/ b2 = true.
    destruct b1; destruct b2; simpl; intuition.
  Qed.

  Lemma subTy'_sound : forall (t2 t1 : ty),
    subTy' t1 t2 = true
    -> forall ctx v,
      hasTy ctx v t1 -> hasTy ctx v t2.
    apply (ty_ind
      (fun t2 => forall t1, subTy' t1 t2 = true ->
	forall ctx v, hasTy ctx v t1 -> hasTy ctx v t2)
      (fun ts2 => forall ts1, subProduct' ts1 ts2 = true ->
	forall ctx v, hasTy ctx v (Product ts1) -> hasTy ctx v (Product ts2)));
    simpl; intuition; try destruct t1; simpl in *; intuition; try discriminate.
    unfold subTy' in H; destruct (int32_eq_dec i0 i); try discriminate;
      rewrite <- e; trivial.
    destruct p; try discriminate; auto.
    unfold subTy' in H0;
      destruct p; eauto.

    destruct m; discriminate.
    destruct m; try discriminate; destruct m0; try discriminate; destruct p; discriminate.
    destruct p; auto; discriminate.

    assert (subProduct' p0 p = true).
    unfold subTy' in H0.
    destruct p0; try destruct p; try destruct m; do 2 try destruct p0; trivial.
    eauto.

    destruct p; auto; discriminate.

    inversion H1.

    destruct p; auto; discriminate.

    unfold subTy' in H1.
    destruct p; try discriminate; destruct m1; try discriminate;
      do 2 (destruct p; try discriminate).
    destruct (or_true _ _ H1); destruct (and_true _ _ H3);
      rewrite <- (ty_eq_dec'_sound1 _ _ H5);
    [destruct (int32_eq_dec i bzero32)
      | destruct (int32_eq_dec i bone32)]; try discriminate;
    rewrite e in H2;
      eauto.

    unfold subTy' in H1.
    destruct (and_true _ _ H1).
    inversion H2; subst;
      inversion H8; subst;
	inversion H11; subst;
	  [rewrite <- (ty_eq_dec'_sound1 _ _ H3)
	    | rewrite <- (ty_eq_dec'_sound1 _ _ H4)];
	  eauto.

    destruct p; try destruct m; do 2 try destruct p; discriminate.

    unfold subTy' in H.
    destruct (eq_nat_dec v1 v); subst; trivial; discriminate.

    destruct p; try destruct m0; do 2 try destruct p; discriminate.

    unfold subTy' in H0.
    destruct (eq_nat_dec v1 v); subst; try discriminate.
    rewrite <- (ty_eq_dec'_sound1 _ _ H0).
    trivial.

    destruct ts1; try discriminate.
    inversion H2; subst.
    simpl in H1.
    destruct (and_true _ _ H1).
    rewrite <- (ty_eq_dec'_sound1 _ _ H3).
    unfold hasTy in *; eauto.
  Qed.

  Hint Resolve subTy'_sound.

  Definition subTy2 : forall (t1 t2 : ty),
    pred_option (forall ctx v,
      hasTy ctx v t1 -> hasTy ctx v t2).
    intros.
    refine ((if subTy' t1 t2 as sub return (subTy' t1 t2 = sub -> _)
      then fun pfEq => PSome _
      else fun pfEq => PNone _) (refl_equal _)); eauto.
  Qed.

  Definition subTy1 : forall (t1 t2 : ty),
    pred_option (forall ctx v,
      hasTy ctx v t1 -> hasTy ctx v t2).
    intros.
    refine (if subTy2 t1 t2
      then PSome _
      else match t2 as t2 return (pred_option (forall ctx v, hasTy ctx v t1 -> hasTy ctx v t2)) with
	     | Recursive x b =>
	       Hsub <-- subTy2 t1 (subst x (Recursive x b) b);;
	       PSome _
	      | _ => PNone _
	   end); intuition.
    apply HT_Recursive.
    apply Hsub; auto.
  Qed.

  Definition subTy : forall (t1 t2 : ty),
    pred_option (forall ctx v,
      hasTy ctx v t1 -> hasTy ctx v t2).
    intros.
    refine (if subTy1 t1 t2
      then PSome _
      else match t1 as t1 return (pred_option (forall ctx v, hasTy ctx v t1 -> hasTy ctx v t2)) with
	     | Recursive x b =>
	       Hsub <-- subTy2 (subst x (Recursive x b) b) t2;;
	       PSome _
	      | _ => PNone _
	   end); intuition.
    inversion H; subst.
    apply Hsub; auto.
  Qed.

  Import Datatypes.

  Lemma btn_four : forall len, bvec_to_nat (nat_to_bvec (S (S (S len))) 4) = 4.
    intros.
    simpl.
    unfold mod2_bool, mod2, div2.
    destruct (Euclid.eucl_dev 2 two_gt_0 4); try (assert False; [omega | tauto]).
    destruct (eq_nat_dec r 0); try (assert False; [omega | tauto]).
    destruct (Euclid.eucl_dev 2 two_gt_0 q); try (assert False; [omega | tauto]).
    destruct (eq_nat_dec r0 0); try (assert False; [omega | tauto]).
    destruct (Euclid.eucl_dev 2 two_gt_0 q0); try (assert False; [omega | tauto]).
    destruct (eq_nat_dec r1 0); try (assert False; [omega | tauto]).
    replace q1 with 0.
    rewrite ntb_invert; trivial.
    induction len; simpl; omega.
    omega.
  Qed.

  Lemma eqMod_refl' : forall m n1 n2,
    n1 = n2
    -> eqMod m n1 n2.
    intros; subst; apply eqMod_refl.
  Qed.

  Lemma four_lt_pow2 : forall len, 4 <= pow2 (S (S len)).
    simpl; intros.
    generalize (pow2_pos len).
    omega.
  Qed.

  Lemma add_sub_four : forall n v,
    bvec_to_nat n >= 4
    -> badd32 n v = badd32 (bsub32 n four) (binc4 v).
    intros.
    unfold badd32, badd, bsub32, bsub, barith.
    apply ntb_mod.
    replace (bvec_to_nat (nat_to_bvec 32 (bvec_to_nat n + bvec_to_nat (bneg four))))
      with (bvec_to_nat n - 4). 
    apply eqMod_trans
      with (bvec_to_nat n - 4 + S (S (S (S (bvec_to_nat v))))).
    apply eqMod_refl'; omega.
    apply eqMod_plus.
    apply eqMod_refl.
    apply eqMod_sym; apply binc4_btn.

    rewrite (ntb_mod 32 (n1 := bvec_to_nat n + bvec_to_nat (bneg four))
      (n2 := bvec_to_nat n - bvec_to_nat four)).
    unfold four; rewrite btn_four.
    rewrite ntb_invert; trivial.
    generalize (btn_upper n); omega.
    apply eqMod_sym; apply bneg_btn_sub.
    unfold four; rewrite btn_four; omega.
  Qed.

  Lemma bzero_inv : forall len (n : Bvector len),
    bvec_to_nat n = 0
    -> n = @bzero len.
    induction len; simpl; intros.
    myVOtac; trivial.
    myVSntac Hn n.
    rewrite Hn in H.
    simpl in H.
    destruct (Vhead n); try discriminate.
    rewrite <- (IHlen (Vtail n)); trivial.
    omega.
  Qed.

  Definition seekInProduct : forall (n : int32) (ts : productTy),
    {ts' : _
      | forall ctx v, hasTy ctx v (Product ts)
	-> hasTy ctx (badd32 n v) (Product ts')}.
    refine (fix seekInProduct (n : int32) (ts : productTy) {struct ts}
      : {ts' : _
	| forall ctx v, hasTy ctx v (Product ts)
	  -> hasTy ctx (badd32 n v) (Product ts')} :=
      match ts as ts' return (ts' = ts -> _) with
	| PT_Nil => fun pfEq => exist _ PT_Nil _
	| PT_Cons t rest => fun pfEq =>
	  let n' := bvec_to_nat n in
	    if eq_nat_dec n' 0
	      then exist _ ts _
	      else if le_lt_dec n' 3
		then exist _ PT_Nil _
		else (ts'' ::: Hts'' <- seekInProduct (bsub32 n four) rest;
           	  exist _ ts'' _)
      end (refl_equal _)); subst; intuition.

    rewrite (bzero_inv n); trivial.
    unfold badd32; rewrite badd_zero'; trivial.

    inversion H; subst.
    rewrite add_sub_four; eauto.
  Qed.

  Definition typeofArith : forall (aop : arith_op) (t1 t2 : ty),
    {t : ty
      | forall ctx v1 v2,
	hasTy ctx v1 t1
	-> hasTy ctx v2 t2
	-> hasTy ctx (eval_arith aop v1 v2) t}.
    intros.
    refine (match aop as aop' return (aop' = aop -> _) with
	      | Add => fun pfEq =>
		match (t1, t2) as ts return (ts = (t1, t2) -> _) with
		  | (Constant n, Product ts) => fun pfEq' =>
		    ts' ::: Hts' <- seekInProduct n ts;
		    exist _ (Product ts') _
		  | _ => fun _ => exist _ tyTop _
		end (refl_equal _)
	      | _ => fun _ => exist _ tyTop _
	    end (refl_equal _)); try injection pfEq'; intros; subst; simpl; intuition.

    inversion H1; eauto.
  Qed.

  Definition typeofCell : forall (t : ty),
    sig_option (fun t' =>
      forall ctx v,
	hasTy ctx v t
	-> ctx v = Some t').
    intros.
    refine (match t as t return (sig_option
      (fun t' => forall ctx v, hasTy ctx v t -> ctx v = Some t')) with
	      | Product (PT_Cons t _) => SSome _ t _
	      | _ => SNone _
	    end); intuition.

    inversion H; trivial.
  Qed.

  Definition considerNeq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx v1,
	hasTy ctx v1 t
	-> v1 <> v
	-> hasTy ctx v1 t'}.
    intros.
    exists t; trivial.
  Qed.

  Lemma zero_one_neq : forall len, @bzero (S len) <> @bone (S len).
    intros.
    unfold bone.
    simpl.
    replace 1 with (S (2 * 0)); trivial.
    rewrite mod2_bool_S2.
    discriminate.
  Qed.

  Definition considerDerefEq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx st v1,
	(forall a,
	  match ctx a with
	    | None => True
	    | Some t2 => hasTy ctx (get_mem32 st a) t2
	  end)
	-> hasTy ctx v1 t
	-> get_mem32 st v1 = v
	-> hasTy ctx v1 t'}.
    intros.
    refine (match t as t' return (t' = t -> _) with
	      | Sum t1 t2 => fun pfEq =>
		if int32_eq_dec v bzero32
		  then exist _ (Product (PT_Cons (Constant bzero32) (PT_Cons t1 PT_Nil))) _
		  else if int32_eq_dec v bone32
		    then exist _ (Product (PT_Cons (Constant bone32) (PT_Cons t2 PT_Nil))) _
		    else exist _ t _
	      | Recursive x b => fun pfEq =>
		match subst x (Recursive x b) b as t'' return (t'' = subst x (Recursive x b) b -> _) with
		  | Sum t1 t2 => fun pfEq' =>
		    if int32_eq_dec v bzero32
		      then exist _ (Product (PT_Cons (Constant bzero32) (PT_Cons t1 PT_Nil))) _
		      else if int32_eq_dec v bone32
			then exist _ (Product (PT_Cons (Constant bone32) (PT_Cons t2 PT_Nil))) _
			else exist _ t _
		  | _ => fun pfEq' => exist _ t _
		end (refl_equal _)
	      | _ => fun pfEq => exist _ t _
	    end (refl_equal _)); subst; intuition.
  
    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.
    rewrite H1 in H2.
    generalize (zero_one_neq (len := 31)).
    unfold bzero32, bone32 in H2; tauto.

    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.
    rewrite H1 in H2.
    generalize (zero_one_neq (len := 31)).
    unfold bzero32, bone32 in H2; tauto.

    inversion H0; subst.
    rewrite <- pfEq' in H5.
    clear H0.
    rename H5 into H0.
    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.
    rewrite H1 in H2.
    generalize (zero_one_neq (len := 31)).
    unfold bzero32, bone32 in H2; tauto.

    inversion H0; subst.
    rewrite <- pfEq' in H5.
    clear H0.
    rename H5 into H0.    
    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.
    rewrite H1 in H2.
    generalize (zero_one_neq (len := 31)).
    unfold bzero32, bone32 in H2; tauto.
  Qed.

  Definition considerDerefNeq : forall (t : ty) (v : int32),
    {t' : ty
      | forall ctx st v1,
	(forall a,
	  match ctx a with
	    | None => True
	    | Some t2 => hasTy ctx (get_mem32 st a) t2
	  end)
	-> hasTy ctx v1 t
	-> get_mem32 st v1 <> v
	-> hasTy ctx v1 t'}.
    intros.
    refine (match t as t' return (t' = t -> _) with
	      | Sum t1 t2 => fun pfEq =>
		if int32_eq_dec v bzero32
		  then exist _ (Product (PT_Cons (Constant bone32) (PT_Cons t2 PT_Nil))) _
		  else if int32_eq_dec v bone32
		    then exist _ (Product (PT_Cons (Constant bzero32) (PT_Cons t1 PT_Nil))) _
		    else exist _ t _
	      | Recursive x b => fun pfEq =>
		match subst x (Recursive x b) b as t'' return (t'' = subst x (Recursive x b) b -> _) with
		  | Sum t1 t2 => fun pfEq' =>
		    if int32_eq_dec v bzero32
		      then exist _ (Product (PT_Cons (Constant bone32) (PT_Cons t2 PT_Nil))) _
		      else if int32_eq_dec v bone32
			then exist _ (Product (PT_Cons (Constant bzero32) (PT_Cons t1 PT_Nil))) _
			else exist _ t _
		  | _ => fun pfEq' => exist _ t _
		end (refl_equal _)		
	      | _ => fun pfEq => exist _ t _
	    end (refl_equal _)); subst; intuition.
  
    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.

    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.

    inversion H0; subst.
    rewrite <- pfEq' in H5.
    clear H0; rename H5 into H0.
    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.

    inversion H0; subst.
    rewrite <- pfEq' in H5.
    clear H0; rename H5 into H0.
    inversion H0; subst; trivial.
    inversion H5; subst.
    add Htag (H v1).
    rewrite H7 in Htag.
    generalize (tyConstSound2 Htag); intuition.
  Qed.
End MemoryTySys.

Module Verifier (Param : DRIVER_REGIONS).
  Module MyParam <: DRIVER_PARAM.
    Module Regions := Param.
    Module Sys := MemoryTySys.
  End MyParam.

  Module Dr := Driver(MyParam).
End Verifier.
