(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Support code for FirstOrder verifiers *)

open AsmUtil
open X86Types
open X86Print
open Ver_types
open Ver_fixed
open FirstOrder_types
open Format


let word_type_to_ty = function
    TyTop -> Top
  | TyConst n -> Const n
  | TyCustom t -> Custom t

let ty_to_word_type = function
    Const n -> TyConst n
  | Custom t -> TyCustom t
  | _ -> TyTop


(** Memory layout parameters *)

let regions_loc = 2
let stackSize = CoqNat.of_int 8186 (* 8K of stack *)
let stackStart = 0x01000000L


(** Record constructors *)

let build_cutpoint_info pc sz rs ss cust =
  {ciPc = pc;
   ciStackSize = sz;
   ciRegs = rs;
   ciSlots = ss;
   ciCust = cust}

let build_function_info rt ag =
  {fiRet = rt;
   fiArgs = ag}


(** Parsing the binary image *)

let module_name =
  if Array.length Sys.argv < 3 then
    failwith "No module name specified on the command line"
  else
    Sys.argv.(2)

let regionsAddr =
  let bits' = skip binaryBits (8 * regions_loc) in
  let regionsAddr, bits' = read_uint bits' 32 in
  regionsAddr - initialPc_int

let eps, initialPc, actualCodeLen, dataStart, dataEnd, bssStart, bssEnd, metaAddr =
  let bits' = skip binaryBits (8 * regionsAddr) in
  let rec try_one bits =
    let name, bits = read_string bits in
    printf "Module: %s@\n" name;
    if name = "" then
      failwith "I couldn't find the module you asked for in the code file"
    else
      let rec read_entryPoints bits eps =
	let epname, bits = read_string bits in
	if epname = "" then begin
	  (*printf "Done with entries@\n";*)
	  eps, bits
	end else begin
	  let epaddr, bits = read_uint bits 32 in
	  (*printf "Read %s : %x@\n" epname epaddr;*)
	  read_entryPoints bits ((epname, epaddr) :: eps)
	end in
      let eps, bits = read_entryPoints bits [] in
      let text_start, bits = read_uint bits 32 in
      let text_end, bits = read_uint bits 32 in
      let data_start, bits = read_uint bits 32 in
      let data_end, bits = read_uint bits 32 in
      let bss_start, bits = read_uint bits 32 in
      let bss_end, bits = read_uint bits 32 in
      let metaAddr, bits = read_uint bits 32 in
      if name = module_name then
	(eps,
	 Int64.of_int text_start, CoqNat.of_int (text_end - text_start),
	 Int64.of_int data_start, Int64.of_int data_end,
	 Int64.of_int bss_start, Int64.of_int bss_end,
	 metaAddr - initialPc_int)
      else
	try_one bits in
  try_one bits'

let _ = printf "Regions: %Lx(%s), [%Lx-%Lx], [%Lx-%Lx], %x@\n"
    initialPc (Big_int.string_of_big_int actualCodeLen)
    dataStart dataEnd bssStart bssEnd metaAddr

module Parse (Param : FIRSTORDER) : FIRSTORDER_PARSE
with type typ = Param.typ and type astat = Param.astat = struct
  type typ = Param.typ
  type astat = Param.astat

  let read_ty bits =
    let kind, bits = read_uint bits 32 in
    match kind with
      0 -> Top, bits
    | 1 ->
	let n, bits = read_uint32 bits 32 in
	Const n, bits
    | 2 ->
	let n, bits = read_uint bits 32 in
	Stack (CoqNat.of_int n), bits
    | 3 -> OldEbp, bits
    | 4 -> RetPtr, bits
    | 5 ->
	let t, bits = Param.read_typ bits in
	Custom t, bits
    | _ -> failwith "Bad type to deserialize"

  open X86Print

  let pp_ty fmt = function
      Top -> fprintf fmt "T"
    | Const n -> fprintf fmt "C(%Ld)" n
    | Stack n -> fprintf fmt "S(%s)" (Big_int.string_of_big_int n)
    | OldEbp -> fprintf fmt "EBP0"
    | RetPtr -> fprintf fmt "RET"
    | Custom t -> Param.pp_typ fmt t

  let rec read_facts regs slots bits =
    let kind, bits = read_uint bits 32 in
    match kind with
      0 -> regs, slots, bits
    | 1 ->
	let rcode, bits = read_uint bits 32 in
	let ty, bits = read_ty bits in
	let r = X86Util.int_to_reg32 rcode in

	(*printf "Register %a : %a\n" pp_reg32 r pp_ty ty;*)
	
	read_facts (Cons (Pair (r, ty), regs)) slots bits
    | 2 ->
	let sl, bits = read_uint bits 32 in
	let ty, bits = read_ty bits in

	read_facts regs (Cons (Pair (CoqNat.of_int sl, ty), slots)) bits
    | _ -> failwith "Bad fact to deserialize"

  and read_cutpoint bits =
    let pc, bits' = read_uint bits 32 in
    let ss, bits' = read_uint bits' 32 in
    let regs, slots, bits' = read_facts Nil Nil bits' in
    let absIn, bits' = Param.read_astat bits' in
    let absOut, bits' = Param.read_astat bits' in
    {ciPc = ApcConst (Int64.of_int pc);
     ciStackSize = Big_int.big_int_of_int ss;
     ciRegs = regs;
     ciSlots = slots;
     ciCust = absIn}, absOut, bits'

  let four = CoqNat.of_int 4

  type entry_info = {
      eiRet : typ word_type;
      eiAbsIn : astat;
      eiAbsOut : astat
    }
  let entry_table = Hashtbl.create 100

  let function_table = Hashtbl.create 100

  let cutpoints = begin
    let bits' = skip binaryBits (8 * metaAddr) in

    let current_ret_loc = ref Big_int.zero_big_int in
    let current_arg_count = ref 0 in

    let rec read bits' conts cutpoints =
      let kind, bits' = read_uint bits' 32 in
      match kind with
	0 -> cutpoints
      | 1 ->
	  let cp, absOut, bits' = read_cutpoint bits' in
	  
	  let rec findEax = function
	      Cons (Pair (EAX, t), _) -> t
	    | Cons (_, rest) -> findEax rest
	    | Nil -> Top in

	  let rt = findEax cp.ciRegs in

	  (*let rec findArgs args = function
	     Cons (Pair (n, _), rest) -> findArgs (Big_int.max_big_int args n) rest
	     | _ -> args in

	     let args = findArgs Big_int.zero_big_int cp.ciSlots in*)

	  let rec findArgs args = function
	      Cons (Pair (n, ty), rest) -> findArgs (Cons (ty, args)) rest
	    | _ -> args in

	  let args = findArgs Nil cp.ciSlots in

	  begin match cp.ciPc with
	    ApcConst pc ->
	      Hashtbl.add function_table pc
		{fiRet = rt;
		 fiArgs = args}
	  | _ -> ()
	  end;

	  let alen = CoqNat.of_int (4 * coq_length args) in
	  let conts = Cons ({ciPc = ApcRet;
			     ciStackSize = Big_int.zero_big_int;
			     ciRegs = Cons (Pair (EAX, rt),
					    Cons (Pair (EBP, OldEbp),
						  Cons (Pair (ESP, Stack alen),
							Nil)));
			     ciSlots = Nil;
			     ciCust = absOut}, Nil) in

	  let ret_loc = Big_int.add_big_int alen four in
	  current_ret_loc := ret_loc;

	  let cp = {cp with
		    ciRegs = Cons (Pair (EBP, OldEbp),
				   Cons (Pair (ESP, Stack ret_loc), Nil));
		    ciSlots = Cons (Pair (ret_loc, RetPtr), cp.ciSlots)} in

	  begin match cp.ciPc with
	    ApcConst pc ->
	      Format.printf "Function at %a@\n" pp_addr pc;
	      Hashtbl.add entry_table pc
		{eiRet = ty_to_word_type rt;
		 eiAbsIn = cp.ciCust;
		 eiAbsOut = absOut}
	  | _ -> ()
	  end;

	  read bits' conts (Cons (Pair (conts, cp), cutpoints))
      | 2 ->
	  let cp, absOut, bits' = read_cutpoint bits' in

	  begin match cp.ciPc with
	    ApcConst pc -> Format.printf "Cutpoint at %a@\n" pp_addr pc
	  | _ -> ()
	  end;

	  read bits' conts (Cons (Pair (conts, cp), cutpoints))
      | _ -> failwith "Unknown cutpoint declaration type" in
    read bits' Nil Nil
  end

  let functions pc =
    try
      Some (Hashtbl.find function_table pc)
    with Not_found -> None

  let rec map_to_coq f = function
      [] -> Nil
    | h::t -> Cons (f h, map_to_coq f t)

  let entries = map_to_coq (fun (name, addr) ->
    try
      let addr = Int64.of_int addr in
      let ei = Hashtbl.find entry_table addr in
      {epPc = addr;
       epStackSize = CoqNat.of_int 8182;
       epArgs = Nil;
       epRet = ei.eiRet;
       epAbsIn = ei.eiAbsIn;
       epAbsOut = ei.eiAbsOut}
    with Not_found ->
      failwith (sprintf "Entry point %s not found as a function in this module" name)) eps
end

