(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Support types for FirstOrder verifiers *)

open X86Types
open Ver_types


(** {6 Abstract states} *)

(** Abstract types of word-sized values *)
type 'a ty =
  | Top                              (** Any value *)
  | Const of bool CoqBvector.bvector (** A known constant value *)
  | Stack of CoqNat.nat              (** Pointer to a constant stack slot *)
  | RetPtr                           (** Return pointer *)
  | OldEbp                           (** Saved value of EBP *)
  | Custom of 'a                     (** Custom type *)

(** Simpler types that the verifier extensions need to handle *)
type 'a word_type =
  | TyTop
  | TyConst of bool CoqBvector.bvector
  | TyCustom of 'a

(** Abstract program counters *)
type apc =
    ApcConst of bool CoqBvector.bvector (** Known program counter *)
  | ApcRet                              (** Return pointer *)


(** {6 Metadata} *)

(** Module entry points (functions) *)
type ('typ, 'astat) entryPoint = {
    epPc : bool CoqBvector.bvector;
    epStackSize : CoqNat.nat;
    epArgs : 'typ word_type coqlist;
    epRet : 'typ word_type;
    epAbsIn : 'astat;
    epAbsOut : 'astat
  }

(** Program point invariant *)
type ('a, 'b) cutpoint_info = {
    ciPc : apc;                                 (** Program counter *)
    ciStackSize : CoqNat.nat;                   (** Lower bound on available
					           stack space *)
    ciRegs : (reg32, 'a ty) prod coqlist;       (** Register types *)
    ciSlots : (CoqNat.nat, 'a ty) prod coqlist; (** Stack slot types *)
    ciCust : 'b                                 (** Custom data *)
  }

(** Function information *)
type 'a function_info = {
    fiRet : 'a ty;         (** Return type *)
    fiArgs : 'a ty coqlist (** Argument types *)
  }


(** {6 Customization-specific code, via a functor} *)

(** Information needed to read metadata for a particular verifier *)
module type FIRSTORDER = sig
  type typ
  (** Custom types *)
  type astat
  (** Additional verification state *)

  val read_typ : AsmUtil.bits -> typ * AsmUtil.bits

  val pp_typ : typ X86Print.printer

  val read_astat : AsmUtil.bits -> astat * AsmUtil.bits
end

(** An interface to that metadata *)
module type FIRSTORDER_PARSE = sig
  type typ
  type astat

  val cutpoints : ((typ, astat) cutpoint_info coqlist, (typ, astat) cutpoint_info) prod coqlist
  (** All program point invariants *)

  val functions : bool CoqBvector.bvector -> typ function_info option
  (** Look for information on a function that begins at a particular address *)

  val entries : (typ, astat) entryPoint coqlist
  (** Module entry points *)

  val pp_ty : typ ty X86Print.printer
end

