(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Support code for the extracted clunky Kernel verifier *)

open AsmUtil
open X86Types
open X86Print
open Ver_types
open Ver_fixed
open Kernel_types


(** Memory layout parameters *)

let regions_loc = 0x00100000
let cutpoints_loc = 0x00200000
let stackSize = CoqNat.of_int 8186 (* 8K of stack *)
let stackStart = 0x01000000L


(** Parsing the binary image *)

let actualCodeLen =
  let bits' = skip binaryBits (8 * regions_loc) in
  let code_end, bits' = read_uint bits' 32 in
  let actualCodeLen = CoqNat.of_int (code_end - initialPc_int) in
  (*Printf.printf "actualCodeLen = %x = %x - %x\n" (code_end - initialPc_int) code_end initialPc_int;*)
  actualCodeLen

let read_ty bits =
  let kind, bits = read_uint bits 32 in
  match kind with
    0 -> Top, bits
  | 1 ->
      let n, bits = read_uint32 bits 32 in
      Const n, bits
  | 2 ->
      let n, bits = read_uint bits 32 in
      Stack (CoqNat.of_int n), bits
  | _ -> failwith "Bad type to deserialize"

open Format
open X86Print

let pp_ty fmt = function
    Top -> printf "T"
  | Const n -> printf "C(%Ld)" n
  | Stack n -> printf "S(%s)" (Big_int.string_of_big_int n)
  | OldEbp -> printf "EBP0"
  | RetPtr -> printf "RET"

let rec read_facts regs slots bits =
  let kind, bits = read_uint bits 32 in
  match kind with
    0 -> regs, slots, bits
  | 1 ->
      let rcode, bits = read_uint bits 32 in
      let ty, bits = read_ty bits in
      let r = X86Util.int_to_reg32 rcode in

      (*printf "Register %a : %a\n" pp_reg32 r pp_ty ty;*)
      
      read_facts (Cons (Pair (r, ty), regs)) slots bits
  | 2 ->
      let sl, bits = read_uint bits 32 in
      let ty, bits = read_ty bits in

      read_facts regs (Cons (Pair (CoqNat.of_int sl, ty), slots)) bits
  | _ -> failwith "Bad fact to deserialize"

and read_cutpoint bits =
  let pc, bits' = read_uint bits 32 in
  let ss, bits' = read_uint bits' 32 in
  let regs, slots, bits' = read_facts Nil Nil bits' in
  {ciPc = ApcConst (Int64.of_int pc);
   ciStackSize = Big_int.big_int_of_int ss;
   ciRegs = regs;
   ciSlots = slots}, bits'

let four = CoqNat.of_int 4

let function_table = Hashtbl.create 100

let cutpoints = begin
  let bits' = skip binaryBits (8 * cutpoints_loc) in
  let rec read bits' conts cutpoints =
    let kind, bits' = read_uint bits' 32 in
    match kind with
      0 -> cutpoints
    | 1 ->
	let cp, bits' = read_cutpoint bits' in
	
	let rec findEax = function
	    Cons (Pair (EAX, t), _) -> t
	  | Cons (_, rest) -> findEax rest
	  | Nil -> Top in

	let rt = findEax cp.ciRegs in

	(*let rec findArgs args = function
	    Cons (Pair (n, _), rest) -> findArgs (Big_int.max_big_int args n) rest
	  | _ -> args in

	let args = findArgs Big_int.zero_big_int cp.ciSlots in*)

	let rec findArgs args = function
	    Cons (Pair (n, ty), rest) -> findArgs (Cons (ty, args)) rest
	  | _ -> args in

	let args = findArgs Nil cp.ciSlots in

	begin match cp.ciPc with
	  ApcConst pc ->
	    Hashtbl.add function_table pc
	      {fiRet = rt;
	       fiArgs = args}
	| _ -> ()
	end;

	let alen = CoqNat.of_int (4 * coq_length args) in
	let conts = Cons ({ciPc = ApcRet;
			   ciStackSize = Big_int.zero_big_int;
			   ciRegs = Cons (Pair (EAX, rt),
					  Cons (Pair (EBP, OldEbp),
						Cons (Pair (ESP, Stack alen),
						      Nil)));
			   ciSlots = Nil}, Nil) in

	let ret_loc = Big_int.add_big_int alen four in
	let cp = {cp with
		  ciRegs = Cons (Pair (EBP, OldEbp),
				 Cons (Pair (ESP, Stack ret_loc), Nil));
		  ciSlots = Cons (Pair (ret_loc, RetPtr), cp.ciSlots)} in
	read bits' conts (Cons (Pair (conts, cp), cutpoints))
    | 2 ->
	let cp, bits' = read_cutpoint bits' in
	read bits' conts (Cons (Pair (conts, cp), cutpoints))
    | _ -> failwith "Unknown cutpoint declaration type" in
  read bits' Nil Nil
end

let functions pc =
  try
    Some (Hashtbl.find function_table pc)
  with Not_found -> None

