(* Copyright (c) 2005, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Functionality used by all extracted verifiers *)

open AsmUtil
open X86Types
open X86Print
open Ver_types

module TotalMap (T : ORDERED_SET) : TOTAL_MAP with type key = T.t =
 struct
   type key = T.t

   module Param = struct
     type t = T.t
     let compare x y =
       if T.eq_dec x y then
	 0
       else if T.lt_dec x y then
	 -1
       else
	 1
   end

   module M = Map.Make(Param)

   type 'x map = 'x * 'x M.t
	 
   let empty d = (d, M.empty)
  
   let sel (d, m) k =
     try
       M.find k m
     with Not_found -> d
  
   let upd (d, m) k v = (d, M.add k v m)

   let eq_dec p m1 m2 =
     try
       M.iter (fun k v -> if not (p (sel m1 k) v) then raise Not_found) (snd m2);
       M.iter (fun k v -> if not (p v (sel m2 k)) then raise Not_found) (snd m1);
       true
     with Not_found -> false

   let subMap p m1 m2 =
     try
       M.iter (fun k v -> if not (p (sel m1 k) v) then raise Not_found) (snd m2);
       M.iter (fun k v -> if not (p v (sel m2 k)) then raise Not_found) (snd m1);
       true
     with Not_found -> false

   let iter f (_, m) = M.iter f m

   let print printDom printRan fmt (d, m) =
     let first = ref true in
     M.iter (fun k v ->
       if v <> d then begin
	 if !first then
	   first := false
	 else
	   Format.fprintf fmt ";@ ";
	 Format.fprintf fmt "@[@[%a@]@ ->@ @[%a@]@]" printDom k printRan v
       end) m

   let mapMap f (d, m) = (f d, M.map f m)
 end

let binaryBits = read_from_file (Sys.argv.(1))

open Format

let pp_print_big_int fmt n =
  pp_print_string fmt (Big_int.string_of_big_int n)

let rec coq_length = function
    Nil -> 0
  | Cons (_, ls) -> 1 + coq_length ls

let initialPc = 0x00100000L
let initialPc_int = 0x00100000

let _ = X86Parse.file_base_offset := initialPc_int

let buildMap (init : 'a) (adder : 'a -> int64 -> int64 -> 'a) : 'a =
  let rec builder byteNum bits map =
    if more bits then
      let b, bits' = read_uint bits 8 in
      let map' =
	if b = 0 then
	  map
	else
	  adder map byteNum (Int64.of_int b) in
      builder (Int64.add byteNum Int64.one) bits' map'
    else
      map in
  builder initialPc binaryBits init

let get_instr_from_bits addr =
  let bits = skip binaryBits (8 * Int64.to_int (Int64.sub addr initialPc)) in
  let ins, bits' = X86Parse.read_instr bits in
  Format.printf "Read instruction[0x%x]: %a@\n@?" (Int64.to_int addr) pp_instr ins;
  Pair (ins, Int64.add initialPc (Int64.of_int (get_byte bits')))

let get_instr_failed () = failwith "Tried to decode instruction in modified memory"
