(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

open Ver_types
open MemoryTypes_types
open MemoryTypes

(* Instantiating and calling the Vanilla verifier *)

module Param = struct
  let minitMem = Int32Map.init 0L
  let codeStart = Ver_fixed.codeStart
  let codeLength = Ver_fixed.codeLength
  let stackStart = Ver_fixed.stackStart
  let stackLength = Ver_fixed.stackLength
  let heapStart = Ver_fixed.heapStart
  let heapLength = Ver_fixed.heapLength
end

module Ver = Verifier(Param)
open Ver
open Dr
open Weak
open Cond
open Stack

module Metadata = struct
  type typ = memoryTy
  type astat = weakState condState

  type 'a coq_list = 'a list
  let coq_list_nil = Nil
  let coq_list_cons = fun x y -> Cons (x, y)
  let coq_list_foldr = fold_right

  type salRegMap = typ stackTy SalRegMap.t
  let salRegMap_empty = SalRegMap.init Top
  let salRegMap_add m r t = SalRegMap.set r t m
  let salRegMap_get m r = SalRegMap.get r m

  type natMap = typ stackTy NatMap.t
  let natMap_empty = NatMap.init Top
  let natMap_add m r t = NatMap.set r t m
  let natMap_elements = NatMap.elements

  type stateDescription = (typ stackTy, (typ, astat) stackState) MemoryTypes.stateDescription
  let build_stateDescription pc regs isRet ss slots cust =
    {sdPc = pc;
     sdRegs = regs;
     sdCust = {aIsRet = isRet;
	       aStackSize = ss;
	       aStackSlots = slots;
	       aCustom = cust}}
  let stateDescription_pc sd = sd.sdPc
  let stateDescription_regs sd = sd.sdRegs
  let stateDescription_slots sd = sd.sdCust.aStackSlots
  let stateDescription_setRegs sd rs = {sd with sdRegs = rs}
  let stateDescription_setSlots sd sls = {sd with sdCust = {sd.sdCust with aStackSlots = sls}}

  type function_info = typ stackTy MemoryTypes.function_info
  let build_function_info pc rt args =
    {fiPc = pc;
     fiRet = rt;
     fiArgs = args}

  type stateList = stateDescription list
  let stateList_nil = Nil
  let stateList_cons x y = Cons (x, y)

  type rootList = (stateList, stateDescription) X86Types.prod list
  let rootList_nil = Nil
  let rootList_cons x y z = Cons (X86Types.Pair (x, y), z)

  let rec read_typ bits =
    let kind, bits = AsmUtil.read_uint bits 32 in
    match kind with
      0 ->
	let n, bits = AsmUtil.read_uint bits 32 in
	Constant (Int64.of_int n), bits
    | 1 ->
	let ts, bits = read_product bits in
	Product ts, bits
    | 2 ->
	let t1, bits = read_typ bits in
	let t2, bits = read_typ bits in
	Sum (t1, t2), bits
    | 3 ->
	let n, bits = AsmUtil.read_uint bits 32 in
	Var (CoqNat.of_int n), bits
    | 4 ->
	let n, bits = AsmUtil.read_uint bits 32 in
	let t, bits = read_typ bits in
	Recursive (CoqNat.of_int n, t), bits
    | _ -> failwith "Bad MemoryTypes type to deserialize"

  and read_product bits =
    let kind, bits = AsmUtil.read_uint bits 32 in
    match kind with
      0 -> PT_Nil, bits
    | 1 ->
	let t, bits = read_typ bits in
	let ts, bits = read_product bits in
	PT_Cons (t, ts), bits
    | _ -> failwith "Bad MemoryTypes product to deserialize"

  let read_astat b =
    {csFlags = FlagMap.init None;
     csCust = SalRegMap.init None}, b
end

module MetadataParser = Ver_fixed.Parser(Metadata)

module Roots = struct
  let roots = MetadataParser.cutpoints
  let functions = MetadataParser.functions
end

module Ver2 = Ver.Dr.Make(Roots)

let _ = if Ver2.verify (CoqNat.of_int 100) then
  print_string "Verification succeeded!  :-)\n"
else
  print_string "Verification failed.  :-(\n"
