(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

open AsmUtil
open X86Types
open X86Print
open Ver_types

(** Functionality used by all extracted verifiers *)

let exeBits = read_from_file (Sys.argv.(1))

let codeStart = 0x00100000L
let codeStart_int = 0x00100000
let codeLength_int = 0x100000
let codeLength = CoqNat.of_int codeLength_int

let stackStart = 0x01001000L
let stackLength_int = 0x1000
let stackLength = CoqNat.of_int stackLength_int

let heapStart = 0x02000000L
let heapLength_int = 0x1000
let heapLength = CoqNat.of_int heapLength_int

let metaAddr = 0x00100000

let _ = X86Parse.file_base_offset := codeStart_int

let decode_instr _ addr =
  let bits = skip exeBits (8 * (Int64.to_int (Int64.sub addr codeStart))) in
  let ins, bits' = X86Parse.read_instr bits in
  Format.printf "Read instruction[0x%x]: %a@\n@?" (Int64.to_int addr) pp_instr ins;
  let r = Pair (ins, Int64.add codeStart (Int64.of_int (get_byte bits'))) in
  r


(** {6 Printing abstract states} *)

let print_big_int fmt n = Format.fprintf fmt "%s" (Big_int.string_of_big_int n)

let print_top () = Format.printf "Top"
let print_int32_option = function
    None -> Format.printf "Anything"
  | Some n -> Format.printf "Constant(%Ld)" n
let print_nostate () = Format.printf "_"

let print_salReg = function
    SX86 r -> pp_reg32 Format.std_formatter r
  | STmp1 -> Format.printf "Tmp1"
  | STmp2 -> Format.printf "Tmp2"

let rec print_salExp = function
    SConst n -> Format.printf "%Ld" n
  | SReg r -> print_salReg r
  | SArith (aop, e1, e2) ->
      Format.printf "(";
      print_salExp e1;
      Format.printf " %a " pp_arith_op aop;
      print_salExp e2;
      Format.printf ")"
  | SShift (sop, e1, e2) ->
      Format.printf "(";
      print_salExp e1;
      Format.printf " <SHIFT> %a)" print_big_int e2

let print_salExp_option = function
    None -> Format.printf "None"
  | Some e -> print_salExp e

let print_weakState iter =
  iter (function Pair (r, t) ->
    if t != None then begin
      Format.printf " ";
      print_salReg r;
      Format.printf " -> ";
      print_salExp_option t;
      Format.printf ";"
    end);
  Format.printf "\n@?"

let print_stackTy print_ty = function
    Top -> Format.printf "Top"
  | Const n -> Format.printf "Const(%Ld)" n
  | Stack n -> Format.printf "Stack(%a)" print_big_int n
  | OldEbp -> Format.printf "OldEbp"
  | RetPtr -> Format.printf "RetPtr"
  | Cust t -> print_ty t

let pp_apc fmt = function
    ApcUnknown -> Format.fprintf fmt "Unknown"
  | ApcConst pc -> Format.fprintf fmt "0x%Lx" pc

let print_apc = pp_apc Format.std_formatter

let print_stateDescription print_ty print_mastate pc default iterRegs abs =
  Format.printf "   PC: %a\n Regs: {" pp_apc pc;
  print_ty default;
  Format.printf "}";
  iterRegs (function Pair (r, t) ->
    if t != default then begin
      Format.printf " ";
      print_salReg r;
      Format.printf " -> ";
      print_ty t;
      Format.printf ";"
    end);
  Format.printf "\n";
  print_mastate abs;
  Format.printf "\n@?"

let print_stackState print_ty print_mastate ss default iterSlots abs =
  Format.printf "SSize: %a\nSlots: {" print_big_int ss;
  print_stackTy print_ty default;
  Format.printf "}";
  iterSlots (function Pair (n, t) ->
    if t != default then begin
      Format.printf " %a -> " print_big_int n;
      print_stackTy print_ty t;
      Format.printf ";"
    end);
  Format.printf "\n";
  print_mastate abs;
  Format.printf "\n@?"

let print_condState print_mastate iterFlags abs =
  Format.printf "Flags: ";
  iterFlags (function Pair (n, t) ->
    match t with
    | None -> ()
    | Some (Pair (Pair (aop, e1), e2)) ->
	Format.printf " %a -> (%a, " pp_flag n pp_arith_op aop;
	print_salExp e1;
	Format.printf ", ";
	print_salExp e2;
	Format.printf ")");
  Format.printf "\n Cust: ";
  print_mastate abs;
  Format.printf "\n@?";


(** {6 Parsing metadata} *)

module Parser (Param : METADATA) : METADATA_PARSE
with type 'a coq_list = 'a Param.coq_list
and type stateDescription = Param.stateDescription
and type function_info = Param.function_info = struct
  type 'a coq_list = 'a Param.coq_list
  type stateDescription = Param.stateDescription
  type function_info = Param.function_info

  open Param

  let pp_salReg fmt = function
      STmp1 -> Format.fprintf fmt "Tmp1"
    | STmp2 -> Format.fprintf fmt "Tmp2"
    | SX86 r -> pp_reg32 fmt r

  let read_ty bits =
    let kind, bits = read_uint bits 32 in
    match kind with
      0 -> Top, bits
    | 1 ->
	let n, bits = read_uint32 bits 32 in
	Const n, bits
    | 2 ->
	let n, bits = read_uint bits 32 in
	Stack (CoqNat.of_int n), bits
    | 3 ->
	OldEbp, bits
    | 4 ->
	RetPtr, bits
    | 5 ->
	let t, bits = Param.read_typ bits in
	Cust t, bits
    | _ -> failwith "Bad type to deserialize"

  let rec read_facts regs slots bits =
    let kind, bits = read_uint bits 32 in
    match kind with
      0 -> regs, slots, bits
    | 1 ->
	let rcode, bits = read_uint bits 32 in
	let ty, bits = read_ty bits in
	let r = X86Util.int_to_reg32 rcode in

	read_facts (Param.salRegMap_add regs (SX86 r) ty) slots bits
    | 2 ->
	let sl, bits = read_uint bits 32 in
	let ty, bits = read_ty bits in

	read_facts regs (Param.natMap_add slots (CoqNat.of_int sl) ty) bits
    | _ -> failwith "Bad fact to deserialize"

  and read_cutpoint bits =
    let pc, bits' = read_uint bits 32 in
    let ss, bits' = read_uint bits' 32 in
    let regs, slots, bits' = read_facts Param.salRegMap_empty Param.natMap_empty bits' in
    let absIn, bits' = Param.read_astat bits' in
    let absOut, bits' = Param.read_astat bits' in
    let state = build_stateDescription
	(ApcConst (Int64.of_int pc))
	regs
	false
	(Big_int.big_int_of_int ss)
	slots
	absIn in
    state, absOut, bits'

  let four = CoqNat.of_int 4

  let cutpoints, functions = begin
    let bits' = skip exeBits (8 * metaAddr) in

    let rec read bits' conts cutpoints functions =
      let kind, bits' = read_uint bits' 32 in
      match kind with
	0 -> cutpoints, functions
      | 1 ->
	  let cp, absOut, bits' = read_cutpoint bits' in
	  
	  let rt = salRegMap_get (stateDescription_regs cp) (SX86 EAX) in

	  let Pair (_, slots) = natMap_elements (stateDescription_slots cp) in
	  let slots = coq_list_foldr (fun h t -> h :: t) [] slots in
	  let slots = List.sort (fun (Pair (n1, _)) (Pair (n2, _)) ->
	    Big_int.compare_big_int n1 n2) slots in
	  let args' = List.map (fun (Pair (_, t)) -> t) slots in
	  let args = List.fold_right coq_list_cons args' coq_list_nil in

	  (*Format.printf "Function at %a\n@?" pp_apc (stateDescription_pc cp);*)
	  let functions =
	    match stateDescription_pc cp with
	      ApcUnknown -> functions
	    | ApcConst pc ->
		coq_list_cons
		  (build_function_info pc rt args)
		  functions in

	  let alen = CoqNat.of_int (4 * List.length args') in
	  let conts = coq_list_cons
	      (build_stateDescription 
		 ApcUnknown
		 (salRegMap_add
		    (salRegMap_add
		       (salRegMap_add
			  salRegMap_empty (SX86 EAX) rt)
		       (SX86 EBP) OldEbp)
		    (SX86 ESP) (Stack alen))
		 true
		 Big_int.zero_big_int
		 natMap_empty
		 absOut)
	      coq_list_nil in

	  let ret_loc = Big_int.add_big_int alen four in
	  let cp = stateDescription_setRegs cp
	      (salRegMap_add
		 (salRegMap_add salRegMap_empty (SX86 ESP) (Stack ret_loc))
		 (SX86 EBP) OldEbp) in
	  let cp = stateDescription_setSlots cp
	      (natMap_add
		 (stateDescription_slots cp)
		 ret_loc RetPtr) in

	  (*let cp = {cp with
		    ciRegs = Cons (Pair (EBP, OldEbp),
				   Cons (Pair (ESP, Stack ret_loc), Nil));
		    ciSlots = Cons (Pair (ret_loc, RetPtr), cp.ciSlots)} in

	  begin match cp.ciPc with
	    ApcConst pc ->
	      Format.printf "Function at %a@\n" pp_addr pc;
	      Hashtbl.add entry_table pc
		{eiRet = ty_to_word_type rt;
		 eiAbsIn = cp.ciCust;
		 eiAbsOut = absOut}
	  | _ -> ()
	  end;*)

	  let cutpoints = coq_list_cons (Pair (conts, cp)) cutpoints in

	  read bits' conts cutpoints functions
      | 2 ->
	  let cp, absOut, bits' = read_cutpoint bits' in

	  (*Format.printf "Cutpoint at %a@\n" pp_addr (cp.ciPc);*)

	  let cutpoints = coq_list_cons (Pair (conts, cp)) cutpoints in

	  read bits' conts cutpoints functions
      | _ -> failwith "Unknown cutpoint declaration type" in

    read bits' coq_list_nil coq_list_nil coq_list_nil
  end

end

