(* Copyright (c) 2006, Regents of the University of California
 * All rights reserved.
 *
 * Author: Adam Chlipala
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 * - Neither the name of the University of California, Berkeley nor the names of
 *   its contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR 
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *)

(** Types used by extracted verifiers *)

open X86Types

type 'ty stackTy =
    Top 
  | Const of bool CoqBvector.bvector
  | Stack of CoqNat.nat
  | OldEbp
  | RetPtr
  | Cust of 'ty

type salReg =
    SX86 of X86Types.reg32
  | STmp1
  | STmp2

type salExp =
  | SConst of bool CoqBvector.bvector
  | SReg of salReg
  | SArith of arith_op * salExp * salExp
  | SShift of shift_op * salExp * CoqNat.nat

type apc =
    ApcUnknown
  | ApcConst of bool CoqBvector.bvector

(** Information needed to read metadata for a particular verifier *)
module type METADATA = sig
  type typ
  (** Custom types *)
  type astat
  (** Additional verification state *)

  (** Lists *)
  type 'a coq_list
  val coq_list_nil : 'a coq_list
  val coq_list_cons : 'a -> 'a coq_list -> 'a coq_list
  val coq_list_foldr : ('a -> 'b -> 'b) -> 'b -> 'a coq_list -> 'b

  (** Finite maps for registers *)
  type salRegMap
  val salRegMap_empty : salRegMap
  val salRegMap_add : salRegMap -> salReg -> typ stackTy -> salRegMap
  val salRegMap_get : salRegMap -> salReg -> typ stackTy

  (** Finite maps for naturals *)
  type natMap
  val natMap_empty : natMap
  val natMap_add : natMap -> CoqNat.nat -> typ stackTy -> natMap
  val natMap_elements : natMap -> (typ stackTy, (CoqNat.nat, typ stackTy) prod coq_list) prod

  (** Abstract states *)
  type stateDescription
  val build_stateDescription : apc -> salRegMap
    -> bool -> CoqNat.nat -> natMap -> astat -> stateDescription
  val stateDescription_pc : stateDescription -> apc
  val stateDescription_regs : stateDescription -> salRegMap
  val stateDescription_slots : stateDescription -> natMap
  val stateDescription_setRegs : stateDescription -> salRegMap -> stateDescription
  val stateDescription_setSlots : stateDescription -> natMap -> stateDescription

  (** Function information entries *)
  type function_info
  val build_function_info : bool CoqBvector.bvector -> typ stackTy
    -> typ stackTy coq_list -> function_info

  (** Deserializers *)
  val read_typ : AsmUtil.bits -> typ * AsmUtil.bits
  val read_astat : AsmUtil.bits -> astat * AsmUtil.bits
end

(** An interface to that metadata *)
module type METADATA_PARSE = sig
  type 'a coq_list
  type stateDescription
  type function_info

  val cutpoints : (stateDescription coq_list, stateDescription) prod coq_list
  (** All program point invariants *)

  val functions : function_info coq_list
  (** Function type information *)
end
